DROP TABLE IF EXISTS item;
CREATE TABLE item (
  id INT NOT NULL,
  name VARCHAR(50),
  type SMALLINT NOT NULL DEFAULT 1,
  cost INT NOT NULL DEFAULT 0,
  member BOOLEAN NOT NULL DEFAULT FALSE,
  bait BOOLEAN NOT NULL DEFAULT FALSE,
  patched BOOLEAN NOT NULL DEFAULT FALSE,
  legacy_inventory BOOLEAN NOT NULL DEFAULT FALSE,
  vanilla_inventory BOOLEAN NOT NULL DEFAULT FALSE,
  epf BOOLEAN NOT NULL DEFAULT FALSE,
  tour BOOLEAN NOT NULL DEFAULT FALSE,
  release_date DATE NOT NULL,
  treasure BOOLEAN NOT NULL DEFAULT FALSE,
  innocent BOOLEAN NOT NULL DEFAULT FALSE,
  PRIMARY KEY (id)
);

ALTER TABLE item ALTER COLUMN release_date SET DEFAULT now();

COMMENT ON TABLE item IS 'Server item crumbs';

COMMENT ON COLUMN item.id IS 'Unique item ID';
COMMENT ON COLUMN item.name IS 'Item name';
COMMENT ON COLUMN item.type IS 'Item clothing type';
COMMENT ON COLUMN item.cost IS 'Cost of item';
COMMENT ON COLUMN item.member IS 'Is member-only?';
COMMENT ON COLUMN item.bait IS 'Is bait item?';
COMMENT ON COLUMN item.patched IS 'Is item patched?';
COMMENT ON COLUMN item.legacy_inventory IS 'Add to default legacy inventory?';
COMMENT ON COLUMN item.vanilla_inventory IS 'Add to default vanilla inventory?';
COMMENT ON COLUMN item.epf IS 'Is EPF item?';
COMMENT ON COLUMN item.tour IS 'Gives tour status?';
COMMENT ON COLUMN item.release_date IS 'Release date of item';
COMMENT ON COLUMN item.treasure IS 'Is a treasure item?';
COMMENT ON COLUMN item.innocent IS 'Is a innocent item?';

DROP TABLE IF EXISTS postcard;
CREATE TABLE postcard (
  id INT NOT NULL,
  name VARCHAR (50) NOT NULL,
  cost INT NOT NULL DEFAULT 10,
  enabled BOOLEAN NOT NULL DEFAULT FALSE,
  PRIMARY KEY (id)
);

COMMENT ON TABLE postcard IS 'Server postcard crumbs';

COMMENT ON COLUMN postcard.id IS 'Unique postcard ID';
COMMENT ON COLUMN postcard.name IS 'Postcard name';
COMMENT ON COLUMN postcard.cost IS 'Cost of postcard';
COMMENT ON COLUMN postcard.enabled IS 'Can send postcard?';

DROP TABLE IF EXISTS igloo;
CREATE TABLE igloo (
  id INT NOT NULL,
  name VARCHAR(50) NOT NULL,
  cost SMALLINT NOT NULL DEFAULT 0,
  patched BOOLEAN NOT NULL DEFAULT FALSE,
  legacy_inventory BOOLEAN NOT NULL DEFAULT FALSE,
  vanilla_inventory BOOLEAN NOT NULL DEFAULT FALSE,
  PRIMARY KEY(id)
);

COMMENT ON TABLE igloo IS 'Server igloo crumbs';

COMMENT ON COLUMN igloo.id IS 'Unique igloo ID';
COMMENT ON COLUMN igloo.name IS 'Igloo name';
COMMENT ON COLUMN igloo.cost IS 'Cost of igloo';
COMMENT ON COLUMN igloo.patched IS 'Is igloo patched?';
COMMENT ON COLUMN igloo.legacy_inventory IS 'Add to default legacy inventory?';
COMMENT ON COLUMN igloo.vanilla_inventory IS 'Add to default vanilla inventory?';

DROP TABLE IF EXISTS location;
CREATE TABLE location (
  id INT NOT NULL,
  name VARCHAR(50) NOT NULL,
  cost INT NOT NULL DEFAULT 0,
  patched BOOLEAN NOT NULL DEFAULT FALSE,
  legacy_inventory BOOLEAN NOT NULL DEFAULT FALSE,
  vanilla_inventory BOOLEAN NOT NULL DEFAULT FALSE,
  PRIMARY KEY (id)
);

COMMENT ON TABLE location IS 'Server location crumbs';

COMMENT ON COLUMN location.id IS 'Unique location ID';
COMMENT ON COLUMN location.name IS 'Location name';
COMMENT ON COLUMN location.cost IS 'Cost of location';
COMMENT ON COLUMN location.patched IS 'Is location patched?';
COMMENT ON COLUMN location.legacy_inventory IS 'Add to default legacy inventory?';
COMMENT ON COLUMN location.vanilla_inventory IS 'Add to default vanilla inventory?';

DROP TABLE IF EXISTS furniture;
CREATE TABLE furniture (
  id INT NOT NULL,
  name VARCHAR(50) NOT NULL,
  type SMALLINT NOT NULL DEFAULT 1,
  sort SMALLINT NOT NULL DEFAULT 1,
  cost INT NOT NULL DEFAULT 0,
  member BOOLEAN NOT NULL DEFAULT FALSE,
  patched BOOLEAN NOT NULL DEFAULT FALSE,
  legacy_inventory BOOLEAN NOT NULL DEFAULT FALSE,
  vanilla_inventory BOOLEAN NOT NULL DEFAULT FALSE,
  bait BOOLEAN NOT NULL DEFAULT FALSE,
  max_quantity SMALLINT NOT NULL DEFAULT 100,
  innocent BOOLEAN NOT NULL DEFAULT FALSE,
  PRIMARY KEY(id)
);

COMMENT ON TABLE furniture IS 'Server furniture crumbs';

COMMENT ON COLUMN furniture.id IS 'Unique furniture ID';
COMMENT ON COLUMN furniture.type IS 'Furniture type ID';
COMMENT ON COLUMN furniture.sort IS 'Furniture sort ID';
COMMENT ON COLUMN furniture.cost IS 'Cost of furniture';
COMMENT ON COLUMN furniture.member IS 'Is member-only?';
COMMENT ON COLUMN furniture.patched IS 'Is furniture patched?';
COMMENT ON COLUMN furniture.legacy_inventory IS 'Add to default legacy inventory?';
COMMENT ON COLUMN furniture.vanilla_inventory IS 'Add to default vanilla inventory?';
COMMENT ON COLUMN furniture.bait IS 'Is furniture bait?';
COMMENT ON COLUMN furniture.max_quantity IS 'Max inventory quantity';
COMMENT ON COLUMN furniture.innocent IS 'Is furniture innocent?';

DROP TABLE IF EXISTS flooring;
CREATE TABLE flooring (
  id INT NOT NULL,
  name VARCHAR(50),
  cost INT NOT NULL DEFAULT 0,
  patched BOOLEAN NOT NULL DEFAULT FALSE,
  legacy_inventory BOOLEAN NOT NULL DEFAULT FALSE,
  vanilla_inventory BOOLEAN NOT NULL DEFAULT FALSE,
  PRIMARY KEY (id)
);

COMMENT ON TABLE flooring IS 'Server flooring crumbs';

COMMENT ON COLUMN flooring.id IS 'Unique flooring ID';
COMMENT ON COLUMN flooring.name IS 'Flooring name';
COMMENT ON COLUMN flooring.cost IS 'Cost of flooring';
COMMENT ON COLUMN flooring.patched IS 'Is flooring patched?';
COMMENT ON COLUMN flooring.legacy_inventory IS 'Add to default legacy inventory?';
COMMENT ON COLUMN flooring.vanilla_inventory IS 'Add to default vanilla inventory?';

DROP TABLE IF EXISTS card;
CREATE TABLE card (
  id INT NOT NULL,
  name VARCHAR(50) NOT NULL,
  set_id SMALLINT NOT NULL DEFAULT 1,
  power_id SMALLINT NOT NULL DEFAULT 0,
  element CHAR(1) NOT NULL DEFAULT 's',
  color CHAR(1) NOT NULL DEFAULT 'b',
  value SMALLINT NOT NULL DEFAULT 2,
  description VARCHAR(255) NOT NULL DEFAULT '',
  PRIMARY KEY (id)
);

COMMENT ON TABLE card IS 'Server jitsu card crumbs';

COMMENT ON COLUMN card.id IS 'Unique card ID';
COMMENT ON COLUMN card.name IS 'Card name';
COMMENT ON COLUMN card.set_id IS 'Card set ID';
COMMENT ON COLUMN card.power_id IS 'Card power ID';
COMMENT ON COLUMN card.element IS 'Card element';
COMMENT ON COLUMN card.color IS 'Card color';
COMMENT ON COLUMN card.value IS 'Value of card';
COMMENT ON COLUMN card.description IS 'Play description';

DROP TABLE IF EXISTS card_starter_deck;
CREATE TABLE card_starter_deck (
  item_id INT NOT NULL,
  card_id INT NOT NULL,
  quantity SMALLINT NOT NULL DEFAULT 1,
  PRIMARY KEY (item_id, card_id),
  CONSTRAINT card_starter_deck_ibfk_1 FOREIGN KEY (item_id) REFERENCES item (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT card_starter_deck_ibfk_2 FOREIGN KEY (card_id) REFERENCES card(id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE card_starter_deck IS 'Jitsu card starter decks';

COMMENT ON COLUMN card_starter_deck.item_id IS 'Starter deck item ID';
COMMENT ON COLUMN card_starter_deck.card_id IS 'Starter deck card ID';
COMMENT ON COLUMN card_starter_deck.quantity IS 'Card quantity';

DROP TABLE IF EXISTS stamp_group;
CREATE TABLE stamp_group (
  id INT NOT NULL,
  name VARCHAR(50) NOT NULL,
  parent_id INT DEFAULT NULL,
  PRIMARY KEY (id),
  CONSTRAINT stamp_group_ibfk_1 FOREIGN KEY (parent_id) REFERENCES stamp_group (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE stamp_group IS 'Stamp group collections';

COMMENT ON COLUMN stamp_group.id IS 'Unique stamp group ID';
COMMENT ON COLUMN stamp_group.name IS 'Name of stamp group';
COMMENT ON COLUMN stamp_group.parent_id IS 'Parent stamp group ID';

DROP TABLE IF EXISTS stamp;
CREATE TABLE stamp (
  id INT NOT NULL,
  name VARCHAR(50) NOT NULL,
  group_id SMALLINT NOT NULL,
  member BOOLEAN NOT NULL DEFAULT FALSE,
  rank SMALLINT NOT NULL DEFAULT 1,
  description VARCHAR(255) NOT NULL DEFAULT '',
  PRIMARY KEY(id),
  CONSTRAINT stamp_ibfk_1 FOREIGN KEY (group_id) REFERENCES stamp_group (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE stamp IS 'Server stamp crumbs';

COMMENT ON COLUMN stamp.id IS 'Unique stamp ID';
COMMENT ON COLUMN stamp.name IS 'Stamp name';
COMMENT ON COLUMN stamp.group_id IS 'Stamp group ID';
COMMENT ON COLUMN stamp.member IS 'Is member-only?';
COMMENT ON COLUMN stamp.rank IS 'Stamp difficulty ranking';
COMMENT ON COLUMN stamp.description IS 'Stamp description';

DROP TABLE IF EXISTS room;
CREATE TABLE room (
  id INT NOT NULL,
  internal_id SERIAL NOT NULL,
  name VARCHAR(50) NOT NULL,
  member BOOLEAN NOT NULL DEFAULT FALSE,
  max_users SMALLINT NOT NULL DEFAULT 80,
  required_item INT DEFAULT NULL,
  game BOOLEAN NOT NULL DEFAULT FALSE,
  blackhole BOOLEAN NOT NULL DEFAULT FALSE,
  spawn BOOLEAN NOT NULL DEFAULT FALSE,
  stamp_group INT DEFAULT NULL,
  PRIMARY KEY(id),
  CONSTRAINT room_ibfk_1 FOREIGN KEY (required_item) REFERENCES item (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT room_ibfk_2 FOREIGN KEY (stamp_group) REFERENCES stamp_group (id) ON DELETE CASCADE ON UPDATE CASCADE
);

CREATE UNIQUE INDEX room_internal_id ON room (internal_id);

COMMENT ON TABLE room IS 'Server room crumbs';

COMMENT ON COLUMN room.id IS 'Unique room ID';
COMMENT ON COLUMN room.internal_id IS 'Internal room key';
COMMENT ON COLUMN room.name IS 'Room name';
COMMENT ON COLUMN room.member IS 'Is member-only?';
COMMENT ON COLUMN room.max_users IS 'Maximum room users';
COMMENT ON COLUMN room.required_item IS 'Required inventory item';
COMMENT ON COLUMN room.game IS 'Is game room?';
COMMENT ON COLUMN room.blackhole IS 'Is blackhole game room?';
COMMENT ON COLUMN room.spawn IS 'Is spawn room?';

DROP TABLE IF EXISTS character;
CREATE TABLE character (
  id INT NOT NULL,
  name VARCHAR (30) NOT NULL,
  gift_id INT DEFAULT NULL,
  stamp_id INT DEFAULT NULL,
  PRIMARY KEY (id),
  CONSTRAINT character_ibfk_1 FOREIGN KEY (gift_id) REFERENCES item (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT character_ibfk_2 FOREIGN KEY (stamp_id) REFERENCES stamp (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE character IS 'Server character crumbs';

COMMENT ON COLUMN character.id IS 'Unique character ID';
COMMENT ON COLUMN character.name IS 'Character name';
COMMENT ON COLUMN character.gift_id IS 'Character gift item ID';
COMMENT ON COLUMN character.stamp_id IS 'Character stamp ID';

DROP TABLE IF EXISTS epf_com_message;
CREATE TABLE epf_com_message (
  message TEXT NOT NULL,
  character_id INT NOT NULL,
  date TIMESTAMP NOT NULL,
  CONSTRAINT epf_com_message_ibfk_1 FOREIGN KEY (character_id) REFERENCES character(id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE epf_com_message IS 'EPF phone com messages';

COMMENT ON COLUMN epf_com_message.message IS 'Message content';
COMMENT ON COLUMN epf_com_message.character_id IS 'Character ID of message';
COMMENT ON COLUMN epf_com_message.date IS 'Date of message creation';

ALTER TABLE epf_com_message ALTER COLUMN date SET DEFAULT now();

DROP TABLE IF EXISTS puffle_item;
CREATE TABLE puffle_item (
  id INT NOT NULL,
  parent_id INT NOT NULL,
  name VARCHAR(50) NOT NULL DEFAULT '',
  type VARCHAR(10) NOT NULL DEFAULT 'care',
  play_external VARCHAR (10) NOT NULL DEFAULT 'none',
  cost INT NOT NULL DEFAULT 0,
  quantity SMALLINT NOT NULL DEFAULT 1,
  member BOOLEAN NOT NULL DEFAULT FALSE,
  food_effect SMALLINT NOT NULL DEFAULT 0,
  rest_effect SMALLINT NOT NULL DEFAULT 0,
  play_effect SMALLINT NOT NULL DEFAULT 0,
  clean_effect SMALLINT NOT NULL DEFAULT 0,
  PRIMARY KEY (id),
  CONSTRAINT puffle_item_ibfk_1 FOREIGN KEY (parent_id) REFERENCES puffle_item (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE puffle_item IS 'Server puffle care item crumbs';

COMMENT ON COLUMN puffle_item.id IS 'Unique care item ID';
COMMENT ON COLUMN puffle_item.parent_id IS 'Parent care item ID';
COMMENT ON COLUMN puffle_item.name IS 'Care item name';
COMMENT ON COLUMN puffle_item.type IS 'Type of care item';
COMMENT ON COLUMN puffle_item.play_external IS 'External play mode';
COMMENT ON COLUMN puffle_item.cost IS 'Cost of care item';
COMMENT ON COLUMN puffle_item.quantity IS 'Base quantity of purchase';
COMMENT ON COLUMN puffle_item.member IS 'Is member-only?';
COMMENT ON COLUMN puffle_item.food_effect IS 'Effect on puffle food level';
COMMENT ON COLUMN puffle_item.rest_effect IS 'Effect on puffle rest level';
COMMENT ON COLUMN puffle_item.play_effect IS 'Effect on puffle play level';
COMMENT ON COLUMN puffle_item.clean_effect IS 'Effect on puffle clean level';

DROP TABLE IF EXISTS puffle;
CREATE TABLE puffle (
  id INT NOT NULL,
  parent_id SMALLINT DEFAULT NULL,
  name VARCHAR(50) NOT NULL DEFAULT '',
  cost INT NOT NULL DEFAULT 0,
  member BOOLEAN NOT NULL DEFAULT FALSE,
  favourite_food SMALLINT NOT NULL,
  favourite_toy SMALLINT DEFAULT NULL,
  runaway_postcard SMALLINT DEFAULT NULL,
  PRIMARY KEY (id),
  CONSTRAINT puffle_ibfk_1 FOREIGN KEY (parent_id) REFERENCES puffle (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT puffle_ibfk_2 FOREIGN KEY (favourite_food) REFERENCES puffle_item (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT puffle_ibfk_3 FOREIGN KEY (favourite_toy) REFERENCES puffle_item (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT puffle_ibfk_4 FOREIGN KEY (runaway_postcard) REFERENCES postcard (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE puffle IS 'Server puffle crumbs';

COMMENT ON COLUMN puffle.id IS 'Unique puffle ID';
COMMENT ON COLUMN puffle.parent_id IS 'Base color puffle ID';
COMMENT ON COLUMN puffle.name IS 'Puffle name';
COMMENT ON COLUMN puffle.cost IS 'Puffle cost';
COMMENT ON COLUMN puffle.member IS 'Is member-only?';
COMMENT ON COLUMN puffle.favourite_food IS 'Favourite puffle-care item';
COMMENT ON COLUMN puffle.runaway_postcard IS 'Runaway postcard ID';

DROP TABLE IF EXISTS puffle_treasure_item;
CREATE TABLE puffle_treasure_item (
  puffle_id INT NOT NULL,
  item_id INT NOT NULL,
  PRIMARY KEY (puffle_id, item_id),
  CONSTRAINT puffle_treasure_item_ibfk_1 FOREIGN KEY (puffle_id) REFERENCES puffle (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT puffle_treasure_item_ibfk_2 FOREIGN KEY (item_id) REFERENCES item (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE puffle_treasure_item IS 'Puffle digging treasure clothing';

COMMENT ON COLUMN puffle_treasure_item.puffle_id IS 'Puffle type ID';
COMMENT ON COLUMN puffle_treasure_item.item_id IS 'Clothing item ID';

DROP TABLE IF EXISTS puffle_treasure_furniture;
CREATE TABLE puffle_treasure_furniture (
  puffle_id INT NOT NULL,
  furniture_id INT NOT NULL,
  PRIMARY KEY (puffle_id, furniture_id),
  CONSTRAINT puffle_treasure_furniture_ibfk_1 FOREIGN KEY (puffle_id) REFERENCES puffle (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT puffle_treasure_furniture_ibfk_2 FOREIGN KEY (furniture_id) REFERENCES furniture (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE puffle_treasure_furniture IS 'Puffle digging treasure furniture';

COMMENT ON COLUMN puffle_treasure_furniture.puffle_id IS 'Puffle type ID';
COMMENT ON COLUMN puffle_treasure_furniture.furniture_id IS 'Furniture item ID';

DROP TABLE IF EXISTS puffle_treasure_puffle_item;
CREATE TABLE puffle_treasure_puffle_item (
  puffle_id INT NOT NULL,
  puffle_item_id INT NOT NULL,
  PRIMARY KEY (puffle_id, puffle_item_id),
  CONSTRAINT puffle_treasure_puffle_item_ibfk_1 FOREIGN KEY (puffle_id) REFERENCES puffle (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT puffle_treasure_puffle_item_ibfk_2 FOREIGN KEY (puffle_item_id) REFERENCES puffle_item (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE puffle_treasure_puffle_item IS 'Puffle digging treasure puffle care items';

COMMENT ON COLUMN puffle_treasure_puffle_item.puffle_id IS 'Puffle type ID';
COMMENT ON COLUMN puffle_treasure_puffle_item.puffle_item_id IS 'Puffle care item ID';

DROP TABLE IF EXISTS quest;
CREATE TABLE quest (
  id SERIAL,
  name VARCHAR(30) NOT NULL,
  PRIMARY KEY(id)
);

COMMENT ON TABLE quest IS 'Player map quests';

COMMENT ON COLUMN quest.id IS 'Unique quest ID';
COMMENT ON COLUMN quest.name IS 'Short name of quest';

DROP TABLE IF EXISTS quest_award_item;
CREATE TABLE quest_award_item (
  quest_id INT NOT NULL,
  item_id INT NOT NULL,
  PRIMARY KEY (quest_id, item_id),
  CONSTRAINT quest_award_item_ibfk_1 FOREIGN KEY (quest_id) REFERENCES quest (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT quest_award_item_ibfk_2 FOREIGN KEY (item_id) REFERENCES item (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON COLUMN quest_award_item.quest_id IS 'Quest ID';
COMMENT ON COLUMN quest_award_item.item_id IS 'Clothing item ID';

DROP TABLE IF EXISTS quest_award_furniture;
CREATE TABLE quest_award_furniture (
  quest_id INT NOT NULL,
  furniture_id INT NOT NULL,
  quantity SMALLINT NOT NULL DEFAULT 1,
  PRIMARY KEY (quest_id, furniture_id),
  CONSTRAINT quest_award_furniture_ibfk_1 FOREIGN KEY (quest_id) REFERENCES quest (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT quest_award_furniture_ibfk_2 FOREIGN KEY (furniture_id) REFERENCES furniture (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON COLUMN quest_award_furniture.quest_id IS 'Quest ID';
COMMENT ON COLUMN quest_award_furniture.furniture_id IS 'Furniture item ID';

DROP TABLE IF EXISTS quest_award_puffle_item;
CREATE TABLE quest_award_puffle_item (
  quest_id INT NOT NULL,
  puffle_item_id INT NOT NULL,
  quantity SMALLINT NOT NULL DEFAULT 1,
  PRIMARY KEY (quest_id, puffle_item_id),
  CONSTRAINT quest_award_puffle_item_ibfk_1 FOREIGN KEY (quest_id) REFERENCES quest (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT quest_award_puffle_item_ibfk_2 FOREIGN KEY (puffle_item_id) REFERENCES puffle_item (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON COLUMN quest_award_puffle_item.quest_id IS 'Quest ID';
COMMENT ON COLUMN quest_award_puffle_item.puffle_item_id IS 'Puffle care item ID';

DROP TABLE IF EXISTS quest_task;
CREATE TABLE quest_task (
  id SERIAL NOT NULL,
  quest_id INT NOT NULL,
  description VARCHAR(50) NOT NULL,
  room_id INT DEFAULT NULL,
  data VARCHAR(50) DEFAULT NULL,
  PRIMARY KEY (id),
  CONSTRAINT quest_task_ibfk_1 FOREIGN KEY (quest_id) REFERENCES quest (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT quest_task_ibfk_2 FOREIGN KEY (room_id) REFERENCES room (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE quest_task IS 'Player map quest tasks';

COMMENT ON COLUMN quest_task.id IS 'Unique task ID';
COMMENT ON COLUMN quest_task.quest_id IS 'Task quest ID';
COMMENT ON COLUMN quest_task.description IS 'Description of task';
COMMENT ON COLUMN quest_task.room_id IS 'Room ID for completion';

DROP TABLE IF EXISTS dance_song;
CREATE TABLE dance_song (
  id INT NOT NULL,
  name VARCHAR (30) NOT NULL,
  song_length_millis INT NOT NULL,
  song_length INT NOT NULL,
  millis_per_bar INT NOT NULL,
  PRIMARY KEY (id)
);

COMMENT ON TABLE dance_song IS 'Dance contest multiplayer tracks';

COMMENT ON COLUMN dance_song.id IS 'Unique song ID';
COMMENT ON COLUMN dance_song.name IS 'Name of song';
COMMENT ON COLUMN dance_song.song_length_millis IS 'Length of song in milliseconds';
COMMENT ON COLUMN dance_song.song_length IS 'Length of song in beats';
COMMENT ON COLUMN dance_song.millis_per_bar IS 'Milliseconds per song note';

CREATE TABLE room_waddle (
  id INT NOT NULL,
  room_id INT NOT NULL,
  seats SMALLINT NOT NULL DEFAULT 2,
  game VARCHAR(20) NOT NULL,
  PRIMARY KEY (id, room_id),
  CONSTRAINT room_waddle_ibfk_1 FOREIGN KEY (room_id) REFERENCES room (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE room_waddle IS 'Server waddle games';

COMMENT ON COLUMN room_waddle.id IS 'Waddle ID';
COMMENT ON COLUMN room_waddle.room_id IS 'Room ID of waddle';
COMMENT ON COLUMN room_waddle.seats IS 'Number of seats';
COMMENT ON COLUMN room_waddle.game IS 'Game of waddle';

CREATE TABLE room_table (
  id INT NOT NULL,
  room_id INT NOT NULL,
  game VARCHAR(20) NOT NULL,
  PRIMARY KEY (id, room_id),
  CONSTRAINT room_table_ibfk_1 FOREIGN KEY (room_id) REFERENCES room (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE room_table IS 'Server table games';

COMMENT ON COLUMN room_table.id IS 'Table ID';
COMMENT ON COLUMN room_table.room_id IS 'Room ID of table';
COMMENT ON COLUMN room_table.game IS 'Game of table';

CREATE TABLE chat_filter_rule (
  word TEXT,
  filter BOOLEAN NOT NULL DEFAULT FALSE,
  warn BOOLEAN NOT NULL DEFAULT FALSE,
  ban BOOLEAN NOT NULL DEFAULT FALSE,
  PRIMARY KEY(word)
);

CREATE UNIQUE INDEX chat_filter_rule_word ON chat_filter_rule(word);

COMMENT ON COLUMN chat_filter_rule.word IS 'Word to filter';
COMMENT ON COLUMN chat_filter_rule.filter IS 'Hide word from players';
COMMENT ON COLUMN chat_filter_rule.warn IS 'Warn player for word';
COMMENT ON COLUMN chat_filter_rule.ban IS 'Ban player for word';

DROP TABLE IF EXISTS penguin;
CREATE TABLE penguin (
  id SERIAL,
  username VARCHAR(12) NOT NULL,
  nickname VARCHAR(30) NOT NULL,
  password CHAR(60) NOT NULL,
  email VARCHAR(255) NOT NULL,
  registration_date TIMESTAMP NOT NULL,
  active BOOLEAN NOT NULL DEFAULT FALSE,
  safe_chat BOOLEAN NOT NULL DEFAULT FALSE,
  last_paycheck TIMESTAMP NOT NULL,
  minutes_played INT NOT NULL DEFAULT 0,
  moderator BOOLEAN NOT NULL DEFAULT FALSE,
  stealth_moderator BOOLEAN NOT NULL DEFAULT FALSE,
  character INT DEFAULT NULL,
  igloo INT DEFAULT NULL,
  coins INT NOT NULL DEFAULT 500,
  color INT DEFAULT NULL,
  head INT DEFAULT NULL,
  face INT DEFAULT NULL,
  neck INT DEFAULT NULL,
  body INT DEFAULT NULL,
  hand INT DEFAULT NULL,
  feet INT DEFAULT NULL,
  photo INT DEFAULT NULL,
  flag INT DEFAULT NULL,
  permaban BOOLEAN NOT NULL DEFAULT FALSE,
  book_modified SMALLINT NOT NULL DEFAULT 0,
  book_color SMALLINT NOT NULL DEFAULT 1,
  book_highlight SMALLINT NOT NULL DEFAULT 1,
  book_pattern SMALLINT NOT NULL DEFAULT 0,
  book_icon SMALLINT NOT NULL DEFAULT 1,
  agent_status BOOLEAN NOT NULL DEFAULT FALSE,
  field_op_status SMALLINT NOT NULL DEFAULT 0,
  career_medals INT NOT NULL DEFAULT 0,
  agent_medals INT NOT NULL DEFAULT 0,
  last_field_op TIMESTAMP NOT NULL,
  com_message_read_date TIMESTAMP NOT NULL,
  ninja_rank SMALLINT NOT NULL DEFAULT 0,
  ninja_progress SMALLINT NOT NULL DEFAULT 0,
  fire_ninja_rank SMALLINT NOT NULL DEFAULT 0,
  fire_ninja_progress SMALLINT NOT NULL DEFAULT 0,
  water_ninja_rank SMALLINT NOT NULL DEFAULT 0,
  water_ninja_progress SMALLINT NOT NULL DEFAULT 0,
  ninja_matches_won INT NOT NULL DEFAULT 0,
  fire_matches_won INT NOT NULL DEFAULT 0,
  water_matches_won INT NOT NULL DEFAULT 0,
  rainbow_adoptability BOOLEAN NOT NULL DEFAULT FALSE,
  has_dug BOOLEAN NOT NULL DEFAULT FALSE,
  puffle_handler BOOLEAN NOT NULL DEFAULT FALSE,
  nuggets SMALLINT NOT NULL DEFAULT 0,
  walking INT DEFAULT NULL,
  opened_playercard BOOLEAN NOT NULL DEFAULT FALSE,
  special_wave BOOLEAN NOT NULL DEFAULT FALSE,
  special_dance BOOLEAN NOT NULL DEFAULT FALSE,
  special_snowball BOOLEAN NOT NULL DEFAULT FALSE,
  map_category SMALLINT NOT NULL DEFAULT 0,
  status_field INT NOT NULL DEFAULT 0,
  timer_active BOOLEAN NOT NULL DEFAULT FALSE,
  timer_start TIME NOT NULL DEFAULT '00:00:00',
  timer_end TIME NOT NULL DEFAULT '23:59:59',
  timer_total INTERVAL NOT NULL DEFAULT '01:00:00',
  grounded BOOLEAN NOT NULL DEFAULT FALSE,
  approval_en BOOLEAN NOT NULL DEFAULT FALSE,
  approval_pt BOOLEAN NOT NULL DEFAULT FALSE,
  approval_fr BOOLEAN NOT NULL DEFAULT FALSE,
  approval_es BOOLEAN NOT NULL DEFAULT FALSE,
  approval_de BOOLEAN NOT NULL DEFAULT FALSE,
  approval_ru BOOLEAN NOT NULL DEFAULT FALSE,
  rejection_en BOOLEAN NOT NULL DEFAULT FALSE,
  rejection_pt BOOLEAN NOT NULL DEFAULT FALSE,
  rejection_fr BOOLEAN NOT NULL DEFAULT FALSE,
  rejection_es BOOLEAN NOT NULL DEFAULT FALSE,
  rejection_de BOOLEAN NOT NULL DEFAULT FALSE,
  rejection_ru BOOLEAN NOT NULL DEFAULT FALSE,
  PRIMARY KEY (id),
  CONSTRAINT penguin_ibfk_1 FOREIGN KEY (color) REFERENCES item (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT penguin_ibfk_2 FOREIGN KEY (head) REFERENCES item (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT penguin_ibfk_3 FOREIGN KEY (face) REFERENCES item (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT penguin_ibfk_4 FOREIGN KEY (neck) REFERENCES item (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT penguin_ibfk_5 FOREIGN KEY (body) REFERENCES item (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT penguin_ibfk_6 FOREIGN KEY (hand) REFERENCES item (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT penguin_ibfk_7 FOREIGN KEY (feet) REFERENCES item (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT penguin_ibfk_8 FOREIGN KEY (photo) REFERENCES item (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT penguin_ibfk_9 FOREIGN KEY (flag) REFERENCES item (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT penguin_ibfk_10 FOREIGN KEY (character) REFERENCES character (id) ON DELETE CASCADE ON UPDATE CASCADE
);

CREATE INDEX penguin_email ON Penguin(email);
CREATE UNIQUE INDEX penguin_username ON Penguin(username);

ALTER TABLE penguin ALTER COLUMN registration_date SET DEFAULT now();
ALTER TABLE penguin ALTER COLUMN last_paycheck SET DEFAULT now();
ALTER TABLE penguin ALTER COLUMN last_field_op SET DEFAULT now();
ALTER TABLE penguin ALTER COLUMN com_message_read_date SET DEFAULT now();

ALTER SEQUENCE penguin_id_seq RESTART WITH 101;

COMMENT ON TABLE penguin IS 'Penguins';

COMMENT ON COLUMN penguin.id IS 'Unique penguin ID';
COMMENT ON COLUMN penguin.username IS 'Penguin login name';
COMMENT ON COLUMN penguin.nickname IS 'Penguin display name';
COMMENT ON COLUMN penguin.password IS 'Password hash';
COMMENT ON COLUMN penguin.email IS 'User Email address';
COMMENT ON COLUMN penguin.registration_date IS 'Date of registration';
COMMENT ON COLUMN penguin.active IS 'email activated';
COMMENT ON COLUMN penguin.last_paycheck IS 'EPF previous paycheck';
COMMENT ON COLUMN penguin.minutes_played IS 'Total minutes connected';
COMMENT ON COLUMN penguin.moderator IS 'Is user moderator?';
COMMENT ON COLUMN penguin.character IS 'Character ID';
COMMENT ON COLUMN penguin.igloo IS 'Penguin active igloo ID';
COMMENT ON COLUMN penguin.coins IS 'Penguin coins';
COMMENT ON COLUMN penguin.color IS 'Penguin color ID';
COMMENT ON COLUMN penguin.head IS 'Penguin head item ID';
COMMENT ON COLUMN penguin.face IS 'Penguin face item ID';
COMMENT ON COLUMN penguin.neck IS 'Penguin neck item ID';
COMMENT ON COLUMN penguin.body IS 'Penguin body item ID';
COMMENT ON COLUMN penguin.hand IS 'Penguin hand item ID';
COMMENT ON COLUMN penguin.feet IS 'Penguin feet item ID';
COMMENT ON COLUMN penguin.photo IS 'Penguin background ID';
COMMENT ON COLUMN penguin.flag IS 'Penguin pin ID';
COMMENT ON COLUMN penguin.permaban IS 'Is penguin banned forever?';
COMMENT ON COLUMN penguin.book_modified IS 'Is book cover modified?';
COMMENT ON COLUMN penguin.book_color IS 'Stampbook cover color';
COMMENT ON COLUMN penguin.book_highlight IS 'Stampbook highlight color';
COMMENT ON COLUMN penguin.book_pattern IS 'Stampbook cover pattern';
COMMENT ON COLUMN penguin.book_icon IS 'Stampbook cover icon';
COMMENT ON COLUMN penguin.agent_status IS 'Is penguin EPF agent?';
COMMENT ON COLUMN penguin.field_op_status IS 'Is field op complete?';
COMMENT ON COLUMN penguin.career_medals IS 'Total career medals';
COMMENT ON COLUMN penguin.agent_medals IS 'Current medals';
COMMENT ON COLUMN penguin.last_field_op IS 'Date of last field op';
COMMENT ON COLUMN penguin.com_message_read_date IS 'Recent agent message read';
COMMENT ON COLUMN penguin.ninja_rank IS 'Ninja rank';
COMMENT ON COLUMN penguin.ninja_progress IS 'Ninja progress';
COMMENT ON COLUMN penguin.fire_ninja_rank IS 'Fire ninja rank';
COMMENT ON COLUMN penguin.fire_ninja_progress IS 'Fire ninja progress';
COMMENT ON COLUMN penguin.water_ninja_rank IS 'Water ninja rank';
COMMENT ON COLUMN penguin.water_ninja_progress IS 'Water ninja progress';
COMMENT ON COLUMN penguin.ninja_matches_won IS 'CardJitsu matches won';
COMMENT ON COLUMN penguin.fire_matches_won IS 'JitsuFire matches won';
COMMENT ON COLUMN penguin.water_matches_won IS 'JitsuWater matces won';
COMMENT ON COLUMN penguin.rainbow_adoptability IS 'Rainbow puffle adoptability status';
COMMENT ON COLUMN penguin.has_dug IS 'Puffle digging boolean';
COMMENT ON COLUMN penguin.puffle_handler IS 'Has met puffle handler?';
COMMENT ON COLUMN penguin.nuggets IS 'Golden puffle nuggets';
COMMENT ON COLUMN penguin.walking IS 'Walking puffle ID';
COMMENT ON COLUMN penguin.opened_playercard IS 'Has player opened playercard?';
COMMENT ON COLUMN penguin.map_category IS 'Currently selected map category';
COMMENT ON COLUMN penguin.status_field IS 'New player status field';
COMMENT ON COLUMN penguin.timer_active IS 'Is egg-timer active?';
COMMENT ON COLUMN penguin.timer_start IS 'Egg-timer start time';
COMMENT ON COLUMN penguin.timer_end IS 'Egg-timer end time';
COMMENT ON COLUMN penguin.timer_total IS 'Egg-timer total play time';
COMMENT ON COLUMN penguin.grounded IS 'Is player grounded?';
COMMENT ON COLUMN penguin.approval_en IS 'English username approval';
COMMENT ON COLUMN penguin.approval_pt IS 'Portuguese username approval';
COMMENT ON COLUMN penguin.approval_fr IS 'French username approval';
COMMENT ON COLUMN penguin.approval_es IS 'Spanish username approval';
COMMENT ON COLUMN penguin.approval_de IS 'Dutch username approval';
COMMENT ON COLUMN penguin.approval_ru IS 'Russian username approval';

DROP TABLE IF EXISTS activation_key;
CREATE TABLE activation_key (
  penguin_id INT NOT NULL,
  activation_key CHAR(255) NOT NULL,
  PRIMARY KEY (penguin_id, activation_key),
  CONSTRAINT activation_key_ibfk_1 FOREIGN KEY (penguin_id) REFERENCES penguin(id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE activation_key IS 'Penguin activation keys';

COMMENT ON COLUMN activation_key.penguin_id IS 'Penguin ID';
COMMENT ON COLUMN activation_key.activation_key IS 'Penguin activation key';

DROP TABLE IF EXISTS permission;
CREATE TABLE permission (
  id SERIAL NOT NULL,
  name VARCHAR(50) NOT NULL,
  enabled BOOLEAN NOT NULL DEFAULT TRUE,
  PRIMARY KEY (id)
);

CREATE UNIQUE INDEX permission_name ON permission(name);

COMMENT ON TABLE permission IS 'Registered server permissions';

COMMENT ON COLUMN permission.id IS 'Unique permission ID';
COMMENT ON COLUMN permission.name IS 'Unique permission identifier';

DROP TABLE IF EXISTS penguin_permission;
CREATE TABLE penguin_permission (
  penguin_id INT NOT NULL,
  permission_id INT NOT NULL,
  PRIMARY KEY (penguin_id),
  CONSTRAINT penguin_permission_ibfk_1 FOREIGN KEY (penguin_id) REFERENCES penguin(id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT penguin_permission_ibfk_2 FOREIGN KEY (permission_id) REFERENCES permission(id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE penguin_permission IS 'Penguin permissions';

COMMENT ON COLUMN penguin_permission.penguin_id IS 'Penguin ID';
COMMENT ON COLUMN penguin_permission.permission_id IS 'Penguin permission ID';

DROP TABLE IF EXISTS report;
CREATE TABLE report (
  id SERIAL NOT NULL,
  penguin_id INT NOT NULL,
  reporter_id INT NOT NULL,
  report_type SMALLINT NOT NULL DEFAULT 0,
  date TIMESTAMP NOT NULL,
  server_id INT NOT NULL,
  room_id INT NOT NULL,
  PRIMARY KEY(id),
  CONSTRAINT report_ibfk_1 FOREIGN KEY (penguin_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT report_ibfk_2 FOREIGN KEY (reporter_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT report_ibfk_3 FOREIGN KEY (room_id) REFERENCES room (id) ON DELETE CASCADE ON UPDATE CASCADE
);

ALTER TABLE report ALTER COLUMN date SET DEFAULT now();

COMMENT ON TABLE report IS 'Player reports';

COMMENT ON COLUMN report.id IS 'Unique report ID';
COMMENT ON COLUMN report.penguin_id IS 'Reported penguin ID';
COMMENT ON COLUMN report.reporter_id IS 'Reporting penguin ID';
COMMENT ON COLUMN report.report_type IS 'Report type ID';
COMMENT ON COLUMN report.date IS 'Date of report';

DROP TABLE IF EXISTS ban;
CREATE TABLE ban (
  penguin_id INT NOT NULL,
  issued TIMESTAMP NOT NULL,
  expires TIMESTAMP NOT NULL,
  moderator_id INT DEFAULT NULL,
  reason SMALLINT NOT NULL,
  comment TEXT DEFAULT NULL,
  message TEXT DEFAULT NULL,
  PRIMARY KEY (penguin_id, issued, expires),
  CONSTRAINT ban_ibfk_1 FOREIGN KEY (penguin_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT ban_ibfk_2 FOREIGN KEY (moderator_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE
);

CREATE INDEX ban_moderator_id ON ban (moderator_id);

ALTER TABLE ban ALTER COLUMN issued SET DEFAULT now();
ALTER TABLE ban ALTER COLUMN expires SET DEFAULT now();

COMMENT ON TABLE ban IS 'Penguin ban records';

COMMENT ON COLUMN ban.penguin_id IS 'Banned penguin ID';
COMMENT ON COLUMN ban.issued IS 'Issue date';
COMMENT ON COLUMN ban.expires IS 'Expiry date';
COMMENT ON COLUMN ban.moderator_id IS 'Moderator penguin ID';
COMMENT ON COLUMN ban.reason IS 'Ban reason';
COMMENT ON COLUMN ban.comment IS 'Ban comment';
COMMENT ON COLUMN ban.message IS 'Banned for message';

DROP TABLE IF EXISTS warning;
CREATE TABLE warning (
  penguin_id INT NOT NULL,
  issued TIMESTAMP NOT NULL,
  expires TIMESTAMP NOT NULL,
  PRIMARY KEY (penguin_id, issued, expires),
  CONSTRAINT warning_ibfk_1 FOREIGN KEY (penguin_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE warning IS 'Penguin moderator warnings';

COMMENT ON COLUMN warning.penguin_id IS 'Warning penguin ID';
COMMENT ON COLUMN warning.issued IS 'Warning issue date';
COMMENT ON COLUMN warning.expires IS 'Warning expiry date';

DROP TABLE IF EXISTS buddy_list;
CREATE TABLE buddy_list (
  penguin_id INT NOT NULL,
  buddy_id INT NOT NULL,
  best_buddy BOOLEAN NOT NULL DEFAULT FALSE,
  PRIMARY KEY (penguin_id,buddy_id),
  CONSTRAINT buddy_list_ibfk_1 FOREIGN KEY (penguin_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT buddy_list_ibfk_2 FOREIGN KEY (buddy_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE
);

CREATE INDEX buddy_id ON buddy_list (buddy_id);

COMMENT ON TABLE buddy_list IS 'Penguin buddy relationships';

DROP TABLE IF EXISTS buddy_request;
CREATE TABLE buddy_request (
  penguin_id INT NOT NULL,
  requester_id INT NOT NULL,
  PRIMARY KEY (penguin_id, requester_id),
  CONSTRAINT buddy_request_ibfk_1 FOREIGN KEY (penguin_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT buddy_request_ibfk_2 FOREIGN KEY (requester_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE buddy_request IS 'Penguin buddy requests';

DROP TABLE IF EXISTS character_buddy;
CREATE TABLE character_buddy (
  penguin_id INT NOT NULL,
  character_id SMALLINT NOT NULL,
  best_buddy BOOLEAN NOT NULL DEFAULT FALSE,
  PRIMARY KEY (penguin_id, character_id),
  CONSTRAINT character_buddy_ibfk_1 FOREIGN KEY (penguin_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT character_buddy_ibfk_2 FOREIGN KEY (character_id) REFERENCES character (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE character_buddy IS 'Penguin character buddies';

DROP TABLE IF EXISTS cover_stamp;
CREATE TABLE cover_stamp (
  penguin_id INT NOT NULL,
  stamp_id INT NOT NULL,
  x SMALLINT NOT NULL DEFAULT 0,
  y SMALLINT NOT NULL DEFAULT 0,
  rotation SMALLINT NOT NULL DEFAULT 0,
  depth SMALLINT NOT NULL DEFAULT 0,
  PRIMARY KEY (penguin_id, stamp_id),
  CONSTRAINT cover_stamp_ibfk_1 FOREIGN KEY (penguin_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT cover_stamp_ibfk_2 FOREIGN KEY (stamp_id) REFERENCES stamp (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE cover_stamp IS 'Stamps placed on book cover';

COMMENT ON COLUMN cover_stamp.penguin_id IS 'Unique penguin ID';
COMMENT ON COLUMN cover_stamp.stamp_id IS 'Cover stamp ID';
COMMENT ON COLUMN cover_stamp.x IS 'Cover X position';
COMMENT ON COLUMN cover_stamp.y IS 'Cover Y position';
COMMENT ON COLUMN cover_stamp.rotation IS 'Stamp cover rotation';
COMMENT ON COLUMN cover_stamp.depth IS 'Stamp cover depth';

DROP TABLE IF EXISTS cover_item;
CREATE TABLE cover_item (
  penguin_id INT NOT NULL,
  item_id INT NOT NULL,
  x SMALLINT NOT NULL DEFAULT 0,
  y SMALLINT NOT NULL DEFAULT 0,
  rotation SMALLINT NOT NULL DEFAULT 0,
  depth SMALLINT NOT NULL DEFAULT 0,
  PRIMARY KEY (penguin_id, item_id),
  CONSTRAINT cover_item_ibfk_1 FOREIGN KEY (penguin_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT cover_item_ibfk_2 FOREIGN KEY (item_id) REFERENCES item (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE cover_item IS 'Items placed on book cover';

COMMENT ON COLUMN cover_item.penguin_id IS 'Unique penguin ID';
COMMENT ON COLUMN cover_item.item_id IS 'Cover item ID';
COMMENT ON COLUMN cover_item.x IS 'Cover X position';
COMMENT ON COLUMN cover_item.y IS 'Cover Y position';
COMMENT ON COLUMN cover_item.rotation IS 'Stamp cover rotation';
COMMENT ON COLUMN cover_item.depth IS 'Stamp cover depth';

DROP TABLE IF EXISTS penguin_card;
CREATE TABLE penguin_card (
  penguin_id INT NOT NULL,
  card_id INT NOT NULL,
  quantity SMALLINT NOT NULL DEFAULT 1,
  member_quantity SMALLINT NOT NULL DEFAULT 0,
  PRIMARY KEY (penguin_id, card_id),
  CONSTRAINT penguin_card_ibfk_1 FOREIGN KEY (penguin_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT penguin_card_ibfk_2 FOREIGN KEY (card_id) REFERENCES card (id) ON DELETE CASCADE ON UPDATE CASCADE
);

CREATE INDEX penguin_card_penguin_id ON penguin_card(penguin_id);

COMMENT ON TABLE penguin_card IS 'Penguin Card Jitsu decks';

COMMENT ON COLUMN penguin_card.penguin_id IS 'Owner penguin ID';
COMMENT ON COLUMN penguin_card.card_id IS 'Card type ID';
COMMENT ON COLUMN penguin_card.quantity IS 'Quantity owned';
COMMENT ON COLUMN penguin_card.member_quantity IS 'Quantity owned as member';

DROP TABLE IF EXISTS penguin_furniture;
CREATE TABLE penguin_furniture (
  penguin_id INT NOT NULL,
  furniture_id SMALLINT NOT NULL,
  quantity SMALLINT NOT NULL DEFAULT 1,
  PRIMARY KEY (penguin_id, furniture_id),
  CONSTRAINT penguin_furniture_ibfk_1 FOREIGN KEY (penguin_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT penguin_furniture_ibfk_2 FOREIGN KEY (furniture_id) REFERENCES furniture (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE penguin_furniture IS 'Penguin owned furniture';

COMMENT ON COLUMN penguin_furniture.penguin_id IS 'Owner penguin ID';
COMMENT ON COLUMN penguin_furniture.furniture_id IS 'Furniture item ID';
COMMENT ON COLUMN penguin_furniture.quantity IS 'Quantity owned';

DROP TABLE IF EXISTS penguin_flooring;
CREATE TABLE penguin_flooring (
  penguin_id INT NOT NULL,
  flooring_id SMALLINT NOT NULL,
  PRIMARY KEY(penguin_id, flooring_id),
  CONSTRAINT penguin_flooring_ibfk_1 FOREIGN KEY (penguin_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT penugin_flooring_ibfk_2 FOREIGN KEY (flooring_id) REFERENCES flooring (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE penguin_flooring IS 'Penguin owned furniture';

COMMENT ON COLUMN penguin_flooring.penguin_id IS 'Owner penguin ID';
COMMENT ON COLUMN penguin_flooring.flooring_id IS 'Flooring item ID';

DROP TABLE IF EXISTS penguin_igloo_room;
CREATE TABLE penguin_igloo_room (
  id SERIAL,
  penguin_id INT NOT NULL,
  type INT NOT NULL,
  flooring INT NOT NULL,
  music SMALLINT NOT NULL DEFAULT 0,
  location INT NOT NULL,
  locked BOOLEAN NOT NULL DEFAULT TRUE,
  competition BOOLEAN NOT NULL DEFAULT FALSE,
  PRIMARY KEY (id),
  CONSTRAINT igloo_room_ibfk_1 FOREIGN KEY (penguin_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT igloo_room_ibfk_2 FOREIGN KEY (type) REFERENCES igloo (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT igloo_room_ibfk_3 FOREIGN KEY (flooring) REFERENCES flooring (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT igloo_room_ibfk_4 FOREIGN KEY (location) REFERENCES location (id) ON DELETE CASCADE ON UPDATE CASCADE
);

ALTER TABLE penguin ADD CONSTRAINT penguin_ibfk_11 FOREIGN KEY (igloo) REFERENCES penguin_igloo_room (id) ON DELETE CASCADE ON UPDATE CASCADE;

COMMENT ON TABLE penguin_igloo_room IS 'Penguin igloo settings';

COMMENT ON COLUMN penguin_igloo_room.id IS 'Unique igloo ID';
COMMENT ON COLUMN penguin_igloo_room.penguin_id IS 'Owner penguin ID';
COMMENT ON COLUMN penguin_igloo_room.type IS 'Igloo type ID';
COMMENT ON COLUMN penguin_igloo_room.flooring IS 'Igloo flooring ID';
COMMENT ON COLUMN penguin_igloo_room.music IS 'Igloo music ID';
COMMENT ON COLUMN penguin_igloo_room.locked IS 'Is igloo locked?';
COMMENT ON COLUMN penguin_igloo_room.competition IS 'Is entered in competition?';

DROP TABLE IF EXISTS igloo_like;
CREATE TABLE igloo_like (
  igloo_id INT NOT NULL,
  player_id INT NOT NULL,
  count SMALLINT NOT NULL DEFAULT 1,
  date TIMESTAMP NOT NULL,
  PRIMARY KEY (igloo_id, player_id),
  CONSTRAINT igloo_like_ibfk_1 FOREIGN KEY (igloo_id) REFERENCES penguin_igloo_room (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT igloo_like_ibfk_2 FOREIGN KEY (player_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE
);

ALTER TABLE igloo_like ALTER COLUMN date SET DEFAULT now();

COMMENT ON TABLE igloo_like IS 'Player igloo likes';

COMMENT ON COLUMN igloo_like.igloo_id IS 'Igloo unique ID';
COMMENT ON COLUMN igloo_like.player_id IS 'Liker unique ID';
COMMENT ON COLUMN igloo_like.count IS 'Number of likes';
COMMENT ON COLUMN igloo_like.date IS 'Date of like';


DROP TABLE IF EXISTS penguin_location;
CREATE TABLE penguin_location (
  penguin_id INT NOT NULL,
  location_id INT NOT NULL,
  PRIMARY KEY (penguin_id, location_id),
  CONSTRAINT penguin_location_ibfk_1 FOREIGN KEY (penguin_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT penguin_location_ibfk_2 FOREIGN KEY (location_id) REFERENCES location (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE penguin_location IS 'Penguin owned locations';

COMMENT ON COLUMN penguin_location.penguin_id IS 'Owner penguin ID';
COMMENT ON COLUMN penguin_location.location_id IS 'Location ID';

DROP TABLE IF EXISTS igloo_furniture;
CREATE TABLE igloo_furniture (
  igloo_id INT NOT NULL,
  furniture_id INT NOT NULL,
  x SMALLINT NOT NULL DEFAULT 0,
  y SMALLINT NOT NULL DEFAULT 0,
  frame SMALLINT NOT NULL DEFAULT 0,
  rotation SMALLINT NOT NULL DEFAULT 0,
  PRIMARY KEY (igloo_id, furniture_id, x, y, frame, rotation),
  CONSTRAINT igloo_furniture_ibfk_1 FOREIGN KEY (igloo_id) REFERENCES penguin_igloo_room (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT igloo_furniture_ibfk_2 FOREIGN KEY (furniture_id) REFERENCES furniture (id) ON DELETE CASCADE ON UPDATE CASCADE
);

CREATE INDEX igloo_furniture_igloo_id ON igloo_furniture(igloo_id);

COMMENT ON TABLE igloo_furniture IS 'Furniture placed inside igloos';

COMMENT ON COLUMN igloo_furniture.igloo_id IS 'Furniture igloo ID';
COMMENT ON COLUMN igloo_furniture.furniture_id IS 'Furniture item ID';
COMMENT ON COLUMN igloo_furniture.x IS 'Igloo X position';
COMMENT ON COLUMN igloo_furniture.y IS 'Igloo Y position';
COMMENT ON COLUMN igloo_furniture.frame IS 'Furniture frame ID';
COMMENT ON COLUMN igloo_furniture.rotation IS 'Furniture rotation ID';

DROP TABLE IF EXISTS penguin_igloo;
CREATE TABLE penguin_igloo (
  penguin_id INT NOT NULL,
  igloo_id INT NOT NULL,
  PRIMARY KEY (penguin_id, igloo_id),
  CONSTRAINT penguin_igloo_ibfk_1 FOREIGN KEY (penguin_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT penguin_igloo_ibfk_2 FOREIGN KEY (igloo_id) REFERENCES igloo (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE penguin_igloo IS 'Penguin owned igloos';

COMMENT ON COLUMN penguin_igloo.penguin_id IS 'Owner penguin ID';
COMMENT ON COLUMN penguin_igloo.igloo_id IS 'Igloo ID';

DROP TABLE IF EXISTS penguin_track;
CREATE TABLE penguin_track (
  id SERIAL NOT NULL,
  name VARCHAR(12) NOT NULL DEFAULT '',
  owner_id INT NOT NULL,
  sharing BOOLEAN NOT NULL DEFAULT FALSE,
  pattern TEXT NOT NULL,
  PRIMARY KEY (id),
  CONSTRAINT penguin_track_ibfk_1 FOREIGN KEY (owner_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE penguin_track IS 'Penguin SoundStudio tracks';

COMMENT ON COLUMN penguin_track.id IS 'Unique track ID';
COMMENT ON COLUMN penguin_track.name IS 'Name of track';
COMMENT ON COLUMN penguin_track.owner_id IS 'Owner penguin ID';
COMMENT ON COLUMN penguin_track.sharing IS 'Is song shared?';
COMMENT ON COLUMN penguin_track.pattern IS 'Song data';

DROP TABLE IF EXISTS track_like;
CREATE TABLE track_like (
  penguin_id INT NOT NULL,
  track_id INT NOT NULL,
  date TIMESTAMP NOT NULL,
  PRIMARY KEY (track_id, penguin_id, date),
  CONSTRAINT track_like_ibfk_1 FOREIGN KEY (penguin_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT track_like_ibfk_2 FOREIGN KEY (track_id) REFERENCES penguin_track (id) ON DELETE CASCADE ON UPDATE CASCADE
);

CREATE INDEX track_like_track_id ON track_like (track_id);

ALTER TABLE track_like ALTER COLUMN date SET DEFAULT now();

COMMENT ON TABLE track_like IS 'SoundStudio likes';

COMMENT ON COLUMN track_like.track_id IS 'Liked track ID';
COMMENT ON COLUMN track_like.penguin_id IS 'Liker penguin ID';
COMMENT ON COLUMN track_like.date IS 'Timestamp of like';

DROP TABLE IF EXISTS penguin_game_data;
CREATE TABLE penguin_game_data (
  penguin_id INT NOT NULL,
  room_id INT NOT NULL,
  index INT DEFAULT NULL,
  data TEXT DEFAULT '',
  PRIMARY KEY (penguin_id, room_id, index),
  CONSTRAINT penguin_game_data_ibfk_1 FOREIGN KEY (penguin_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT penguin_game_data_ibfk_2 FOREIGN KEY (room_id) REFERENCES room (id) ON DELETE CASCADE ON UPDATE CASCADE
);

DROP TABLE IF EXISTS ignore_list;
CREATE TABLE ignore_list (
  penguin_id INT NOT NULL,
  ignore_id INT NOT NULL,
  PRIMARY KEY (penguin_id, ignore_id),
  CONSTRAINT ignore_list_ibfk_1 FOREIGN KEY (penguin_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT ignore_list_ibfk_2 FOREIGN KEY (ignore_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE
);

CREATE INDEX ignore_list_ignore_id ON ignore_list (ignore_id);

COMMENT ON TABLE ignore_list IS 'Penguin ignore relationships';

DROP TABLE IF EXISTS penguin_item;
CREATE TABLE penguin_item (
  penguin_id INT NOT NULL,
  item_id SMALLINT NOT NULL,
  PRIMARY KEY (penguin_id, item_id),
  CONSTRAINT penguin_item_ibfk_1 FOREIGN KEY (penguin_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT penguin_item_ibfk_2 FOREIGN KEY (item_id) REFERENCES item (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE penguin_item IS 'Penguin owned clothing items';

COMMENT ON COLUMN penguin_item.penguin_id IS 'Owner penguin ID';
COMMENT ON COLUMN penguin_item.item_id IS 'Clothing item ID';

DROP TABLE IF EXISTS login;
CREATE TABLE login (
  id SERIAL,
  penguin_id INT NOT NULL,
  date TIMESTAMP NOT NULL,
  ip_hash CHAR(128) NOT NULL,
  minutes_played INT NOT NULL DEFAULT 0,
  PRIMARY KEY (id),
  CONSTRAINT login_ibfk_1 FOREIGN KEY (penguin_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE
);

ALTER TABLE login ALTER COLUMN date SET DEFAULT now();

COMMENT ON TABLE login IS 'Penguin login records';

COMMENT ON COLUMN login.id IS 'Unique login ID';
COMMENT ON COLUMN login.penguin_id IS 'Login penguin ID';
COMMENT ON COLUMN login.date IS 'Login date';
COMMENT ON COLUMN login.ip_hash IS 'Connection IP address';
COMMENT ON COLUMN login.minutes_played IS 'Minutes played for session';

DROP TABLE IF EXISTS cfc_donation;
CREATE TABLE cfc_donation (
  penguin_id INT NOT NULL,
  coins INT NOT NULL,
  charity SMALLINT NOT NULL,
  date TIMESTAMP NOT NULL,
  PRIMARY KEY (penguin_id, charity, date),
  CONSTRAINT cfc_donation_ibfk_1 FOREIGN KEY (penguin_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE
);

ALTER TABLE cfc_donation ALTER COLUMN date SET DEFAULT now();

COMMENT ON TABLE cfc_donation IS 'CFC charity donations';

COMMENT ON COLUMN cfc_donation.penguin_id IS 'Donator penguin ID';
COMMENT ON COLUMN cfc_donation.coins IS 'Donation coin amount';
COMMENT ON COLUMN cfc_donation.charity IS 'Donation charity or cause';
COMMENT ON COLUMN cfc_donation.date IS 'Date of donation';

DROP TABLE IF EXISTS penguin_postcard;
CREATE TABLE penguin_postcard (
  id SERIAL,
  penguin_id INT NOT NULL,
  sender_id INT DEFAULT NULL,
  postcard_id SMALLINT NOT NULL,
  send_date TIMESTAMP NOT NULL,
  details VARCHAR(255) NOT NULL DEFAULT '',
  has_read BOOLEAN NOT NULL DEFAULT FALSE,
  PRIMARY KEY (id),
  CONSTRAINT penguin_postcard_ibfk_1 FOREIGN KEY (penguin_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT penguin_postcard_ibfk_2 FOREIGN KEY (sender_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT penguin_postcard_ibfk_3 FOREIGN KEY (postcard_id) REFERENCES postcard (id) ON DELETE CASCADE ON UPDATE CASCADE
);

ALTER TABLE penguin_postcard ALTER COLUMN send_date SET DEFAULT now();

CREATE INDEX penguin_postcard_penguin_id ON penguin_postcard(penguin_id);
CREATE INDEX penguin_postcard_sender_id ON penguin_postcard(sender_id);

COMMENT ON TABLE penguin_postcard IS 'Sent postcards';

COMMENT ON COLUMN penguin_postcard.id IS 'Unique postcard ID';
COMMENT ON COLUMN penguin_postcard.penguin_id IS 'Postcard type ID';
COMMENT ON COLUMN penguin_postcard.sender_id IS 'Sender penguin ID';
COMMENT ON COLUMN penguin_postcard.postcard_id IS 'Postcard type ID';
COMMENT ON COLUMN penguin_postcard.send_date IS 'Postcard type ID';
COMMENT ON COLUMN penguin_postcard.details IS 'Postcard details';
COMMENT ON COLUMN penguin_postcard.has_read IS 'Is read?';

DROP TABLE IF EXISTS penguin_puffle;
CREATE TABLE penguin_puffle (
  id SERIAL,
  penguin_id INT NOT NULL,
  puffle_id INT NOT NULL,
  name VARCHAR(16) NOT NULL,
  adoption_date TIMESTAMP NOT NULL,
  food SMALLINT NOT NULL DEFAULT 100,
  play SMALLINT NOT NULL DEFAULT 100,
  rest SMALLINT NOT NULL DEFAULT 100,
  clean SMALLINT NOT NULL DEFAULT 100,
  hat INT DEFAULT NULL,
  backyard BOOLEAN DEFAULT FALSE,
  has_dug BOOLEAN DEFAULT FALSE,
  PRIMARY KEY (id),
  CONSTRAINT penguin_puffle_ibfk_1 FOREIGN KEY (penguin_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT penguin_puffle_ibfk_2 FOREIGN KEY (puffle_id) REFERENCES puffle (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT penguin_puffle_ibfk_3 FOREIGN KEY (hat) REFERENCES puffle_item (id) ON DELETE CASCADE ON UPDATE CASCADE
);

ALTER TABLE penguin_puffle ALTER COLUMN adoption_date SET DEFAULT now();

COMMENT ON TABLE penguin_puffle IS 'Adopted puffles';

COMMENT ON COLUMN penguin_puffle.id IS 'Unique puffle ID';
COMMENT ON COLUMN penguin_puffle.penguin_id IS 'Owner penguin ID';
COMMENT ON COLUMN penguin_puffle.puffle_id IS 'Puffle type ID';
COMMENT ON COLUMN penguin_puffle.name IS 'Puffle name';
COMMENT ON COLUMN penguin_puffle.adoption_date IS 'Date of adoption';
COMMENT ON COLUMN penguin_puffle.food IS 'Puffle health %';
COMMENT ON COLUMN penguin_puffle.play IS 'Puffle hunger %';
COMMENT ON COLUMN penguin_puffle.rest IS 'Puffle rest %';
COMMENT ON COLUMN penguin_puffle.clean IS 'Puffle clean %';
COMMENT ON COLUMN penguin_puffle.hat IS 'Puffle hat item ID';
COMMENT ON COLUMN penguin_puffle.backyard IS 'Is in backyard?';
COMMENT ON COLUMN penguin_puffle.has_dug IS 'Has dug?';

ALTER TABLE penguin ADD CONSTRAINT penguin_ibfk_12 FOREIGN KEY (walking) REFERENCES penguin_puffle (id) ON DELETE CASCADE ON UPDATE CASCADE;

DROP TABLE IF EXISTS penguin_puffle_item;
CREATE TABLE penguin_puffle_item (
  penguin_id INT NOT NULL,
  item_id INT NOT NULL,
  quantity SMALLINT NOT NULL DEFAULT 1,
  PRIMARY KEY (penguin_id, item_id),
  CONSTRAINT penguin_puffle_item_ibfk_1 FOREIGN KEY (penguin_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT penguin_puffle_item_ibfk_2 FOREIGN KEY (item_id) REFERENCES puffle_item (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE penguin_puffle_item IS 'Owned puffle care items';

COMMENT ON COLUMN penguin_puffle_item.penguin_id IS 'Owner penguin ID';
COMMENT ON COLUMN penguin_puffle_item.item_id IS 'Puffle care item ID';
COMMENT ON COLUMN penguin_puffle_item.quantity IS 'Quantity owned';

DROP TABLE IF EXISTS redemption_code;
CREATE TABLE redemption_code (
  id SERIAL,
  code VARCHAR(16) NOT NULL,
  type VARCHAR(8) NOT NULL DEFAULT 'BLANKET',
  coins INT NOT NULL DEFAULT 0,
  expires TIMESTAMP DEFAULT NULL,
  uses INT DEFAULT NULL,
  PRIMARY KEY (id)
);

CREATE UNIQUE INDEX redemption_code_code ON redemption_code (code);

COMMENT ON TABLE redemption_code IS 'Redemption codes';

COMMENT ON COLUMN redemption_code.id IS 'Unique code ID';
COMMENT ON COLUMN redemption_code.code IS 'Redemption code';
COMMENT ON COLUMN redemption_code.type IS 'Code type';
COMMENT ON COLUMN redemption_code.coins IS 'Code coins amount';
COMMENT ON COLUMN redemption_code.expires IS 'Expiry date';
COMMENT ON COLUMN redemption_code.uses IS 'Number of uses';

DROP TABLE IF EXISTS redemption_book;
CREATE TABLE redemption_book (
  id INT NOT NULL,
  name VARCHAR(255) NOT NULL,
  PRIMARY KEY (id)
);

COMMENT ON TABLE redemption_book IS 'Redemption books';

COMMENT ON COLUMN redemption_book.id IS 'Unique redemption book ID';
COMMENT ON COLUMN redemption_book.name IS 'Book name';

DROP TABLE IF EXISTS redemption_book_word;
CREATE TABLE redemption_book_word (
  book_id INT NOT NULL,
  page SMALLINT NOT NULL DEFAULT 1,
  line SMALLINT NOT NULL DEFAULT 1,
  word_number SMALLINT NOT NULL DEFAULT 1,
  answer VARCHAR(20) NOT NULL,
  PRIMARY KEY(book_id, page, line, word_number),
  CONSTRAINT redemption_book_word_ibfk_1 FOREIGN KEY (book_id) REFERENCES redemption_book (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE redemption_book_word IS 'Redemption book answers';

COMMENT ON COLUMN redemption_book_word.book_id IS 'Redemption book ID';
COMMENT ON COLUMN redemption_book_word.page IS 'Page number inside book';
COMMENT ON COLUMN redemption_book_word.line IS 'Line number of page';
COMMENT ON COLUMN redemption_book_word.word_number IS 'The nth word on the line';
COMMENT ON COLUMN redemption_book_word.answer IS 'The correct word';

DROP TABLE IF EXISTS penguin_redemption_code;
CREATE TABLE penguin_redemption_code (
  penguin_id INT NOT NULL,
  code_id INT NOT NULL,
  PRIMARY KEY (penguin_id, code_id),
  CONSTRAINT penguin_redemption_code_ibfk_1 FOREIGN KEY (penguin_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT penguin_redemption_code_ibfk_2 FOREIGN KEY (code_id) REFERENCES redemption_code (id) ON DELETE CASCADE ON UPDATE CASCADE
);

CREATE INDEX penguin_redemption_code_code_id ON penguin_redemption_code (code_id);

COMMENT ON TABLE penguin_redemption_code IS 'Redeemed codes';

COMMENT ON COLUMN penguin_redemption_code.penguin_id IS 'Unique penguin ID';
COMMENT ON COLUMN penguin_redemption_code.code_id IS 'Unique code ID';


DROP TABLE IF EXISTS penguin_redemption_book;
CREATE TABLE penguin_redemption_book (
  penguin_id INT NOT NULL,
  book_id INT NOT NULL,
  PRIMARY KEY (penguin_id, book_id),
  CONSTRAINT penguin_redemption_book_ibfk_1 FOREIGN KEY (penguin_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT penguin_redemption_book_ibfk_2 FOREIGN KEY (book_id) REFERENCES redemption_book (id) ON DELETE CASCADE ON UPDATE CASCADE
);

CREATE INDEX penguin_redemption_book_book_id ON penguin_redemption_book (book_id);

COMMENT ON TABLE penguin_redemption_book IS 'Redeemed book codes';

COMMENT ON COLUMN penguin_redemption_book.penguin_id IS 'Unique penguin ID';
COMMENT ON COLUMN penguin_redemption_book.book_id IS 'Unique book ID';


DROP TABLE IF EXISTS redemption_award_card;
CREATE TABLE redemption_award_card (
  code_id INT NOT NULL,
  card_id INT DEFAULT NULL,
  PRIMARY KEY (code_id, card_id),
  CONSTRAINT redemption_award_ibfk_1 FOREIGN KEY (code_id) REFERENCES redemption_code (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT redemption_award_ibfk_2 FOREIGN KEY (card_id) REFERENCES card (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE redemption_award_card IS 'Redemption code card awards';

COMMENT ON COLUMN redemption_award_card.code_id IS 'Unique code ID';
COMMENT ON COLUMN redemption_award_card.card_id IS 'Code card ID';

DROP TABLE IF EXISTS redemption_award_item;
CREATE TABLE redemption_award_item (
  code_id INT NOT NULL,
  item_id INT DEFAULT NULL,
  PRIMARY KEY (code_id, item_id),
  CONSTRAINT redemption_award_item_ibfk_1 FOREIGN KEY (code_id) REFERENCES redemption_code (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT redemption_award_item_ibfk_2 FOREIGN KEY (item_id) REFERENCES item (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE redemption_award_item IS 'Redemption code item awards';

COMMENT ON COLUMN redemption_award_item.code_id IS 'Unique code ID';
COMMENT ON COLUMN redemption_award_item.item_id IS 'Code item ID';

DROP TABLE IF EXISTS redemption_award_furniture;
CREATE TABLE redemption_award_furniture (
  code_id INT NOT NULL,
  furniture_id INT DEFAULT NULL,
  PRIMARY KEY (code_id, furniture_id),
  CONSTRAINT redemption_award_furniture_ibfk_1 FOREIGN KEY (code_id) REFERENCES redemption_code (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT redemption_award_furniture_ibfk_2 FOREIGN KEY (furniture_id) REFERENCES furniture (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE redemption_award_furniture IS 'Redemption code furniture awards';

COMMENT ON COLUMN redemption_award_furniture.code_id IS 'Unique code ID';
COMMENT ON COLUMN redemption_award_furniture.furniture_id IS 'Code igloo furniture ID';

DROP TABLE IF EXISTS redemption_award_igloo;
CREATE TABLE redemption_award_igloo (
  code_id INT NOT NULL,
  igloo_id INT DEFAULT NULL,
  PRIMARY KEY (code_id, igloo_id),
  CONSTRAINT redemption_award_igloo_ibfk_1 FOREIGN KEY (code_id) REFERENCES redemption_code (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT redemption_award_igloo_ibfk_2 FOREIGN KEY (igloo_id) REFERENCES igloo (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE redemption_award_igloo IS 'Redemption code igloo awards';

COMMENT ON COLUMN redemption_award_igloo.code_id IS 'Unique code ID';
COMMENT ON COLUMN redemption_award_igloo.igloo_id IS 'Code igloo ID';

DROP TABLE IF EXISTS redemption_award_location;
CREATE TABLE redemption_award_location (
  code_id INT NOT NULL,
  location_id INT DEFAULT NULL,
  PRIMARY KEY (code_id, location_id),
  CONSTRAINT redemption_award_location_ibfk_1 FOREIGN KEY (code_id) REFERENCES redemption_code (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT redemption_award_location_ibfk_2 FOREIGN KEY (location_id) REFERENCES location (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE redemption_award_location IS 'Redemption code location awards';

COMMENT ON COLUMN redemption_award_location.code_id IS 'Unique code ID';
COMMENT ON COLUMN redemption_award_location.location_id IS 'Code igloo location ID';

DROP TABLE IF EXISTS redemption_award_flooring;
CREATE TABLE redemption_award_flooring (
  code_id INT NOT NULL,
  flooring_id INT DEFAULT NULL,
  PRIMARY KEY (code_id, flooring_id),
  CONSTRAINT redemption_award_flooring_ibfk_1 FOREIGN KEY (code_id) REFERENCES redemption_code (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT redemption_award_flooring_ibfk_2 FOREIGN KEY (flooring_id) REFERENCES flooring (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE redemption_award_flooring IS 'Redemption code flooring awards';

COMMENT ON COLUMN redemption_award_flooring.code_id IS 'Unique code ID';
COMMENT ON COLUMN redemption_award_flooring.flooring_id IS 'Code igloo flooring ID';

DROP TABLE IF EXISTS redemption_award_puffle;
CREATE TABLE redemption_award_puffle (
  code_id INT NOT NULL,
  puffle_id INT DEFAULT NULL,
  PRIMARY KEY (code_id, puffle_id),
  CONSTRAINT redemption_award_puffle_ibfk_1 FOREIGN KEY (code_id) REFERENCES redemption_code (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT redemption_award_puffle_ibfk_2 FOREIGN KEY (puffle_id) REFERENCES puffle (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE redemption_award_puffle IS 'Redemption code puffle awards';

COMMENT ON COLUMN redemption_award_puffle.code_id IS 'Unique code ID';
COMMENT ON COLUMN redemption_award_puffle.puffle_id IS 'Code puffle ID';

DROP TABLE IF EXISTS redemption_award_puffle_item;
CREATE TABLE redemption_award_puffle_item (
  code_id INT NOT NULL,
  puffle_item_id INT DEFAULT NULL,
  PRIMARY KEY (code_id, puffle_item_id),
  CONSTRAINT redemption_award_puffle_item_ibfk_1 FOREIGN KEY (code_id) REFERENCES redemption_code (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT redemption_award_puffle_item_ibfk_2 FOREIGN KEY (puffle_item_id) REFERENCES puffle_item (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE redemption_award_puffle_item IS 'Redemption code puffle care item awards';

COMMENT ON COLUMN redemption_award_puffle_item.code_id IS 'Unique code ID';
COMMENT ON COLUMN redemption_award_puffle_item.puffle_item_id IS 'Code puffle care item ID';

DROP TABLE IF EXISTS penguin_stamp;
CREATE TABLE penguin_stamp (
  penguin_id INT NOT NULL,
  stamp_id INT NOT NULL,
  recent BOOLEAN NOT NULL DEFAULT TRUE,
  PRIMARY KEY (penguin_id, stamp_id),
  CONSTRAINT stamp_ibfk_1 FOREIGN KEY (penguin_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT stamp_ibfk_2 FOREIGN KEY (stamp_id) REFERENCES stamp (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE penguin_stamp IS 'Penguin earned stamps';

COMMENT ON COLUMN penguin_stamp.penguin_id IS 'Stamp penguin ID';
COMMENT ON COLUMN penguin_stamp.stamp_id IS 'Stamp ID';
COMMENT ON COLUMN penguin_stamp.recent IS 'Is recently earned?';
COMMENT ON COLUMN penguin_stamp.recent IS 'Is recently earned?';

DROP TABLE IF EXISTS penguin_membership;
CREATE TABLE penguin_membership (
  penguin_id INT NOT NULL,
  start TIMESTAMP NOT NULL, 
  expires TIMESTAMP DEFAULT NULL,
  start_aware BOOLEAN DEFAULT FALSE,
  expires_aware BOOLEAN DEFAULT FALSE,
  expired_aware BOOLEAN DEFAULT FALSE,
  PRIMARY KEY(penguin_id, start),
  CONSTRAINT penguin_membership_ibfk_1 FOREIGN KEY (penguin_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE penguin_membership IS 'Penguin membership records';

COMMENT ON COLUMN penguin_membership.penguin_id IS 'Penguin ID of membership';
COMMENT ON COLUMN penguin_membership.start IS 'Start time of membership';
COMMENT ON COLUMN penguin_membership.expires IS 'End time of membership';

DROP TABLE IF EXISTS penguin_quest_task;
CREATE TABLE penguin_quest_task (
  task_id INT NOT NULL,
  penguin_id INT NOT NULL,
  complete BOOLEAN NOT NULL DEFAULT FALSE,
  PRIMARY KEY (task_id, penguin_id),
  CONSTRAINT penguin_quest_task_ibfk_1 FOREIGN KEY (task_id) REFERENCES quest_task (id) ON DELETE CASCADE ON UPDATE CASCADE,
  CONSTRAINT penguin_quest_task_ibfk_2 FOREIGN KEY (penguin_id) REFERENCES penguin (id) ON DELETE CASCADE ON UPDATE CASCADE
);

COMMENT ON TABLE penguin_quest_task IS 'Completed quest tasks';

COMMENT ON COLUMN penguin_quest_task.task_id IS 'Completed task ID';
COMMENT ON COLUMN penguin_quest_task.penguin_id IS 'Task penguin ID';

INSERT INTO item (id, name, type, cost, member, bait, patched, legacy_inventory, vanilla_inventory, epf, tour, release_date) VALUES
 (1, 'Blue', 1, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2, 'Green', 1, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3, 'Pink', 1, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4, 'Black', 1, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5, 'Red', 1, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6, 'Orange', 1, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (7, 'Yellow', 1, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (8, 'Dark Purple', 1, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9, 'Brown', 1, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10, 'Peach', 1, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11, 'Dark Green', 1, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (12, 'Light Blue', 1, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (13, 'Lime Green', 1, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14, 'Sensei Gray', 1, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15, 'Aqua', 1, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16, 'Arctic White', 1, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (17, 'Dot Lavender', 1, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (101, 'Black Sunglasses', 3, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (102, 'Dark Vision Goggles', 3, 10, TRUE, FALSE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (103, '3D Glasses', 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (104, 'Ninja Mask', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (105, 'Eyepatch', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (106, 'Mask', 3, 100, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (107, 'Blue Sunglasses', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (108, 'Blue Sunglasses', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (109, 'Free Red Sunglasses', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (110, 'Red Sunglasses', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (111, 'Designer Glasses', 3, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (112, 'Funny-Face Glasses', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (113, 'Black Glasses', 3, 225, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (114, 'Brown Glasses', 3, 225, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (116, 'Snowflake Mask', 3, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (117, 'Diva Sunglasses', 3, 225, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (118, 'White Diva Sunglasses', 3, 225, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (119, 'Monocle', 3, 160, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (120, 'Alien Mask', 3, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (121, 'Green Sunglasses', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (122, 'Robot Helmet', 3, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (123, 'Blue Superhero Mask', 3, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (124, 'Pink Superhero Mask', 3, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (125, 'Aviator Sunglasses', 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (126, 'Big Brow', 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (127, 'Blue Tiki Mask', 3, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (128, 'Pink Tiki Mask', 3, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (130, 'Earring', 3, 75, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (131, 'Green Snorkel', 3, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (132, 'Blue Snorkel', 3, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (133, 'Red Face Paint', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (134, 'Blue Face Paint', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (135, 'Yellow Snorkel', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (136, 'Ski Goggles', 3, 480, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (137, 'Swim Goggles', 3, 220, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (138, 'Swirly Glasses', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (139, 'Pink Snorkel', 3, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (152, 'Rockhoppers Eyebrows', 3, 65535, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (161, 'Rockhoppers Beard', 4, 65535, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (162, 'Santa Beard', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (166, 'Green Bowtie', 4, 60, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (167, 'Green Necktie', 4, 125, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (168, 'Parrot', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (169, 'Green Scarf', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (170, 'Halloween Scarf', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (171, 'Hawaiian Lei', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (172, 'Yellow Scarf', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (173, 'Christmas Scarf', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (174, 'Boa', 4, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (175, 'Pink Scarf', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (176, 'Black Tie', 4, 125, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (177, 'Striped Tie', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (178, 'Shell Necklace', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (179, 'Blue Scarf', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (180, 'Snare Drum', 4, 380, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (181, 'Pearl Necklace', 4, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (182, 'Pendant Necklace', 4, 110, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (183, 'Bling Bling Medallion', 4, 600, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (184, 'Star Necklace', 4, 280, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (185, 'Candy Necklace', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (186, 'Ruffle Collar', 4, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (187, 'Cheesy Necktie', 4, 125, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (188, 'Mullet Necktie', 4, 125, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (189, 'Smiley Necktie', 4, 125, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (190, 'Primitive Necklace', 4, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (191, 'Life Vest', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (192, 'Bandana', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (193, 'Blue Lei', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (194, 'Whistle', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (195, 'Camera', 4, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (196, 'Green Tabard', 4, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (197, 'Orange Tabard', 4, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (198, 'Purple Tabard', 4, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (199, 'Backstage Pass', 4, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (201, 'Red Shirt', 5, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (202, 'Blue Polo Shirt', 5, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (203, 'Butterfly T-Shirt', 5, 200, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (204, 'Astro Barrier T-Shirt', 5, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (205, 'I Love My Puffle T-Shirt', 5, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (206, 'Red Shirt', 5, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (207, 'Snowflake T-Shirt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (208, 'Sailor''s Shirt', 5, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (209, 'Flannel Shirt', 5, 240, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (210, 'Star T-Shirt', 5, 200, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (211, 'Hawaiian Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (212, 'Grass Skirt', 5, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (213, 'Orange Snowsuit', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (214, 'Black Bowtie', 4, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (215, 'Life Jacket', 5, 320, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (216, 'Free Bowtie', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (217, 'Cowboy Vest', 5, 420, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (218, 'Quilted Vest', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (219, 'Blue Duffle Coat', 5, 750, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (220, 'Hockey Stick', 6, 280, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (221, 'Black Hoodie', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (222, 'Pink Hoodie', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (223, 'Green Hoodie', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (224, 'Purple Hoodie', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (225, 'Red Turtleneck', 5, 440, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (226, 'Green Turtleneck', 5, 440, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (227, 'Bee Costume', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (228, 'Princess Costume', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (229, 'Blue Turtleneck', 5, 440, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (230, 'Orange Jumper', 5, 450, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (231, 'Pirate Belt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (232, 'Green Suede Jacket', 5, 700, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (233, 'Red Electric Guitar', 6, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (234, 'Acoustic Guitar', 6, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (235, 'Yellow Raincoat', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (236, 'Poncho', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (237, 'Red Suede Jacket', 5, 700, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (238, 'Pastel Suede Jacket', 5, 700, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (239, 'Wetsuit', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (240, 'Pizza Apron', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (241, 'Purple Suede Jacket', 5, 700, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (242, 'Victorian Dress', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (243, 'Black Parka', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (244, 'Ghost Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (245, 'Orange Space Suit', 5, 700, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (246, 'Skeleton', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (247, 'Clown Suit', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (248, 'Kimono', 5, 460, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (249, 'Pink Snowsuit', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (250, 'Ice Cream Apron', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (251, 'Jean Jacket', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (252, 'Pink Dress', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (253, 'Purple Dress', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (254, 'Red Cheerleader Outfit', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (255, 'Blue Cheerleader Outfit', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (256, 'Ballerina', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (257, 'Spring Dress', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (258, 'Pirate Dress', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (259, 'Shamrock Dress', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (260, 'Seafoam Dress', 5, 580, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (261, 'Black Suit', 5, 650, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (262, 'Coffee Apron', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (263, 'Pizza Apron', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (264, 'Red Pirate Dress', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (265, 'Fishing Vest', 5, 480, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (266, 'Alien Costume', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (267, 'Red Shorts', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (268, 'Purple Space Suit', 5, 700, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (269, 'Pink Swimsuit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (270, 'Purple Bikini', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (271, 'Yellow Bikini', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (272, 'Blue Shorts', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (273, 'Aqua Bikini', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (274, 'Inflatable Duck', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (275, 'Overalls', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (276, 'Robot Suit', 5, 800, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (277, 'Red Hockey Jersey', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (278, 'Blue Hockey Jersey', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (279, 'Pink Letterman Jacket', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (280, 'Green Letterman Jacket', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (281, 'Black Letterman Jacket', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (282, 'Lighthouse Shirt', 5, 750, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (283, 'Pink Duffle Coat', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (284, 'Elf Suit', 5, 350, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (285, 'Admiral Jacket', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (286, 'Flamenco Dress', 5, 640, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (287, 'Matador Outfit', 5, 520, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (288, 'Striped Overalls', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (289, 'Long Johns', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (290, 'Victorian Jacket', 5, 650, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (291, 'Leprechaun Tuxedo', 5, 560, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (292, 'Green Duck', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (293, 'Tuba', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (294, 'Tuxedo', 5, 800, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (295, 'Captain''s Coat', 5, 650, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (296, 'Red Letterman Jacket', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (297, 'Lifeguard Shirt', 5, 180, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (298, 'Sport Life Jacket', 5, 480, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (299, 'Firefighter Jacket', 5, 380, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (301, 'Red Cape', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (302, 'Blue Cape', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (303, 'Black Cape', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (304, 'Lime Green Cape', 4, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (305, 'Magicians Cape', 4, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (306, 'Red Backpack', 4, 410, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (307, 'Messenger Bag', 4, 290, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (308, 'Scuba Tank', 4, 480, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (309, 'Bee Wings', 4, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (310, 'Fairy Wings', 4, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (312, 'Blue Backpack', 4, 410, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (313, 'Hiking Backpack', 4, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (314, 'Pink Backpack', 4, 410, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (315, 'Pink Cape', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (316, 'Supply Bag', 4, 700, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (317, 'Blue Mail Bag', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (321, 'Flashing Lure Fishing Rod', 6, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (322, 'Gold Wristwatch', 6, 175, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (323, 'Silver Watch', 6, 80, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (324, 'Pot O''Gold', 6, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (325, 'Water Wings', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (326, 'Paddle Ball', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (327, 'Lasso', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (328, 'Bell', 6, 50, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (329, 'Orange Pompoms', 6, 120, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (330, 'Red Pompoms', 6, 120, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (331, 'Pumpkin Basket', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (332, 'Magic Wand', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (333, 'Blue Pompoms', 6, 120, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (334, 'Cane', 6, 170, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (335, 'Pair of Maracas', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (336, 'Lollipop', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (337, 'Festive Maracas', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (338, 'Black Electric Guitar', 6, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (339, 'Friendship Bracelet', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (340, 'Drumsticks', 6, 120, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (341, 'Flower Basket', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (342, 'Snow Shovel', 6, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (343, 'Violin', 6, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (344, 'Life Ring', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (345, 'Blue Water Wings', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (346, 'Fishing Rod', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (347, 'Flashlight', 6, 120, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (348, 'Marshmallow Stick', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (349, 'Tennis Racket', 6, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (350, 'Cotton Candy', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (351, 'Brown Shoes', 7, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (352, 'Black Sneakers', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (353, 'Ballet Shoes', 7, 180, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (354, 'Clown Shoes', 7, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (355, 'Cement Shoes', 7, 5, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (357, 'Blue Sneakers', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (358, 'Black Dress Shoes', 7, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (359, 'Pink Bow Dress Shoes', 7, 420, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (360, 'Running Shoes', 7, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (361, 'Snowshoes', 7, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (362, 'Green Flippers', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (363, 'Yellow Sandals', 7, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (364, 'Blue Flippers', 7, 225, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (365, 'Winter Boots', 7, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (366, 'Bunny Slippers', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (367, 'Yellow Flippers', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (368, 'Cowboy Boots', 7, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (369, 'Ice Skates', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (370, 'Elf Shoes', 7, 170, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (371, 'Skateboard', 7, 320, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (372, 'Hiking Boots', 7, 325, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (373, 'Yellow Rubber Boots', 7, 280, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (374, 'Pirate Boots', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (375, 'Pink Sandals', 7, 160, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (376, 'Brown Sandals', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (377, 'Blue Rollerskates', 7, 260, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (378, 'Pink Rollerskates', 7, 260, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (379, 'Wool Socks', 7, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (380, 'Fuzzy Boots', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (381, 'Green Rubber Boots', 7, 280, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (382, 'Pink Flippers', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (383, 'Blue Flower Sandals', 7, 160, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (384, 'Green Sandals', 7, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (385, 'Cleats', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (386, 'Pink Sneakers', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (387, 'Orange Sneakers', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (388, 'Stardust Slippers', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (401, 'Sombrero', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (402, 'Black Toque', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (403, 'Hard Hat', 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (404, 'Tan Cowboy Hat', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (405, 'Green Ball Cap', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (406, 'Pink Ball Cap', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (407, 'Red Propeller Cap', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (408, 'Flower Hat', 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (409, 'Graduation Cap', 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (410, 'Queen''s Crown', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (411, 'Festive Sombrero', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (412, 'Tiara', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (413, 'Party Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (414, 'Santa Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (415, 'Pilgrim Hat', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (417, 'Brown Fedora', 2, 225, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (418, 'Pink Toque', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (419, 'Russian Hat', 2, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (420, 'Black Toque', 2, 75, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (421, 'Pink Toque', 2, 75, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (422, 'Newspaper Hat', 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (423, 'Top Hat', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (424, 'Chef Hat', 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (425, 'Shamrock Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (426, 'Jester Hat', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (427, 'Pink Bunny Ears', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (428, 'Tour Guide Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, TRUE, now()),
 (429, 'Miners Helmet', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (430, '2nd Year Party Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (431, 'Blue Propeller Cap', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (432, 'Safari Hat', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (433, 'Black Cowboy Hat', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (434, 'Pink Cowgirl Hat', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (435, 'Red Ball Cap', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (436, 'Blue Ball Cap', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (437, 'Straw Hat', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (438, 'Blue Toque', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (439, 'Park Ranger Hat', 2, 210, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (440, 'Sailor''s Cap', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (441, 'Admirals Hat', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (443, 'Wizard Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (444, '1st Year Party Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (445, 'Green Toque', 2, 120, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (446, 'Fishing Hat', 2, 220, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (447, 'Elf Hat', 2, 150, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (448, 'Winter Cape', 2, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (449, 'Fruit Headdress', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (450, 'Earflap Cap', 2, 180, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (451, 'Roman Helmet', 2, 800, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (452, 'Viking Helmet', 2, 750, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (453, 'Football Helmet', 2, 360, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (454, 'Bike Helmet', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (455, 'Pink Viking Helmet', 2, 650, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (456, 'Blue Viking Helmet', 2, 1200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (457, 'Divers Helmet', 2, 650, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (458, 'Space Helmet', 2, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (460, 'Gold Viking Helmet', 2, 1500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (461, 'Hockey Helmet', 2, 220, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (462, 'Red Football Helmet', 2, 360, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (463, 'Blue Football Helmet', 2, 360, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (464, 'Snowboard Helmet', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (465, 'Firefighter Hat', 2, 130, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (466, 'Angel Halo', 2, 0, FALSE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (467, 'Pink Pom Pom Toque', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (468, 'Alien Antenna', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (469, 'Umbrella Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (470, 'Feathered Hat', 2, 280, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (471, 'Reindeer Antlers', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (472, 'Bee Antennae', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (473, 'Winged Viking Helmet', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (474, 'Clown Hair', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (475, 'Countess Steeple Hat', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (476, 'Bard Hat', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (477, 'Court Jester Hat', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (478, 'Coral Crown', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (479, 'Zeus The Moose Head', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (480, 'Bird Mascot Head', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (481, 'Headphones', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (482, 'Stocking Cap', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (483, 'Blue Earmuffs', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (484, 'Pink Earmuffs', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (485, 'Bonnet', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (486, 'Feathered Tiara', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (487, 'Princess Hat', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (488, 'Head Band', 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (489, 'Director''s Hat', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (490, 'Pirate Bandana', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (491, 'Ice Crown', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (492, 'Floppy Hat', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (493, 'Tricorn Hat', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (494, 'Golf Hat', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (495, 'Marching Band Hat', 2, 360, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (496, 'Blue Bunny Ears', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (497, 'Sailor Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (498, 'Puffle Bandana', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (499, 'Flower Headdress', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (500, 'Canada', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (501, 'USA', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (502, 'Australia', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (503, 'United Kingdom', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (504, 'Belgium', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (505, 'Brazil', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (506, 'China', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (507, 'Denmark', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (508, 'Finland', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (509, 'France', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (510, 'Germany', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (511, 'Israel', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (512, 'Japan', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (513, 'Korea', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (514, 'Netherlands', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (515, 'Norway', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (516, 'Poland', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (517, 'Russia', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (518, 'Spain', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (519, 'Sweden', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (520, 'Switzerland', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (521, 'Turkey', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (522, 'Mexico', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (523, 'New Zealand', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (524, 'Ireland', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (525, 'Portugal', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (526, 'South Africa', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (527, 'India', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (528, 'Italy', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (529, 'Belize', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (530, 'Egypt', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (531, 'Hungary', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (532, 'Antarctica', 8, 20, FALSE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (533, 'Argentina', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (534, 'Jamaica', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (535, 'Chile', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (536, 'Colombia', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (537, 'Puerto Rico', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (538, 'Peru', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (539, 'Venezuela', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (540, 'Costa Rica', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (541, 'Guatemala', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (542, 'Singapore', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (543, 'Malaysia', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (544, 'Philippines', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (545, 'Haiti', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (546, 'Dominican Republic', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (547, 'Uruguay', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (548, 'Ecuador', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (550, 'Shamrock', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2006-03-17'),
 (551, 'Music Note', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2006-03-31'),
 (552, 'Plant', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2006-04-14'),
 (553, 'Pizza Slice', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2006-04-28'),
 (554, 'Balloon', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2006-05-12'),
 (555, 'Mining Lantern', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2006-05-26'),
 (556, 'Beach Ball', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2006-06-09'),
 (557, 'Sun', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2006-06-23'),
 (558, 'Horse Shoe', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2006-07-07'),
 (559, 'Astro-Barrier Ship', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2006-07-21'),
 (560, 'Hockey Stick', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2006-08-04'),
 (561, 'Soccer Ball', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2006-08-18'),
 (562, 'Pencil', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2006-09-01'),
 (563, 'Lighthouse', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2006-09-15'),
 (564, 'Telescope', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-07-20'),
 (565, 'Jolly Roger Flag', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2006-10-13'),
 (566, 'Pumpkin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2006-10-27'),
 (567, 'Jet Pack', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2006-11-10'),
 (568, 'Life Ring', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2006-11-24'),
 (569, 'Bonfire', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2006-12-08'),
 (570, 'Christmas Tree', 8, 50, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2006-12-14'),
 (571, 'Candy Cane', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2007-12-22'),
 (572, 'Apple', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2007-01-05'),
 (573, 'Cactus', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2007-01-19'),
 (574, 'Teddy Bear', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2007-02-02'),
 (575, 'Ice Block', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2007-02-16'),
 (576, 'Shrimp', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2007-03-02'),
 (577, 'Pot O'' Gold', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2007-03-16'),
 (578, 'Cardboard Box', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2007-03-30'),
 (579, 'Microphone Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2007-04-13'),
 (580, 'Gem', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2007-04-27'),
 (581, 'Tulip', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2007-05-11'),
 (582, 'Starfish', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2007-05-25'),
 (583, 'Surfboard', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2007-06-08'),
 (584, 'Picnic Basket', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2007-06-22'),
 (585, 'Water Droplet', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2007-07-06'),
 (586, 'Cart', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2007-07-20'),
 (587, 'Butterfly', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2007-08-03'),
 (588, 'Tent', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2007-08-17'),
 (589, 'Baseball', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2007-08-31'),
 (590, 'Jellyfish', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2007-09-14'),
 (591, 'Circus Tent', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2007-09-21'),
 (592, 'Ping Pong Paddle', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2007-09-28'),
 (593, 'Hairbrush', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2007-10-12'),
 (594, 'Spider', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2007-10-26'),
 (595, 'UFO', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2007-11-09'),
 (596, 'Needle', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2007-11-23'),
 (597, 'Holly', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2007-12-07'),
 (598, 'Wreath', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2008-12-21'),
 (599, 'Red Snow Shovel', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2008-01-04'),
 (600, 'Sombrero', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2008-01-18'),
 (601, 'Rowboat', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2008-02-01'),
 (602, 'Anchor', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2008-02-15'),
 (603, 'Golden Wheel', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2008-02-19'),
 (604, 'Aqua Grabber Sub', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2008-02-19'),
 (605, 'Book', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2008-03-14'),
 (606, 'Crayon', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2008-03-28'),
 (607, 'Pyramid', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2008-04-11'),
 (608, 'Treasure Chest', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2008-04-25'),
 (609, 'Rockhopper''s Key', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2008-04-28'),
 (610, 'Goblet', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2008-05-09'),
 (611, 'Anvil', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2008-05-23'),
 (612, 'Icecream Cone', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2008-06-06'),
 (613, 'Basketball', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2008-06-20'),
 (614, 'Firework Rocket', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2008-07-04'),
 (615, 'Treble Clef', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2008-07-18'),
 (616, 'Record', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2008-08-01'),
 (617, 'Dodgeball', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2008-08-15'),
 (618, '150th Newspaper', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2008-08-30'),
 (619, 'Magnifying Glass', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2008-09-12'),
 (620, 'Ruby', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2010-06-10'),
 (621, 'Yellow Balloon', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2008-09-26'),
 (622, 'Lollipop', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2008-09-26'),
 (623, 'Microscope', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2008-10-10'),
 (624, '3rd Anniversary Cake', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2008-10-24'),
 (625, 'Blue Snow Shovel', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2008-11-07'),
 (626, 'Snowflake Tile', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2008-11-21'),
 (627, 'Snow Fort', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2008-12-05'),
 (628, 'Present', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2009-12-19'),
 (629, 'Gingerbread Man', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2009-01-02'),
 (630, 'Taco', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2009-01-16'),
 (631, 'Lily', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2009-01-30'),
 (632, 'Puffle O''s', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2009-02-13'),
 (633, 'O-Berry', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2009-02-27'),
 (634, 'Lucky Coin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2009-03-13'),
 (635, 'Top Hat', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2009-03-27'),
 (636, 'Chocolate Bunny', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2009-04-10'),
 (637, 'Tree', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2009-04-24'),
 (638, 'King''s Crown', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2009-05-08'),
 (639, 'Ice Cream Sundae', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2009-05-22'),
 (640, 'Safari Hat', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2009-06-05'),
 (641, 'Watermelon', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2009-06-19'),
 (642, 'Dojo Lantern', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2009-07-03'),
 (644, 'Toy Sailboat', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2009-07-31'),
 (645, 'Sand Castle', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2009-08-14'),
 (646, 'Koi Fish', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2009-08-28'),
 (647, 'Padlock', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2009-09-25'),
 (648, 'Beach Umbrella', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2009-07-17'),
 (649, 'Cotton Candy', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2009-09-04'),
 (650, 'The Rocker', 2, 750, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (651, 'The Spikester', 2, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (652, 'The Disco', 2, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (653, 'The Funster', 2, 500, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (654, 'The Beekeeper', 2, 750, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (655, 'The Sportster', 2, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (656, 'The Sunstriker', 2, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (657, 'The Flutterby', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (658, 'The Sidetied', 2, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (659, 'Migrator Mascot Head', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (660, 'Cocoa Bunny Ears', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (661, 'The Cleo', 2, 650, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (662, 'The Surf Knot', 2, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (663, 'Gigantic St. Patrick''s Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (664, 'Green Bunny Ears', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (665, 'Alaska''s Explorer Hat', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (666, 'Pharaoh Headdress', 2, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (667, 'King''s Crown', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (668, 'Knight Helmet', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (669, 'Woodsman''s Hat', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (670, 'The Firestriker', 2, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (671, 'The Sidetied Too', 2, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (672, 'The Spikette', 2, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (673, 'Emerald Hat', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (674, 'Red Hard Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (675, 'The Fern Fuzz', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (676, 'Cheap Time Travel Hat', 2, 10, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (677, 'The Suprema Diva', 2, 650, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (678, 'The Messiness', 2, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (701, 'Flame Surfboard', 6, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (702, 'Daisy Surfboard', 6, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (703, 'Silver Surfboard', 6, 800, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (711, 'Yellow Arrow Wakeboard', 6, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (712, 'Pink Striped Wakeboard', 6, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (713, 'Pink Purse', 6, 300, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (715, 'Gold Bracelets', 6, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (716, 'Crook & Flail', 6, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (717, 'Baseball Glove', 6, 300, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (718, 'Crystal Staff', 6, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (719, 'Basketball', 6, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (720, 'Blue Football Jersey', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (721, 'Red Football Jersey', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (722, 'Referee Jersey', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (723, 'Green Shield', 6, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (724, 'Orange Shield', 6, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (725, 'Purple Shield', 6, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (726, 'Football', 6, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (727, 'Soccer Ball', 6, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (728, 'Spikester Cuffs', 6, 170, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (729, 'Blue Electric Bass', 6, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (730, 'Acoustic Sunburst Guitar', 6, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (731, 'Pink Electric Guitar', 6, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (732, 'Microphone', 6, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (739, 'Puffle Trivia Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2017-01-31'),
 (750, 'Blue Puffle', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (751, 'Pink Puffle', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (752, 'Black Puffle', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (753, 'Green Puffle', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (754, 'Purple Puffle', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (755, 'Red Puffle', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (756, 'Yellow Puffle', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (757, 'White Puffle', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (758, 'Orange Puffle', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (759, 'Brown Puffle', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (760, 'Countess Dress', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (761, 'Bard Outfit', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (762, 'Court Jester Outfit', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (763, 'Fish Costume', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (764, 'Winter Poncho', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (765, 'Shadow Guy Costume', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (766, 'Gamma Gal Costume', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (767, 'Squidzoid Costume', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (768, 'Mermaid Costume', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (769, 'White Admiral Jacket', 5, 620, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (770, 'Safety Vest', 5, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (771, 'Blue Track Jacket', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (772, 'Migrator Mascot Body', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (773, 'Penguin Band Hoodie', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (774, 'Seashell Belt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (775, 'Red Soccer Jersey', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (778, 'Blue Soccer Jersey', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (780, 'Daisy Zippered Hoodie', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (781, 'Striped Rugby Shirt', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (782, 'Cocoa Bunny Costume', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (783, 'Casual Suit Jacket', 5, 650, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (784, 'Yellow Sun Dress', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (785, 'Ski Patrol Jacket', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (786, 'Sienna Explorer Outfit', 5, 650, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (787, 'Beige Explorer Outfit', 5, 650, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (788, 'Pharaoh Costume', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (789, 'Mummy Costume', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (790, 'Black and Red Sailor Shirt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (791, 'Red Baseball Uniform', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (792, 'Blue Baseball Uniform', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (793, 'Royal Robe', 4, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (794, 'Knight Costume', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (795, 'Dragon Costume', 5, 1000, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (796, 'Emerald Dress', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (797, 'Queen''s Dress', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (798, 'Blacksmith Apron', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (799, 'Squire Outfit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (800, 'Spy Phone', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (801, 'Mission 1 Medal', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (802, 'Letter from Aunt Arctic', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (803, 'Mission 2 Medal', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (804, 'Letter from G', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (805, 'Mission 3 Medal', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (806, 'Card from Dancing Penguin', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (807, 'Forest Map', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (808, 'Mission 4 Medal', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (809, 'Handy Penguin Award', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (810, 'Mission 5 Medal', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (811, 'Box of Pizza', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (812, 'Coins For Change Card', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (813, 'Mission 6 Medal', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (814, 'Magnet Blueprints', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (815, 'Mission 7 Medal', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (816, 'Blue Pennant', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (817, 'Mission 8 Medal', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (818, 'Cool Gift', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (819, 'Mission 9 Medal', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (820, 'Box of Chocolates', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (821, 'Starter Deck', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (822, 'Mission 10 Medal', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (823, 'Employee of the Month Trophy', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (830, 'Sport Bikini', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (831, 'Green Hawaiian Shorts', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (832, 'Sarong', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (833, 'Divers Suit', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (834, 'Tiger Cave Toga', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (835, 'Zebra Cave Toga', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (836, 'Red Basketball Jersey', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (837, 'Blue Basketball Jersey', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (838, 'Striped Sash Dress', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (839, 'Pop Music Shirt', 5, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (840, 'Orange Rocker Shirt', 5, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (841, 'Supernova Suit', 5, 850, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (842, 'Dazzle Dress', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (843, 'Black Cowboy Shirt', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (844, 'Pink Cowgirl Shirt', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (845, 'Gray Pirate Coat', 5, 650, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (846, 'Conductor''s Suit', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (848, 'Pop Girl T-Shirt', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (849, 'Caribbean T-Shirt', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (850, 'Western T-Shirt', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (851, 'Sparkle Rock T-Shirt', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (852, 'Classical T-Shirt', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (853, 'Rocker T-Shirt', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (854, 'Music Jam T-Shirt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (901, 'Heart Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (902, 'Flower Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (903, 'Nightime Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (904, 'Clouds Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (905, 'Cut-out Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (906, 'Target Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (907, 'Camo Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (908, 'Lined Paper Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (909, 'Pink Hawaiian Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (910, 'Blue Hawaiian Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (911, 'Oasis Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (912, 'Beach Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (913, 'Western Sunset Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (914, 'Corral Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (915, 'Saloon Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (916, 'Stagecoach Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (917, 'Band Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (918, 'Soccer Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (919, 'Hockey Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (920, 'Basketball Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (921, 'Gymnastic Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (922, 'Blue Vines Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (923, 'Brick Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (924, 'Underwater Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (925, 'Puffle Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (926, 'Lighthouse Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (927, 'Autumn Leaves Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (928, 'Pumpkin Patch Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (929, 'Igloo Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (930, 'Launch Pad Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (931, 'Zigzag Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (932, 'Winter Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (933, 'Gift Wrap Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (934, 'Christmas Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (935, 'Snowflakes Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (936, 'Ice Fishing Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (937, 'Disco Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (938, 'Mexican Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (939, 'Aurora Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (940, 'Snowman Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (941, 'Lace Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (942, 'Melting Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (943, 'Springtime Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (944, 'Pizza Splat Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (945, 'Sea Monster Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (946, 'Music Score Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (947, 'Piano Keys Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (948, 'Instruments Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (949, 'Bookshelves Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (950, 'Balloon Arch Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (951, 'Hydro Hopper Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (952, 'Picnic Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (953, 'Surfboards Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (954, 'Fish Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (955, 'Campfire Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (956, 'Jigsaw Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (957, 'Water Balloon Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (958, 'Splash Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (959, 'Rockhopper Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (960, 'Treasure Map Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (961, 'Tent Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (962, 'Lemonade Stand Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (963, 'Fireflies Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (964, 'Camping Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (965, 'Baseball Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (966, 'Apples Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (967, 'Dock Jumping Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (968, 'Wakeboard Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (969, 'Sunset', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (970, 'Teddy Bear Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (971, 'Spooky Trees Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (972, 'Igloo Door Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (973, 'Candy Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (974, 'Football Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (975, 'Mountaintop Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (976, 'Jetpack Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (977, 'Snapshots Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (978, 'Christmas Trees Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (979, 'Plaid Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (980, 'Blossom Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (981, 'Blue Stars Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (982, 'Bamboo Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (983, 'Sunflowers', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (984, 'Space Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (985, 'Emotes Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (986, 'Snow Forts Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (987, 'Adventure Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (988, 'Castle Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (989, 'Tabard Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (990, 'Twelfth Fish Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (991, 'Fresco Waves', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (992, 'Volcanic', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (993, 'Sports Equipment Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (994, 'Karaoke Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (995, 'Cityscape Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (996, 'Band Autograph', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (997, 'Beach Chairs Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (998, 'Gym Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (999, 'Blueprints Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1004, 'Denim Hip Hop Hat', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1005, 'Green Paisley Bandana', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1006, 'Tweed Hat', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1007, 'The Tuft', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1008, 'The Dizzy', 2, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1009, 'Grey Fedora', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1010, 'The Movie Star', 2, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1011, 'Blue Felt Hat', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1012, 'Doorman''s Cap', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1013, 'Frankenpenguin Hat', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1014, 'Faery Hair', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1015, 'Rad Scientist Wig', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1016, 'Alien Thinking Cap', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1017, 'Brown Skater Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1018, 'Pink Skater Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1019, '3rd Year Party Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1020, 'Chullo Hat', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1021, 'The Chill', 2, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1022, 'The Flouncy', 2, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1023, 'Prince Redhood Hat', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1024, 'Beautiful Braid', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1025, 'Fairytails', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1026, 'Snowman Head', 2, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1027, 'The Desert Rose', 2, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1028, 'The Freestyle', 2, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1029, 'The Shamrocker', 2, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1030, 'Starlit Sombrero', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1032, 'Cadence Hair', 2, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1033, 'Cadence Shoes', 7, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1034, 'Cadence''s Wristbands', 6, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1035, 'Mini Sombrero', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1036, 'Goalie Helmet', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1037, 'Swim Cap and Goggles', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1038, 'Yellow Ball Cap', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1039, 'The Befluttered', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1040, 'Blue Top Hat', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1041, 'The Starlette', 2, 700, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1042, 'The Sidewinder', 2, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1043, 'The Flickstar', 2, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1044, 'Aunt Arctic''s Hat', 2, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1045, 'Pink Visor', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1046, 'White Head Band', 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1047, 'Baseball Helmet', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1048, 'The Belle', 2, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1049, 'The Shenanigan', 2, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1050, 'The Posh', 2, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1051, 'Black Graduation Cap', 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1052, 'Golden Knight''s Helmet', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1053, 'Amethyst Hat', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1054, 'Topaz Hat', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1055, 'The Shimmer', 2, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1056, 'Black Viking Helmet', 2, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1057, 'Jewelled Viking Helmet', 2, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1058, 'Outback Hat', 2, 175, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1059, 'The Electric', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1060, 'The Flow', 2, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1061, 'The Waves', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1062, 'Green Safari Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1063, 'The Sunray', 2, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1064, 'The Shock Wave', 2, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1065, 'Dazzling Blue Top Hat', 2, 475, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1066, 'The Tousled', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1067, 'The Vibrant', 2, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1069, 'Green Headphones', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1072, 'Girl''s Pilot Cap', 2, 330, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1073, 'Pilot Cap', 2, 310, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1074, 'Red Paisley Bandanna', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1075, 'Green Propeller Cap', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1076, 'The Blue Lagoon', 2, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1077, 'The Surf', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1078, 'Green Football Helmet', 2, 360, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1079, 'The Sidetied Strikes Back', 2, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1080, 'Stripey Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1081, 'Fuzzy Viking Hat', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1082, 'Cosmic Star Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1083, 'Ring Master Hat', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1084, 'The Prep', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1085, 'The Chill Out', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1086, 'Fiery Helmet', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1087, 'Helmet of Oceans', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1091, 'Snow Fairy Hair', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1092, 'The Royal', 2, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1093, 'Blizzard Wizard Hat', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1094, 'Ladybug Antennae', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1095, 'Pumpkin Head', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1096, 'Pumpkin Antennae', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1097, 'Ogre Ears', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1098, '4th Year Party Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1099, 'Multi-Colored Chullo', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1100, 'Yellow Toque', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1101, 'The Flutter', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1102, 'Buckle Hat', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1103, 'The Sidekick', 2, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1104, 'Blue Pompom Toque', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1105, 'Reindeer Head', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1106, 'Humbug Hat', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1109, 'The Glimmer', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1110, 'The Vintage', 2, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1111, 'Brown Teal Cap', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1112, 'The Frost', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1113, 'The Brunette', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1114, 'Monkey King Mask', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1115, 'The Phoenix', 2, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1116, 'Funny Pig Hat', 2, 275, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1117, 'Guardian Dog Hat', 2, 320, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1119, 'Candy Cane Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1120, 'Puffle Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1121, 'The Elegant', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1122, 'The Scarlet Star', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1123, 'The Short & Sweet', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1124, 'The Part', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1125, 'Cumberband Hat', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1126, 'Puffle Cap', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1127, 'Golden Feather Crown', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1129, 'Press Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1131, 'King Jester Hat', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1133, 'Green Hard Hat', 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1134, 'Straw Gardening Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1135, 'Brown Striped Fedora', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1136, 'Green Cap', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1137, 'White Cocoa Bunny Ears', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1138, 'The Band', 2, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1139, 'The Trend', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1142, 'The Fair Maiden', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1143, 'The Damsel', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1144, 'The Duchess', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1145, 'Duke''s Hat', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1146, 'Iron Helmet', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1148, 'EPF Earpiece', 2, 2, FALSE, TRUE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (1149, 'Delta Fedora', 2, 10, TRUE, TRUE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (1150, 'The Alpha', 2, 10, TRUE, TRUE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (1151, 'The Bonny Curls', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1152, 'First Mate''s Hat', 2, 425, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1153, 'Striped Pirate Bandanna', 2, 375, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1154, 'Swashbuckler''s Hat', 2, 430, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1155, 'Commander''s Hat', 2, 375, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1156, 'The Razzmatazz', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1157, 'The Aquamarine', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1158, 'The Adventurer', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1159, 'Pirate Bandanna', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1160, 'The Squid Lid', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1161, 'The Sunburst', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1162, 'The Flip', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1163, 'The Golden Waves', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1164, 'Blue Headphones', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1167, 'Yellow Climbing Helmet', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1168, 'Red Climbing Helmet', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1169, 'Chilly Trek Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1170, 'Alpha Headset', 2, 2, TRUE, TRUE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (1171, 'Delta Headset', 2, 2, TRUE, TRUE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (1172, 'Red Mohawk', 2, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1173, 'The Summer Tussle', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1174, 'Jam Cap', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1175, 'The Sun Rays', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1176, 'Balloon Flower Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1177, 'Red and Yellow Stripey Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1178, 'Blue Fuzzy Viking Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1179, 'Green Cosmic Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1180, 'Magician''s Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1183, 'Clown Wig', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1184, 'Magical Crown', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1185, 'The Count', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1186, 'Witch''s Hat', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1187, 'The Lady Frankenpenguin', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1188, 'Cleo Headdress', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1189, 'Orange Blob', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1190, 'Green Cyclops', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1191, 'Purple Bat Wings', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1192, '5th Year Party Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1193, 'The Side Swept', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1194, 'The Skater', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1195, 'The Chic', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1196, 'Bucket Hat', 2, 75, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1199, 'The Cosmic', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1201, 'Comm Helmet', 2, 15, TRUE, TRUE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (1202, 'Elf Pigtails', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1203, 'Blue Goggles', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1204, 'Yellow Goggles', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1205, 'Blue Fuzzy Hat', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1206, 'Santa Hat', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1207, 'Globe Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1208, 'Expedition Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1210, 'The Twister', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1211, 'Blue Toque', 2, 80, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1212, 'The Explorer', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1213, 'Holiday Toque', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1214, 'The Bolero', 2, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1215, 'Black Hockey Helmet', 2, 220, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1217, 'Optic Headset', 2, 15, TRUE, TRUE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (1218, 'The Tuned In', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1219, 'The Ebony', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1220, 'The Fancy Free', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1221, 'Khaki Rocker Cap', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1222, 'Brown Puffle Hat', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1225, 'Blue Puffle Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1226, 'The Boho', 2, 375, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1227, 'The Sweep', 2, 360, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1228, 'The Overgrown', 2, 375, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1229, 'The Violet Beret', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1230, 'The Storm', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1231, 'The Bolt', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1232, 'The Quicksilver', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1233, 'The Cinder', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1234, 'Green Viking Helmet', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1235, 'Cadence''s Puffle', 6, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1236, 'Purple Propeller Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1237, 'Delivery Hat', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1238, 'The Cotton Candy', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1239, 'The Country Gal', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1240, 'Brown Cowboy Hat', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1241, 'Astro Helmet', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1242, 'Dark Cocoa Bunny Ears', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1245, 'Sapphire Princess Hat', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1246, 'Gold Princess Hat', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1247, 'Ruby Princess Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1248, 'King''s Blue Crown', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1249, 'The Duchess II', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1250, 'The Forest Maiden', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1251, 'The Viscount', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1252, 'Village Jester Hat', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1253, 'White Knight Helmet', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1256, 'Outback Exploring Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1258, 'The Daydream', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1259, 'The Sunny Side', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1260, 'The Shore Thing', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1261, 'The Certain Something', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1262, 'The Funkadelic', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1263, 'One Man Band Hat', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1264, 'G Billy Bandana', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1265, 'The Rock Bandana', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1266, 'Music Jam Cap', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1267, 'The Hairspray', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1268, 'The Fiesta', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1269, 'The Cabana', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1272, 'Sunset Crown', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1273, 'Petey K Toque and Hair', 2, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1274, 'Stompin'' Bob Fowhawk Hair', 2, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1275, 'G Billy Hair and Bandana', 2, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1276, 'The Shipshape', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1277, 'The Pirateer', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1278, 'The Lookout', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1279, 'The Navy Officer', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1280, 'The Buccaneer', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1281, 'The Cerulean Lagoon', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1282, 'The Rascal', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1283, 'The Continental', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1284, 'Purple Wizard Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1285, 'New Player Red Baseball Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, TRUE, FALSE, FALSE, now()),
 (1286, 'The Gnarly Helmet', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1287, 'The Rad Helmet', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1288, 'The Flying Fishtail', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1289, 'The Sub Zero', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1290, 'The Head-turner', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1291, 'The Supersonic', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1292, 'Alpine Helmet', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1293, 'The Rocket', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1294, 'Infrared Tracker', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1295, 'The Downtown', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1296, 'The Clown-Around', 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1297, 'The Abracadabra', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1298, 'The Forget Me Knot', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1299, 'Mime Beret', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1300, 'The Trapeze Artist', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1301, 'The Jingle Jangle', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1302, 'The Sundae Surprise', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1303, 'The Balloonist', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1304, 'Inky Squid Lid', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1305, 'The Fairy Princess', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1306, 'The Witch Hazel', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1307, 'Werewolf Mask', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1308, '6th Anniversary Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1309, 'Unicorn Horn', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1315, 'The Fire Flicker', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1316, 'The Water Ripple', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1317, 'The Sushi Master', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1318, 'The Tremor', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1319, 'The Bad Hair Day', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1320, 'The Side Swirl', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1321, 'Fire Headband', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1322, 'Water Headband', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1323, 'Glowing Pumpkin Head', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1324, 'The Chillax', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1325, 'The Claus', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1326, 'The Claus-ette', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1327, 'The Chestnut', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1328, 'The Evergreen', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1329, 'The Pirouette', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1330, 'The Nutcracker', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1331, 'Springy Santa Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1332, 'CFC Cap', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1333, 'The Jingle Bell', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1334, 'The Tree Topper', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1335, 'The Hornament Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1336, 'The Twizzle', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1337, 'The Triple Lutz', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1338, 'The Free Skate', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1339, 'Tallest Haircut', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1340, 'The Sweet Knot', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1341, 'The Classic', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1342, 'Heavy Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1350, 'The Stunning', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1351, 'The Striking', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1352, 'The Flitter Flutter', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1353, 'The Ensemble', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1354, 'The Bright Side', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1355, 'Crew Cap', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1356, 'The Corsair Captain', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1357, 'Black Swashbuckler''s Hat', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1358, 'Rising Sun Crown', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1359, 'Viking Lord Helmet', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1360, 'The Strawberry', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1361, 'The Blueberry', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1362, 'The Watermelon', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1363, 'The Blackberry', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1364, 'The Lime', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1365, 'The Banana', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1366, 'The Grape', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1367, 'The Coconut', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1368, 'The Orange', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1369, 'The Pear', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1370, 'The Red Licorice', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1371, 'The Frosting', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1372, 'The Bubblegum', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1373, 'The Black Licorice', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1374, 'The Peppermint', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1375, 'The Honeycomb', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1376, 'The Sprinkles', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1377, 'The Whipped Cream', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1378, 'The Gummy', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1379, 'The Chocolate', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1380, 'Polka Dot Puffle Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1381, 'The Tunage', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1382, 'The Auburn', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1383, 'Box Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1384, 'PH Hat/Hair', 2, 0, FALSE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1385, 'The Tresses', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1386, 'The Chilled', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1387, 'The Snow Drift', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1388, 'Yellow Bunny Ears', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1389, 'Safari Helmet', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1394, 'The Reverie', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1395, 'The Funktastic', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1396, 'The Mystical', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1397, 'Spiked Warrior Helm', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1398, 'Winged Crimson Helm', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1399, 'The Majesty', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1400, 'Noble Helmet', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1401, 'The Giddy', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1402, 'The Thrill-Billy', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1404, 'The Chief', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1405, 'The Pro', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1406, 'The Newsie', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1407, 'The Reporter', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1408, 'The Foxtale', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1409, 'The Quarterback', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1410, 'The Light', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1411, 'The Dark', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1412, 'Iron Man Cowl', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1413, 'The Ms. Marvel', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1414, 'Nova Helmet', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1415, 'Loki Horns', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1416, 'Captain America Cowl', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1417, 'The Hawkeye', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1418, 'Thor Helmet', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1419, 'THE HULK SMASH', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1420, 'The Black Widow', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1421, 'Cop Cap', 2, 350, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1422, 'Press Cap', 2, 150, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1424, 'The Headspin', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1425, 'The Iconic', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1426, 'The Elusive', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1427, 'The Crowd Surf', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1428, 'The Stylish', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1429, 'The Handsome', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1430, 'Music Splatter Hat', 2, 150, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1431, 'The Blackbird', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1432, 'The Step Up', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1433, 'The Roll Bandana', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1437, 'The Lovely', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1438, 'The Fever', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1439, 'The Lavish', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1440, 'The Sandy Side', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1441, 'Apple Headband', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1442, 'Watermelon Headband', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1443, 'Pineapple Headband', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1444, 'Grape Headband', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1445, 'Watermelon Mohawk', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1446, 'Apple Headdress', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1447, 'Pineapple Headdress', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1448, 'Grape Headdress', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1449, 'The Melon Head', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1450, 'The Apple A Day', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1451, 'The Grape Vine', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1452, 'The Pineapple Crown', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1453, 'The Midnight', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1454, 'The Blast', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1476, 'Rockhopper''s Hat Aug 2012', 2, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1478, 'The Delight', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1480, 'The Mischief Maker', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1481, 'The Taa Daa', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1482, 'The All Dolled Up', 2, 140, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1483, 'The Cursed', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1484, 'The Punked Out', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1485, 'The Punk', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1486, 'The Grim', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1487, 'The Masquerade', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1488, 'The Gentleman', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1492, '7th Anniversary Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1493, 'Mystic Headdress', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1494, 'The Nine Lives', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1495, 'The Chamberlain', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1496, 'The Parlormaid', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1497, 'The Mysterious', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1498, 'The Secured', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1499, 'The Tough Enough', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1500, 'The Impossible', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1501, 'The Dashing', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1502, 'Puffle Care Cap', 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1504, 'The Golden Locks', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1505, 'The Gelled', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1506, 'Dark Blue Beanie', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1508, 'Sweet Ski Beanie', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1509, 'Cool Ski Beanie', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1510, 'Green Bobsled Helmet', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1511, 'Red Bobsled Helmet', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1512, 'The Craftsman', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1513, 'The Snowfall', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1514, 'The Present', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1515, 'Reindeer Handler Hat', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1516, 'Blue Lightbulb', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1517, 'Pink Lightbulb', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1518, 'Green Lightbulb', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1519, 'Yellow Lightbulb', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1520, 'Red Riding Hood', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1521, 'Rockhopper''s Santa Hat', 2, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1522, 'The Right On', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1523, 'The Too Cool', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1524, 'The Smashing', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1525, 'The Cruising', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1526, 'The Super Fly', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1527, 'The Sea Breeze', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1528, 'The Twilight Sky', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1529, 'Great Horn Mask', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1530, 'The Gale Storm', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1531, 'Caveguin Helmet', 2, 150, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1532, 'The First Fire', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1533, 'Prehistoric Helmet', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1534, 'Saber Toothed Helmet', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1535, 'The Yub Nub', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1536, 'The Grub Grub', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1537, 'Shale Cap', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1538, 'Rocky Headphones', 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1539, 'Green Ball Cap', 2, 200, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1540, 'The Free Ride', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1541, 'The Right Direction', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1543, 'The Concrete Wave', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1544, 'The Glamorous', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1545, 'Driver''s Cap', 2, 150, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1546, 'The Aspiring', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1547, 'The Lead', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1548, 'The All Star', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1549, 'The Among Stars', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1551, 'The Right Stuff', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1552, 'The Prom Perfect', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1553, 'The Double O', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1554, 'The Soulful Diva', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1555, 'Film Director Cap', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1556, 'The Willow Wisp', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1557, 'The Casual Classic', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1558, 'The Rainbow', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1559, 'The Tussled', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1560, 'Puffle Hotel Cap', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1562, 'Aunt Arctic''s Hat Feb 2013', 2, 0, FALSE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1563, 'The Electric Shock', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1564, 'The Stomping', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1565, 'Ooze Mask', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1566, 'Tornado Mask', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1567, 'The Sonic Blast', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1568, 'The Latest Scoop', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1569, 'The Headlines', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1570, 'Captain Marvel Helmet', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1571, 'Sif Helmet', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1572, 'The Mandarin', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1573, 'Odin Helmet', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1574, 'Iron Patriot Helmet', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1575, 'Silver Centurion Helmet', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1576, 'Mark 42 Helmet', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1577, 'Heartbreaker Helmet', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1578, 'War Machine Helmet', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1579, 'Nightclub Helmet', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1580, 'News Room Hat', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1581, 'Blizzard Helmet', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1582, 'Black Ice Headband', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1583, 'The Flurry', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1584, 'Patty''s Hat', 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1585, 'Police Helmet', 2, 250, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1586, 'The Sashimi Chef', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1587, 'The Snow Fall', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1588, 'Snow Headband', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1589, 'The Wind Chill', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1590, 'Iron Spider Mask', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1591, 'A-Bomb Mask', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1592, 'The Red Hulk', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1593, 'The She-Hulk', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1594, 'The Skaar', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1595, 'The Styled Messy', 2, 350, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1596, 'The Sleek Chic', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1597, 'Purple Toque', 2, 150, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1598, 'The Heartthrob', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1599, 'The Adored', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1600, 'Snow Beta Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1601, 'Sulley Mask', 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1602, 'Johnny Mask', 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1603, 'Randy Mask', 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1604, 'Hardscrabble Mask', 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1605, 'Brock Mask', 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1606, 'Claire Mask', 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1607, 'Don Mask', 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1608, 'George Mask', 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1609, 'Squishy Mask', 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1610, 'Squishy''s Mom Mask', 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1611, 'PNK Mask', 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1612, 'Chet Mask', 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1613, 'JOX Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1614, 'ROR Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1615, 'PNK Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1616, 'OK Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1617, 'One Eye CDA Mask', 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1618, 'Two Eye CDA Mask', 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1619, 'Three Eye CDA Mask', 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1620, 'The Free Spirit', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1621, 'The Keepin'' It Real', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1622, 'Sensei 2013 CJ Snow Player Card', 2, 0, FALSE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1623, 'Pink Monster Fins', 2, 100, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1624, 'Blue Monster Head Fin', 2, 100, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1625, 'Purple Monster Do', 2, 100, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1626, 'The Three Horned Green', 2, 100, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1627, 'Blue Monster Diva Mask', 2, 100, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1629, 'Monster Horns', 2, 100, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1630, 'The Slick', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1631, 'The Cap it & Forget it', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1632, 'The Summer Sidetied', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1633, 'The Easy Breezy', 2, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1634, 'The County Fair', 2, 250, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1635, 'C-3PO Mask', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1636, 'Chewbacca Mask', 2, 175, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1637, 'Darth Vader Helmet', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1638, 'The Han Solo', 2, 175, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1639, 'The Princess Leia', 2, 175, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1640, 'Stormtrooper Helmet', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1641, 'The Padawan', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1642, 'X-wing Helmet', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1643, 'Rebel Helmet', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1644, 'Boba Fett Helmet', 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1645, 'The Skywalker', 2, 175, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1646, 'Greedo Mask', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1647, 'Tusken Raider Mask', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1648, 'Jawa Mask', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1650, 'Imperial Officer Hat', 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1652, 'Emperor Palpatine Mask', 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1653, 'The Fusion', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1655, 'The Electric Tunage', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1656, 'The Flipster', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1657, 'The Side Sweep', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1658, 'The Drama', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1660, 'The Cheechee', 2, 175, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1661, 'The Lela', 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1662, 'The Butchy', 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1663, 'The Brady', 2, 100, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1664, 'The Tanner', 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1665, 'The McKenzie', 2, 175, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1666, 'The Seacat', 2, 175, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1667, 'The Big Momma', 2, 175, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1668, 'Butchy''s Bike Helmet', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1669, 'Lela''s Bike Helmet', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1670, 'The Blonde Beehive', 2, 175, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1671, 'The Giggles', 2, 175, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1672, 'Retro Cap', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1682, 'Brady', 2, 0, FALSE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1683, 'McKenzie', 2, 0, FALSE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1684, 'The Maiden', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1685, 'The Elfin', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1686, 'The Snow Princess', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1687, 'The Villager', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1688, 'Knightly Helmet', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1689, 'The Classy', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1690, 'The Back-to-School', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1691, 'Apprentice Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1692, 'RH''s Jacket Player Card', 2, 0, FALSE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1693, 'Epic Knight Helmet', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1695, 'Horned Ogre Head', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1696, 'Tusked Ogre Head', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1697, 'The Enchantment', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1698, 'Wizardly Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1701, 'Cadence''s Aug 2013 Hair', 2, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1702, 'Swamp Monster Mask', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1704, 'The Rebel', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1705, 'The Countess', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1706, 'The Spider Silk', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1707, 'The Creepy Count', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1710, 'The Punktails', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1711, 'Springy Pumpkins', 2, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1712, 'Bat Wing Headband', 2, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1713, 'Wrapper Head', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1714, 'Tomb King Hat', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1715, 'The Villain Do', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1716, '8th Anniversary Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1717, 'Pumpkin Cap', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1718, 'Pumpkin Head', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1720, 'Rookie''s Hat Oct 2013', 2, 9999, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1721, 'The Gretel', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1722, 'The Huntress', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1723, 'The Idol', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1724, 'The Dude', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1725, 'The Shock Top', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1726, 'Cozy Hat', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1727, 'The Curly Pie', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1730, 'Head Lamp', 2, 100, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1731, 'Snow Goggles', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1732, 'Penguin Brain Box', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1733, 'The Esquire', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1734, 'Solid Gold Viking Helmet', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1735, 'Gold Top Hat', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1736, 'PH''s Hat (Golden Puffle)', 2, 0, FALSE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1737, 'Herbert Mascot Card Nov 2013', 2, 9999, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1738, 'The Snow Day', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1739, 'Snow Day Toque', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1740, 'The Holiday Topper', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1741, 'Caroler''s Bonnet', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1745, 'Toy Robot Helmet', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1748, 'Jack-in-the-Box Hat', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1749, 'Holiday Conductor Hat', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1750, 'Train Engineer Hat', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1751, 'Jolly Roger Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1752, 'Teddy Bear Head', 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1753, 'Rockhopper''s Santa Hat', 2, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1754, 'Sea Foam Earmuffs', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1755, 'The Tiki Tiki', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1756, 'Primal Headdress', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1757, 'The Top Knot', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1758, 'The Aboo Aboo', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1759, 'Redhead Headphones', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1760, 'The Eco-Brunette', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1761, 'The Ugga Ugga', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1762, 'The Courageous', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1763, 'The Baba Baba', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1764, 'Primal Helmet', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1765, 'Rock Hard Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1766, 'Stone Chef Hat', 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1767, 'Prehistoric Beta Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1768, 'The Bonytail', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1769, 'The Side Bonytail', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1770, 'Tour Guide Headband', 2, 9999, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1771, 'Dino Horns', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1772, 'Welcome Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1774, 'The Bob 3000', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1775, 'The Spike 3000', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1776, 'The Ringlets', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1777, 'Pork Pie Hat', 2, 900, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1778, 'Carnival Jester Hat', 2, 1000, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1779, 'Carnival Party Hat', 2, 1000, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1780, 'Twinkle Twinkle Hat', 2, 1000, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1781, 'Cool Down Hat', 2, 1000, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1782, 'Puffle Wrangler Hat', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1783, 'Rookie''s Hat New 2014 Player Card', 2, 9999, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1784, 'The Jetsetter', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1785, 'The High Flyer', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1786, 'Straw Sun Hat', 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1787, 'The Getaway', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1788, 'Fozzie Bear Head', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1789, 'Kermit the Frog Head', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1790, 'Miss Piggy Head', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1791, 'The Great Gonzo Head', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1792, 'Beaker Head', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1793, 'Animal Head', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1795, 'Sam Eagle Head', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1796, 'Swedish Chef Head', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1797, 'Sombrero Cordobés', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1798, 'Alpine Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1799, 'Beret', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1800, 'Bowler Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1801, 'Turban', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1802, 'Mexican Sombrero', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1803, 'Sambista', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1804, 'Ushanka', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1805, 'Kermit''s Mascot Player Card', 2, 9999, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1806, 'AA', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1807, 'AB', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1808, 'AC', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1809, 'AD', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1810, 'AE', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1811, 'AF', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1812, 'AG', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1813, 'AH', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1814, 'AI', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1815, 'AJ', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1816, 'AK', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1817, 'AL', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1818, 'AM', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1819, 'AN', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1820, 'AO', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1821, 'AP', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1822, 'AQ', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1823, 'AR', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1824, 'AS', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1825, 'AT', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1826, 'AU', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1827, 'AV', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1828, 'AW', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1829, 'AX', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1830, 'AY', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1831, 'AZ', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1832, 'BA', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1834, 'Constantine Head', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1835, 'The Indie-go', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1836, 'The Shade Do', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1837, 'The Up Sweep', 2, 250, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1838, 'The Archaeologist', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1839, 'Junior Explorer Hat', 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1840, 'The Orange Tomcat', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1841, 'The Orange Kittytail', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1842, 'The Blue Doggone', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1843, 'The Blue Floppy Ears', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1844, 'The Goldilocks', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1845, 'The Lux', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1846, 'The Rainbow Curls', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1847, 'The Rainbow Sweep', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1848, 'Orange Tabby Headband', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1849, 'Blue Border Collie Headband', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1850, 'Pufflescape Ball Cap', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1851, 'Vet''s Head Mirror', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1852, 'Lavender Bunny Ears', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1853, 'Bear Ears', 2, 50, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1855, 'The Mullet', 2, 300, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1856, 'Donut Conqueror', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1857, 'Salami Sun Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1858, 'The Plunge', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1859, 'The Sleepy Head', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1861, 'The Astro Pomp', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1862, 'The Wave', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1863, 'The Metropolitan', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1864, 'The Futura', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1865, 'The Fashion Fluff', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1866, 'The Tubular', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1867, 'The Holo Headphones', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1868, 'The Transmitter', 2, 100, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1869, 'Final Frontier Helmet', 2, 100, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1870, 'Zaryax VI Hat', 2, 150, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1871, 'Asteroid Head', 2, 150, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1872, 'Cyborg Hat', 2, 100, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1873, 'The Galaxy Gal', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1874, 'Protobot Helmet', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1875, 'Gary 3000 Helmet (Robo Party)', 2, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1876, 'The Fan Fro', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1877, 'Penguin Cup Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1879, 'The Soccer Star', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1880, 'Graduation Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1881, 'Police Hat', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1884, 'The Frontman', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1885, 'The Rows', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1886, 'The Summer Jam', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1887, 'The Flower Child', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1888, 'The Replay', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1889, 'Loose Toque', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1890, 'Cruise Captain Hat', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1891, 'Beat Dropper Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1894, 'Go-Karter Helmet', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1895, 'The Coronation', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1896, 'The Winter Traveler', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1897, 'The Crowned Queen', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1898, 'The Ice Queen', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1899, 'The Oaken', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1900, 'The Ice Seller', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1901, 'The Thirteenth Son', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1902, 'Troll Mask', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1903, 'Rustic Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1904, 'Personal Flurry', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1905, 'Ice Party Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1906, 'Spiked Skateboard Helmet', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1907, 'Skater Pigtails', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1908, 'Class Clown Helmet', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1910, 'Security Guard Hat', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1911, 'The Skate Shopper', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1912, 'The Kickflip', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1913, 'The Platinum', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1914, 'Pro Skater Helmet', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1915, 'Brain Freeze', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1916, 'The Darkslide', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1917, 'Sam', 2, 0, FALSE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1918, 'Franky''s Playercard 2014', 2, 9999, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1919, 'G Billy''s Playercard 2014', 2, 9999, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1920, 'Stompin'' Bob Playercard 2014', 2, 9999, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1921, 'Pete K''s Playercard 2014', 2, 9999, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1924, 'Gary''s Glasses (evergreen)', 3, 9999, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1925, 'The Shadow Tail', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1926, 'The Surf Star', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1927, 'Werewolf Ears', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1928, 'The Enchantress', 2, 275, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1929, 'The Shadowy Spike', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1930, 'The Spider Poof', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1931, 'Bellhop Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1932, 'Magic Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1933, '9th Anniversary Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1934, 'Orange Skate Spike Helmet', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1935, 'Brunette Skater Pigtails', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1936, 'Straw Fedora', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1937, 'The Strawberry Braid', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1938, 'The Popstar', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1939, 'The Savvy', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1940, 'The Pirate Queen', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1941, 'The Pearl', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1942, 'The Duelist', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1943, 'The Paradise Punk', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1944, 'The Swabbie', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1945, 'The Bosun', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1946, 'The Castaway', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1947, 'PH Wild Puffle Playercard', 2, 9999, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1948, 'The Washed-Up', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1949, 'The Scallywag', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1950, 'Golden Pirate Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1951, 'The Continental', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1952, 'The Squid Lid', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1953, 'Festive Hat', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1954, 'Purple Lightbulb', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1955, 'The Winter Land', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1956, 'The Tundra', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1957, 'The Cold Snap', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1958, 'The Summit', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1959, 'First Merry Walrus Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1960, 'Blue Holiday Cap', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1961, 'Big Bell', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1962, 'The Festive Fluffy', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1963, 'Crystal Puffle Earmuffs', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1964, 'Merry Walrus', 2, 9999, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1965, 'The Ezra', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1966, 'The Kanan', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1967, 'The Sabine', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1968, 'The Zeb', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1969, 'The Hera', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1970, 'The Inquisitor', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1971, 'The Kallus', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1972, 'Vizago Mask', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1973, 'Wookiee Mask', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1974, 'PH puffle wild player card', 2, 9999, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1975, 'Minister Tua''s Helmet', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1976, 'Imperial Trooper Helmet', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1977, 'StormtrooperBOT', 2, 9999, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1978, 'KananBOT', 2, 9999, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1979, 'The Party Style', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1980, 'The Fuchsia Fringe', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1981, 'The Platinum Artist', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1982, 'The Electro Rocker', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1983, 'The Punkhawk', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1984, 'The Rocker-chic', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1985, 'The Golden Sweep', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1986, 'The Hip-Hop', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1987, 'Green Raccoon Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1988, 'Red Deer Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1989, 'White Rabbit Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1990, 'Yellow Unicorn Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1991, 'Puffle Guide Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1992, 'The Undercut', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1993, 'The Retro', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1995, 'The Legend', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1996, 'Golden Goggles', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1997, 'The Spring Bun', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1998, 'The Frozen Flowers', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (1999, 'Remove Head Item', 2, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2001, 'Green Giant Sunglasses', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2002, 'Yellow Giant Sunglasses', 3, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2003, 'Black Scuba Mask', 3, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2004, 'Blue Alien Mask', 3, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2007, 'Aunt Arctic''s Glasses', 3, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2009, 'Senseis Beard', 3, 0, FALSE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2010, 'Gold Sunglasses', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2011, 'Aviator Goggles', 3, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2012, 'Curly Mustache', 3, 50, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2013, 'Lava Mask', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2014, 'Humbug Spectacles', 3, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2016, 'Funny Pig Snout', 3, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2017, 'Guardian Dog Muzzle', 3, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2018, 'Green Face Paint', 3, 15, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2019, 'Yellow Face Paint', 3, 15, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2020, 'Spy Goggles', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2021, 'Delta Shades', 3, 5, TRUE, TRUE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (2022, 'Alpha Shades', 3, 5, TRUE, TRUE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (2023, 'Adventure Face Paint', 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2025, 'Torrent Mask', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2026, 'Red Face Paint', 3, 15, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2027, 'Blue Face Paint', 3, 15, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2028, 'Santa Beard', 3, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2029, 'Gray Beard', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2030, 'Rookies Sunglasses', 3, 0, FALSE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2031, 'Petey K Shades', 3, 65, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2032, 'Pink Diva Shades', 3, 125, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2034, 'Petey K Shades', 3, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2035, 'Castaway Face Paint', 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2036, 'Sleet Stopper', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2037, 'Snow Stopper', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2038, 'Slush Stopper', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2039, 'Mime Face Paint', 3, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2040, 'Mustachio', 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2043, 'The Burly Beard', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2044, 'Blue Aviator Shades', 3, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2045, 'Blue Starglasses', 3, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2046, 'Indigo Sunglasses', 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2047, 'Sunset Diva Glasses', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2048, 'Emerald Aviators', 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2049, 'Pink Face Paint', 3, 15, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2050, 'Black Face Paint', 3, 15, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2051, 'Purple Face Paint', 3, 15, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2052, 'White Face Paint', 3, 15, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2053, 'Orange Face Paint', 3, 15, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2054, 'Brown Face Paint', 3, 15, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2057, 'Golden Shades', 3, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2059, 'Pink Starglasses', 3, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2060, 'Mask of Justice', 3, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2061, 'Fiendish Mask', 3, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2062, 'Valiant Mask', 3, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2063, 'Sinister Mask', 3, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2064, 'Iron Fist Mask', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2065, 'Hawkeye-wear', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2066, 'Nick Fury Eyepatch', 3, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2068, 'Orange Aviator Shades', 3, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2069, 'Giant White Sunglasses', 3, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2070, 'Watermelon Tiki Paint', 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2071, 'Apple Tiki Paint', 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2072, 'Grape Tiki Paint', 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2073, 'Pineapple Tiki Paint', 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2074, 'Real Teal Sunglasses', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2075, 'Star Shades', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2080, 'Creepy Doll Makeup', 3, 120, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2081, 'Spectacles', 3, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2082, 'Vampire Fangs', 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2083, 'Ghost Goggles', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2084, 'The Mystery', 3, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2085, 'The Golden Secret', 3, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2086, 'The Phantom', 3, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2087, 'Gary''s Glasses', 3, 0, FALSE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2088, 'Skull Mask', 3, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2089, 'Butler Brow', 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2090, 'Old Maid Makeup', 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2091, 'Mystic Makeup', 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2092, 'Cat Eyes', 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2093, 'Curly Mustache', 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2094, 'Shady Shades', 3, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2095, 'Sleek Shades', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2096, 'Stealthy Shades', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2097, 'Smoke Goggles', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2100, 'Craftsman Spectacles', 3, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2101, 'Majestic Makeup', 3, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2102, 'Gary''s Glasses (Prehistoric Party)', 3, 0, FALSE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2103, 'Prehistoric Tusks', 3, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2104, 'Caveguin Face Paint', 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2105, 'Biggest Brow', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2106, 'Grillz', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2107, '2 Cool Glasses', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2108, 'Diva Glam Eyes', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2109, 'Kitty Cat Eyes', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2110, 'Wide Awake Eyes', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2111, 'Starlet Eyes', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2112, 'Red Nose', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2113, 'Gary''s Glasses Feb 2013', 3, 0, FALSE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2114, 'Black Mask', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2115, 'Stomp Mask', 3, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2116, 'Telekinesis Mask', 3, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2117, 'Sonic Blast Mask', 3, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2119, 'Icy Mask', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2120, 'Cat Burglar  Mask', 3, 100, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2121, 'Police Aviators', 3, 150, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2122, 'Art Costume', 3, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2124, 'Leia''s Lashes', 3, 75, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2125, 'Cat Eye Sunglasses', 3, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2126, 'Bubble Gum', 3, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2128, 'Gary''s Glasses (sept 2013 PC)', 3, 9000, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2129, 'Countess Makeup', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2130, 'Count Mask', 3, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2131, 'Ghoulish Mask', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2132, 'The Long Haired Ghoul', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2134, 'G-Tech Glasses', 3, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2135, 'Survival Snorkel', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2136, 'Gold Diva Shades', 3, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2137, 'Gold D Glasses', 3, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2138, 'Goldstache', 3, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2139, 'Golden Unicorn Horn', 3, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2141, 'Jack-in-the-Box Face', 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2142, 'Holiday Conductor Specs', 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2143, 'Mud Makeup', 3, 75, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2144, 'Stone Shades', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2145, 'Cavegeek Glasses', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2146, 'Primal Blue Face Paint', 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2147, 'Primal Pink Face Paint', 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2148, 'Primal Green Face Paint', 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2149, 'Pixel Shades', 3, 1500, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2150, 'Starshine Makeup', 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2151, 'Interstellar Makeup', 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2152, 'Metropolis Makeup', 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2153, 'Utopia Makeup', 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2154, 'Hot Sauce Face Paint', 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2155, 'Fluffies Face Paint', 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2156, 'Space Squids Face Paint', 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2157, 'Sharks Face Paint', 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2158, 'Indie Rocker Glasses', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2159, 'Royal Eyelashes', 3, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2160, 'Icy Eyelashes', 3, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2161, 'Lab Goggles', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2162, 'Vampire Grin', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2163, 'Skeleton Mask', 3, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2164, 'Grey Glasses', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2165, 'Fancy Lashes', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2166, 'Winged Eyeliner', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2167, 'Rhinestone Sunglasses', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2168, 'Red Pixel Glasses', 3, 100, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2169, 'Delta Shades', 3, 0, TRUE, FALSE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (2170, 'Alpha Shades', 3, 0, TRUE, FALSE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (2171, 'Daisy Glasses', 3, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2172, 'Spy Visor', 3, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2173, 'Gold Chandelier Earrings', 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2174, 'Lashful Eyes', 3, 75, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2176, 'Thin Mustache', 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2177, 'Mal''s Makeup', 3, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2178, 'Clown Face Paint', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2179, 'Robot Antenna', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2180, 'Western Mustache', 3, 50, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2181, 'Teacher Mustache', 3, 50, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2182, 'Artist Mustache', 3, 50, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2183, 'Glam Glasses', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2184, 'Black Diva Shades', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2185, 'Silver Aviators', 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (2999, 'Remove Face Item', 3, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3000, 'Beaded Necklace', 4, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3001, 'Jade Necklace', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3002, 'Silver Whistle', 4, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3003, 'Gold Medal', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3004, 'Vinyl Messenger Bag', 4, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3005, 'Skinny Blue Tie', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3006, 'Faery Wings', 4, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3007, 'Yellow Cape', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3008, 'Twee''s Wings', 4, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3009, 'Green Parrot', 4, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3011, 'Cadence Scarf', 4, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3012, 'Blue Striped Scarf', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3013, 'Pink Designer Scarf', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3014, 'Accordion', 4, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3015, 'White Bowtie', 4, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3016, 'Petey''s Accordian', 4, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3018, 'Green Trendy Scarf', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3019, 'Blue Designer Scarf', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3020, 'Green Hooded Cloak', 4, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3021, 'Torc', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3022, 'Compass', 4, 124, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3023, 'Tropical Bird', 4, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3024, 'All Access Pass', 4, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3025, 'Golden Fairy Wings', 4, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3026, 'Big White Scarf', 4, 175, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3027, 'Test Glider', 4, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3028, 'Jet Pack', 4, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3029, 'Shell Collar', 4, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3030, 'Flower Messenger Bag', 4, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3031, 'Flame Messenger Bag', 4, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3032, 'Amulet', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3033, 'Snow Fairy Wings', 4, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3034, 'Butterfly Wings', 4, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3035, 'Blue Striped Scarf', 4, 175, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3036, 'Moth Wings', 4, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3037, 'White Fuzzy Beard', 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3038, 'Tri-color Scarf', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3039, 'Pom Pom Scarf', 4, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3040, 'Smitten Scarf', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3041, 'Phoenix Wings', 4, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3043, 'Checkered Tote', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3044, 'Green Cotton Scarf', 4, 175, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3046, 'Gold Chain', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3047, 'Noble Horse', 4, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3048, 'Conga Drums', 4, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3049, 'Surf Necklace', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3050, 'All Access Pass', 4, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3052, 'Blue Climbing Rope', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3053, 'Yellow Climbing Rope', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3054, 'Expedition Backpack', 4, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3055, 'H2O Pack', 4, 15, TRUE, TRUE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (3056, 'Magician''s Cloak', 4, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3058, 'The Count''s Cloak', 4, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3061, 'Sat-Pack', 4, 15, TRUE, TRUE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (3062, 'Santa''s Present Bag', 4, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3063, 'Tech Satchel', 4, 10, TRUE, TRUE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (3064, 'Popcorn Tray', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3065, 'Blue Patched Bag', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3066, 'Polka-Dot Bandanna', 4, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3067, 'Royal Blue Robe', 4, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3068, 'Red Tabard', 4, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3069, 'Yellow Tabard', 4, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3070, 'Blue Tabard', 4, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3071, 'White Noble Horse', 4, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3073, 'Blue Power Cell', 4, 20, TRUE, FALSE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (3074, 'Raindrop Necklace', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3075, 'Pink Zebra Scarf', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3076, 'Bumblebee Scarf', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3077, 'Noteworthy Necklace', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3078, 'Orange Feather Boa', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3079, 'All-Access Pass', 4, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3080, 'Blue Snare Drum', 4, 380, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3081, 'Blue Accordion', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3082, 'Petey K Accordion', 4, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3083, 'Tundra Board', 4, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3084, 'Electric Pink Snowboard', 4, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3085, 'Blast-Off Board', 4, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3086, 'Abracadabra Cape', 4, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3087, 'Squid Hug', 4, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3088, 'Purple Butterfly Wings', 4, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3089, 'Pegasus Wings', 4, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3090, 'Raven Wings', 4, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3091, 'Purple Rugby Scarf', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3092, 'Seeing Spots Scarf', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3093, 'Supreme Toy Sack', 4, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3094, 'Magenta Scarf', 4, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3095, 'Bronze Music Note Necklace', 4, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3096, 'Blue Scuba Tank', 4, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3097, 'Aqua Bead Necklace', 4, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3098, 'Cornflower Scarf', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3099, 'Red Cotton Scarf', 4, 175, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3100, 'White Feather Boa', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3101, 'Silver Star Necklace', 4, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3102, 'Green Cape', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3103, 'Yellow Cape', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3104, 'Purple Cape', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3105, 'White Cape', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3106, 'Orange Cape', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3107, 'Brown Cape', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3108, 'Red Feather Boa', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3109, 'Blue Feather Boa', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3110, 'Pink Feather Boa', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3111, 'Black Feather Boa', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3112, 'Green Feather Boa', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3113, 'Yellow Feather Boa', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3114, 'Purple Feather Boa', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3115, 'Brown Feather Boa', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3116, 'Green Zebra Scarf', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3117, 'Purple Beaded Necklace', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3118, 'Sunset Scarves', 4, 175, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3119, 'Golden Wings', 4, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3120, 'Battle Cape', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3121, 'Video Camera', 4, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3122, 'Feather Necklace', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3123, 'Pauldrons of Justice', 4, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3124, 'Fiendish Pauldrons', 4, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3125, 'Hawkeye Quiver & Bow', 6, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3126, 'Toothy Necklace', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3127, 'Kukui Nut Necklace', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3128, 'Clam Shell Collar', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3133, 'Green Butterfly Wings', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3135, 'Brown Bat Wings', 4, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3136, 'Sea Foam Pearls', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3138, 'Plasma Laser', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3139, 'Leather Satchel', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3140, 'Puffle Care Sash', 4, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3141, 'Gold Charm Necklace', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3144, 'Heather Grey Scarf', 4, 175, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3145, 'Chestnut Checker Scarf', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3146, 'Sleigh Bells', 4, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3147, 'Cozy Orange Scarf', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3148, 'Thick Hide Cloak', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3150, 'Prehistoric Necklace', 4, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3151, 'Great Bone Cloak', 4, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3152, 'Stone Bow Tie', 4, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3153, 'Stony Necklace', 4, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3154, 'Red Bandana', 4, 150, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3155, 'Lavender Knit Scarf', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3156, 'Black Shoulder Bag', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3157, 'Digital Camera', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3158, 'Movie Camera', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3159, 'Spa Towel', 4, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3160, 'Tusk''s Cloak', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3161, 'Nautical Necklace', 4, 100, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3162, 'Tags', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3163, 'Daisy Chain', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3164, 'Hardscrabble''s Wings', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3166, '14K Fish Necklace', 4, 110, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3167, 'Green Satchel', 4, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3168, 'Yellow Monster Wings', 4, 100, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3169, 'Jedi Cloak', 4, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3170, 'Black Jedi Cloak', 4, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3171, 'Rebel Reward Medal', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3172, 'Obi-Wan Cloak', 4, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3173, 'Sandtrooper Pauldron', 4, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3174, 'McKenzie''s Necklace', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3175, 'Friends Forever Lei', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3176, 'Enchanted Dragon Wings', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3177, 'Enchanted Wings', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3178, 'Medieval Cloak', 4, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3179, 'Edgy Necklace', 4, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3180, 'Vampire Cape', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3181, 'Dog Tags', 4, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3183, 'Gold Puffle Chain', 4, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3184, 'Stethoscope', 4, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3185, 'Golden Lei', 4, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3186, 'Royal Golden Robe', 4, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3187, 'Gold Jet Pack', 4, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3189, 'Green Plaid Scarf', 4, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3190, 'Toy Sack', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3191, 'Marching Band Drum', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3192, 'Snowflake Bandana', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3193, 'Jack-in-the-Box Ruffle', 4, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3194, 'Rockhopper''s Holiday 2013 Beard', 4, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3195, 'Stone Necklace', 4, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3197, 'Constantine''s Cloak', 4, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3198, 'Tourist Camera', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3199, 'Fringed Purse', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3200, 'Explorer''s Bag', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3201, 'O''Berry Necklace', 4, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3202, 'Gold Feather Boa', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3203, 'Rainbow Feather Boa', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3205, 'Gold Cape', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3206, 'Rainbow Cape', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3207, 'Space Cadet Jetpack', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3208, 'Penguin Cup VIP Pass', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3209, 'Unbeatable Keeper Costume', 4, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3210, 'DJ Headphones', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3211, 'Music Cruise Pass', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3212, 'Coronation Necklace', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3213, 'Troll Cape', 4, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3214, 'Coral Shoulder Bag', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3215, 'Skate Satchel', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3216, 'Limited Edition Backpack', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3217, 'Leopard Print Scarf', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3218, 'Jagged Purple Necklace', 4, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3219, 'Shadow Wings', 4, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3221, 'Wheel Necklace', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3222, 'Autumn Scarf', 4, 175, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3224, 'String of Shells', 4, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3225, 'Leather Purse', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3226, 'Houndstooth Bag', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3227, 'Plaid Shell Purse', 4, 175, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3228, 'Knitted Scarf', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3229, 'Checkered Scarf', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3230, 'Platinum Note Necklace', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3231, 'Gold Necklace', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3232, 'Dubstep Puffle Bling', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3233, 'Green Butterfly Wings', 4, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3234, 'Disgust Neckerchief', 4, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3235, 'Beach Chain', 4, 55, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3236, 'Jellyfish Necklace', 4, 55, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3238, 'Orange Tie', 4, 55, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3239, 'Priceless Necklace', 4, 55, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3240, 'Jet Pack', 4, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3241, 'Infinity Scarf', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3242, 'Coffee Bean Necklace', 4, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3243, 'Classic Car', 4, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3244, 'Cozy Blue Scarf', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3245, 'Double Stranded Necklace', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3246, 'Wilderness Tent', 4, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3247, 'Island Lei', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3248, 'Tattered Wings', 4, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3249, 'VIP Pass', 4, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3250, 'Music Costume', 4, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3251, 'Holiday Tie', 4, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3252, 'Jolly Holly Wreath', 4, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3253, 'Winter Tent', 4, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3254, 'Winter Camo Jetpack', 4, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3255, 'Holi-lei', 4, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3256, 'Holiday Handbag', 4, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3257, 'Sunken Treasure', 4, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (3999, 'Remove Neck Item', 4, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4000, 'Glacier Suit', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4001, 'Sunset Dress', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4002, 'Blue Cheerleading Sweater', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4003, 'Red Cheerleading Sweater', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4004, 'Red Track Jacket', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4005, 'Red Track & Field Uniform', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4006, 'Blue Track & Field Uniform', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4007, 'Layered Lavender Outfit', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4008, 'Sea Green Outfit', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4009, 'Blue Hoodie', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4010, 'Tweed Coat', 5, 650, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4011, 'Painter''s Overalls', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4012, 'Detective''s Coat', 5, 650, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4013, 'Blue Zoot Suit', 5, 700, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4014, 'Doorman''s Jacket', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4015, 'Dark Detective''s Coat', 5, 650, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4016, 'Frankenpenguin Costume', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4017, 'Faery Costume', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4018, 'Rad Scientist Costume', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4019, 'Sea-Worthy Dress', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4020, 'Sea-Worthy Suit', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4021, 'Blue Alien Costume', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4022, 'Garys Coat', 5, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4023, 'Green Vest', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4024, 'Purple Vest', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4025, 'White Ninja Belt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4026, 'Yellow Ninja Belt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4027, 'Orange Ninja Belt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4028, 'Green Ninja Belt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4029, 'Blue Ninja Belt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4030, 'Red Ninja Belt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4031, 'Purple Ninja Belt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4032, 'Brown Ninja Belt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4033, 'Black Ninja Belt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4034, 'Ninja Outfit', 5, 1000, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4035, 'White Parka', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4036, 'Prince Redhood''s Costume', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4037, 'Grumpunzel''s Dress', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4038, 'Twee''s Costume', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4039, 'Big Bad Wool Suit', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4040, 'Purple Figure Skating Dress', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4041, 'Snowman Body', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4042, 'Gingerbread Costume', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4043, 'Yellow Winter Jacket', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4044, 'Green Winter Jacket', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4045, 'Baker''s Apron', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4046, 'Red Hoodie', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4047, 'Spikester Threads', 5, 650, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4048, 'Freestyle Threads', 5, 650, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4049, 'Yellow Fiesta Dress', 5, 640, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4050, 'Blue Torero Suit', 5, 520, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4053, 'Red Goalie Gear', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4054, 'Blue Goalie Gear', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4055, 'Racing Swimsuit', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4056, 'Mountain Climber Gear', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4057, 'Yellow Shirt', 5, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4058, 'Blue Shirt', 5, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4059, 'Green Shirt', 5, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4060, 'Pink Shirt', 5, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4061, 'Green Striped Rugby Shirt', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4062, 'Pink Striped Rugby Shirt', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4063, 'Pet Shop Apron', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4064, 'Swashbuckler Outfit', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4065, 'Blue Tuxedo', 5, 800, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4066, 'Lavender Gown', 5, 800, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4067, 'Blue Dazzle Dress', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4068, 'Black Party Dress', 5, 700, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4069, 'White Tuxedo', 5, 800, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4070, 'Pink Tennis Outfit', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4071, 'Orange Tennis Outfit', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4072, 'Green Baseball Uniform', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4075, 'White Gi', 5, 800, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4076, 'Blue CP Hoodie Outfit', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4077, 'Green Quilted Coat', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4078, 'Golden Knight''s Armor', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4079, 'Blue Doublet', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4080, 'Amethyst Dress', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4081, 'Topaz Dress', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4082, 'Blue Dragon Costume', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4083, 'Lady''s Gown', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4084, 'Green Tunic', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4085, 'Pink Viking Dress', 5, 475, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4086, 'Brown Viking Costume', 5, 475, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4087, 'Purple Wetsuit', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4088, 'Green Soccer Jersey', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4089, 'Red Hiking Outfit', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4090, 'Pink Hiking Outfit', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4091, 'Goalie Jersey', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4092, 'Yellow Summer Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4093, 'Orange Hawaiian Outfit', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4094, 'Blue Star Swimsuit', 5, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4095, 'Blue Board Shorts', 5, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4096, 'Black Hawaiian Shorts', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4097, 'Green Flower Bikini', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4098, 'Pop Princess Outfit', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4099, 'Pink Polka-dot Dress', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4100, 'Rocker Outfit', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4101, 'Dazzling Blue Tux', 5, 850, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4102, 'Electro T-Shirt', 5, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4103, 'Sound Wave T-Shirt', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4106, 'Fairy Flight Outfit', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4107, 'Pilot''s Jacket', 5, 380, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4108, 'Orange Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4109, 'Splatter T-Shirt', 5, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4110, 'Merpenguin Fin', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4111, 'Lobster Costume', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4112, 'Wise Fish Costume', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4113, 'Blue Racing Bathing Suit', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4114, 'Green Wetsuit', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4115, 'Green Football Jersey', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4116, 'Green Cheerleader Outfit', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4117, 'Girl''s Sweater Vest', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4118, 'Boy''s Sweater Vest', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4119, 'Ring Master Outfit', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4120, 'Magma Coat', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4121, 'Waterfall Coat', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4123, 'Snow Fairy Dress', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4126, 'Santa Suit', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4127, 'Butterfly Dress', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4128, 'Blizzard Wizard Robe', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4129, 'Ladybug Suit', 5, 330, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4130, 'Purple Carapace', 5, 330, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4131, 'Fuzzy Experiment', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4132, 'Goldsmith Apron', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4133, 'Black Whirlpool Snowsuit', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4134, 'Pink Winter Peacoat', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4135, 'Snail Costume', 5, 700, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4136, 'Gnome Costume', 5, 375, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4137, 'Pretty as a Petal Dress', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4138, 'Spider Costume', 5, 680, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4139, 'Green Figure Skating Dress', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4140, 'Pink Figure Skating Dress', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4141, 'Yeti Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4142, 'Silly Scarecrow Suit', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4143, 'Green Hockey Jersey', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4144, 'Green Goalie Gear', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4145, 'Reindeer Costume', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4146, 'Humbug Coat', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4147, 'Tree Costume', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4149, 'Cozy Winter Coat', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4154, 'Beta Hat T-Shirt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4155, 'Blue-Sky Dress', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4156, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4157, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4158, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4159, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4160, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4161, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4162, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4163, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4164, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4165, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4166, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4167, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4168, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4169, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4170, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4171, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4172, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4173, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4174, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4175, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4176, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4177, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4178, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4179, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4180, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4181, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4186, 'Puffle Pullover', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4187, 'Pink Quilted Coat', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4188, 'Orange Vest', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4189, 'Purple Spring Dress', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4190, 'Monkey King Costume', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4191, 'Phoenix Dress', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4192, 'Funny Pig Costume', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4193, 'Deep Sea Diving Suit', 5, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4194, 'Guardian Dog Costume', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4195, 'Puffle Jacket', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4196, 'Buttercup Ball Gown', 5, 375, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4197, 'Petal Pattern Dress', 5, 325, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4198, 'Ruffle Dress', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4199, 'Classy T-Shirt', 5, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4200, 'Admirals Coat', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4204, 'Yellow Hockey Jersey', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4205, 'Yellow Cheerleader Outfit', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4206, 'Water Suit 3000', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4207, 'Green Recycle Shirt', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4208, 'Knotted Recycle T-Shirt', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4209, 'White Cocoa Bunny Costume', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4210, 'Rustic Tunic and Skirt', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4211, 'City Top and Jacket', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4215, 'Maiden''s Gown', 5, 425, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4216, 'Damsel''s Dress', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4217, 'Duchess'' Dress', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4218, 'Duke''s Outfit', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4219, 'Iron Armor', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4220, 'Squire Outfit', 5, 325, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4221, 'King''s Outfit', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4222, 'EPF Suit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (4223, 'Delta Suit', 5, 20, TRUE, TRUE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (4224, 'Alpha Suit', 5, 20, TRUE, TRUE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (4225, 'Seafarer''s Gown', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4226, 'First Mate''s Outfit', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4227, 'Pirate Lass', 5, 425, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4228, 'Swashbuckler''s Coat', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4229, 'Commander''s Outfit', 5, 430, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4230, 'Coral Mermaid Costume', 5, 425, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4231, 'Tropical Mermaid Costume', 5, 425, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4232, 'Castaway''s Clothing', 5, 325, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4233, 'Marooned Outfit', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4234, 'Sunset Surf Shorts', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4235, 'Summer Threads', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4236, 'Floral Bikini', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4237, 'Yellow Pop Outfit', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4238, 'Music Jam Shirt', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4241, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4242, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4243, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4244, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4245, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4246, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4247, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4248, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4249, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4250, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4251, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4252, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4253, 'Custom T-shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4254, 'Yellow Expedition Jacket', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4255, 'Blue Expedition Jacket', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4256, 'Khaki Expedition Jacket', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4257, 'Blue Duck', 5, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4258, 'Tactical Gear', 5, 20, TRUE, TRUE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (4259, 'Yellow Hoodie', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4260, 'Waddle On Hoodie', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4261, 'Magenta Sunset Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4262, 'Aqua Striped T-Shirt', 5, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4263, 'Red Tracksuit', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4264, 'Blue Tracksuit', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4265, 'Green Tracksuit', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4266, 'Yellow Tracksuit', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4269, 'Clown Suit', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4270, 'The Count''s Costume', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4271, 'Witch''s Costume', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4272, 'Lady Frankenpenguin''s Dress', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4273, 'Cleo Dress', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4274, 'Blue Letterman Jacket', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4275, 'Puffle Raincoat', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4276, 'Blue Striped Raincoat', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4277, 'Pink Striped Raincoat', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4280, 'Pink Cosmic Dress', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4282, 'Comm Gear', 5, 20, TRUE, TRUE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (4283, 'Blue Snow Jacket', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4284, 'Pink Sled Coat', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4285, 'Puffy Parka', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4286, 'Silly Snowman Sweater', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4287, 'Holly Elf Dress', 5, 475, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4288, 'Santa Suit', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4289, 'Protect The Earth T-Shirt', 5, 100, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4290, 'Provide Medical Help T-Shirt', 5, 100, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4291, 'Build Safe Places T-Shirt', 5, 100, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4292, 'Life Jacket', 5, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4293, 'Checked Crop Coat', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4294, 'Lumberjack Look', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4295, 'Canoeing Outfit', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4296, 'Blue Buckle-Up', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4297, 'Yellow Goalie Gear', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4298, 'Gold Figure Skating Dress', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4299, 'Hockey Referee', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4300, 'Tech Coat', 5, 20, TRUE, TRUE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (4301, 'I Heart My Red Puffle T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4302, 'I Heart My Blue Puffle T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4303, 'I Heart My Pink Puffle T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4304, 'I Heart My Black Puffle T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4305, 'I Heart My Green Puffle T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4306, 'I Heart My Yellow Puffle T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4307, 'I Heart My Purple Puffle T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4308, 'I Heart My White Puffle T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4309, 'I Heart My Orange Puffle T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4310, 'I Heart My Brown Puffle T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4311, 'Flare Hoodie', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4312, 'Green Polka-Dot Dress', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4315, 'Striped Puffle Jacket', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4316, 'Black Vest Outfit', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4317, 'Neon Grid Jacket', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4318, 'Ocean Blue Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4319, 'Jazzercise Top', 5, 330, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4321, 'Custom T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4322, 'Custom T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4323, 'Custom T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4324, 'Custom T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4325, 'Custom T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4326, 'Custom T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4327, 'Custom T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4328, 'Custom T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4329, 'Custom T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4330, 'Custom T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4331, 'Custom T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4332, 'Lightning Gi', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4333, 'Thunder Gi', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4334, 'Crimson Sun Suit', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4335, 'Golden Sun Suit', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4336, 'Tea Ceremony Robes', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4337, 'Ink Ceremony Robes', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4338, 'Delivery Outfit', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4339, 'Box Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4340, 'Technicolor Fairy Dress', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4341, 'Cowboy Outfit', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4342, 'Astro Suit', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4343, 'Dark Cocoa Bunny Costume', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4344, 'African Painted Dog Costume', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4345, 'Snow Leopard Costume', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4346, 'Yellow Soccer Jersey', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4347, 'Yellow Away Soccer Jersey', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4348, 'Red Away Soccer Jersey', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4349, 'Blue Away Soccer Jersey', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4350, 'Green Away Soccer Jersey', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4353, 'Sapphire Princess Dress', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4354, 'Gold Princess Dress', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4355, 'Ruby Princess Dress', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4357, 'Sapphire Duchess Dress', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4358, 'Forest Maiden Gown', 5, 425, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4359, 'Viscount Outfit', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4360, 'Village Jester', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4361, 'Purple Dragon Costume', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4362, 'White Knight Armor', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4365, 'Rookies Shirt', 5, 0, FALSE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4366, 'Rookies Cardboard Box', 4, 0, FALSE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4367, 'Canister Camouflage', 5, 20, TRUE, FALSE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (4368, 'Green Bubble Look', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4369, 'Striped Skater Threads', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4370, 'Neon Electro Hoodie', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4371, 'Layered Bubblegum Look', 5, 330, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4372, 'Rhythm & Purple Blues', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4373, 'One Man Band', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4374, 'Franky Tee', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4375, 'Music Swirl Tee', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4376, 'Summer Suit', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4377, 'Tropical Ruffle Dress', 5, 425, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4381, 'Petey K Black Vest and Jeans', 5, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4382, 'Franky Jacket and Jeans', 5, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4383, 'Stompin'' Bob Black T-Shirt and Jeans', 5, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4384, 'G Billy Plaid Shirt and Jeans', 5, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4385, 'Crow''s Nest Vest', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4386, 'Shipshape Suit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4387, 'Navy Suit', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4388, 'Raggedy Rags', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4389, 'Blue Mermaid Costume', 5, 425, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4390, 'Scallywag Rags', 5, 325, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4391, 'Commander Coat', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4392, 'Blue Crab', 5, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4393, 'Evergreen Snow Vest', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4394, 'Spot-On Snowsuit', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4395, 'Toboggan Suit', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4396, 'Supersonic Speed Suit', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4397, 'Custom T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4398, 'Custom T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4399, 'Custom T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4400, 'Custom T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4401, 'Custom T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4402, 'Custom T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4403, 'Custom T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4404, 'Custom T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4405, 'Custom T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4406, 'Custom T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4407, 'Custom T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4408, 'Custom T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4409, 'Custom T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4410, 'Ice Cream Vendor', 5, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4411, 'Cotton Candy Cut-Offs', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4412, 'Keep in Checks Look', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4413, 'I Heart Pizza T-Shirt', 5, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4414, 'Blue Wheeler', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4415, 'Green Wheeler', 5, 3000, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4416, 'Clown-Around Costume', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4417, 'See You Layer Dress', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4418, 'Balloon Vendor', 5, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4419, 'Rooster Costume', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4420, 'Cow Costume', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4421, 'The Red Racer', 5, 3000, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4422, 'Mime Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4423, 'Strongman Suit', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4424, 'Custom T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4425, 'Custom T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4426, 'Sled Costume', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4427, 'Fairy Princess Dress', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4428, 'Witch Hazel Dress', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4429, 'Horse Costume', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4430, 'Unicorn Costume', 5, 530, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4431, 'Ghoul Detector Jumpsuit', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4432, 'Headless Horseman', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4433, 'Candy Corn Costume', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4434, 'Purple Spider Costume', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4435, 'Werewolf Costume', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4447, 'Fire Kimono', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4448, 'Water Kimono', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4449, 'Sushi Master Uniform', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4450, 'Earthquake Coat', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4451, 'Solar Flare Suit', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4452, 'Mint Condition Coat', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4453, 'Slope Suit', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4454, 'Fire Training Plates', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4455, 'Sage Fish', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4456, 'Ancient Dragon', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4457, 'Snow Monkey', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4458, 'Water Training Plates', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4459, 'Piccadilly Plaid', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4460, 'White Puffle Pullover', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4461, 'Kind of a Big Teal Coat', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4462, 'Claus-ette Outfit', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4463, 'The Peppermint Suit', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4464, 'Nutcracker Costume', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4465, 'Snowflake Costume', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4466, 'Ghost of Yesterday', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4467, 'Ghost of Today', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4468, 'Ghost of Tomorrow', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4469, 'Cranberry Decoration', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4470, 'Gold Decoration', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4471, 'Evergreen Decoration', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4472, 'Cookie Serving Apron', 5, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4473, 'Gingerbread Cookie Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4474, 'Red Away Hockey Jersey', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4475, 'Red Away Goalie Gear', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4476, 'Blue Away Hockey Jersey', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4477, 'Blue Away Goalie Gear', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4478, 'Green Away Hockey Jersey', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4479, 'Green Away Goalie Gear', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4480, 'Yellow Away Hockey Jersey', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4481, 'Yellow Away Goalie Gear', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4482, 'Perfect Twirl Dress', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4483, 'Floating on Ice Dress', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4484, 'Snowy Sky Dress', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4487, 'Hip Red Jacket', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4488, 'Pink Silk Blouse', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4489, 'Green Grid Hoodie', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4490, 'Marine Vest Outfit', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4491, 'Blue Surf Shorts', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4492, 'Pink Stripe Bikini', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4493, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4494, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4495, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4496, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4497, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4498, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4499, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4500, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4501, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4502, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4503, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4504, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4505, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4506, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4507, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4508, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4509, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4510, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4511, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4512, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4513, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4514, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4515, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4516, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4517, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4518, 'Submarine Suit', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4532, 'Bubble Dress', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4533, 'Sharp Black Vest', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4534, 'Layered Marshmallow Look', 5, 330, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4535, 'Winter Threads', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4536, 'Corsair Coat', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4537, 'Swashbuckler''s Leather Coat', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4538, 'Dinosaurus Rex', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4539, 'Ancient Robes', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4540, 'Viking Lord Armor', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4542, 'Red Puffle Costume', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4543, 'Blue Puffle Costume', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4544, 'Pink Puffle Costume', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4545, 'Black Puffle Costume', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4546, 'Green Puffle Costume', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4547, 'Yellow Puffle Costume', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4548, 'Purple Puffle Costume', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4549, 'White Puffle Costume', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4550, 'Orange Puffle Costume', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4551, 'Brown Puffle Costume', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4552, 'White Electro Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4553, 'Purple Polka-dot Dress', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4555, 'PH Outfit', 5, 0, FALSE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4556, 'Purple Shirt n'' Skirt', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4557, 'Grey Shirt n'' Shorts', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4558, 'Old School Hoodie', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4559, 'Urban Top and Jacket', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4560, 'Rock-hopper Costume', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4561, 'Lion Costume', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4572, 'Layered Sunset Outfit', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4573, 'Sandy Shore Outfit', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4575, 'Fairy Gown', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4576, 'Stout Warrior Armor', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4577, 'Chain Mail and Tabard', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4578, 'Crimson Plate Mail', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4579, 'Sky Gown', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4580, 'Onyx Dragon Outfit', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4582, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4583, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4584, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4585, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4586, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4587, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4588, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4589, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4590, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4591, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4592, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4593, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4594, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4595, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4596, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4597, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4598, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4599, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4600, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4601, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4602, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4603, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4604, 'Custom Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4606, 'Pop-n-Lock Music Shirt', 5, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4607, 'Trendy Red Shirt', 5, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4611, 'Gray Pinstripe Suit', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4612, 'Working Clothes', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4613, 'Business Dress', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4614, 'Reporter Outfit', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4615, 'Hipster Threads', 5, 475, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4616, 'Jock Jacket', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4617, 'Justice Suit', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4618, 'Fiendish Suit', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4619, 'Valiant Suit', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4620, 'Sinister Suit', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4621, 'Iron Man Bodysuit', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4622, 'Ms. Marvel Bodysuit', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4623, 'Nova Bodysuit', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4624, 'Iron Fist Bodysuit', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4625, 'Loki Armor', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4626, 'Spider-Man Bodysuit', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4627, 'Venom Symbiote', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4628, 'Captain America Bodysuit', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4629, 'The Lizard Bodysuit', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4630, 'Hawkeye Bodysuit', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4631, 'Thor Armor', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4632, 'HULK BODYSUIT', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4633, 'Nick Fury Coat', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4634, 'Black Widow Bodysuit', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4635, 'Doom Drone', 5, 3000, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4636, 'City''s Finest Uniform', 5, 400, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4637, 'Up To No Good Suit', 5, 350, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4638, 'Green Dance Sweats', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4639, 'Check My Moves Vest', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4640, 'Ice Moves Hoodie', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4641, 'Pink Dance-off Dress', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4642, 'Color Splash Hoodie', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4643, 'Layered Lime Look', 5, 330, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4644, 'Pitch Perfect Shirt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4645, 'Hip Hop Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4646, 'Stand Out Outfit', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4647, 'Green Scene Jacket', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4648, 'LOL Threads', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4649, 'Snow Cone Ruffle Dress', 5, 425, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4660, 'Beach Dress', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4661, 'Big Deal Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4662, 'Aqua Sarong', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4663, 'Green Surf Shorts', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4664, 'Watermelon Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4665, 'Pineapple Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4666, 'Apple Costume', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4667, 'Grape Bunch Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4668, 'Watermelon Jungle Clothing', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4669, 'Apple Tiki Costume', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4670, 'Pineapple Tiki Costume', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4671, 'Grape Tiki Costume', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4672, 'Watermelon Tiki Dress', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4673, 'Apple Tiki Dress', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4674, 'Grape Tiki Dress', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4675, 'Pineapple Tiki Dress', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4676, 'Lemon Lime Stripe Bikini', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4677, 'Street Surf Shorts', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4686, 'Volcano Costume', 5, 700, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4688, 'Shake It Up Outfit', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4689, 'Make Your Mark Outfit', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4715, 'Purple Wheeler', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4716, 'Violet Striped Shirt', 5, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4717, 'Mischief Maker Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4718, 'Creepy Doll Costume', 5, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4719, 'Ghost Pirate Costume', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4720, 'Fashionable Fuchsia Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4721, 'Red Skater Punk Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4722, 'Gravedigger Suit', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4723, 'Extra Cheesy Costume', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4724, 'Brown Bat Costume', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4725, 'Masquerade Ball Gown', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4726, 'Distinguished Suit', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4727, 'Ghost Catcher Uniform', 5, 450, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4732, 'Butler Suit', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4733, 'Maid Outfit', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4734, 'Mystic Robe', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4735, 'Kitty Costume', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4738, 'Tux Redux', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4739, 'Elegant Ensemble', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4740, 'Sturdy Jacket', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4741, 'Green Turtleneck', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4742, 'Rugged Jacket', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4743, 'Red Undercover Shirt', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4744, 'Herbert Disguise', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4745, 'Klutzy Disguise', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4746, 'Deflection Vest', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4747, 'Rappelling Gear', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4749, 'Capture Equipment', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4751, 'Scrubs', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4753, 'Striking Red Jacket', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4754, 'Baby Blue Shoulder Shirt', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4755, 'Black Zip Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4756, 'Warm Red Jacket', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4757, 'Sweet Ski Suit', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4758, 'Cool Ski Suit', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4759, 'Green Bobsled', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4760, 'Red Bobsled', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4761, 'Toymaker Apron', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4762, 'Icy Dress', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4763, 'Penguin Stuffie Costume', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4764, 'Reindeer Handler Uniform', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4765, 'Festive Sweater', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4769, 'Green Tuff Jacket', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4770, 'Fuzzy Coat and Tights', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4771, 'Creepy Cutie T-Shirt', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4772, 'Blue Casual Duds', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4773, 'Green Overshirt', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4774, 'Red Street Smart Outfit', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4775, 'Furry Togs', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4777, 'Furry Shorts', 5, 250, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4778, 'Warm Furry Frock', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4779, 'T Rex Hoodie', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4780, 'Fuzzy Polka Dot Bikini', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4781, 'Fuzzy Blue Vest', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4782, 'Prehistoric Gown', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4783, 'Prehistoric Suit', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4785, 'Laid Back Gray Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4786, 'No Fuss Denim Jacket', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4787, 'Hi-Lo Outfit', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4788, 'Neon Skater Punk Outfit', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4789, 'Gold Letterman Jacket', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4790, 'Space Squid Costume', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4791, 'Stuntman Suit', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4792, 'Stuntwoman Suit', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4793, 'Superfan Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4794, 'Golf Cart', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4795, 'Hip-Hop Fever Outfit', 5, 800, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4796, 'Diamond Gown', 5, 800, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4797, 'Among the Stars Mint Dress', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4798, 'Jazzy Gray Shirt', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4799, 'Shimmer Diva Dress', 5, 375, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4800, 'Prom King Tux', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4801, 'Night Sky Prom Dress', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4802, 'Midnight Glamor Dress', 5, 375, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4803, 'Navy Royale Tux', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4804, 'Classic Blue Dress', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4805, 'Green Fur Lined Jacket', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4806, 'Outdoorsy Jacket', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4807, 'Puffle Mania Shirt', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4808, 'Blue Polo and Denim', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4809, 'Puffle Groomer Outfit', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4810, 'Rainbow Puffle Costume', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4811, 'I Heart My Rainbow Puffle T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4812, 'Rainbow Puffle Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4814, 'Aunt Arctic''s Shirt', 5, 0, FALSE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4815, 'Teleportation Suit', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4816, 'Laser Beam Costume', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4817, 'Stomp Suit', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4818, 'Crystal Shards Suit', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4819, 'Telekinesis Suit', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4820, 'Ooze Suit', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4821, 'Tornado Suit', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4822, 'Sonic Blast Suit', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4823, 'In the Press Outfit', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4824, 'Captain Marvel Bodysuit', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4825, 'Sif Armor', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4826, 'Mandarin Robes', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4827, 'Odin Armor', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4828, 'Iron Patriot Armor', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4829, 'Silver Centurion Armor', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4830, 'Mark 42 Armor', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4831, 'Heartbreaker Armor', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4832, 'War Machine Armor', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4833, 'Nightclub Armor', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4834, 'Coat of Frost', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4835, 'Frozen Armor', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4836, 'Black Ice Training Plates', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4837, 'Snowstorm Gi', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4838, 'Cat Burglar Outfit', 5, 250, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4839, 'Police Gear', 5, 350, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4840, 'Super Hero Hoodie', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4841, 'Hero Hoodie', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4842, 'Robot Hoodie', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4844, 'Sashimi Chef Uniform', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4845, 'Whiteout Gi', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4846, 'Snow Training Plates', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4847, 'Avalanche Suit', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4848, 'Elemental Balance Gi', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4849, 'Snow Kimono', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4850, 'Iron Spider Bodysuit', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4851, 'A-Bomb Bodysuit', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4852, 'Red Hulk Bodysuit', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4853, 'She-Hulk Bodysuit', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4854, 'Skaar Bodysuit', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4855, 'Goes with Everything Shirt', 5, 400, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4856, 'Street Cred Outfit', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4857, 'Downtown Dress', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4858, 'Sulley Costume', 5, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4859, 'Johnny Costume', 5, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4860, 'Mike Costume', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4861, 'Randy Costume', 5, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4862, 'Hardscrabble Costume', 5, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4863, 'Brock Costume', 5, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4864, 'Claire Costume', 5, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4865, 'Don Costume', 5, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4866, 'George Costume', 5, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4867, 'Squishy Costume', 5, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4868, 'Squishy''s Mom Costume', 5, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4869, 'Terry and Terri Costume', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4870, 'PNK Costume', 5, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4871, 'Chet Costume', 5, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4872, 'CDA Suit', 5, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4873, 'JOX Jacket', 5, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4874, 'ROR Jacket', 5, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4875, 'OK Sweater', 5, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4876, 'PNK Sweater', 5, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4877, 'Daisy Daydream Outfit', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4878, 'On The Bright Side Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4879, 'Seaside Denim Jacket and Hoodie', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4880, 'Color Me Cool Outfit', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4881, 'Pretty In Punk Outfit', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4882, 'Beach Ready Outfit', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4883, 'Puffle Hoodie', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4884, 'Electric Summer Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4885, 'Sensei''s Epic Robes', 5, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4886, 'C-3PO Costume', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4887, 'Chewbacca Costume', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4888, 'Darth Vader Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4889, 'Han Solo Costume', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4890, 'Princess Leia Dress', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4891, 'Stormtrooper Costume', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4892, 'Jedi Robes', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4893, 'X-wing Pilot Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4894, 'Rebel Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4895, 'Black Jedi Robes', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4896, 'Boba Fett Costume', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4897, 'Luke Skywalker Robes', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4898, 'Greedo Costume', 5, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4899, 'Tusken Raider Costume', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4900, 'Jawa Costume', 5, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4902, 'Emperor Palpatine Cloak', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4903, 'Imperial Officer Uniform', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4905, 'TIE Fighter Costume', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4906, 'X-wing Costume', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4907, 'Surf ''n Style Outfit', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4908, 'Surf''s Up Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4909, 'Endless Summer Outfit', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4910, 'Boho Style Outfit', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4912, 'Cheechee''s Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4913, 'Lela''s Biker Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4914, 'Lela''s Red Dress', 5, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4915, 'Butchy''s Outfit', 5, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4916, 'Biker Girl Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4917, 'Brady''s Beach Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4918, 'Tanner''s Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4919, 'McKenzie''s Biker Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4920, 'McKenzie''s Beach Outfit', 5, 300, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4921, 'Seacat Blue Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4922, 'Big Momma''s Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4923, 'Giggle''s Pink Outfit', 5, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4938, 'Brady''s Biker Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4939, 'Regal Armor', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4940, 'Emerald Princess Gown', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4941, 'Snow Queen Robe', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4942, 'Village Wear', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4943, 'Knightly Armor', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4944, 'Back-to-School Outfit', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4945, 'Prep Style', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4946, 'Rockhopper''s Jacket', 5, 0, FALSE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4947, 'Epic Knight Armor', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4948, 'Enchanted Dragon', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4949, 'Armored Ogre Costume', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4950, 'Ogre Costume', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4951, 'Enchanted Fairy Dress', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4952, 'Wizardly Jacket', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4955, 'Cadence''s Shirt', 5, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4956, 'Swamp Monster Costume', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4958, 'Puffle Bat Tee', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4959, 'Countess Costume', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4960, 'Perky Punk Outfit', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4961, 'Count Costume', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4962, 'Ghoul Suit', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4963, 'Ghoul Dress', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4964, 'Puffle Pirate Dress', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4965, 'Tomb King Costume', 5, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4966, 'Classic Vampire Costume', 5, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4967, 'Penguin-o''-Lantern', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4969, 'Rookie''s Halloween Shirt', 5, 9999, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4970, 'Blue Off The Shoulder', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4971, 'Black Puffle Don''t Care Tee', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4972, 'Black T-Shirt', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4973, 'Leather Outdoors Jacket', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4974, 'Camo Hoodie', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4975, 'Blue Fade Hoodie', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4976, 'The Rock n'' Rollin''', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4977, 'The Rockin'' Rainbow', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4978, 'Sequined Rocker Vest', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4981, 'Golden Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4982, 'EPF Workout Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4983, 'Elite Tactical Armor', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4986, 'Puffle Field Medic Outfit', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4987, 'Rescue Diver Suit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4988, 'Arctic Camouflage Suit', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4989, 'Yellow and Black Two-tone Hoodie', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4990, 'Blue and White Two-tone Hoodie', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4991, 'Black and Green Two-tone Hoodie', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4992, 'White and Red Two-tone Hoodie', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4993, 'Ancient Gold Dragon', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4994, 'Golden Bumper Car', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4995, 'Klutzy Disguise M', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4996, 'Snow Style Jacket', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4997, 'Winter Toggles Coat', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4998, 'Classy Holiday Outfit', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (4999, 'Remove Body Item', 5, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5000, 'G Billy''s Silver Drumsticks', 6, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5001, 'Mixed Bracelets', 6, 150, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5002, 'Denim Purse', 6, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5003, 'Magnifying Glass', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5004, 'Blue Cotton Candy', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5005, 'Candy Apple', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5006, 'Teddy Bear', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5007, 'Blue Book', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5008, 'MP3K', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5009, 'Lantern', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5010, 'Pink Mittens', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5011, 'Blue Mittens', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5012, 'Hand Gong', 6, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5013, 'Silver Wand', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5014, 'Trumpet', 6, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5015, 'Wingwarmers', 6, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5016, 'Boombox', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5018, 'Goalie Hockey Stick', 6, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5019, 'Stop Watch', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5020, 'Yarr', 6, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5021, 'Toboggan', 6, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5022, 'Penguin Play Award', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5023, 'Cadence''s Boombox', 6, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5024, 'Franky''s Guitar', 6, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5025, 'Bob''s Bass Guitar', 6, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5027, 'Gardening Rake', 6, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5028, 'Golden Shield', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5029, 'Lute', 6, 675, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5030, 'Bracers', 6, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5031, 'Megaphone', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5032, 'Binoculars', 6, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5033, 'Canteen', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5034, 'Trombone', 6, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5035, 'Cowbell', 6, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5036, 'Double Necked Guitar', 6, 1000, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5037, 'Funky Record', 6, 75, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5038, 'Festival Record', 6, 75, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5039, 'Jungle Record', 6, 75, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5040, 'Cloud Wave Bracers', 6, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5042, 'Orange Balloon', 6, 50, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5043, 'Kite', 6, 75, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5044, 'Shell Cuffs', 6, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5045, 'Laptop', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5046, 'Green Racing Sled', 6, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5047, 'Pink Racing Sled', 6, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5048, 'White Pompoms', 6, 120, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5049, 'Cloudy Umbrella', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5050, 'Smitten Mittens', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5051, 'Silver Bell', 6, 75, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5052, 'Teal Bracelet', 6, 75, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5053, 'Black MP3000', 6, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5054, 'Video Camera', 6, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5055, 'Floral Clutch Bag', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5056, 'Brown Leather Cuffs', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5057, 'Water Bottle', 6, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5058, 'Staff and Shield', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5059, 'Royal Scepter', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5060, 'EPF Puffle', 6, 0, FALSE, TRUE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (5061, 'Pirate Arm Bands', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5062, 'Foraged Bracelet', 6, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5063, 'Blue Electric Guitar', 6, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5064, 'Orange Double Necked Guitar', 6, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5065, 'Tambourine', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5066, 'Green Keytar', 6, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5067, 'Purple Electric Bass', 6, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5068, 'Leather Watch', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5069, 'Garden Shovel', 6, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5070, 'Reggaetron Record', 6, 75, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5071, 'Desi-bel Record', 6, 75, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5072, 'House Blend Record', 6, 75, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5073, 'Red Flag', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5074, 'First Aid Kit', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5075, 'Range Finder', 6, 10, TRUE, TRUE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (5076, 'Green Balloon', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5077, 'Ice Cream Cone', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5079, 'Popcorn', 6, 50, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5080, 'Magician''s Wand', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5081, 'Storm Lantern', 6, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5082, 'Cosmic Umbrella', 6, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5083, 'Polka Dot Umbrella', 6, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5084, 'Mop and Bucket', 6, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5087, 'Candy Cane Wing-warmers', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5088, 'Paddle', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5089, 'Tech-book 3000', 6, 15, TRUE, TRUE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (5090, 'Pink MP3000', 6, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5091, 'Green Foam Finger', 6, 75, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5092, 'Red Foam Finger', 6, 75, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5093, 'Blue Foam Finger', 6, 75, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5094, 'Yellow Foam Finger', 6, 75, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5095, 'Gold Shield', 6, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5096, 'Royal Blue Scepter', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5097, 'Red Shield', 6, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5098, 'Yellow Shield', 6, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5099, 'Blue Shield', 6, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5100, 'Gold Staff and Shield', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5101, 'Blue Double Guitar', 6, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5102, 'Stompin Bob Cuff', 6, 175, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5103, 'Green and Blue Maracas', 6, 125, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5104, 'Oil Slick Guitar', 6, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5105, 'Stompin'' Bob Electric Bass', 4, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5106, 'Stompin'' Bob Cuff', 6, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5107, 'G Billy Drum Sticks', 6, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5108, 'Telescope', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5109, 'Treasure Maps', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5110, 'Bangles', 6, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5111, 'Grape Balloon', 6, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5112, 'Bunch of Balloons', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5113, 'Unicycle', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5114, 'Turtle', 6, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5115, 'Every Flavor Ice Cream', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5116, 'Cherry Balloon', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5117, 'Lemon Balloon', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5118, 'Dumbbells', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5119, 'Ghoul Detector 3000', 6, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5120, 'Ghostly Gallop', 6, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5121, 'Monster Muzak', 6, 75, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5122, 'Trick-or-Treat Basket', 6, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5123, 'Witch Wing Warmers', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5126, 'Fire Blossom Fan', 6, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5127, 'Water Lotus Fan', 6, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5128, 'Sushi Tray', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5129, 'Mauve Mittens', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5130, 'Candy Cane Cane', 6, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5131, 'Iron Lantern', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5132, 'Green MP3000', 6, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5133, 'Blue Water Bottle', 6, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5135, 'Pearl Clutch Bag', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5136, 'Pink Pompoms', 6, 120, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5137, 'Black Pompoms', 6, 120, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5138, 'Green Pompoms', 6, 120, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5139, 'Yellow Pompoms', 6, 120, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5140, 'Purple Pompoms', 6, 120, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5141, 'Brown Pompoms', 6, 120, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5144, 'Fireworks Bangle', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5145, 'Off the Cuff', 6, 175, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5146, 'Steel Shield', 6, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5147, 'Crimson Lance and Shield', 6, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5148, 'Skyward Staff', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5150, 'Yellow Bass Guitar', 6, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5151, 'Thunder Blade', 6, 1700, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5153, 'Money Bag', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5154, 'Captain America Shield', 6, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5155, 'Mjolnir', 6, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5156, 'Freezing Super Gloves', 6, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5157, 'Shocking Super Gloves', 6, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5158, 'Cosmic Super Gloves', 6, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5159, 'Purple Boom Box', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5160, 'Golden Microphone', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5161, 'Acid Guitar!', 6, 1000, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5162, 'Gold Bling Bracelet', 6, 175, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5163, 'Sweet Spikester Cuffs', 6, 170, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5164, 'Purple MP3000', 6, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5166, 'Grape Spear', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5167, 'Kiwi Purse', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5176, 'Lime Laptop', 6, 1800, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5177, 'Grey Shovel', 6, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5179, 'Antique Mirror', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5180, 'Pocket Watch', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5181, 'Masquerade Fan', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5183, 'Mint Purse', 6, 175, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5184, 'Steel Watch', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5185, 'Grappling Hook', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5186, 'Rugged Radio', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5187, 'Green Chic Purse', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5188, 'Sweet Mittens', 6, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5189, 'Cool Mittens', 6, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5190, 'Race Car Controller', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5191, 'Holiday Teddy', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5192, 'Hot Chocolate', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5193, 'Checkered Flag', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5194, 'Blue Puffle Stuffie', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5195, 'Red Mittens', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5196, 'Piratey Stuffie', 6, 150, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5197, 'Golden Bangles', 6, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5198, 'Prehistoric Staff', 6, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5199, 'Ugg Club', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5201, 'Trusty Spear', 6, 150, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5202, 'Rockman', 6, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5203, 'The Wheel', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5204, 'Stony Bracelet', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5205, 'Red Longboard', 6, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5206, 'Neon Blue Longboard', 6, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5207, 'Gold Award', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5208, 'Boom Mic', 6, 150, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5209, 'Clapboard', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5210, '500 Carat Diamond Ring', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5211, 'Film Script', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5212, 'Coffee Tray', 6, 100, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5214, 'Navy Clutch', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5216, 'Silver Award', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5217, 'Bronze Award', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5218, 'Diamond Bracelet', 6, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5219, 'Slider Cell', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5220, 'Gourmet O''Berries', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5221, 'Brief Case', 6, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5222, 'Sif''s Sword', 6, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5223, 'Ice Cap Cuffs', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5224, 'Storm Cloud Bracers', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5225, 'Snow Shuriken', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5226, 'Fire Nunchaku', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5227, 'Water Hammer', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5228, 'Brown Leather Watch', 6, 150, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5229, 'Green Slouch Purse', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5230, 'Rainbow Puffle', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5240, 'Electro Blast Gloves', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5241, 'Force Wave Power Gloves', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5242, 'Magma Power Gloves', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5243, 'Crystal Shards Power Gloves', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5244, 'Telekinesis Power Gloves', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5245, 'Ooze Power Gloves', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5246, 'Fireball Gloves', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5247, 'Brainwave Gloves', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5248, 'Robot Remote Control', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5249, 'Construct Power Gloves', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5250, 'Sushi Combo', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5251, 'Snow Rose Fan', 6, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5252, 'Arctic Cuffs', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5332, 'Scare Games Trophy', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5334, 'Emperor Palpatine Cane', 6, 75, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5335, 'Luke''s Lightsaber', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5336, 'Anakin''s Lightsaber', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5337, 'Darth Vader''s Lightsaber', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5338, 'Kloo Horn', 6, 100, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5339, 'Bandfill', 6, 100, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5340, 'Magic Surfboard', 6, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5341, 'McKenzie''s Surfboard', 6, 200, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5342, 'Brady''s Surfboard', 6, 200, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5343, 'Butchy''s Gloves', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5344, 'Black Motorbike', 6, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5345, 'Pink Motorbike', 6, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5346, 'Red Motorbike', 6, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5349, 'Pink Ice Cream', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5350, 'Ukulele', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5351, 'Campfire Marshmallow', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5352, 'Beach Towel', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5353, 'Summer Beach Ball', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5354, 'Tropical Smoothie', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5356, 'Bow and Arrows', 6, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5357, 'Sword of Virtue', 6, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5358, 'Magic Potion', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5359, 'Magical Book', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5360, 'Epic Staff and Shield', 6, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5361, 'Magic Staff', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5362, 'Magical Wand', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5367, 'Staff of Wonder', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5368, 'Countess Bangles', 6, 75, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5369, 'Indigo Pompoms', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5370, 'Trick-or-Treat Bag', 6, 50, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5371, 'Big Lollipop', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5372, 'Search & Rescue Snowmobile', 6, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5373, 'Rescue Off-Roader', 6, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5374, 'Hot Cocoa', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5376, 'Rescue Ring', 6, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5377, 'Thermos', 6, 100, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5378, 'Rock Climbing Rope', 6, 100, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5379, 'Hydro-tester 3000', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5380, 'Bullhorn', 6, 200, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5381, 'Search Flashlight', 6, 200, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5382, 'Golden Off-Roader', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5384, 'Golden Laptop', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5385, 'Golden Guitar', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5386, 'Gold MP3000', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5387, 'Book of Carols', 6, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5388, 'The Holiday Express', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5389, 'The Puffle Premier', 6, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5390, 'The Sweet-tooth Special', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5391, 'The Rainbow Zephyr', 6, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5392, 'Milk and Cookies', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5393, 'Little Red Wagon', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5394, 'Fireproof Gloves', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5395, 'Wooden Wheelie Puffle', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5396, 'Holiday Teddy', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5397, 'Holiday Lantern', 6, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5398, 'Candy Pirate Hook', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5399, 'Santa''s Sleigh', 6, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5400, 'Leather Arm Band', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5401, 'Dino-leather Shield', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5402, 'Dino Claws', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5403, 'Stone Hair Dryer', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5404, 'Prehistoric Sling Shot', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5405, 'Stone Hatchet', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5406, 'Primal Bracers', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5407, 'Prehistoric Kite', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5414, 'Fluffy Stuffie', 6, 1500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5415, 'Green Motorbike', 6, 3500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5416, 'Blue Motorbike', 6, 3500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5417, 'Noble Steed', 6, 1500, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5418, 'Carnival Cane', 6, 750, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5419, 'Blue Balloon Sword', 6, 1500, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5420, 'Retro Joystick', 6, 500, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5421, 'Sea Fishing Rod', 6, 900, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5422, 'Old West Lasso', 6, 700, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5423, 'Popcorn', 6, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5424, 'Stuffed Suitcase', 6, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5425, 'Resort Brochure', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5427, 'Bull Cape', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5428, 'Pink Microphone', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5429, 'Cream Pie', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5430, 'Animal''s Drum Sticks', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5431, 'Beakers', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5432, 'Salad Tongs', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5433, 'Mexican Maracas', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5434, 'Gold Cane', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5435, 'Wrist Wrap', 6, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5436, 'Ball Of Yarn', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5438, 'Throwing Disc', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5439, 'Squeaky Ball', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5440, 'Wind-up Mouse', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5441, 'Zero G Gloves', 6, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5442, 'Bubble Ray Gun', 6, 100, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5443, 'Cybernetic Arm', 6, 150, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5445, 'The Golden Soccer Ball', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5446, 'Penguin Cup Foam Finger', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5447, 'Hot Sauce Vuvuzela', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5448, 'Fluffies Vuvuzela', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5449, 'Space Squids Vuvuzela', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5450, 'Sharks Vuvuzela', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5451, 'Penguin Cup Flag', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5452, '2014 Penguin Cup Green', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5453, '2014 Penguin Cup Red', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5454, '2014 Penguin Cup Yellow', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5455, '2014 Penguin Cup Blue', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5456, 'Penguin Cup Ball', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5457, 'Rescue Buoy', 6, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5458, 'Sushi and Soda', 6, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5460, 'Party Bracelets', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5461, 'MP3000 Bling Edition', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5462, 'Girl Next Door Guitar', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5463, 'Glitter Microphone', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5465, 'Kristoff''s Lute', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5466, 'Arendelle Banner', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5467, 'Reindeer Carrots', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5468, 'Summer Punch', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5469, 'Arendelle Royal Regalia', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5471, 'Sponsored Pro Skateboard', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5472, 'Fluffy Skateboard', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5473, 'Galactic Pink Skateboard', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5474, 'Bad Bear Skateboard', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5475, '', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5476, 'Back-to-School Haul', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5477, 'Undodgeable', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5478, 'School Shopping', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5479, 'Rainbow Board', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5480, 'Ghost Vial', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5481, 'Creepy Candle', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5482, 'Inspector''s Magnifying Glass', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5483, 'Little Jack', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5484, 'Eye of Newt Gumballs', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5485, 'Ghostly Suitcase', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5486, 'Candy Wing Warmers', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5487, 'Witch''s Broom', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5488, 'Puffle Watch', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5489, 'Anchor Charm', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5490, 'Leather Bangle', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5491, 'Puffle Umbrella', 6, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5494, 'Wooden Sword', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5495, 'Wooden Sword', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5496, 'Stinky Cheese Sword', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5497, 'Balloon Blade', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5498, 'Rogue''s Rapier', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5499, 'Jeweled Cutlass', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5500, 'Pirate''s Hook', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5501, 'Squid Scrunch', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5502, 'Deserted Island Slingshot', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5503, 'Pirate Flag', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5504, 'Jailor''s Keys', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5505, 'Shell Bracers', 6, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5506, 'Slider Cell', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5507, 'Candy Cane Cane', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5508, 'Ezra''s Slingshot', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5509, 'Sabine''s Spray Gun', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5510, 'Rebel''s Flag', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5511, 'Blue Lightsaber', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5512, 'Purple Lightsaber', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5513, 'Dual Lightsabers', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5514, 'The Inquisitor''s Lightsaber', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5515, 'Hera''s Blaster', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5516, 'Hydrospanner', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5517, 'Glow Bracelets', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5518, 'Six-string Axe', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5519, 'Black Watch', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5520, 'Pickaxe', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5521, 'Snowgies', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5522, 'Jelly Donut Plush', 6, 75, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5523, 'Big Game Catch', 6, 75, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5524, 'Blue Fluffy Stuffie', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5525, 'Monkey Stuffie', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5526, 'Beach Bracelet', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5527, 'Blue Tropical Cup', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5528, 'Brown Tropical Cup', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5529, 'Happy Memory', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5530, 'Memory Bag', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5531, 'Riley''s Hockey Stick', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5535, 'Card-Jitsu Carryall', 6, 75, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5536, 'Spiky Dubstep Purse', 6, 75, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5537, 'Strawberry Cake Purse', 6, 75, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5538, 'Compact', 6, 75, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5539, 'Charm Bangle', 6, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5540, 'Makeup Palette', 6, 75, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5542, 'Toasty Cocoa', 6, 150, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5543, 'Mal''s Spellbook', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5544, 'Dude Stuffie', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5545, 'Fairy Godmother''s Wand', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5546, 'Auradon Knights Stick', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5547, 'Magic Hand Mirror', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5548, 'Trophy', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5549, 'Spectral Sweet', 6, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5550, 'Clean Sweeper', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5551, 'Decaf Coffee', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5552, 'Spelling Test', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5553, 'Pink Flamingo', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5554, 'Novelty Bug', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5555, 'Old Sweater', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5556, 'Beard Trimmer', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5557, 'Toy UFO', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5558, 'Planet Lantern', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5559, 'Candy Cane Hook', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5560, 'My Spoon', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5561, 'Hand Lens', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5562, 'Detective''s Case', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5563, 'Violet Boom Box', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5564, 'Black Puffle Balloon', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5565, 'Blue Puffle Balloon', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5566, 'Brown Puffle Balloon', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5567, 'Green Puffle Balloon', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5568, 'Orange Puffle Balloon', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5569, 'Pink Puffle Balloon', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5570, 'Purple Puffle Balloon', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5571, 'Rainbow Puffle Balloon', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5572, 'Red Puffle Balloon', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5573, 'White Puffle Balloon', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5574, 'Yellow Puffle Balloon', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5575, 'Gold Puffle Balloon', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5576, 'On The Hook', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5577, 'Wooden Guitar', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5578, 'Iced Donut', 6, 75, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5579, 'School Shopping', 6, 75, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5580, 'Megaphone', 6, 145, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5581, 'Squid Scrunch', 6, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5583, 'Electric Green Guitar', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5584, 'Bag of Coffee', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5585, 'Stinky Cheese Wheel', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5586, 'Dancer Bracelets', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5587, 'Favorite Fluffy Sandwich', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5588, 'Fruitcake', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5589, 'Bag of Coal', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5590, 'Squid Sticks', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (5999, 'Remove Hand Item', 6, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6000, 'Frankys Red Sneakers', 7, 54, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6001, 'Sunset Dress Shoes', 7, 420, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6002, 'Green Running Shoes', 7, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6003, 'Pink Checkered Shoes', 7, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6004, 'Black Checkered Shoes', 7, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6005, 'Blue Cuckoo Ka-Shoes', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6006, 'Purple Cuckoo Ka-Shoes', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6007, 'Brown Pirate Boots', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6008, 'Black Zoot Shoes', 7, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6009, 'Green Bunny Slippers', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6010, 'Big Bad Wool Hooves', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6011, 'Figure Skates', 7, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6012, 'Geta Sandals', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6013, 'Hockey Skates', 7, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6015, 'White Dress Shoes', 7, 420, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6016, 'Blue Stardust Slippers', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6017, 'Tennis Shoes', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6019, 'Plated Shoes', 7, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6020, 'Pointy Shoes', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6021, 'Green Hiking Boots', 7, 325, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6022, 'Black Cowboy Boots', 7, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6024, 'Burgundy Buckle Shoes', 7, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6025, 'Flame Sandals', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6026, 'Wave Sandals', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6029, 'Ladybug Shoes', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6030, 'Frankenfeet', 7, 220, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6031, 'Vintage Boots', 7, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6032, 'Dark Brown Fuzzy Boots', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6033, 'Monkey King Feet', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6034, 'Guardian Dog Feet', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6035, 'Untied Sneakers', 7, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6036, 'Red Sneakers', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6037, 'Green Sneakers', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6038, 'Yellow Sneakers', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6039, 'Pink Canvas Shoes', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6040, 'Gray Boots', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6042, 'Delta Sneaks', 7, 15, TRUE, TRUE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (6043, 'Alpha Pumps', 7, 10, TRUE, TRUE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (6044, 'Nautical Boots', 7, 275, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6045, 'Commander''s Boots', 7, 275, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6047, 'Yellow Hiking Shoes', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6048, 'Red Hiking Shoes', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6049, 'Tactical Boots', 7, 10, TRUE, TRUE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (6050, 'Orange Checkered Shoes', 7, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6052, 'Clown Shoes', 7, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6053, 'Blue Striped Rubber Boots', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6054, 'Pink Striped Rubber Boots', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6057, 'Comm Boots', 7, 10, TRUE, TRUE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (6058, 'Snowboard Boots', 7, 275, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6059, 'Pink Sparkly Snowshoes', 7, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6060, 'Lumberjack Boots', 7, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6063, 'Blue Canvas Shoes', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6064, 'Green Checkered Shoes', 7, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6065, 'Astro Boots', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6067, 'Village Jester Shoes', 7, 170, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6068, 'Blue Dragon Feet', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6069, 'Purple Dragon Feet', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6072, 'Sneak-ers', 7, 10, TRUE, FALSE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (6073, 'Purple High Tops', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6074, 'Orange High Tops', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6075, 'Blue High Tops', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6076, 'Red High Tops', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6079, 'Franky Purple Kicks', 7, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6080, 'G Billy Cowboy Boots', 7, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6081, 'Peak Boots', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6082, 'Summit Boots', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6083, 'Rooster Feet', 7, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6084, 'Sparkly Emerald Shoes', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6085, 'Horse Hooves', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6086, 'Stompin'' Boots', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6087, 'Green Dragon Feet', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6088, 'Orange Frankenfeet', 7, 220, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6089, 'Werewolf Feet', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6090, 'Swamp Monster', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6091, 'Unicorn Hooves', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6092, 'Seismic Sandals', 7, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6093, 'Sunburst Sandals', 7, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6094, 'Brown Snow Bunny Boots', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6095, 'Ancient Dragon Feet', 7, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6096, 'Snow Monkey Feet', 7, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6097, 'Magenta Sandals', 7, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6098, 'White Sandals', 7, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6100, 'Sparkly Sea Foam Slippers', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6101, 'Untied Violet Sneakers', 7, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6102, 'Groovy Shoes', 7, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6103, 'Red Checkered Shoes', 7, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6104, 'Blue Checkered Shoes', 7, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6105, 'Yellow Checkered Shoes', 7, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6106, 'Purple Checkered Shoes', 7, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6107, 'White Checkered Shoes', 7, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6108, 'Brown Checkered Shoes', 7, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6109, 'Green High Tops', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6110, 'Purple Sandals', 7, 160, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6111, 'Box Shoes', 7, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6112, 'Brown Canvas Shoes', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6113, 'Tan Boots', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6115, 'Plum Slippers', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6116, 'Plate Boots', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6117, 'Onyx Dragon Feet', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6118, 'Neon Pink Sneakers', 7, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6119, 'Slate Untied Sneakers', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6120, 'Boots of Justice', 7, 275, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6121, 'Bad Boots', 7, 275, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6123, 'Golden Shiny Shoes', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6124, 'Purple Sneakers', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6126, 'Cream Sandals', 7, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6134, 'Squeak-Proof Shoes', 7, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6135, 'Jade Stilettos', 7, 275, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6136, 'Sturdy Green Boots', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6137, 'Rugged Boots', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6138, 'Sturdy Brown Boots', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6139, 'Anti Lava Boots', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6140, 'Shiny Slippers', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6141, 'Black Top Sneakers', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6142, 'Sweet Ski Boots', 7, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6143, 'Cool Ski Boots', 7, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6144, 'Festive Socks', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6145, 'Reindeer Handler Boots', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6146, 'Leather Moccasins', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6147, 'Squiggle Shoes', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6148, 'Brown Flats', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6149, 'Silver Stilettos', 7, 275, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6150, 'Jet Stilettos', 7, 275, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6151, 'Brown Furry Boots', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6152, 'Gray Furry Boots', 7, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6153, 'Purple & Yellow Laced Hightops', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6154, 'Pink & Purple Laced Hightops', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6155, 'White Untied Sneakers', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6156, 'Hipster Hightops', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6157, 'Break a Leg Cast', 7, 150, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6158, 'Puffle Trainers', 7, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6159, 'Brown Leather Boots', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6160, 'Stomp Shoes', 7, 275, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6161, 'Telekinesis Shoes', 7, 275, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6162, 'Ooze Shoes', 7, 275, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6163, 'Glacial Sandals', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6164, 'Cold Snap Sandals', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6166, 'Icicle Sandals', 7, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6167, 'Blue Skater Shoes', 7, 200, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6168, 'White Sneaks', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6169, 'Cotton Sandals', 7, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6170, 'Seafoam Slip Ons', 7, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6171, 'Brown Flip Flops', 7, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6172, 'Sandy Flip Flops', 7, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6173, 'Blue & White Flips Flops', 7, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6174, 'Lime Slip Ons', 7, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6175, 'Pink Flip Flops', 7, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6176, 'Gold Sneakers', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6177, 'Sensei''s Epic Sandals', 7, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6179, 'Rebel Boots', 7, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6180, 'Gold Sparkle Loafers', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6181, 'Purple Flower Sandals', 7, 160, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6182, 'Cheechee''s Boots', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6183, 'Lela''s Boots', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6184, 'Red Biker Shoes', 7, 275, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6185, 'Pink Biker Shoes', 7, 275, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6186, 'Butchy''s Boots', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6191, 'McKenzie''s Beach Shoes', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6192, 'Brady''s Sneakers', 7, 250, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6193, 'Brady''s Shoes', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6194, 'Villager Shoes', 7, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6196, 'Epic Knight Shoes', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6197, 'Enchanted Feet', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6200, 'Black Studded Shoes', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6201, 'Count Shoes', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6202, 'Swamp Monster Feet', 7, 50, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6203, 'Heavy Duty Boots', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6204, 'Grey B-Boy Sneakers', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6207, 'Golden Sneakers', 7, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6208, 'Amber Stilettos', 7, 275, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6209, 'Gold Bunny Slippers', 7, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6210, 'Snowy Night Boots', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6211, 'Woodland Reindeer Hooves', 7, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6212, 'Toy Robot Feet', 7, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6213, 'Polished Shoes', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6216, 'High Seas Boots', 7, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6217, 'White Patent Shoes', 7, 900, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6218, 'Black Hiking Boots', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6219, 'Gold Checkered Shoes', 7, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6220, 'Rainbow Checkered Shoes', 7, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6221, 'Sled Tube Boots', 7, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6222, 'Water Power Shoes', 7, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6223, 'Fire Power Shoes', 7, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6224, 'Wind Power Shoes', 7, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6225, 'Lightning Power Shoes', 7, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6226, 'Anna''s Traveling Shoes', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6227, 'Elsa''s Ice Queen Shoes', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6228, 'Oaken''s Shoes', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6229, 'Kristoff''s Boots', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6230, 'Sven Hooves', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6231, 'Hans'' Boots', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6232, 'Troll Feet', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6233, 'Coral Sneakers', 7, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6234, 'Sparkle Sneakers', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6235, 'Sneaky Sneakers', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6236, 'Sparkly Amber Shoes', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6237, 'Black Kicks', 7, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6238, 'Harvest Orange Slippers', 7, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6239, 'Pumpkin Sneakers', 7, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6240, 'Black Flats', 7, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6241, 'Ball and Chain', 7, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6242, 'Octolegs', 7, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6243, 'Festive Footwear', 7, 170, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6244, 'Purple Snuggly Boots', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6245, 'Beige Hiking Boots', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6246, 'Gray Booties', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6247, 'Trek Boots', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6248, 'Festive Socks', 7, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6249, 'Sky Blue Sneakers', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6250, 'Strawberry Sneakers', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6251, 'Kiwi Sneakers', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6252, 'Peach Sneakers', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6253, 'Strapped Sandals', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6255, 'Royal Blue Sneakers', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6256, 'Lime Green Sneakers', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6257, 'Hot Pink Sneakers', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6258, 'Sunshine Sandals', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6259, 'Rainbow Sandals', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6260, 'Pink and White Sandals', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6261, 'Glitzy Sandals', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6262, 'Lime Green Sandals', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6263, 'Fearful Shoes', 7, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6264, 'Disgusted Shoes', 7, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6265, 'Sad Shoes', 7, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6266, 'Delta Sneaks', 7, 0, TRUE, FALSE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (6267, 'Alpha Pumps', 7, 0, TRUE, FALSE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (6268, 'Mal''s Boots', 7, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6269, 'Evie''s Boots', 7, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6270, 'Jay''s Boots', 7, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6271, 'Carlos'' Boots', 7, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6272, 'Prince Ben''s Shoes', 7, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6273, 'Audrey''s Shoes', 7, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6274, 'Candy Corn Shoes', 7, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6275, 'Licorice Shoes', 7, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6276, 'Snapper Slippers', 7, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6277, 'Red Winter Boots', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6278, 'Arlo''s Feet', 7, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6279, 'Butch''s Feet', 7, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6281, 'Caramel Flats', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6282, 'Hikers', 7, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6283, 'Black Puffle Slippers', 7, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6284, 'Blue Puffle Slippers', 7, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6285, 'Brown Puffle Slippers', 7, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6286, 'Green Puffle Slippers', 7, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6287, 'Orange Puffle Slippers', 7, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6288, 'Pink Puffle Slippers', 7, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6289, 'Ghost Puffle Slippers', 7, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6290, 'Rainbow Puffle Slippers', 7, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6291, 'Red Puffle Slippers', 7, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6292, 'White Puffle Slippers', 7, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6293, 'Yellow Puffle Slippers', 7, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6294, 'Gold Puffle Slippers', 7, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6295, 'Canvas Cloud Shoes', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6296, 'Light Up Shoes', 7, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6297, 'Pink Fuzzy Boots', 7, 275, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6298, 'Blue Winter Boots', 7, 275, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (6999, 'Remove Feet Item', 7, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (7000, '101 Days Of Fun', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2009-09-11'),
 (7001, 'Fire', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2009-09-14'),
 (7002, 'Castle', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2009-09-22'),
 (7003, 'Football', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2009-10-09'),
 (7004, '4th Anniversary Cake', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2009-10-23'),
 (7005, 'Toboggan', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2009-11-06'),
 (7006, 'Magic Phial', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2009-11-13'),
 (7007, 'Hot Chocolate', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2009-11-20'),
 (7008, 'Christmas Bell', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2009-12-04'),
 (7009, 'Snowman', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2010-12-18'),
 (7010, 'Fireworks', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2010-01-02'),
 (7011, 'Puffer Fish', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2010-01-15'),
 (7012, 'Amethyst', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2010-01-22'),
 (7013, 'Speaker', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2010-01-29'),
 (7014, 'Feather', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2010-02-11'),
 (7015, 'Wagon', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2010-02-25'),
 (7016, 'Moss Key', 8, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2010-03-15'),
 (7017, 'Boot', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2010-03-11'),
 (7018, 'Cupcake', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2010-03-25'),
 (7019, 'Frog', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2010-04-09'),
 (7020, 'Leaf', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2010-04-22'),
 (7021, 'Shield', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2010-05-06'),
 (7022, 'Cream Soda Barrel', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2010-05-20'),
 (7023, 'Seashell', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2010-06-03'),
 (7024, 'Ruby', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (7025, 'Recycle', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2010-04-22'),
 (7026, 'Light Bulb', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2010-06-17'),
 (7027, 'Toothbrush', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-05-19'),
 (7028, 'Candy Apple Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2010-06-30'),
 (7029, 'Tambourine Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2010-07-15'),
 (7030, 'Compass Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2010-07-29'),
 (7031, 'Carabiner Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2010-08-12'),
 (7032, 'Popcorn Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2010-09-02'),
 (7033, 'Igloo Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2010-08-26'),
 (7034, 'Fair Ticket Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2010-09-09'),
 (7035, 'Sandwich Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2010-09-23'),
 (7036, 'Gold Feather Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2010-10-07'),
 (7037, 'Carrot Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2010-10-07'),
 (7038, 'Bat Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2010-10-21'),
 (7039, 'Water Tap', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2010-11-04'),
 (7040, 'Water Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2010-11-18'),
 (7041, 'Snowflakes Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2010-12-03'),
 (7042, 'Snow Globe Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2010-12-16'),
 (7043, 'Sleeping Bag Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-01-13'),
 (7044, 'Fire Extinguisher', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-01-27'),
 (7045, 'Ruby Brooch', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-12-28'),
 (7046, 'Party Favors Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-12-30'),
 (7047, 'Red Puffle Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-02-10'),
 (7048, 'Blue Puffle Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-02-10'),
 (7049, 'Green Puffle Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-02-24'),
 (7050, 'Black Puffle Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-02-24'),
 (7051, 'Viking Ship Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-02-24'),
 (7052, 'Purple Puffle Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-03-10'),
 (7053, 'White Puffle Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-03-24'),
 (7054, 'Yellow Puffle Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-03-24'),
 (7055, 'Orange Puffle Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-04-07'),
 (7056, 'Brown Puffle Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-04-07'),
 (7057, 'Pink Puffle Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-03-10'),
 (7058, 'Quartz Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-04-14'),
 (7059, 'Savanna Tree Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-04-20'),
 (7060, 'Medieval Shield', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-05-06'),
 (7061, 'Brazier Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-05-19'),
 (7062, 'Mermaid Shell Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2013-06-05'),
 (7063, 'Eye Patch Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-07-20'),
 (7064, 'Telescope', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (7066, 'Tropical Feather', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-07-20'),
 (7067, 'Old Key', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-07-20'),
 (7068, 'Enchanted Feather Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-02-04'),
 (7069, 'Red Electric Guitar Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-06-02'),
 (7070, 'Stereo Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-06-16'),
 (7071, 'EPF Badge', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-05-31'),
 (7072, 'Tiki Mask Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-07-06'),
 (7073, 'Gold Anchor Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-07-13'),
 (7074, 'Tropical Bird', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-07-27'),
 (7075, 'Expedition Key', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-08-25'),
 (7076, 'AC 3000 Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-08-25'),
 (7077, 'Checkered Flag Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-08-12'),
 (7078, 'Polar Paw Print', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-08-25'),
 (7079, 'Purple Balloon Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-09-21'),
 (7080, 'Balloon Bunch Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-09-07'),
 (7081, 'Milkshake', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-09-21'),
 (7083, 'Snow Cone Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-09-21'),
 (7084, 'Fence Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-10-04'),
 (7085, 'Crystal Ball Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-10-19'),
 (7086, 'Hot Sauce Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-11-02'),
 (7087, 'Blue Fish Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-11-16'),
 (7088, 'Bonsai Tree Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-12-01'),
 (7089, 'Reindeer Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-12-14'),
 (7090, 'Candy Swirl Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-12-28'),
 (7091, 'Milk ''N Cookies Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-12-22'),
 (7092, 'Holiday Stocking Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-12-19'),
 (7093, 'Candy Cane Duo Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-12-16'),
 (7094, 'CFC 2011 Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2011-12-14'),
 (7095, 'Liechtenstein', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (7096, 'Austria', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (7097, 'Conch Shell', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-01-11'),
 (7098, 'Beach Chair', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-01-25'),
 (7099, 'Camera', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-02-08'),
 (7100, 'Helm', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-02-23'),
 (7101, 'Shipwreck Beacon', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-02-23'),
 (7102, 'Cheese', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-03-07'),
 (7103, 'Cake', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-03-22'),
 (7104, 'Easter Basket', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-04-04'),
 (7105, 'Forest', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-04-18'),
 (7106, 'Round Ruby', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2013-08-14'),
 (7107, 'Black Helm', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-05-02'),
 (7108, 'Scorn Crest', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-05-16'),
 (7109, 'Bean Bag', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-05-30'),
 (7110, 'Crown of the Dragon King', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-05-16'),
 (7111, 'Superhero Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-06-13'),
 (7112, 'Supervillain Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-06-28'),
 (7113, 'Pink Mystery Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-07-25'),
 (7114, 'Dub-step Puffle', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-07-11'),
 (7115, 'Spotlight', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-07-25'),
 (7116, 'Sweet Star', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-07-18'),
 (7117, 'Mega Star', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-07-19'),
 (7118, 'Super Star', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-07-20'),
 (7119, 'Gold Star', 8, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-07-21'),
 (7120, 'Banana Peel', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-08-08'),
 (7121, 'Fruit Combo', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-08-22'),
 (7124, 'Windmill', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-09-05'),
 (7125, 'Triple Scoop', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-09-19'),
 (7126, 'Puffle Bat Key', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-10-03'),
 (7127, 'Gariwald Family Crest', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-10-31'),
 (7128, '7th Anniversary Hat Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-10-17'),
 (7129, 'Heavy Weights', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-11-14'),
 (7130, 'First Aid', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-11-19'),
 (7131, 'Square Ruby', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-10-31'),
 (7132, 'Herbert Security Clearance 1', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-11-14'),
 (7133, 'Herbert Security Clearance 2', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-11-15'),
 (7134, 'Herbert Security Clearance 3', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-11-17'),
 (7135, 'Herbert Security Clearance 4', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-11-19'),
 (7136, 'Herbert Security Clearance 5', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-11-22'),
 (7137, 'Herbert Security Clearance 6', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-11-24'),
 (7138, 'Holiday Cookies', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2013-12-26'),
 (7139, 'Snowflake', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2012-12-12'),
 (7140, 'CFC 2012 Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2013-12-19'),
 (7141, 'Dino Footprint', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2013-01-09'),
 (7142, 'Stone Hammer', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2013-01-23'),
 (7143, 'Shooting Star Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2013-02-06'),
 (7144, 'Golden Award Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2013-02-20'),
 (7146, 'Red Nose Comedy Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2013-03-06'),
 (7147, 'Rainbow Puffle Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2013-03-20'),
 (7148, 'Romania', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (7149, 'Bunny Rabbit Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2013-04-04'),
 (7150, 'Robot Villain Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2013-04-17'),
 (7151, 'Paper Lantern Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2013-05-02'),
 (7152, 'Elemental Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2013-05-15'),
 (7153, 'Snow Shuriken Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2013-05-29'),
 (7154, 'Neon Flamingo Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2013-06-12'),
 (7155, 'One Eyed Puffle Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2013-06-26'),
 (7156, 'Text Book Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2013-07-11'),
 (7157, 'R2-D2 Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2013-07-24'),
 (7158, 'Rebel Reward Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2013-08-01'),
 (7159, 'Snow Gem Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2013-05-22'),
 (7160, 'Calculator Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2013-08-07'),
 (7161, 'Friends Forever Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2013-08-21'),
 (7162, 'Coconut Smoothie Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2013-09-04'),
 (7163, 'Fluffy Crest Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2013-09-18'),
 (7164, 'Prank Fangs Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2013-10-02'),
 (7165, 'Tombstone', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2013-10-16'),
 (7166, 'Scissors', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2013-11-01'),
 (7167, 'Herbertech Logo', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2013-11-13'),
 (7168, 'Puffle Medic', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2013-11-27'),
 (7169, 'Ornament Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2013-12-11'),
 (7170, 'Penguin Nests Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2013-12-19'),
 (7171, 'Clean Water Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2013-12-20'),
 (7172, 'Holiday Gift', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2013-12-26'),
 (7173, '2013 CFC donation pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2013-12-18'),
 (7174, 'Stone Scissors', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2014-01-08'),
 (7175, 'Dino Snack', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2014-01-22'),
 (7176, 'Pixel Puffle', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2014-02-05'),
 (7177, 'Space Cart Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2014-02-19'),
 (7178, 'Pirate Skull', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2014-02-19'),
 (7179, 'Desert Cactus', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2014-02-19'),
 (7180, 'Lion Statue', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2014-03-05'),
 (7181, 'Pinata Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2014-03-19'),
 (7182, 'Greece', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (7183, 'Croatia', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (7184, 'Czech Republic', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (7185, 'Pakistan', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (7186, 'Latvia', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (7187, 'Slovenia', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (7188, 'Gibraltar', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (7189, 'Malta', 8, 20, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (7190, 'Picnic Bundle', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2014-05-14'),
 (7191, 'Puffle Park Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2014-04-16'),
 (7192, 'Neon O''Berry', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2014-04-30'),
 (7193, 'Cream Pie', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2014-04-02'),
 (7194, 'Space Academy Crest', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2014-05-28'),
 (7195, 'Orange Slice', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2014-06-11'),
 (7196, 'MVP', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2014-06-25'),
 (7197, 'Penguin Cup 2014 Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2014-06-18'),
 (7198, 'Life Preserver', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2014-07-09'),
 (7199, 'Cruise Ship', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2014-07-23'),
 (7200, 'Arendell Crest Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2014-08-06'),
 (7201, 'Snowman Puffle Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2014-08-20'),
 (7202, 'Fruit Smoothie Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2014-09-03'),
 (7203, 'Lemon Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2014-09-17'),
 (7204, 'Anniversary Balloons', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-10-14'),
 (7205, 'Ghost Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2014-10-29'),
 (7206, 'Scout Scarf', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2014-10-01'),
 (7207, 'Wooden Sword Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2014-11-12'),
 (7208, 'Crab Lock Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2014-11-26'),
 (7209, 'CFC 2014', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2014-12-17'),
 (7210, 'Merry Walrus', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2014-12-10'),
 (7211, 'Blue Crystal Puffle', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2014-12-22'),
 (7212, 'Rebels Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-01-07'),
 (7213, 'Imperial Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-01-21'),
 (7214, 'Headphones Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-02-04'),
 (7215, 'DJ K-Dance Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-02-18'),
 (7216, 'Pop Star Pin', 8, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '1970-01-01'),
 (7217, 'Rock Star Pin', 8, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '1970-01-01'),
 (7218, 'Dance Pin', 8, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '1970-01-01'),
 (7219, 'Dubstep DJ Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '1970-01-01'),
 (7220, 'Glitterpants Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-03-12'),
 (7221, 'Fishing Rod Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-04-01'),
 (7222, 'Puffle Guide Badge', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-03-18'),
 (7223, 'Pi Pie Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-03-18'),
 (7224, 'Spring Flower Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-04-29'),
 (7225, '?????? Footprint', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-04-15'),
 (7226, 'Clocktower Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-04-23'),
 (7227, 'Bracelet Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-04-23'),
 (7228, 'Flower Bouquet Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-04-23'),
 (7229, 'Deluxe Sandwich Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-04-23'),
 (7230, 'Portrait Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-04-23'),
 (7231, 'Anna and Elsa Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-04-23'),
 (7232, 'Arendelle Globe Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-04-23'),
 (7233, 'Old Fishing Rod Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-04-23'),
 (7234, 'Sparkling Headdress Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-04-23'),
 (7235, 'Birthday Cake Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-04-23'),
 (7236, 'Pixel Penguin Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-05-13'),
 (7237, 'Medieval Monster Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-05-27'),
 (7238, 'Beach Day', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-06-11'),
 (7239, 'Sunglasses', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-06-24'),
 (7240, 'Issue 500', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-05-20'),
 (7241, 'Inside Out', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-07-22'),
 (7242, 'Memory Orb', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-07-08'),
 (7243, 'Glam Cam', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-08-05'),
 (7244, 'Golden Hanger Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-08-19'),
 (7245, 'Royal Scepter Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-09-02'),
 (7246, 'Royal Crown Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-09-16'),
 (7247, 'Descendants Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-09-30'),
 (7248, 'Anniversary Balloons', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (7249, 'Happy Halloween Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-10-28'),
 (7250, 'Mustache Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-11-10'),
 (7251, 'Area 501 Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-11-25'),
 (7252, 'Holiday Happiness Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-12-09'),
 (7253, 'CFC Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-12-16'),
 (7254, 'Crystal Puffle Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2015-12-23'),
 (7255, 'The Good Dinosaur Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-01-06'),
 (7256, 'Power Fragment', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-02-24'),
 (7257, 'Stinky Cheese Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-02-03'),
 (7258, 'Fluffy Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-01-20'),
 (7259, 'Herbertech Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-02-17'),
 (7260, 'Puffle Balloon Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-03-02'),
 (7261, 'Puffle Medal Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-03-16'),
 (7262, 'Wild Puffle Treehouse Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-03-30'),
 (7263, 'Black O''berry Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-03-25'),
 (7264, 'Blue O''berry Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-03-17'),
 (7265, 'Brown O''berry Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-03-22'),
 (7266, 'Green O''berry Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-03-18'),
 (7267, 'Orange O''berry Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-03-23'),
 (7268, 'Pink O''berry Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-03-21'),
 (7269, 'Purple O''berry Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-03-24'),
 (7270, 'Rainbow O''berry Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-03-19'),
 (7271, 'Red O''berry Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-03-16'),
 (7272, 'White O''berry Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-03-27'),
 (7273, 'Yellow O''berry Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-03-20'),
 (7274, 'Gold O''berry Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-03-26'),
 (7275, 'Globe Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-04-13'),
 (7276, 'Sasquatch Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-05-04'),
 (7277, 'Zootopia Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-06-08'),
 (7278, 'Pawpsicle Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-05-19'),
 (7279, 'Little Rodentia Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-06-23'),
 (7280, 'Finding Dory Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-07-13'),
 (7281, 'Coffee Pot Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-07-06'),
 (7282, 'Tire Home Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-07-20'),
 (7283, 'Mancala Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-08-04'),
 (7284, 'Breakfast Bun', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-08-17'),
 (7285, 'Pizza Blueprint', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-08-31'),
 (7286, 'Argyle Pattern Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-09-14'),
 (7287, 'CP Decal Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-09-28'),
 (7288, 'Rubber Ducky Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-10-19'),
 (7289, 'Find Four Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-11-02'),
 (7290, 'Franky''s Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-11-11'),
 (7291, 'Stompin'' Bob''s Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-11-13'),
 (7292, 'G Billy''s Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-11-19'),
 (7293, 'Petey K''s Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-11-02'),
 (7294, 'DJ Cadence''s Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-11-05'),
 (7295, 'Klutzy''s Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-11-16'),
 (7296, 'Dub Step''s Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-11-08'),
 (7297, 'Keeper''s Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-11-22'),
 (7298, 'Jam Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-11-30'),
 (7299, 'Reunion Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-12-04'),
 (7300, 'Gingerbread House Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-12-08'),
 (7301, 'Nutcracker Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-12-11'),
 (7302, 'Beard Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-12-18'),
 (7303, 'CFC 2016 Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-11-30'),
 (7304, 'Holiday Ornament Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (7305, 'Candle Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2016-12-15'),
 (7306, 'Wink Emoji Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2017-01-04'),
 (7307, 'Passport Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2017-01-04'),
 (7308, 'Community Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2017-01-31'),
 (7309, 'Puffle Trivia Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (7310, 'Spy Trivia Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2017-02-03'),
 (7311, 'Party Trivia Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2017-02-04'),
 (7312, 'Items Trivia Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2017-02-06'),
 (7313, 'Ninja Trivia Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2017-02-08'),
 (7314, 'Games Trivia Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2017-02-10'),
 (7315, 'History Trivia Pin', 8, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, '2017-02-11'),
 (8000, 'Wood EPF Badge', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (8001, 'Ice EPF Badge', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (8002, 'Bronze EPF Badge', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (8003, 'Silver EPF Badge', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (8004, 'Gold EPF Badge', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (8005, 'EPF Certificate', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (8006, 'Fire Booster Deck', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (8007, 'Mission 11 Medal', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (8008, 'Gift from Dot', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (8009, 'EPF Phone', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (8010, 'Water Booster Deck', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (8011, 'Snow Starter Deck', 10, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9000, 'Scarlet Leaves Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9001, 'Noir Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9002, 'Detective Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9003, 'Carousel Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9004, 'Stormy Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9005, 'Planet Y Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9006, 'Giant Pumpkin Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9007, 'Gary Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9008, 'Ice Cave Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9009, 'Fairy Fables Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9010, 'Enchanted Forest Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9011, 'Christmas Carol Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9013, 'Pinata', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9014, 'Neon Mania Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9015, 'Cadence''s Autograph', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9016, 'Pet Shop Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9017, 'Leprechaun Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9018, 'Awards Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9019, 'Rockhopper Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9020, 'Penguin Awards Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9021, 'Aunt Arctic''s Autograph', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9022, 'Bright Flowers Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9023, 'Princess Castle Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9024, 'Tree House Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9025, 'Castle Hallway Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9026, 'Exclusive Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9027, 'Viking Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9028, 'Overgrown Ship Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9029, 'Rockhopper Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9030, 'Adventure Party Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9031, 'Tug Boat Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9032, 'Tiki Coffee Shop Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9033, 'Aqua Disco Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9034, 'The Melody Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9035, 'Purple Stars', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9036, 'Sensei''s Autograph', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9037, 'Purple Wave Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9038, 'Tropical Palm Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9039, 'Underwater Adventure Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9040, 'Football Field Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9041, 'Fair Background', 9, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9042, 'Fair Beacon Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9043, 'Green Plaid Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9044, 'Ghost Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9045, 'Pumpkin Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9046, 'Winter Wonderland Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9047, 'Garden Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9048, 'Holiday Wreath Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9049, 'Candy Cane Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9050, 'Coins for Change Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9051, 'Paint Splatter Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9052, 'Ocean Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9053, 'Coral Reef Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9054, 'Puffle Collage Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9055, 'Phoenix Queen''s Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9056, 'Secrets of the Bamboo Forest Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9057, 'Hidden Treasure Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9058, 'Interview Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9059, 'Recycle Decal Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9060, 'Dragon''s Lair Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9061, 'Adventure Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9062, 'Treasure Cove Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9063, 'A Day At The Beach Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9064, 'You''re A Star Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9065, 'Cadence Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9066, 'Cave Entrance Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9067, 'Whirlpool Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9068, 'Top Of The Mountain Background', 9, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9074, 'Circus Tent Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9075, 'Rockhopper Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9076, 'Haunted House Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9077, 'Candy Forest Path Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9078, 'Gary Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9079, 'Rainy Day Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9080, 'Water Tank Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9081, 'Holiday Fireplace Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9082, 'Skating at Dusk Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9083, 'Coins For Change Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9084, 'Ice Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9085, 'Neon Paint Splatter Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9086, 'Pile of Puffles Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9087, 'Pet Shop Puffles Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9088, 'Alien World Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9089, 'Desert Plane', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9090, 'Soda Beach', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9091, 'Medieval Arena', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9092, 'Cave Gate', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9093, 'Medieval Banners', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9094, 'Island Grove Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9095, 'Rookie''s Giveaway', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9098, 'Safari Park Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9099, 'Wall of Speakers Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9100, 'Music Jam Guitars Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9101, 'Penguin Band Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9102, 'Emerald Leaves', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9103, 'Adventure Camp', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9104, 'Snowboard BG', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9105, 'Summit BG', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9106, 'New Player Background', 9, 0, FALSE, FALSE, FALSE, FALSE, TRUE, FALSE, FALSE, now()),
 (9107, 'Fair Forts', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9108, 'Ball Pit Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9109, 'Ranch', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9110, 'Swamp Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9111, 'Trick Or Treat BG', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9112, 'New Gary Giveaway', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9113, 'Aunt Arctic Giveaway', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9114, 'Candy Ghost BG', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9115, 'Step Right Up Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9116, 'Bamboo River Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9117, 'Cherry Blossom Orchard Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9118, 'Lava Rock Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9119, 'All Wrapped Up Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9120, 'Balcony Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9121, 'Santa Seat', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9122, 'Holiday Tree Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9123, 'North Pole Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9124, 'Holiday Magic Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9125, 'Sweet Treat Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9126, 'Cleaned Up Dock BG', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9127, 'Rookie Anvil BG', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9128, 'In the Deep BG', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9129, 'Beach Day BG', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9130, 'Fashion Show', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9131, 'Red and Purple Plaid', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9132, 'Rainy Day', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9134, 'First Prize Puffle', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9135, 'PH Giveaway Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9136, 'Wildlife Den', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9137, 'Wizard''s Workshop', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9138, 'Dragon Peak', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9139, 'Gary Medieval Giveaway', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9140, 'Justice Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9141, 'Sinister Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9142, 'Press Conference', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9143, 'Mug Shot', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9144, 'Aunt Arctic Giveaway', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9145, 'Spotlight Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9146, 'Music Zone', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9147, 'Center Stage', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9148, 'Yellow Kahuna', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9149, 'Red Kahuna', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9150, 'Green Kahuna', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9155, 'Cadence''s New Giveaway', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9156, 'Rocky and Cece Giveaway', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9157, 'Ancient Cavern', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9158, 'Tropical Getaway', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9160, 'Rockhopper''s Tropical Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9167, 'Ghostamatron Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9168, 'Gary''s Giveaway', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9169, 'Haunted Mansion Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9170, 'Grim Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9171, 'Blackout', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9172, 'Secure System', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9173, 'Herbert''s Giveaway Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9174, 'Big Cozy Chair', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9175, 'Cookie Workshop', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9176, 'Cozy Fireplace', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9177, 'Rockhopper''s Holiday Giveaway', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9178, 'Frost Bite Cavern', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9179, 'Cliff Lookout', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9180, 'Tar Pits', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9181, 'Gary''s Prehistoric Giveaway', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9182, 'Green Screen', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9183, 'Studio Lot', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9184, 'Superstar', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9185, 'Action Scene', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9186, 'Space Squid', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9187, 'Club Penguin High', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9188, 'Comedy Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9189, 'Aunt Arctic''s Gift', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9190, 'Gary''s Explosive Giveaway', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9191, 'Puffle Hotel Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9192, 'Patio View Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9193, 'PH Giveaway', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9194, 'Hero Carrier Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9195, 'Villain Entrance Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9196, 'Aunt Arctic''s Press Giveaway', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9197, 'Operation Hot Sauce', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9198, 'Bamboo Grove', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9199, 'Snowy Mountain', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9200, 'Sensei''s Bonsai Giveaway', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9201, 'University Gates', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9202, 'Monster Arcade', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9203, 'Tatooine Sandcrawler', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9204, 'Tatooine''s Desert', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9205, 'Death Star Plans', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9206, 'Yavin 4', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9207, 'Dark Side Giveaway', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9208, 'Garage Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9209, 'Summer Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9210, 'Sunset Beach Background', 9, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9211, 'McKenzie and Brady''s Giveaway', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9212, 'Summer Jam Giveaway', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9213, 'Fairy Hut', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9214, 'Village Bakery', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9215, 'Rockhopper''s Evergreen Giveaway', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9216, 'Gary''s Potion Giveaway', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9217, 'Cadence''s Aug 2013 Giveaway', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9233, 'Spectral Sweets', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9234, 'Rookie''s Halloween Giveaway', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9235, 'This Old Town', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9236, 'Night-Time Forest', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9237, 'Puffle Vet', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9238, 'PH''s Golden Puffle Giveaway', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9239, 'Herbert''s Hoard Giveaway', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9240, 'CFC Train Track Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9242, 'Wildlife Corridor Protector', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9243, 'Habitat Preserver', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9244, 'Rockhopper''s 2013 Holiday Giveaway', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9245, 'Dino Chow Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9246, 'Prehistoric Barber Shop', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9247, 'Tumbleweed Town Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9248, 'BZZORT! Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9249, 'Daily Spinner Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9250, 'Rookie''s Giveaway', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9251, 'Penguin Band Giveaway', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9252, 'Kermit the Frog''s Giveaway', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9253, 'Walk of Fame', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9254, 'Brazilian Mural', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9255, 'Penguin Of The Week', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9256, 'Constantine Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9257, 'Muppets World Tour', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9258, 'Puffle Park Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9259, 'Save the Streams Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9260, 'Spoiler Alert Giveaway', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9261, 'Funny Hat Week Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9262, 'O''Blaster Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9263, 'Gary 3000 Giveaway', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9264, 'Fluffy Sandwich Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9265, 'Soccer Ball Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9266, 'School Dance Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9267, 'Go Team Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9268, 'Music Cruise Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9269, 'Anchors Aweigh Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9270, 'DJ Cole Giveaway', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9271, 'Zendaya Giveaway', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9272, 'Sabrina Giveaway', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9273, 'Violetta Giveaway', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9274, 'Snowflakes Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9275, 'Summer Snowman Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9276, 'Elsa''s Giveaway', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9277, 'Mystery Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9278, 'Mad Science Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9279, 'Skatepark Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9281, 'Ghostly Gates Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9282, 'Eerie Library Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9283, 'Pink Flamingos', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9284, 'Autumn Plaid Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9285, 'PH wild puffle player card giveaway', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9286, 'Merry Walrus', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9287, 'Evergreen Trees', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9288, 'Snowflake Wrapping', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9289, 'Rebels Takeover', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9290, 'Stormtrooper Legion', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9291, 'Join the Crew', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9292, 'SoundStudio Superstar', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9293, 'Hitting the Floor', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9294, 'Wild O''berry Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9295, 'Puffle Creature Party', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9296, 'Puffle Print Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9297, 'Owl The Things Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9298, 'Fluffy Stuffie Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9299, 'Cloud Forest Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9300, 'Tropical Bird Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9301, 'Digital Defender', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9302, 'Emotional Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9303, 'Dot''s Giveaway', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9304, 'Fashion Festival Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9305, 'Palace Background', 9, 60, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9306, 'Back-up Power', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9307, 'Emergency Thrusters', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9308, 'Gamma Shields', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9309, 'Gyro Stabilizer', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9310, 'Reef Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9311, 'Finding Dory Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9312, '11th Anniversary Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9313, 'Yarr''s Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9314, 'Clothing Designer Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9315, 'Elemental Balance', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (9316, 'Puffle Poster', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10101, 'Black Sunglasses', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10105, 'Eye Patch', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10106, 'Mask', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10107, 'Blue Sunglasses', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10110, 'Red Sunglasses', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10113, 'Black Glasses', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10120, 'Alien Mask', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10121, 'Green Sunglasses', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10123, 'Blue Superhero Mask', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10124, 'Pink Superhero Mask', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10125, 'Aviator Sunglasses', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10132, 'Blue Snorkel', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10133, 'Red Face Paint', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10134, 'Blue Face Paint', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10136, 'Ski Goggles', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10162, 'White Beard', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10180, 'Snare Drum', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10181, 'Pearl Necklace', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10182, 'Pendant', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10186, 'Ruffle Collar', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10194, 'Whistle', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10195, 'Reporter Camera', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10205, 'I Love My Puffle T-Shirt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10207, 'Snowflake T-Shirt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10210, 'Star T-Shirt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10213, 'Orange Snowsuit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10214, 'Black Bowtie', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10217, 'Cowboy Vest', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10220, 'Hockey Stick', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10227, 'Bee Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10228, 'Princess Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10233, 'Red Electric Guitar', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10236, 'Poncho', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10237, 'Red Suede Jacket', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10239, 'Wetsuit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10244, 'Ghost Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10245, 'Orange Space Suit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10246, 'Skeleton Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10247, 'Clown Suit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10254, 'Red Cheerleader Outfit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10255, 'Blue Cheerleader Outfit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10256, 'Ballerina', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10259, 'Shamrock Dress', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10260, 'Seafoam Dress', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10261, 'Black Suit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10262, 'Coffee Apron', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10263, 'Pizza Apron', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10265, 'Fishing Vest', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10266, 'Alien Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10267, 'Red Shorts', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10268, 'Astronaut Suit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10276, 'Robot Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10277, 'Red Hockey Jersey', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10278, 'Blue Hockey Jersey', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10282, 'Lighthouse Shirt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10284, 'Elf Suit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10286, 'Flamenco Dress', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10289, 'Long Johns', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10291, 'Leprechaun Tuxedo', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10295, 'Captain''s Coat', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10296, 'Red Letterman Jacket', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10297, 'Lifeguard Shirt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10298, 'Sport Life Jacket', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10299, 'Firefighter Jacket', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10301, 'Red Cape', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10302, 'Blue Cape', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10306, 'Red Backpack', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10307, 'Messenger Bag', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10308, 'Scuba Tank', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10309, 'Bee Wings', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10310, 'Fairy Wings', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10313, 'Hiking Backpack', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10315, 'Pink Cape', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10316, 'Supply Bag', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10323, 'Silver Watch', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10327, 'Lasso', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10330, 'Red Pompom', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10331, 'Pumpkin Basket', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10332, 'Magic Wand', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10333, 'Blue Pompoms', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10334, 'Cane', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10335, 'Pair of Maracas', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10338, 'Black Electric Guitar', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10340, 'Drumsticks', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10344, 'Life Ring', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10346, 'Fishing Rod', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10347, 'Flashlight', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10352, 'Black Sneakers', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10353, 'Ballet Shoes', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10357, 'Blue Sneakers', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10358, 'Black Dress Shoes', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10364, 'Blue Flippers', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10366, 'Bunny Slippers', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10367, 'Yellow Flippers', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10368, 'Cowboy Boots', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10370, 'Elf Shoes', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10372, 'Hiking Boots', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10373, 'Rubber Boots', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10374, 'Pirate Boots', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10376, 'Brown Sandals', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10380, 'Fuzzy Boots', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10385, 'Soccer Cleats', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10387, 'Orange Sneakers', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10401, 'Sombrero', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10403, 'Hard Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10404, 'Tan Cowboy Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10407, 'Red Propeller Cap', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10409, 'Graduation Cap', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10412, 'Tiara', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10414, 'Santa Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10424, 'Chef Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10425, 'Shamrock Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10429, 'Miners Helmet', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10432, 'Safari Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10435, 'Red Ball Cap', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10436, 'Blue Ball Cap', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10439, 'Park Ranger Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10446, 'Fishing Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10449, 'Fruit Headdress', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10452, 'Viking Helmet', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10458, 'Space Helmet', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10460, 'Gold Viking Helmet', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10461, 'Hockey Helmet', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10462, 'Red Football Helmet', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10463, 'Blue Football Helmet', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10464, 'Snowboard Helmet', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10465, 'Firefighter Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10471, 'Reindeer Antlers', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10472, 'Bee Antennae', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10473, 'Winged Helmet', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10476, 'Bard Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10477, 'Court Jester Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10478, 'Coral Crown', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10481, 'Headphones', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10482, 'Stocking Cap', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10487, 'Princess Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10488, 'Head Band', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10493, 'Tricorn Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10494, 'Golf Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10495, 'Marching Band Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10498, 'Puffle Bandana', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10499, 'Flower Headdress', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10651, 'The Spikester', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10653, 'The Funster', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10655, 'The Sportster', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10656, 'The Sunstriker', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10658, 'The Sidetied', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10660, 'Cocoa Bunny Ears', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10662, 'The Surf Knot', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10666, 'Pharaoh Headdress', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10667, 'King''s Crown', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10668, 'Knight Helmet', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10670, 'The Firestriker', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10671, 'The Sidetied Too', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10715, 'Gold Bracelets', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10717, 'Baseball Glove', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10718, 'Crystal Staff', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10719, 'Basketball', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10720, 'Blue Football Jersey', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10721, 'Red Football Jersey', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10722, 'Referee Jersey', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10724, 'Orange Shield', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10726, 'Football', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10727, 'Soccer Ball', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10728, 'Spikester Cuffs', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10729, 'Blue Electric Bass', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10761, 'Bard Outfit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10762, 'Court Jester', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10763, 'Fish Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10765, 'Shadow Guy Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10766, 'Gamma Gal Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10767, 'Squidzoid Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10768, 'Mermaid Outfit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10770, 'Safety Vest', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10773, 'Penguin Band Hoodie', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10775, 'Red Soccer Jersey', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10778, 'Blue Soccer Jersey', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10782, 'Cocoa Bunny Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10785, 'Ski Patrol Jacket', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10787, 'Beige Explorer Outfit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10788, 'Pharaoh Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10789, 'Mummy Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10791, 'Red Baseball Uniform', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10792, 'Blue Baseball Uniform', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10793, 'Royal Robe', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10794, 'Knight''s Armor', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10795, 'Dragon Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10799, 'Squire Outfit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10830, 'Sport Bikini', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10836, 'Red Basketball Jersey', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10837, 'Blue Basketball Jersey', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10839, 'Pop Music Shirt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10840, 'Orange Rocker Shirt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10842, 'Dazzle Dress', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (10845, 'Gray Pirate Coat', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11010, 'The Movie Star', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11013, 'Frankenpenguin Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11014, 'Faery Wig', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11015, 'Rad Scientist Wig', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11017, 'Brown Skater Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11018, 'Pink Skater Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11020, 'Chullo Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11026, 'Snowman Head', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11031, 'Kitty Kat Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11040, 'Blue Top Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11042, 'The Sidewinder', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11053, 'Amethyst Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11070, 'Burgundy Beanie', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11071, 'Canvas Rocker Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11075, 'Green Propeller Cap', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11083, 'Ring Master Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11089, 'The Carefree', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11090, 'Blue Skater Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11091, 'Snow Fairy Hair', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11093, 'Blizzard Wizard Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11094, 'Ladybug Antennae', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11105, 'Reindeer Head', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11108, 'The Tussle', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11115, 'The Phoenix Hair', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11125, 'Cumberband Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11128, 'Green Skater Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11133, 'Green Hard Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11134, 'Straw Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11137, 'White Cocoa Bunny Ears', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11139, 'The Trend', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11140, 'The Strawberry Chill', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11141, 'The Buttercup', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11143, 'The Damsel', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11146, 'Iron Helmet', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11153, 'Striped Pirate Bandana', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11154, 'Swashbuckler''s Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11165, 'The Rustic', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11166, 'Blue Striped Fedora', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11172, 'Red Mohawk', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11181, 'Aqua Hipster Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11182, 'The Sunny', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11191, 'Purple Bat Wings R', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11197, 'Pink and Yellow Chullo', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11198, 'Green Pom Pom Toque', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11203, 'Blue Goggles', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11223, 'The Short and Snappy', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11224, 'The Firey Flare', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11225, 'Blue Puffle Hat R', 2, 9, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11230, 'The Storm R', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11234, 'Green Viking Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11243, 'The Scenester', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11244, 'The Brilliant', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11254, 'The Raven', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11255, 'The Jammin''', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11302, 'The Sundae Surprise', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11303, 'The Balloonist', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11310, 'Zebra Tails', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11311, 'Freaky Tiki Headdress', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11312, 'Armored Viking Helmet', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11313, 'The Nocturnal', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11314, 'Nutcracker Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11331, 'Springy Santa Hat R', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11334, 'The Tree Topper', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11335, 'The Hornament Hat R', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11343, 'Country Jester Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11344, 'The Highlander', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11345, 'The Curly Lochs', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11346, 'The Ocean Sunset', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11347, 'The Corsair', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11348, 'The Blue Shipshape', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11349, 'Rainbow Puffle Toque', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11390, 'Forest Fairy', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11391, 'The Summer Breeze', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11392, 'The Wind Swept', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11393, 'The Rose Weave', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11403, 'The Sumo', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11423, 'Stealth Tracker', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11434, 'Green Indie Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11435, 'Tiki Tiki Headdress', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11436, 'Purple Headphones', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11455, 'Watermelon Helmet', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11456, 'Orange MP3000', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11473, 'Hot Cross Bun', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11474, 'The Thrill', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11475, 'Gray Werewolf Head', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11477, 'Happy Birthday Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11503, 'EPF Comm Headset', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11507, 'Fish Dog Vendor Cap', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11534, 'Saber Toothed Helmet', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11550, 'Neon Green Helmet', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11561, 'Puffle Ball Cap', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11586, 'The Sashimi Chef R', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11659, 'EVA Space Helmet', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11673, 'The Boho', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11674, 'The Wisp', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11675, 'The Side Slick', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11676, 'The Wave Runner', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11677, 'The Safari', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11678, 'The Sass', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11679, 'The Beatster', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11680, 'The Urban', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11681, 'The Spike n'' Style', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11699, 'Blue Sweat Band', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11700, 'Yellow Sweat Band', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11719, '5 Year Celebration Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11730, 'Head Lamp', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11751, 'Jolly Roger Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11833, 'French 5th Year Party Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11854, 'Blue Border Collie Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11860, 'Blast Off Hat R', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11878, 'Superfan Hat R', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11882, 'The Up and About', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11883, 'The Sandy', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (11922, 'Spanish Fifth Year Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (12003, 'Black Scuba Mask', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (12005, 'Red Ski Goggles', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (12006, 'Black Diva Sunglasses', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (12008, 'Star Glasses', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (12012, 'Curly Moustache', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (12018, 'Green Face Paint', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (12019, 'Yellow Face Paint', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (12024, 'Silver Aviators', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (12041, 'Viking Beard', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (12042, 'Grim Glasses', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (12055, 'Hot Sauce Helmet', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (12056, 'Comm Headset', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (12058, 'Celadon Alien Mask', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (12060, 'Mask of Justice', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (12076, 'Strawberry Sunglasses', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (12078, 'Zombie Mask', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (12079, 'The BRRAAAAAIIINS', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (12099, 'Crimson Shades', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (12127, 'Hipster Glasses', 3, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (13002, 'Silver Whistle', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (13003, 'Gold Medal', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (13006, 'Faery Wings', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (13007, 'Yellow Cape', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (13010, 'Candycane Scarf', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (13017, 'Blue Boa', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (13020, 'Green Hooded Cloak', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (13028, 'Jet Pack', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (13033, 'Snow Fairy Wings', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (13045, 'Pastel Beaded Necklace', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (13051, 'Green Checkered Tote', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (13054, 'Expedition Backpack', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (13057, 'Blue Cotton Scarf', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (13059, 'Yellow Designer Scarf', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (13060, 'Green and Blue Scarf', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (13072, 'Emerald Necklace', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (13117, 'Purple Beaded Necklace', 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (13132, 'Snow Launcher', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (13142, 'Bling Bling Necklace', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (13143, 'Fish Dog Vendor Pouch', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14004, 'Red Track Jacket R', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14016, 'Frankenpenguin Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14017, 'Faery Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14018, 'Rad Scientist Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14035, 'White Parka', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14040, 'Purple Figure Skating Dress', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14041, 'Snowman Body', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14046, 'Red Hoodie', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14051, 'Blue Daisy Hoodie', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14052, 'Crosshatch Hoodie', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14063, 'Pet Shop Apron', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14065, 'Blue Tuxedo', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14073, 'Black Penguin Band Hoodie', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14074, 'Tie Dye Shirt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14075, 'White Gi R', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14080, 'Amethyst Dress', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14082, 'Blue Dragon Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14084, 'Green Tunic', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14088, 'Green Soccer Jersey', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14091, 'Goalie Jersey', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14102, 'Electro T-Shirt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14104, 'Green Peacoat', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14105, 'Dark Denim Jacket', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14111, 'Lobster Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14116, 'Green Cheerleader Outfit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14119, 'Ring Master Outfit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14123, 'Snow Fairy Dress', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14124, 'Leafy T-Shirt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14125, 'Black Skater Shirt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14126, 'Santa Suit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14128, 'Blizzard Wizard Robes', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14129, 'Ladybug Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14138, 'Spider Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14141, 'Yeti Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14142, 'Silly Scarecrow Suit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14143, 'Green Hockey Jersey', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14145, 'Reindeer Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14147, 'Tree Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14150, 'Beta Grid Sweater', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14151, 'Puffle Flame Hoodie', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14152, 'Pink Kimono', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14153, 'Tri-Color Rugby Shirt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14191, 'Phoenix Dress', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14199, 'Classy T-Shirt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14201, 'Green Crosshatched Hoodie', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14202, 'Surf Sweater', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14203, 'Vintage Hoodie', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14204, 'Yellow Hockey Jersey', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14205, 'Yellow Cheerleader', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14206, 'Water Suit 3000', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14207, 'Green Recycle T-Shirt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14209, 'White Cocoa Bunny Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14212, 'Crab Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14213, 'Green Skater Shirt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14214, 'Pastel Petal Dress', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14216, 'Damsel''s Dress', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14219, 'Iron Armor', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14220, 'Squire''s Outfit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14222, 'EPF Suit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14227, 'Pirate Lass', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14228, 'Swashbuckler Coat', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14239, 'Rustic Tunic and Skirt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14240, 'Trendy Shirt Blue', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14259, 'Yellow Hoodie', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14264, 'Blue Tracksuit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14267, 'Leather Pilots Jacket', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14268, 'Red Polka Dot Dress', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14278, 'Purple Whirl Snowsuit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14279, 'Blue Winter Jacket', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14313, 'Fluorescent Freestyle Threads', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14314, 'Casual Summer Threads', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14332, 'Lightning Gi R', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14351, 'Dapper Sweater Vest', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14352, 'Chic Sweater Vest', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14361, 'Purple Dragon Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14363, 'Golden Quilted Coat', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14364, 'Puffle Skate Shirt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14436, 'UK Hoodie', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14437, 'Yellow Wheeler', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14438, 'Elephant Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14439, 'Jolly Roger Dress', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14440, 'Freaky Tiki Outfit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14441, 'Viking Captain Tunic', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14442, 'Classy Vampire Outfit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14443, 'Present Outfit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14444, 'Holiday Lights', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14445, 'Nutcracker Outfit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14446, 'Scarlet Admiral Coat', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14465, 'Snowflake Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14485, 'Tranquility Robe', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14486, 'Red Rocker T-Shirt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14519, 'Loch Ness Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14520, 'Country Jester', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14521, 'Emerald Kilt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14522, 'Shamrock Skirt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14523, 'Blueberry Bunny Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14524, 'Ocean Bloom Mermaid', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14525, 'Sea Snail', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14526, 'Deep Diver Suit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14527, 'Blue Buccaneer Dress', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14528, 'Helmsman Vest', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14529, 'Orange Pufflemania Jacket', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14530, 'The Trenchcoat Reporter', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14531, 'The Explorer Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14541, 'Club Shirt (UK redemption shirt)', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14554, 'Treasure Chest Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14562, 'Undercover Ops Suit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14563, 'Hot Sauce Gi', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14564, 'Glowing Grid Jacket', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14565, 'Water Lily Dress', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14566, 'Flit Hoodie', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14567, 'Skater Threads', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14568, 'Hard Rocker Outfit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14569, 'Purple Cloud Look', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14570, 'Speed Boat', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14571, 'Bubblegum Cheerleader Outfit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14574, 'Classy Agent Suit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14581, 'Celadon Alien Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14605, 'Major Tunage Shirt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14608, 'Hazard Suit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14609, 'Sumo Belt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14650, 'Valley Robes', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14651, 'Mountain Robes', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14652, 'Mega Suit 3000', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14653, 'Rainbow Puffle Shirt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14654, 'Ornate Band', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14655, 'Stone Ninja Suit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14656, 'Forest Ninja Suit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14657, 'Puffle Scuffle Hoodie', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14658, 'Easy Listening Shirt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14659, 'Dubstep Puffle T-Shirt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14678, 'Raspberry Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14679, 'Green Grape Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14680, 'Orange Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14681, 'Strawberry Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14682, 'Kiwi Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14683, 'Orange Citrus Hoodie', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14684, 'Cherry T', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14685, 'Tribal Dress', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14687, 'Toga', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14703, 'Elite Body Armor', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14704, 'Zombie Shirt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14705, 'Ghoul Outfit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14706, 'Breezy Dress', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14707, 'Red Stylin'' Jacket', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14708, 'Gray Werewolf Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14709, 'Clear Sky Suit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14710, 'Crashing Wave Gi', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14713, 'Fire Ninja Gi', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14714, 'Neon Skeleton Hoodie', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14728, 'Snowstorm Suit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14729, 'Blue Hip Hop Hoodie', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14730, 'Pink Hip Hop Hoodie', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14737, 'Green Spider Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14748, 'Kendo Suit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14750, 'Delux Tux', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14776, 'Bone Gray Togs', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14784, 'Dig Team Gear', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14813, 'White and Black Puffle Sweater', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14843, 'Incredible Hero Suit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14844, 'Sashimi Chef Uniform R', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14901, 'Gold Puffle Hoodie', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14904, 'TIE Fighter Pilot Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14907, 'Percy Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14911, 'EVA Space Suit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14926, 'Cargo Jacket', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14927, 'Smooth Stylin'' Outfit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14928, 'Color Me Bold Outfit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14929, 'Classic Summer Outfit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14930, 'Summer Frills Outfit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14931, 'A Splash of Summer Dress', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14932, 'Green Dubstep Puffle Tee', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14933, 'Baby Blue Blouse', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14934, 'Volcanic Robes', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14935, 'Snow Peak Robes', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14936, 'Smooth Styles', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14937, 'Leafy Hoodie', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14953, 'Yellow Sweatsuit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14954, 'Grey Tracksuit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14968, 'R-8-Legged Monstrosity', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (14969, 'Blue Blend Hoodie', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15005, 'Candy Apple', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15007, 'Blue Book', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15008, 'MP3000', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15017, 'Keytar', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15026, 'Stuffed Bunny', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15031, 'Megaphone', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15032, 'Binoculars', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15036, 'Double Necked Guitar', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15040, 'Cloud Wave Bracers R', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15041, 'Fuzzy Wrist Bands', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15045, 'Laptop', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15046, 'Racing Sled Green', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15048, 'White Pompoms', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15049, 'Cloudy Umbrella', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15053, 'Black MP3000', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15056, 'Brown Leather Cuffs R', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15058, 'Staff and Shield', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15061, 'Pirate Arm Bands_R', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15066, 'Green Keytar', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15079, 'Popcorn', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15085, 'Green Mittens', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15086, 'Grey Mittens', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15112, 'Bunch of Balloons', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15116, 'Cherry Balloon', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15120, 'eReader', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15124, 'Striped Pompoms', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15125, 'Freaky Tiki Staff', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15134, 'Country Jester Staff', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15142, 'Purple Boom Box', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15143, 'Hard Rock Guitar', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15149, 'Saxophone', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15159, 'Purple Boom Box R', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15165, 'Tiki Shield', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15193, 'Checkered Flag', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15194, 'Blue Puffle Stuffie R', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15199, 'Ugg Club', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15200, 'Triceratops Shield', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15215, 'Neon Green Longboard', 6, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15250, 'Sushi Combo R', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15333, 'Overflowing Suitcase', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15363, 'Blue Hand Weights', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15364, 'Tennis Gear', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15365, 'MP3000 Workout', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15366, 'Pink Hand Weights', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15367, 'Kloo Horn', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15373, 'Rescue Off-Roader', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15383, 'BFF Bracelet', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (15459, 'Leather Bracelets', 6, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16008, 'Zoot Shoes', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16009, 'Green Bunny Slippers', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16011, 'Figure Skates', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16013, 'Hockey Skates', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16015, 'White Dress Shoes', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16018, 'Light Up Shoes', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16023, 'Canvas Cloud Shoes', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16028, 'Beaded Slippers', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16029, 'Ladybug Shoes', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16035, 'Red Untied Sneakers', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16036, 'Red Sneakers', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16037, 'Green Sneakers', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16038, 'Yellow Sneakers', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16041, 'Purple Checkered Shoes', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16046, 'Brown Canvas Shoes', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16051, 'Green Untied Sneakers', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16055, 'Pink Fuzzy Boots', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16056, 'Blue Winter Boots', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16061, 'Loose Yellow Sneaks', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16066, 'Blue Buckle Shoes', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16069, 'Purple Dragon Feet', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16070, 'Green Light Up Shoes', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16071, 'Cherry Pumps', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16099, 'Black Flippers', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16114, 'Undercover Shoes', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16125, 'Aqua Sneakers', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16131, 'Sweet Shoes', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16132, 'Gray Werewolf Feet', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16143, 'Cool Ski Boots', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16187, 'Red and White Sneaks', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16188, 'Grey Loafers', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16189, 'Electric Purple Runners', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16190, 'Brown Leather Sandals', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16198, 'Barefood Runners', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (16199, 'Neon Sneakers', 7, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (19069, 'Go Green Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (19070, 'Go Yellow Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (19071, 'Go Blue Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (19072, 'Go Red Background', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (19126, 'Top Trumps Exclusive', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (19151, 'Very Cherry', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (19152, 'Crazy Kiwi', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (19153, 'Fruit Frenzy', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (19154, 'Wacky Watermelon', 9, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21000, 'Rainbow Toque', 2, 300, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21001, 'The Thundercloud', 2, 300, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21002, 'The Starlight', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21003, 'The Do-Re-Mi', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21004, 'The Wistful', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21005, 'The Firecracker', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21006, 'The Silver Sweep', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21007, 'The Fiery Flair', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21008, 'The Graceful', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21009, 'The Flame', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21010, 'The Riot', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21011, 'Rainbow Winged Helm', 2, 100, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21012, 'The Handyman', 2, 100, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21013, 'The Shadow Guy', 2, 200, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21014, 'The Gamma Gal', 2, 200, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21015, 'Rainbow Cap', 2, 300, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21016, 'The Wave Washed', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21017, 'The Sunshine', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21019, 'The Curl King', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21020, 'The Buzz Star', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21021, 'The Anger', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21022, 'The Joyful', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21023, 'The Sadness', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21024, 'The Fear', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21025, 'The Disgust', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21026, 'Riley''s Helmet', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21027, 'Bing Bong''s Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21029, 'EPF Earpiece', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (21030, 'Delta Fedora', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (21031, 'The Alpha', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (21032, 'Dot''s Up-Do MASCOT', 2, 9999, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21033, 'The Mal', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21034, 'The Evie', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21035, 'The Jay', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21036, 'The Carlos', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21037, 'The Prince Ben', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21038, 'The Audrey', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21039, '10th Anniversary Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21040, 'The Neon Night', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21041, 'The Firebrand', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21042, 'The Black Kitty', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21043, 'The Scary Gary', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21044, 'The Arctic-bot', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21045, 'The Bothopper', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21046, 'The DJ-bot', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21047, 'The Dot-bot', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21048, 'The Sensei-bot', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21049, 'The Bot Handler', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21050, 'The Rookie-bot', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21051, 'The Herbot', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21052, 'Tinfoil Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21053, 'The Ember', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21054, 'The Showstopper', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21055, 'The Rider', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21056, 'The Chaser', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21057, 'The Shine', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21058, 'The Blue Ombre', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21059, 'The Lavender Ombre', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21060, 'The Rose Ombre', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21061, 'EPF Space Helmet', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21062, 'The Gift', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21063, 'Green Jolly Roger', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21064, 'Holiday Spring Hat', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21065, 'Forester Hat', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21066, 'The Slopes', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21067, 'The Night Sky', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21068, 'The Spot', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21071, 'The Noir', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21072, 'The Valorous', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21073, 'The Rough Cut', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21074, 'Brown Skater Cap', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21075, 'Kitty Kat Toque', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21076, 'Cranberry Beanie', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21077, 'Dance Battle Cap', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21078, 'T-Rex Puffle Cap', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21079, 'Deer Puffle Cap', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21080, 'Brown Puffle Cap', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21081, 'Raccoon Puffle Cap', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21082, 'Orange Alien Cap', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21083, 'Stegosaurus Puffle Cap', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21084, 'Ghost Puffle Cap', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21085, 'Rainbow Puffle Cap', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21086, 'Triceratops Puffle Cap', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21087, 'Rabbit Puffle Cap', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21088, 'Unicorn Puffle Cap', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21089, 'Gold Puffle Cap', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21090, 'The Fluffy', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21091, 'Earth Hat', 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21094, 'Wooden Party Hat', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21095, 'Nick Wilde Mask', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21096, 'Judy Hopps Mask', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21097, 'Bellwether Mask', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21098, 'Clawhauser Mask', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21099, 'Lionheart Mask', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21100, 'Chief Bogo Mask', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21101, 'Flash Mask', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21102, 'Donut Destroyer', 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21103, 'Pink Skater Hat', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21104, 'Burgundy Beanie', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21105, 'The Buttercup', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21106, 'Retro Cap', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21107, 'The Firey Flare', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21108, 'The Up and About', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21109, 'The Washed-Up', 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21110, 'The Ocean Sunset', 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21111, '11th Anniversary Hat', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21112, 'The Dark Whorl', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21113, 'The Spooky Spike', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21114, 'The Pink and Pointy', 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21115, 'The Tour Bandana', 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21116, 'Bamboo Hat', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21117, 'Outback Holiday Hat', 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (21118, 'Iceberg Tipper', 2, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24000, 'Caroler''s Dress', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24001, 'Snowy Night Parka', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24002, 'Woodland Reindeer Costume', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24003, 'Ice Princess Dress', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24004, 'Blue Snowflake  Bauble', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24005, 'White Snowflake Bauble', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24006, 'Shortbread Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24008, 'Toy Robot Costume', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24011, 'Yellow Penguin Stuffie', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24012, 'Jack-in-the-Box Outfit', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24013, 'Holiday Conductor Uniform', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24014, 'Train Engineer Outfit', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24015, 'Teddy Bear Costume', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24016, 'Rockhopper''s Santa Jacket', 5, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24017, 'Flower Palm Dress', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24018, 'Feather Fringe', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24019, 'Cavegeek Outfit', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24020, 'Zebra-print Dress', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24021, 'Collared Blue Shirt', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24022, 'Pattern Fashionista', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24023, 'The Insta-Hero', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24024, 'Prehistoric Leisure Suit', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24025, 'Glam Glam Gown', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24026, 'Purple Fur Vest', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24027, 'Powder Blue Bone Dress', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24028, 'Leopard-print Two-piece', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24029, 'Allosaurus Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24030, 'Archaeologist Outfit', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24031, 'Stone Age Expedition Shorts', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24032, 'Cozy Toga', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24033, 'Magma Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24034, 'Caveguin Pizza Apron', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24036, 'Galactic Space Suit', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24037, 'Pixel Puffle Tee', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24038, 'Lucky Yellow Hoodie', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24039, 'Puffle Wrangler Outfit', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24040, 'High Seas Coat', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24041, 'Carnival Barker Outfit', 5, 2000, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24042, 'Online Safety Sweater', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24043, 'CP Air Uniform', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24044, 'Floral Shirt', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24045, 'Fozzie Bear Costume', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24046, 'Kermit the Frog Costume', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24047, 'Miss Piggy Costume', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24048, 'The Great Gonzo Costume', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24049, 'Bull Target Costume', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24050, 'Beaker Costume', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24051, 'Animal Costume', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24053, 'Sam Eagle Costume', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24054, 'Swedish Chef Costume', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24055, 'Striped Hoodie and Jacket', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24056, 'T-Rex Puffle Tee', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24057, 'Outback Traveler Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24058, 'Junior Explorer Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24059, 'Galactic Hoodie', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24060, 'Colored Hearts Hoodie', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24061, 'Rainbow Paint Hoodie', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24062, 'I Heart My Gold Puffle T-Shirt', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24063, 'Blue Border Collie Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24064, 'Orange Tabby Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24065, 'Fire Hydrant Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24067, 'Puffle Vet Coat', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24068, 'Puffle Trainer Outfit', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24069, 'O''berry Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24070, 'Adventurous Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24071, 'Loyalty Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24072, 'Inventive Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24073, 'Zany Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24074, 'Playful Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24075, 'Intense Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24076, 'Sporty Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24077, 'Gentle Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24078, 'Creative Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24079, 'Fabulous Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24080, 'Regal Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24081, 'Lucky Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24082, 'Orange Tabby Costume', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24083, 'Blue Border Collie Costume', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24084, 'Pufflescape Ball', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24085, 'Scratching Post Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24086, 'Starship Suit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24087, 'Zero G Dress', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24088, 'Interstellar Gown', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24089, 'Stellar Suit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24090, 'Fluffy Gown', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24091, 'Sled Tube Suit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24092, 'Holo DJ Suit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24093, 'Asteroid Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24094, 'UFO Costume', 5, 850, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24095, 'Planet Costume', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24096, 'Star Out Costume', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24097, 'Final Frontier Suit', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24098, 'We Come in Peace Costume', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24099, 'Planet Zeta  Costume', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24100, 'Protobot Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24102, 'Blue Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24103, 'Blue Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24104, 'Blue Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24105, 'Blue Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24106, 'Red Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24107, 'Red Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24108, 'Red Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24109, 'Red Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24110, 'Yellow Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24111, 'Yellow Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24112, 'Yellow Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24113, 'Yellow Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24114, 'Green Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24115, 'Green Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24116, 'Green Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24117, 'Green Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24118, 'Blue Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24119, 'Blue Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24120, 'Blue Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24121, 'Blue Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24122, 'Red Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24123, 'Red Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24124, 'Red Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24125, 'Red Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24126, 'Yellow Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24127, 'Yellow Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24128, 'Yellow Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24129, 'Yellow Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24130, 'Green Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24131, 'Green Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24132, 'Green Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24133, 'Green Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24134, 'Referee Uniform', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24137, 'Victory Splash Jug', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24138, 'Penguin Cup Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24139, 'Yellow Keeper Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24140, 'Red Keeper Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24141, 'Blue Keeper Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24142, 'Green Keeper Kit', 5, 100, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24143, 'Coach Outfit', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24146, 'Thrift Shop Style', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24147, 'Indie Rocker Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24148, 'Lifesaver Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24149, 'Shark Costume', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24150, 'Cruise Host', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24151, 'Cruise Captain Jacket', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24152, 'Rugby Team Shirt', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24153, 'Punk Fluffy Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24154, 'Here Comes Treble', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24155, 'City Lights T-Shirt', 5, 325, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24156, 'Beat Dropper Outfit', 5, 375, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24157, 'Urban Diva Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24159, 'Go-Karter Suit', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24173, 'Marshmallow Costume', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24174, 'Olaf''s Costume', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24175, 'Anna''s Coronation Dress', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24176, 'Anna''s Traveling Clothes', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24177, 'Elsa''s Coronation Dress', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24178, 'Elsa''s Ice Queen Dress', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24179, 'Oaken''s Bunad', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24180, 'Sven Costume', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24181, 'Kristoff''s Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24182, 'Hans'' Uniform', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24183, 'Troll Costume', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24184, 'Blizzard', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24185, 'Reindeer Floaty', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24186, 'Solid Ice', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24187, 'Hans'' Horse', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24188, 'Knitted Wool Sweater', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24189, 'Kristoff''s Sleigh', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24190, 'Icy Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24191, 'Tank Top Skater Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24192, 'Security Guard Uniform', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24193, 'Skate-Style Hoodie', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24194, 'Punk Puffle Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24195, 'Coming Up Roses', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24196, 'Summer Skater Tank', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24198, 'Inspired Artist', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24199, 'Sharks'' Mascot Costume', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24200, 'Sharks'' Training Outfit', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24201, 'For Science!', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24202, 'Clerk Outfit', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24203, 'High School Hoodie Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24204, 'Orange Skater Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24207, 'Lemon Button Blouse', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24208, 'Fall Leather Jacket', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24209, 'Enchantdress', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24210, 'Halloween Tee', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24211, 'Happy Punk O''Ween', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24212, 'Ghost Puffle Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24213, 'Spectral Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24214, 'Web Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24215, 'Plaid Hoodie Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24216, 'Autumn Floral Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24217, 'High Seas Fashion', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24218, 'Salty Dog Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24219, 'First Maiden Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24220, 'Pearl Mermaid Costume', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24221, 'Ship Shape Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24222, 'Captain''s Greatcoat', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24223, 'Lost Sailor''s Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24224, 'Merguin Fin', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24225, 'Pirette', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24226, 'Island Scraps', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24227, 'Deserted Beachwear', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24228, 'Lost Lynx', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24229, 'Crock Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24230, 'Master Pirate Outfit', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24231, 'Commander Coat', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24232, 'Klutzy Disguise', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24233, 'Snow Day Parka', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24234, 'Green Parka', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24235, 'Orange Fade Hoodie', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24236, 'Think Pink Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24237, 'Purple Parka', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24238, 'Tree Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24239, 'Merry Walrus Suit', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24240, 'Jingle Apron', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24241, 'Blue Crystal Puffle Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24243, 'Ezra''s Outfit', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24244, 'Kanan''s Outfit', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24245, 'Sabine''s Outfit', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24246, 'Zeb''s Outfit', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24247, 'Hera''s Outfit', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24248, 'The Inquisitor''s Outfit', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24249, 'Agent Kallus'' Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24250, 'Vizago Costume', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24251, 'Wookiee Costume', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24252, 'Minister Tua''s Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24253, 'Business Fashion', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24254, 'Glam Star Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24255, 'Orange Puffer Vest', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24256, 'Black Leather Jacket', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24257, 'Pop Star Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24258, 'Dance Crew Outfit', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24259, 'Dubstep DJ Outfit', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24260, 'SoundStudio Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24261, 'Gold Puffle Costume', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24262, 'Far Out Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24263, 'Cray On Hoodie', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24264, 'Green Raccoon Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24265, 'Red Deer Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24266, 'White Rabbit Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24267, 'Yellow Unicorn Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24268, 'Puffle Guide Uniform', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24270, 'Plaidasaurus Shirt', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24271, 'Owl I Want Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24274, 'Anna''s Birthday Dress', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24275, 'Elsa''s Spring Dress', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24276, 'Kristoff''s Party Outfit', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24277, 'Rainbow Car', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24278, 'Rainbow Smirk Hoodie', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24279, 'Springtime Sass', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24280, 'Pixelhopper Shirt', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24281, 'White Pixel Puffle Tee', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24282, 'Green Pixel Puffle Tee', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24283, 'Purple Pixel Puffle Tee', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24284, 'Pink Pixel Puffle Tee', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24285, 'Gold Pixel Puffle Tee', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24286, 'Red Pixel Puffle Tee', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24287, 'Black Pixel Puffle Tee', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24288, 'Yellow Pixel Puffle Tee', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24289, 'Brown Pixel Puffle Tee', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24290, 'Orange Pixel Puffle Tee', 5, 250, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24291, 'Rap Battler', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24292, 'Lime Lyricist', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24293, 'Mint Beach Dress', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24294, 'Coral Beach Dress', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24295, 'Shadow Guy Reboot', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24296, 'Starfish Costume', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24297, 'Gamma Gal Reboot', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24298, 'Ancient Green Dragon', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24299, 'Caramel Apple Costume', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24300, 'Lacy Shorts Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24302, 'After Surf Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24303, 'Ocean Cool Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24304, 'Seaside Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24306, 'Angry Outfit', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24307, 'Joyful Dress', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24308, 'Sad Outfit', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24309, 'Frightened Outfit', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24310, 'Disgusted Dress', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24311, 'Bing Bong Outfit', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24312, 'Rainbow Unicorn Outfit', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24313, 'Delta Suit', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (24314, 'Alpha Suit', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, TRUE, FALSE, now()),
 (24315, 'Dot''s Evergreen Outfit MASCOT', 5, 9999, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24316, 'Teal Tee', 5, 250, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24317, 'Green Tee', 5, 250, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24318, 'Pink Tee', 5, 250, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24319, 'Mal''s Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24320, 'Evie''s Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24321, 'Jay''s Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24322, 'Carlos'' Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24323, 'Prince Ben''s Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24324, 'Audrey''s Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24325, 'Skele-Tee', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24326, 'Puffle Style', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24327, 'Bright Tights Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24328, 'Antarctic Sun Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24329, 'Zipper Jacket', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24330, 'EPF Spacesuit', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24331, 'Holiday Puffle Story', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24332, 'Blue Penguin Stuffie Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24333, 'Gift Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24334, 'Red Parka', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24335, 'Arlo Costume', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24336, 'Butch Costume', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24337, 'Spot''s Outfit', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24338, 'The Good Dinosaur Hoodie', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24339, 'Spot of Tea', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24340, 'Mirror Ball Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24342, 'Shoe in', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24344, 'Haber-dashing', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24345, 'Grey Puffle Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24346, 'Black and White Pizza', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24347, 'Noir Detective Coat', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24348, 'The Driller Outfit', 4, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24349, 'Cranberry Lemon Outfit', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24350, 'Lumber Style', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24351, 'Blue Crosshatch Sweater', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24352, 'Rugby Team Shirt', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24353, 'Penguin Band Sweater', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24354, 'Blue Trendy Top', 5, 300, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24355, 'Dinosaur Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24356, 'Grey Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24357, 'Black O''berry Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24358, 'Blue O''berry Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24359, 'Brown O''berry Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24360, 'Green O''berry Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24361, 'Orange O''berry Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24362, 'Pink O''berry Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24363, 'Purple O''berry Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24364, 'Rainbow O''berry Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24365, 'Red O''berry Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24366, 'White O''berry Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24367, 'Yellow O''berry Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24368, 'Freestyle Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24369, 'Gold O''berry Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24370, 'Brown Bear Costume', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24372, 'Sasquatch Outfit', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24373, 'Bush Disguise', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24374, 'Wooden Canoe', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24375, 'Nick Wilde Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24376, 'Judy Hopps Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24377, 'Bellwether Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24378, 'Clawhauser Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24379, 'Leather Pilots Jacket', 5, 380, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24380, 'Elephant Costume', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24381, 'Blueberry Bunny Costume', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24382, 'Yeti Costume', 5, 600, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24383, 'Red Polka Dot Dress', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24384, 'Blue Winter Jacket', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24385, 'Lionheart Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24386, 'Chief Bogo Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24387, 'Flash Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24388, 'Purple Whirl Snowsuit', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24389, 'Green Peacoat', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24390, 'Black Penguin Band Hoodie', 5, 350, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24391, 'Watermelon Costume', 5, 450, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24392, 'Pastel Petal Dress', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24393, 'Nemo Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24394, 'Destiny Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24395, 'Dory Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24396, 'Hank Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24397, 'Bailey Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24398, 'Crush Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24399, 'Deep Diver Suit', 5, 500, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24400, 'Speed Boat', 5, 550, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24401, 'Ocean Bloom Mermaid', 5, 425, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24402, 'Red Dread', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24403, 'Spider Queen Dress', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24404, 'Franky Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24405, 'G Billy Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24406, 'Stompin'' Bob Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24407, 'Petey K Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24408, 'DJ Cadence Hoodie', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24409, 'Reindeer Puffle Sweater', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24410, 'Rad Rhinestones', 5, 400, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24411, 'Hoodie For Change', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24412, 'Aunt Arctic Sweater', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24413, 'Party Blaster Costume', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24414, 'Alumni Jacket', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24415, 'Island Tube', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24416, 'Alumni Jacket', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24417, 'Alumni Jacket', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24418, 'Alumni Jacket', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24419, 'Alumni Jacket', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24420, 'Alumni Jacket', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24421, 'Alumni Jacket', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24422, 'Alumni Jacket', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24423, 'Alumni Jacket', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24424, 'Alumni Jacket', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (24425, 'Alumni Jacket', 5, 0, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (34043, 'Bear Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (34052, 'Rowlf Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (34102, 'Monkey Costume', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (34135, 'Soccer Ball Costume R', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (34144, 'Daisies and Denim', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (34145, 'Around Town Outfit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (34158, 'Sunset Party Outfit', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (34174, 'Olaf''s CostumeR', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (34175, 'Merry Walrus Medal', 4, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (34206, 'Be Heard Campaign Shirt', 5, 0, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, now()),
 (90000, 'Elsa''s Braid', 2, 0, TRUE, TRUE, FALSE, FALSE, FALSE, FALSE, FALSE, now());

INSERT INTO item (id, name, type, cost, member, bait) VALUES
 (115, 'Gary''s Glasses', 2, 0, FALSE, TRUE),
 (442, 'Rockhopper''s Hat', 1, 0, FALSE, TRUE),
 (1000, 'Franky''s Cowboy Hat', 1, 0, FALSE, TRUE),
 (1001, 'G Billy''s Cowboy Hat', 1, 0, FALSE, TRUE),
 (1002, 'Stompin Bob''s Cowboy Hat', 1, 0, FALSE, TRUE),
 (1003, 'Petey K''s Old Cowboy Hat', 1, 0, TRUE, TRUE),
 (1068, 'Sensei''s Hat', 1, 0, FALSE, TRUE),
 (1257, 'Rookies Hat', 1, 0, FALSE, TRUE),
 (1458, 'Rocky''s Hair', 1, 0, FALSE, TRUE),
 (1459, 'CeCe''s Hair', 1, 0, FALSE, TRUE),
 (1460, 'Cadence''s 2012 Hair', 1, 0, FALSE, TRUE),
 (1654, 'Darth Herbert''s Costume', 1, 0, TRUE, TRUE),
 (2000, 'Petey K''s Old Sunglasses', 2, 0, FALSE, TRUE),
 (3130, 'Rocky''s Necklace', 3, 0, FALSE, TRUE),
 (3131, 'CeCe Necklace', 3, 0, FALSE, TRUE),
 (4148, 'Sensei''s Fire Suit', 4, 0, FALSE, TRUE),
 (4281, 'Sensei''s Water Suit', 4, 0, FALSE, TRUE),
 (4690, 'Rocky Outfit', 4, 0, FALSE, TRUE),
 (4691, 'Cece Outfit', 4, 0, FALSE, TRUE),
 (4712, 'Rockhopper''s Belt', 4, 0, FALSE, TRUE),
 (4752, 'Herberts Eyebrows', 2, 0, TRUE, TRUE),
 (5168, 'Rocky''s Bracelet', 5, 0, FALSE, TRUE),
 (5169, 'CeCe''s Bracelet', 5, 0, FALSE, TRUE),
 (6078, 'Petey K''s Shoes', 6, 0, FALSE, TRUE),
 (6127, 'Rocky''s Shoes', 6, 0, FALSE, TRUE),
 (6128, 'CeCe''s Shoes', 6, 0, FALSE, TRUE);
 
INSERT INTO card (id, name, set_id, power_id, element, color, value, description) VALUES
 (1, 'CART SURFER', 1, 0, 'f', 'b', 3, ''),
 (2, 'COFFEE SHOP', 1, 0, 'f', 'g', 2, ''),
 (3, 'ASTRO-BARRIER', 1, 0, 'f', 'g', 8, ''),
 (4, 'HOT CHOCOLATE', 1, 0, 'f', 'o', 3, ''),
 (5, 'LANDING PAD', 1, 0, 'f', 'p', 4, ''),
 (6, 'PIZZA CHEF', 1, 0, 'f', 'p', 6, ''),
 (7, 'PAINT BY LETTERS', 1, 0, 'f', 'r', 2, ''),
 (8, 'MINE', 1, 0, 'f', 'r', 7, ''),
 (9, 'CONSTRUCTION WORKER', 1, 0, 'f', 'y', 2, ''),
 (10, 'JET PACK ADVENTURE', 1, 0, 'f', 'y', 5, ''),
 (11, 'GIFT SHOP', 1, 0, 's', 'b', 3, ''),
 (12, 'HIKING IN THE FOREST', 1, 0, 's', 'g', 2, ''),
 (13, 'RESCUE SQUAD', 1, 0, 's', 'g', 5, ''),
 (14, 'PET SHOP', 1, 0, 's', 'o', 3, ''),
 (15, 'SKI VILLAGE', 1, 0, 's', 'p', 4, ''),
 (16, 'ICE HOCKEY', 1, 0, 's', 'p', 8, ''),
 (17, 'SKI HILL', 1, 0, 's', 'r', 2, ''),
 (18, 'SNOWBALL FIGHT', 1, 0, 's', 'r', 6, ''),
 (19, 'SNOW FORTS', 1, 0, 's', 'y', 2, ''),
 (20, 'SOCCER', 1, 0, 's', 'y', 7, ''),
 (21, 'BEACH', 1, 0, 'w', 'b', 3, ''),
 (22, 'FOOTBALL', 1, 0, 'w', 'b', 5, ''),
 (23, 'BASEBALL', 1, 0, 'w', 'g', 2, ''),
 (24, 'EMERALD PRINCESS', 1, 0, 'w', 'g', 8, ''),
 (25, 'BEAN COUNTERS', 1, 0, 'w', 'o', 3, ''),
 (26, 'MANHOLE COVER', 1, 0, 'w', 'p', 4, ''),
 (27, 'NEWSPAPER ARCHIVES', 1, 0, 'w', 'p', 6, ''),
 (28, 'UNDERGROUND POOL', 1, 0, 'w', 'r', 2, ''),
 (29, 'SCUBA DIVING', 1, 0, 'w', 'r', 7, ''),
 (30, 'ICE FISHING', 1, 0, 'w', 'y', 2, ''),
 (31, 'CASE OF THE MISSING COINS', 1, 0, 'f', 'b', 2, ''),
 (32, 'CRATES AND BOXES', 1, 0, 'f', 'b', 5, ''),
 (33, 'ANVIL SMITH', 1, 0, 'f', 'g', 4, ''),
 (34, 'HALLOWEEN', 1, 0, 'f', 'o', 6, ''),
 (35, 'KNIGHT', 1, 0, 'f', 'p', 3, ''),
 (36, 'BLACK PUFFLE', 1, 0, 'f', 'r', 5, ''),
 (37, 'MEDIEVAL PARTY', 1, 0, 'f', 'y', 4, ''),
 (38, 'BLUE PUFFLE', 1, 0, 's', 'b', 2, ''),
 (39, 'PINK PUFFLE', 1, 0, 's', 'b', 5, ''),
 (40, 'GREEN PUFFLE', 1, 0, 's', 'g', 4, ''),
 (41, 'PUFFLE FURNITURE', 1, 0, 's', 'o', 6, ''),
 (42, 'PURPLE PUFFLE', 1, 0, 's', 'p', 3, ''),
 (43, 'RED PUFFLE', 1, 0, 's', 'r', 5, ''),
 (44, 'YELLOW PUFFLE', 1, 0, 's', 'y', 4, ''),
 (45, 'SUMMER PARTY', 1, 0, 'w', 'b', 2, ''),
 (46, 'GRAY FISH', 1, 0, 'w', 'g', 4, ''),
 (47, 'O BERRY', 1, 0, 'w', 'o', 6, ''),
 (48, 'LIFE RING', 1, 0, 'w', 'p', 3, ''),
 (49, 'JOLLY ROGER FLAG', 1, 0, 'w', 'r', 5, ''),
 (50, 'COINS', 1, 0, 'w', 'y', 5, ''),
 (51, 'GARY THE GADGET GUY', 1, 0, 'f', 'b', 6, ''),
 (52, 'BUOYS', 1, 0, 'f', 'g', 5, ''),
 (53, 'YARR', 1, 0, 'f', 'o', 4, ''),
 (54, 'KLUTZY', 1, 0, 'f', 'p', 8, ''),
 (55, 'HERBERT P. BEAR', 1, 0, 'f', 'r', 3, ''),
 (56, 'HOT SAUCE', 1, 0, 'f', 'y', 7, ''),
 (57, 'SNOW CASTLE', 1, 0, 's', 'b', 6, ''),
 (58, 'AUNT ARCTIC', 1, 0, 's', 'g', 7, ''),
 (59, 'CHRISTMAS CAROLS', 1, 0, 's', 'o', 4, ''),
 (60, 'BETA PARTY', 1, 0, 's', 'o', 8, ''),
 (61, 'HIDDEN ITEMS', 1, 0, 's', 'p', 6, ''),
 (62, 'SECRET AGENT', 1, 0, 's', 'r', 3, ''),
 (63, 'SNOW SHOVEL', 1, 0, 's', 'y', 5, ''),
 (64, 'THE MIGRATOR', 1, 0, 'w', 'b', 6, ''),
 (65, 'CAPTAIN ROCKHOPPER', 1, 0, 'w', 'g', 5, ''),
 (66, 'ICE CREAM', 1, 0, 'w', 'o', 4, ''),
 (67, 'MULLET', 1, 0, 'w', 'p', 8, ''),
 (68, 'UNDERGROUND FLOOD', 1, 0, 'w', 'r', 3, ''),
 (69, 'FLUFFY THE FISH', 1, 0, 'w', 'y', 4, ''),
 (70, 'SHARK', 1, 0, 'w', 'y', 7, ''),
 (71, 'FLOORING UPGRADE', 1, 13, 'f', 'r', 9, 'When this is scored, Snow cannot be played next round.'),
 (72, 'GADGET ROOM', 1, 16, 'f', 'b', 9, 'When this is played, Water cards become Fire for this round.'),
 (73, 'JACKHAMMER', 1, 1, 'f', 'y', 10, 'When this card is played, lower values win ties the next round.'),
 (74, 'JET PACK', 1, 4, 'f', 'o', 10, 'When this is scored, discard one Opponent''s Snowball card.'),
 (75, 'THIN ICE', 1, 7, 'f', 'p', 11, 'When this is scored, discard one Opponent''s Red card.'),
 (76, 'NIGHT VISION GOGGLES', 1, 10, 'f', 'g', 11, 'When this is scored, discard one Opponent''s Yellow card.'),
 (77, 'MIGHTY PLUNGER', 1, 2, 'f', 'r', 12, 'When this is scored, your card gets +2 for the next round'),
 (78, 'DRAGON', 1, 3, 'f', 'b', 12, 'When this is scored, your Opponent''s card get -2 for the next round'),
 (79, 'SPACE ADVENTURE', 1, 15, 's', 'r', 9, 'When this is scored, Water cannot be played next round.'),
 (80, 'TOUR GUIDE', 1, 18, 's', 'b', 9, 'When this is played, Fire cards become Snowball for this round.'),
 (81, 'SLED RACING', 1, 1, 's', 'g', 10, 'When this card is played, lower values win ties the next round.'),
 (82, 'PUFFLE ROUND-UP', 1, 5, 's', 'g', 10, 'When this is scored, discard one Opponent''s Water card.'),
 (83, 'AC3000', 1, 9, 's', 'p', 11, 'When this is scored, discard one Opponent''s Green card.'),
 (84, 'SNOW GLOBE IGLOO', 1, 12, 's', 'y', 11, 'When this is scored, discard one Opponent''s Purple card.'),
 (85, 'SHADOW GUY & GAMMA GAL', 1, 2, 's', 'o', 12, 'When this is scored, your card gets +2 for the next round'),
 (86, 'AVALANCHE', 1, 3, 's', 'o', 12, 'When this is scored, your Opponent''s card get -2 for the next round'),
 (87, 'WATER PARTY', 1, 14, 'w', 'r', 9, 'When this is scored, Fire cannot be played next round.'),
 (88, 'CATCHIN'' WAVES', 1, 17, 'w', 'b', 9, 'When this is played, Snowball cards become Water for this round.'),
 (89, 'FIREFIGHTER', 1, 1, 'w', 'y', 10, 'When this card is played, lower values win ties the next round.'),
 (90, 'AQUA GRABBER', 1, 6, 'w', 'o', 10, 'When this is scored, discard one Opponent''s Fire card.'),
 (91, 'OCTOPUS', 1, 8, 'w', 'p', 11, 'When this is scored, discard one Opponent''s Blue card.'),
 (92, 'CACTUS', 1, 11, 'w', 'g', 11, 'When this is scored, discard one Opponent''s Orange card.'),
 (93, 'TIPPING THE ICEBERG', 1, 2, 'w', 'p', 12, 'When this is scored, your card gets +2 for the next round'),
 (94, 'NINJA', 1, 3, 'w', 'y', 12, 'When this is scored, your Opponent''s card get -2 for the next round'),
 (95, 'WHITE PUFFLE', 2, 10, 's', 'b', 12, 'When this is scored, discard one Opponent''s Yellow card.'),
 (96, 'Yarr', 2, 4, 'f', 'r', 11, 'When this is scored, discard one Opponent''s Snowball card.'),
 (97, 'KEEPER OF THE BOILER ROOM', 2, 9, 'w', 'g', 11, 'When this is scored, discard one Opponent''s Green card.'),
 (98, 'BLAST', 2, 0, 'f', 'r', 8, ''),
 (99, 'FLARE', 2, 0, 'f', 'o', 7, ''),
 (100, 'CHIRP', 2, 0, 's', 'y', 6, ''),
 (101, 'POP', 2, 0, 'w', 'p', 5, ''),
 (102, 'LOOP', 2, 0, 'f', 'o', 7, ''),
 (103, 'FLIT', 2, 0, 's', 'g', 6, ''),
 (104, 'BOUNCER', 2, 0, 'w', 'b', 5, ''),
 (105, 'ED MCCOOL AND DESTRUCTO', 2, 0, 'f', 'b', 4, ''),
 (106, 'NIBLET', 2, 0, 'w', 'g', 2, ''),
 (107, 'SNOWY NIGHT', 2, 0, 's', 'p', 3, ''),
 (108, 'PUFFLES ON THE LOOSE', 2, 0, 'f', 'g', 4, ''),
 (109, 'CADENCE AND THE KEEPER', 2, 0, 'w', 'g', 3, ''),
 (110, 'GOLDEN PUFFLE', 2, 0, 's', 'y', 2, ''),
 (111, 'METEOR PUFFLE', 2, 0, 'f', 'r', 5, ''),
 (112, 'RAINBOW PUFFLE', 2, 0, 'w', 'p', 4, ''),
 (113, 'ABOMINABLE SNOW PUFFLE ', 2, 0, 's', 'o', 3, ''),
 (201, 'Club Penguin Times', 3, 0, 'f', 'b', 2, ''),
 (202, 'Can of Worms', 3, 0, 'f', 'b', 3, ''),
 (203, 'Winter Fiesta Party', 3, 0, 'f', 'p', 4, ''),
 (204, 'Invisible Ninja Suit', 3, 0, 'f', 'o', 5, ''),
 (205, 'Penguin Mail', 3, 0, 'f', 'r', 6, ''),
 (206, 'Bumble Bee', 3, 0, 'f', 'y', 6, ''),
 (207, 'Pizzatron 3000', 3, 0, 'f', 'o', 7, ''),
 (208, 'Card-Jitsu', 3, 0, 'f', 'r', 8, ''),
 (209, 'Adopt-A-Puffle', 3, 0, 's', 'b', 4, ''),
 (210, 'Rock & Roll Suit', 3, 0, 's', 'o', 4, ''),
 (211, 'Construction', 3, 0, 's', 'p', 5, ''),
 (212, 'Hidden Pins', 3, 0, 's', 'g', 6, ''),
 (213, 'Treasure Hunt Game', 3, 0, 's', 'y', 7, ''),
 (214, 'DJ3K', 3, 0, 's', 'g', 8, ''),
 (215, 'Surf Boards', 3, 0, 'w', 'g', 2, ''),
 (216, 'Beach Chair Chilling', 3, 0, 'w', 'g', 3, ''),
 (217, 'Telescope', 3, 0, 'w', 'r', 4, ''),
 (218, 'Fish Costume', 3, 0, 'w', 'y', 4, ''),
 (219, 'Crow''s Nest', 3, 0, 'w', 'o', 5, ''),
 (220, 'The Cove', 3, 0, 'w', 'b', 6, ''),
 (221, 'Captain''s Quarters', 3, 0, 'w', 'o', 6, ''),
 (222, 'Iceberg', 3, 0, 'w', 'b', 8, ''),
 (223, 'The Lounge', 3, 0, 'f', 'p', 3, ''),
 (224, 'Rockhopper''s Key', 3, 0, 'f', 'g', 4, ''),
 (225, 'Penguins That Time Forgot', 3, 0, 'f', 'y', 5, ''),
 (226, 'Team Blue Rally', 3, 0, 'f', 'o', 6, ''),
 (227, 'Jet Pack Adventure', 3, 0, 'f', 'r', 7, ''),
 (228, 'Ruby and the Ruby', 3, 0, 's', 'r', 2, ''),
 (229, 'Snowball Fight', 3, 0, 's', 'o', 3, ''),
 (230, 'Snowball Press', 3, 0, 's', 'p', 5, ''),
 (231, 'Music Jam', 3, 0, 's', 'y', 6, ''),
 (232, 'Puffle Chew-Chew', 3, 0, 's', 'g', 7, ''),
 (233, 'Fairy Fables', 3, 0, 'w', 'g', 2, ''),
 (234, 'Jelly Fish', 3, 0, 'w', 'y', 3, ''),
 (235, 'Pirate', 3, 0, 'w', 'r', 4, ''),
 (236, 'Fish Bait', 3, 0, 'w', 'p', 5, ''),
 (237, 'Hydro Hopper', 3, 0, 'w', 'b', 7, ''),
 (238, 'Puffle O''s', 3, 0, 'w', 'b', 7, ''),
 (239, 'Fall Fair Games', 3, 0, 'f', 'p', 4, ''),
 (240, 'Thinga-ma-jig 3000', 3, 0, 'f', 'g', 5, ''),
 (241, 'Rad Scientist', 3, 0, 's', 'r', 3, ''),
 (242, 'Quip and Qua', 3, 0, 's', 'b', 5, ''),
 (243, 'Twee', 3, 0, 's', 'p', 6, ''),
 (244, 'Coins For Change', 3, 0, 's', 'g', 8, ''),
 (245, 'Surfing with the Sharks', 3, 0, 'w', 'r', 5, ''),
 (246, 'Dance Contest', 3, 0, 'w', 'p', 6, ''),
 (247, 'Purple Octopus', 3, 0, 'w', 'p', 7, ''),
 (248, 'Giant Squid', 3, 0, 'w', 'b', 8, ''),
 (249, 'Penguin Band', 3, 16, 'f', 'b', 9, 'When this is played, Water cards become Fire for this round.'),
 (250, 'Dojo Sketch', 3, 13, 'f', 'y', 10, 'When this is scored, Snow cannot be played next round.'),
 (251, 'Flare Flinger 3000', 3, 7, 'f', 'p', 11, 'When this is scored, discard one Opponent''s Red card.'),
 (252, 'Herbert & Klutzy', 3, 11, 'f', 'r', 12, 'When this is scored, discard one Opponent''s Orange card.'),
 (253, 'Coffee Delivery Truck', 3, 15, 's', 'o', 9, 'When this is scored, Water cannot be played next round.'),
 (254, 'Anniversary Party', 3, 18, 's', 'b', 10, 'When this is played, Fire cards become Snowball for this round.'),
 (255, 'Halloween Party', 3, 5, 's', 'y', 11, 'When this is scored, discard one Opponent''s Water card.'),
 (256, 'Sensei', 3, 12, 's', 'p', 12, 'When this is scored, discard one Opponent''s Purple card.'),
 (257, 'Jackhammers', 3, 8, 'w', 'p', 9, 'When this is scored, discard one Opponent''s Blue card.'),
 (258, 'Quest for the Golden Puffle', 3, 14, 'w', 'y', 10, 'When this is scored, Fire cannot be played next round.'),
 (259, 'Storm', 3, 6, 'w', 'o', 11, 'When this is scored, discard one Opponent''s Fire card.'),
 (260, 'Rockhopper', 3, 17, 'w', 'b', 12, 'When this is played, Snowball cards become Water for this round.'),
 (301, 'Camera', 4, 0, 'f', 'o', 2, ''),
 (302, 'Wigs', 4, 0, 'f', 'b', 3, ''),
 (303, 'Pizza Parlor', 4, 0, 'f', 'r', 4, ''),
 (304, 'Coffee Machine', 4, 0, 'f', 'o', 5, ''),
 (305, 'Costume Trunk', 4, 0, 'f', 'r', 6, ''),
 (306, 'Pet Shop Staff', 4, 0, 'f', 'y', 6, ''),
 (307, 'Stompin'' Bob', 4, 0, 'f', 'o', 7, ''),
 (308, 'G Billy', 4, 0, 'f', 'r', 8, ''),
 (309, 'Outdoor Igloo', 4, 0, 's', 'b', 4, ''),
 (310, 'Ticket Booth', 4, 0, 's', 'o', 4, ''),
 (311, 'Ridge Run', 4, 0, 's', 'p', 5, ''),
 (312, 'Clock Tower', 4, 0, 's', 'g', 6, ''),
 (313, 'White Puffles', 4, 0, 's', 'y', 7, ''),
 (314, 'Dojo Courtyard', 4, 0, 's', 'g', 8, ''),
 (315, 'Rocking Horse', 4, 0, 'w', 'g', 2, ''),
 (316, 'Penguin Style', 4, 0, 'w', 'g', 3, ''),
 (317, 'Stage Director', 4, 0, 'f', 'r', 4, ''),
 (318, 'Conductor', 4, 0, 'w', 'y', 4, ''),
 (319, 'Beacon', 4, 0, 'w', 'o', 5, ''),
 (320, 'Discovering The Underground', 4, 0, 'f', 'o', 6, ''),
 (321, 'Franky', 4, 0, 'w', 'o', 6, ''),
 (322, 'Petey K.', 4, 0, 'w', 'b', 8, ''),
 (323, 'Boombox', 4, 0, 'f', 'p', 3, ''),
 (324, 'Mild Mannered Reporter', 4, 0, 'f', 'g', 4, ''),
 (325, 'The Twelfth Fish', 4, 0, 'f', 'y', 5, ''),
 (326, 'April Fool''s Day Party', 4, 0, 'f', 'o', 6, ''),
 (327, 'Plasma Glow Wave', 4, 0, 'f', 'r', 7, ''),
 (328, 'Cowbell', 4, 0, 's', 'r', 2, ''),
 (329, 'O Berry Bush', 4, 0, 's', 'o', 3, ''),
 (330, 'Chester The Time Traveler', 4, 0, 's', 'r', 3, ''),
 (331, '101 Days of Fun', 4, 0, 's', 'r', 5, ''),
 (332, 'Color Vote', 4, 0, 's', 'y', 6, ''),
 (333, 'Christmas Party', 4, 0, 's', 'g', 7, ''),
 (334, 'Viking Opera', 4, 0, 'w', 'g', 2, ''),
 (335, 'Water Balloons', 4, 0, 'w', 'y', 3, ''),
 (336, 'Tuba', 4, 0, 'w', 'r', 4, ''),
 (337, 'Rockhopper''s Plants', 4, 0, 'w', 'p', 5, ''),
 (338, 'Adventure Party', 4, 0, 'w', 'b', 7, ''),
 (339, 'Penguin Play Awards', 4, 0, 'f', 'o', 7, ''),
 (340, 'All-Access Pass', 4, 0, 'f', 'p', 4, ''),
 (341, 'Cadence''s Autograph', 4, 0, 'f', 'g', 5, ''),
 (342, 'Treasure Book', 4, 0, 's', 'b', 5, ''),
 (343, 'Black Hoodie', 4, 0, 's', 'p', 6, ''),
 (344, 'Propellor Cap', 4, 0, 's', 'g', 8, ''),
 (345, 'Dessert Pizza', 4, 0, 'w', 'r', 5, ''),
 (346, 'Black Electric Guitar', 4, 0, 'w', 'r', 6, ''),
 (347, 'Mastering The Elements', 4, 0, 'w', 'p', 7, ''),
 (348, 'Elite Penguin Force', 4, 0, 'w', 'b', 8, ''),
 (349, 'Switchbox 3000', 4, 7, 'f', 'b', 11, 'When this is scored, discard one Opponent''s Red card.'),
 (350, 'Medieval Party', 4, 8, 'f', 'y', 10, 'When this is scored, discard one Opponent''s Blue card.'),
 (351, 'Cadence', 4, 3, 'f', 'r', 11, 'When this is scored, your Opponent''s card get -2 for the next round'),
 (352, 'Cloud Wave', 4, 9, 'f', 'o', 12, 'When this is scored, discard one Opponent''s Green card.'),
 (353, 'Ghost Sheet', 4, 3, 's', 'r', 12, 'When this is scored, your Opponent''s card get -2 for the next round'),
 (354, 'Cart Surfer', 4, 12, 's', 'g', 10, 'When this is scored, discard one Opponent''s Purple card.'),
 (355, 'Keytar', 4, 11, 's', 'p', 9, 'When this is scored, discard one Opponent''s Orange card.'),
 (356, 'X-treme Jetpack', 4, 2, 's', 'y', 9, 'When this is scored, your card gets +2 for the next round'),
 (357, 'Box Dimension', 4, 2, 'w', 'b', 9, 'When this is scored, your card gets +2 for the next round'),
 (358, 'Gold Viking Helmet', 4, 10, 'w', 'p', 10, 'When this is scored, discard one Opponent''s Yellow card.'),
 (359, 'Gary The Gadget Guy', 4, 2, 'w', 'o', 11, 'When this is scored, your card gets +2 for the next round'),
 (360, 'Rockhopper''s Cannon', 4, 3, 'w', 'g', 12, 'When this is scored, your Opponent''s card get -2 for the next round'),
 (401, 'Multi-Player', 5, 0, 'w', 'r', 2, ''),
 (402, 'Fire Battle', 5, 0, 'f', 'b', 3, ''),
 (403, 'Fired Up', 5, 0, 's', 'p', 4, ''),
 (404, 'Goldsmith', 5, 0, 'w', 'g', 2, ''),
 (405, 'Scroll', 5, 0, 'f', 'y', 3, ''),
 (406, 'Volcano', 5, 0, 's', 'o', 4, ''),
 (407, 'Ready For Battle', 5, 0, 'w', 'b', 2, ''),
 (408, 'Flying Flippers Emporium', 5, 0, 'f', 'p', 3, ''),
 (409, 'Duel', 5, 0, 's', 'g', 4, ''),
 (410, 'Concentrate', 5, 0, 'w', 'o', 2, ''),
 (411, 'Interior Construction', 5, 0, 'f', 'r', 3, ''),
 (412, 'Elements Hideout', 5, 0, 's', 'y', 4, ''),
 (413, 'Scavenger Hunt', 5, 0, 'w', 'r', 5, ''),
 (414, 'Fire Wheel', 5, 0, 'f', 'b', 6, ''),
 (415, 'Amulet Construction', 5, 0, 's', 'p', 5, ''),
 (416, 'Fire Dance', 5, 0, 'w', 'g', 6, ''),
 (417, 'Sketched Action', 5, 0, 'f', 'y', 5, ''),
 (418, 'Fire Guardian', 5, 0, 's', 'o', 6, ''),
 (419, 'Hot Sauce Reserve', 5, 0, 'w', 'r', 5, ''),
 (420, 'Sketched: Volcano', 5, 0, 'f', 'b', 6, ''),
 (421, 'Orange Sky', 5, 0, 'w', 'r', 7, ''),
 (422, 'Sketched: Enter', 5, 0, 'w', 'b', 8, ''),
 (423, 'Amulet', 5, 0, 'f', 'p', 9, ''),
 (424, 'A New Journey', 5, 0, 'f', 'g', 7, ''),
 (425, 'Fire Door & Key', 5, 0, 's', 'y', 8, ''),
 (426, 'Sensei Fire', 5, 0, 's', 'o', 9, ''),
 (427, 'Fire Ninja', 5, 4, 'f', 'r', 12, 'When this is scored, discard one Opponent''s Snowball card.'),
 (501, 'Monkey King', 6, 0, 'f', 'b', 2, ''),
 (502, 'Jet Pack Pilot', 6, 0, 'f', 'b', 3, ''),
 (503, 'Fairy Princess', 6, 0, 'f', 'p', 4, ''),
 (504, 'Puffle Grooming', 6, 0, 'f', 'o', 5, ''),
 (505, 'Halloween Costumes', 6, 0, 'f', 'r', 6, ''),
 (506, 'Costume Show', 6, 0, 'f', 'y', 6, ''),
 (507, 'Soccer Match', 6, 0, 'f', 'o', 7, ''),
 (508, 'Bean Counters', 6, 0, 'f', 'b', 2, ''),
 (509, 'Earth Day Recycling', 6, 0, 'f', 'b', 3, ''),
 (510, 'Penguin Play Awards', 6, 0, 'f', 'p', 4, ''),
 (511, 'Goal!', 6, 0, 'f', 'o', 5, ''),
 (512, 'Bamboo Forest Sketch', 6, 0, 'f', 'r', 6, ''),
 (513, 'The Mine Cave is Safe!', 6, 0, 'f', 'y', 6, ''),
 (514, 'Mine Cave Construction', 6, 0, 'f', 'o', 7, ''),
 (515, 'Guardian Dog', 6, 0, 's', 'b', 4, ''),
 (516, 'The Fair', 6, 0, 's', 'o', 4, ''),
 (517, 'Hockey Gear', 6, 0, 's', 'p', 5, ''),
 (518, 'Puffle Show', 6, 0, 's', 'g', 6, ''),
 (519, 'Holiday Gifts', 6, 0, 's', 'y', 7, ''),
 (520, 'Holiday Activities', 6, 0, 's', 'g', 8, ''),
 (521, 'Stadium', 6, 0, 's', 'b', 4, ''),
 (522, 'Igloo Items', 6, 0, 's', 'o', 4, ''),
 (523, 'Music Jam', 6, 0, 's', 'p', 5, ''),
 (524, 'Elite Penguin Sketch', 6, 0, 's', 'g', 6, ''),
 (525, 'Festival of Flight', 6, 0, 'w', 'g', 2, ''),
 (526, 'Test Glider', 6, 0, 'w', 'g', 3, ''),
 (527, 'Puffle Paddle', 6, 0, 'w', 'r', 4, ''),
 (528, 'Fancy Formal Wear', 6, 0, 'w', 'y', 4, ''),
 (529, 'The Docks', 6, 0, 'w', 'o', 5, ''),
 (530, 'Orange Puffle', 6, 0, 'w', 'b', 6, ''),
 (531, 'Community Tree', 6, 0, 'w', 'o', 6, ''),
 (532, 'Community Garden', 6, 0, 'w', 'b', 8, ''),
 (533, 'Island Adventure Party', 6, 0, 'w', 'g', 2, ''),
 (534, 'Shops in Igloos', 6, 0, 'w', 'g', 3, ''),
 (535, 'Seafarer''s Gown', 6, 0, 'w', 'r', 4, ''),
 (536, 'Music Jam Dancers', 6, 0, 'w', 'y', 4, ''),
 (537, 'Music Jam Rock Stage', 6, 0, 'w', 'o', 5, ''),
 (538, 'Music Jam Pop Stage', 6, 0, 'w', 'b', 6, ''),
 (539, 'Underwater Adventure', 6, 0, 'w', 'o', 6, ''),
 (540, 'Norman Swarm', 6, 0, 'w', 'b', 8, ''),
 (541, 'Puffle Rescuer', 6, 0, 'w', 'g', 2, ''),
 (542, 'Everyday Phoning Facility', 6, 0, 'w', 'g', 3, ''),
 (543, 'Jet Packs', 6, 0, 'f', 'p', 3, ''),
 (544, 'Pumpkin Igloo', 6, 0, 'f', 'g', 4, ''),
 (545, 'Treasure Hunt Adventure', 6, 0, 'f', 'y', 5, ''),
 (546, 'Destroyed PSA HQ', 6, 0, 'f', 'o', 6, ''),
 (547, 'Captains', 6, 0, 'f', 'r', 7, ''),
 (548, 'April Fools', 6, 0, 'f', 'r', 7, ''),
 (549, 'The Great Snow Maze', 6, 0, 's', 'r', 2, ''),
 (550, 'Jacques Hammer', 6, 0, 's', 'o', 3, ''),
 (551, 'Throne of April Fools', 6, 0, 's', 'p', 5, ''),
 (552, 'The Test', 6, 0, 's', 'g', 7, ''),
 (553, 'Cloud Maker 3000', 6, 0, 'w', 'g', 2, ''),
 (554, 'Super Exclusives', 6, 0, 'w', 'y', 3, ''),
 (555, 'Rockslide at the Mine', 6, 0, 'w', 'r', 4, ''),
 (556, 'Mer-penguins', 6, 0, 'w', 'p', 5, ''),
 (557, 'Puffer Fish', 6, 0, 'w', 'b', 7, ''),
 (558, 'Garden Hose', 6, 0, 'w', 'b', 7, ''),
 (559, 'Fixing the Mine Shack', 6, 0, 'w', 'b', 7, ''),
 (560, 'Puffle Circus', 6, 0, 'f', 'p', 4, ''),
 (561, 'Puffle Play Sketch', 6, 0, 'f', 'g', 5, ''),
 (562, 'Dragon Statue', 6, 0, 'f', 'p', 4, ''),
 (563, 'Popcorn Party', 6, 0, 'f', 'g', 5, ''),
 (564, 'White Puffle Room', 6, 0, 's', 'r', 3, ''),
 (565, 'Green Puffle Sketch', 6, 0, 's', 'b', 5, ''),
 (566, 'Fishbowl Igloo', 6, 0, 's', 'p', 6, ''),
 (567, 'Ship Battle Adventure', 6, 0, 'w', 'r', 5, ''),
 (568, 'Elite Penguin Force', 6, 0, 'w', 'p', 6, ''),
 (569, 'Rockhopper''s Notice Board', 6, 0, 'w', 'p', 7, ''),
 (570, 'Squid Lid', 6, 0, 'w', 'b', 8, ''),
 (571, 'Hidden Lake', 6, 0, 'w', 'b', 8, ''),
 (572, 'Orange Puffle Wagon', 6, 16, 'f', 'b', 9, 'When this is played, Water cards become Fire for this round.'),
 (573, 'Haunted House', 6, 13, 'f', 'y', 10, 'When this is scored, Snow cannot be played next round.'),
 (574, 'Fire Dragon', 6, 7, 'f', 'p', 11, 'When this is scored, discard one Opponent''s Red card.'),
 (575, 'Puffle Rescue: Underground', 6, 11, 'f', 'r', 12, 'When this is scored, discard one Opponent''s Orange card.'),
 (576, 'Double Necked Guitar', 6, 15, 's', 'o', 9, 'When this is scored, Water cannot be played next round.'),
 (577, 'Puffle Rescue: Ice Floes', 6, 18, 's', 'b', 10, 'When this is played, Fire cards become Snowball for this round.'),
 (578, 'Spy Phone', 6, 12, 's', 'y', 11, 'When this is scored, discard one Opponent''s Purple card.'),
 (579, 'Dance Contest', 6, 5, 's', 'p', 12, 'When this is scored, discard one Opponent''s Water card.'),
 (580, 'Crab Costume', 6, 8, 'w', 'p', 9, 'When this is scored, discard one Opponent''s Blue card.'),
 (581, 'Puffle Washer', 6, 14, 'w', 'y', 10, 'When this is scored, Fire cannot be played next round.'),
 (582, 'Green Jackhammer', 6, 6, 'w', 'o', 11, 'When this is scored, discard one Opponent''s Fire card.'),
 (583, 'The Veggie Villain', 6, 17, 'w', 'b', 12, 'When this is played, Snowball cards become Water for this round.'),
 (584, 'Island Lifter 3000', 6, 16, 'f', 'b', 9, 'When this is played, Water cards become Fire for this round.'),
 (585, 'Flare', 6, 13, 'f', 'y', 10, 'When this is scored, Snow cannot be played next round.'),
 (586, 'Popcorn Device', 6, 7, 'f', 'p', 11, 'When this is scored, discard one Opponent''s Red card.'),
 (587, 'Field-Ops', 6, 11, 'f', 'r', 12, 'When this is scored, discard one Opponent''s Orange card.'),
 (588, 'Recipe for Disaster', 6, 15, 's', 'o', 9, 'When this is scored, Water cannot be played next round.'),
 (589, 'Elite Agent', 6, 18, 's', 'b', 10, 'When this is played, Fire cards become Snowball for this round.'),
 (590, 'Aunt Arctic Sketch', 6, 12, 's', 'y', 11, 'When this is scored, discard one Opponent''s purple card.'),
 (591, 'Yeti Cave', 6, 5, 's', 'p', 12, 'When this is scored, discard one Opponent''s Water card.'),
 (592, 'Shocktopus', 6, 8, 'w', 'p', 9, 'When this is scored, discard one Opponent''s Blue card.'),
 (593, 'Puffle Rescue: Underwater', 6, 14, 'w', 'y', 10, 'When this is scored, Fire cannot be played next round.'),
 (594, 'Hidden Lake Lock', 6, 6, 'w', 'o', 11, 'When this is scored, discard one Opponent''s Fire card.'),
 (595, 'Knight''s Quest', 6, 17, 'w', 'b', 12, 'When this is played, Snowball cards become Water for this round.'),
 (601, 'Spooky Igloo Contest', 7, 0, 'f', 'b', 2, ''),
 (602, 'Halloween Countdown!', 7, 0, 'f', 'b', 3, ''),
 (603, 'Stormy Skies', 7, 0, 'f', 'p', 4, ''),
 (604, '5th Anniversary Party', 7, 0, 'f', 'o', 5, ''),
 (605, '5th Anniversary Cake', 7, 0, 'f', 'r', 6, ''),
 (606, '5th Anniversary Yearbook', 7, 0, 'f', 'y', 6, ''),
 (607, 'Snowball Fight!', 7, 0, 'f', 'o', 7, ''),
 (608, 'Find Four', 7, 0, 'f', 'b', 2, ''),
 (609, 'Cozy Ski Lodge', 7, 0, 'f', 'b', 3, ''),
 (610, 'Holiday Decorations!', 7, 0, 'f', 'p', 4, ''),
 (611, 'Book Room Adventures', 7, 0, 'f', 'o', 5, ''),
 (612, 'Halloween Forest', 7, 0, 'f', 'r', 6, ''),
 (613, 'Brown Puffle Rocket', 7, 0, 'f', 'y', 6, ''),
 (614, 'Fun at the Fair!', 7, 0, 'f', 'o', 7, ''),
 (615, 'Ski Hill Challenge', 7, 0, 's', 'b', 4, ''),
 (616, 'Holiday Party Skating', 7, 0, 's', 'o', 4, ''),
 (617, 'Treasure!', 7, 0, 's', 'p', 5, ''),
 (618, 'Lumberjack Look', 7, 0, 's', 'g', 6, ''),
 (619, 'Pink Sparkle Snowshoes', 7, 0, 's', 'y', 7, ''),
 (620, 'Holiday Tree 2010', 7, 0, 's', 'g', 8, ''),
 (621, 'Fill the Lighthouse!', 7, 0, 's', 'b', 4, ''),
 (622, 'Wilderness Machine', 7, 0, 's', 'o', 4, ''),
 (623, 'Expedition Boat', 7, 0, 's', 'p', 5, ''),
 (624, 'Wilderness Expedition Cave', 7, 0, 's', 'g', 6, ''),
 (625, 'Water Party Prep', 7, 0, 'w', 'g', 2, ''),
 (626, 'Water Dance', 7, 0, 'w', 'g', 3, ''),
 (627, 'Stamp Collectors', 7, 0, 'w', 'r', 4, ''),
 (628, 'Explore the Wilderness', 7, 0, 'w', 'y', 4, ''),
 (629, 'Coins for Change 2010!', 7, 0, 'w', 'o', 5, ''),
 (630, 'Penguin Colors', 7, 0, 'w', 'b', 6, ''),
 (631, 'Fairy Fables', 7, 0, 'w', 'o', 6, ''),
 (632, 'Penguins That Time Forgot', 7, 0, 'w', 'b', 8, ''),
 (633, 'EPF Sketch', 7, 0, 'w', 'g', 2, ''),
 (634, 'Cove Stories', 7, 0, 'w', 'g', 3, ''),
 (635, 'Orange Circus Puffle', 7, 0, 'w', 'r', 4, ''),
 (636, 'Hydro Hopper', 7, 0, 'w', 'y', 4, ''),
 (637, 'Igloo Party!', 7, 0, 'w', 'o', 5, ''),
 (638, 'Life Underwater', 7, 0, 'w', 'b', 6, ''),
 (639, 'Herbert''s Monologue', 7, 0, 'w', 'o', 6, ''),
 (640, 'Enemy Bots', 7, 0, 'w', 'b', 8, ''),
 (641, 'Paint By Letters', 7, 0, 'w', 'g', 2, ''),
 (642, 'Secret Agent', 7, 0, 'w', 'g', 3, ''),
 (643, 'Fireworks', 7, 0, 'f', 'p', 3, ''),
 (644, 'EPF Tech Agent', 7, 0, 'f', 'g', 4, ''),
 (645, 'Popcorn Explosion', 7, 0, 'f', 'y', 5, ''),
 (646, 'Walk Your Puffle', 7, 0, 'f', 'o', 6, ''),
 (647, 'Sumo Smash', 7, 0, 'f', 'r', 7, ''),
 (648, 'inventive Brown Puffle', 7, 0, 'f', 'r', 7, ''),
 (649, 'Holiday Party Tree', 7, 0, 's', 'r', 2, ''),
 (650, 'EPF Enemies', 7, 0, 's', 'o', 3, ''),
 (651, 'Tall Tree Maze', 7, 0, 's', 'p', 5, ''),
 (652, 'Sled ''n Slide', 7, 0, 's', 'g', 7, ''),
 (653, 'Underwater at the Stage', 7, 0, 'w', 'g', 2, ''),
 (654, 'Wilderness Expedition Travel', 7, 0, 'w', 'y', 3, ''),
 (655, 'Catch the Mullet!', 7, 0, 'w', 'r', 4, ''),
 (656, 'Great Puffle Circus', 7, 0, 'w', 'p', 5, ''),
 (657, 'Intense Black Puffle', 7, 0, 'w', 'b', 7, ''),
 (658, 'Brown Puffle Storyboard', 7, 0, 'w', 'b', 7, ''),
 (659, 'Java Jump', 7, 0, 'w', 'b', 7, ''),
 (660, 'Master the Elements', 7, 0, 'f', 'p', 4, ''),
 (661, 'Fairy Fables Stage', 7, 0, 'f', 'g', 5, ''),
 (662, 'Creative Puffles', 7, 0, 'f', 'p', 4, ''),
 (663, 'Goofy Green Puffle', 7, 0, 'f', 'g', 5, ''),
 (664, 'EPF Test Room', 7, 0, 's', 'r', 3, ''),
 (665, 'Snowball Battle', 7, 0, 's', 'b', 5, ''),
 (666, 'Fabulous Purple Puffle', 7, 0, 's', 'p', 6, ''),
 (667, 'Water Party Sketch', 7, 0, 'w', 'r', 5, ''),
 (668, 'Zany Orange Puffle', 7, 0, 'w', 'p', 6, ''),
 (669, 'Sporty Pink Puffle', 7, 0, 'w', 'p', 7, ''),
 (670, 'Wii Game Day goal!', 7, 0, 'w', 'b', 8, ''),
 (671, 'Epic White Puffle', 7, 0, 'w', 'b', 8, ''),
 (672, 'Awesome Blue Puffle', 7, 0, 'f', 'b', 9, ''),
 (673, 'feed a _puffle', 7, 0, 'f', 'y', 9, ''),
 (674, 'Ninja Battle', 7, 0, 'f', 'b', 2, ''),
 (675, 'The Mystery of Sensei', 7, 0, 'f', 'p', 4, ''),
 (676, 'Master the Elements', 7, 0, 'f', 'o', 5, ''),
 (677, 'Brown Puffle', 7, 0, 'f', 'r', 6, ''),
 (678, 'Fire Wave', 7, 0, 'f', 'y', 6, ''),
 (679, 'Black Puffles in Cart Surfer', 7, 0, 'f', 'o', 7, ''),
 (680, 'Mountain Expedition', 7, 0, 's', 'b', 4, ''),
 (681, 'Climbing the Mountain', 7, 0, 's', 'o', 4, ''),
 (682, 'Top of Toughest Mountain', 7, 0, 's', 'p', 5, ''),
 (683, 'Cart Surfer Stamps', 7, 0, 's', 'g', 6, ''),
 (684, 'Stamp Book', 7, 0, 's', 'y', 7, ''),
 (685, 'Stormy Sky ', 7, 0, 's', 'g', 8, ''),
 (686, 'Water Wave Sketches', 7, 0, 's', 'b', 4, ''),
 (687, 'Storm Baffles Gary', 7, 0, 's', 'o', 4, ''),
 (688, 'Water Ninja', 7, 0, 'w', 'g', 3, ''),
 (689, 'The First Rainstorm', 7, 0, 'w', 'r', 4, ''),
 (690, 'Ninja Face-off', 7, 0, 'w', 'y', 4, ''),
 (691, 'Earning the Water Suit', 7, 0, 'w', 'o', 5, ''),
 (692, 'Water Pumping Gadget', 7, 0, 'w', 'b', 6, ''),
 (693, 'Preparing the Water Dojo', 7, 0, 'w', 'o', 6, ''),
 (694, 'Gate to the Water Dojo', 7, 0, 'w', 'b', 8, ''),
 (695, 'Water Suit', 7, 0, 'w', 'g', 2, ''),
 (696, 'Sign From The Elements', 7, 0, 'w', 'g', 3, ''),
 (697, 'Sensei - Master of Water', 7, 0, 'w', 'r', 4, ''),
 (698, 'Water Ninja Poster Sketch', 7, 0, 'w', 'y', 4, ''),
 (699, 'Water Ninja Sketch', 7, 0, 'w', 'o', 5, ''),
 (700, 'Water Beats Fire Sketch', 7, 0, 'f', 'p', 3, ''),
 (701, 'Blast', 7, 0, 'f', 'r', 8, ''),
 (702, 'Flare', 7, 0, 'f', 'o', 7, ''),
 (703, 'Chirp', 7, 0, 's', 'y', 6, ''),
 (704, 'Pop', 7, 0, 'w', 'p', 5, ''),
 (705, 'Loop', 7, 0, 'f', 'o', 7, ''),
 (706, 'Flit', 7, 0, 's', 'g', 6, ''),
 (707, 'Bouncer', 7, 0, 'w', 'b', 5, ''),
 (708, 'Ed McCool and Destructo', 7, 0, 'f', 'b', 4, ''),
 (709, 'Niblet', 7, 0, 'w', 'g', 2, ''),
 (710, 'Snowy Night', 7, 0, 's', 'p', 3, ''),
 (711, 'Puffles on the Loose', 7, 0, 'f', 'g', 4, ''),
 (712, 'Cadence and the Keeper', 7, 0, 'w', 'g', 3, ''),
 (713, 'Golden puffle', 7, 0, 's', 'y', 2, ''),
 (714, 'Meteor puffle', 7, 0, 'f', 'r', 5, ''),
 (715, 'Rainbow puffle', 7, 0, 'w', 'p', 4, ''),
 (716, 'Abominable snow puffle', 7, 0, 's', 'o', 3, ''),
 (717, 'Black Puffle', 7, 0, 'f', 'r', 5, ''),
 (718, 'Blue Puffle', 7, 0, 's', 'b', 2, ''),
 (719, 'Pink Puffle', 7, 0, 's', 'b', 5, ''),
 (720, 'Green Puffle', 7, 0, 's', 'g', 4, ''),
 (721, 'Purple Puffle', 7, 0, 's', 'p', 3, ''),
 (722, 'Red Puffle', 7, 0, 's', 'r', 5, ''),
 (723, 'Yellow Puffle', 7, 0, 's', 'y', 4, ''),
 (724, 'Puffle Round Up', 7, 5, 's', 'g', 10, 'When this is scored, discard one opponent''s Water card.'),
 (725, 'Fire Ninja Force', 7, 11, 'f', 'r', 12, 'When this is scored, discard one opponent''s Orange card.'),
 (726, 'Halloween Swamp', 7, 15, 's', 'o', 9, 'When this is scored, Water cannot be played next round.'),
 (727, 'Prepare for Pumpkins', 7, 17, 's', 'b', 10, 'When this is played, Snowball cards become Fire for this round'),
 (728, 'Finding Rockhopper', 7, 12, 's', 'y', 11, 'When this is scored, discard one opponent''s Purple card.'),
 (729, 'Elite Puffle: Flare', 7, 5, 's', 'p', 12, 'When this is scored, discard one opponent''s Water card.'),
 (730, 'Giant Pearl', 7, 8, 'w', 'p', 9, 'When this is scored, discard one opponent''s Blue card.'),
 (731, 'Water Stamps', 7, 14, 'w', 'y', 10, 'When this is scored, Fire cannot be played next round.'),
 (732, 'EPF Agents', 7, 6, 'w', 'o', 11, 'When this is scored, discard one opponent''s Fire card.'),
 (733, 'Puffle Rescue', 7, 17, 'w', 'b', 12, 'When this is played, Snowball cards become Water for this round.'),
 (734, 'Fire Dojo Challenge', 7, 16, 'f', 'b', 9, 'When this is played, Water cards become Fire for this round.'),
 (735, 'System Defender', 7, 13, 'f', 'y', 10, 'When this is scored, Snow cannot be played next round.'),
 (736, 'Adventurous Red Puffle', 7, 7, 'f', 'p', 11, 'When this is scored, discard one opponent''s Red card.'),
 (737, 'G and the Machine', 7, 11, 'f', 'r', 12, 'When this is scored, discard one opponent''s Orange card.'),
 (738, 'Enemies of the System', 7, 15, 's', 'o', 9, 'When this is scored, Water cannot be played next round.'),
 (739, 'Holiday Dojo Lights', 7, 18, 's', 'b', 10, 'When this is played, Fire cards become Snowball for this round.'),
 (740, 'Wii Rollin'' Riot', 7, 12, 's', 'y', 11, 'When this is scored, discard one opponent''s Purple card.'),
 (741, 'EPF Under Attack', 7, 5, 's', 'p', 12, 'When this is scored, discard one opponent''s Water card.'),
 (742, 'Elite Comm Class Agents', 7, 8, 'w', 'p', 9, 'When this is scored, discard one opponent''s Blue card.'),
 (743, 'Brown Puffle Cliff', 7, 14, 'w', 'y', 10, 'When this is scored, Fire cannot be played next round.'),
 (744, 'Coins for Change Goal!', 7, 6, 'w', 'o', 11, 'When this is scored, discard one opponent''s Fire card.'),
 (745, 'Sensei Water Battle ', 7, 14, 'w', 'b', 9, 'When this is played, Fire cards become Water for this round.'),
 (746, 'Green Puffles in Jet Pack Adve', 7, 13, 'f', 'y', 10, 'When this is scored, Snow cannot be played next round.'),
 (747, 'Water Dojo', 7, 7, 'w', 'p', 11, 'When this is scored, discard one opponent''s Red card.'),
 (748, 'White Puffle', 7, 10, 's', 'b', 12, 'When this is scored, discard one opponent''s Yellow card.'),
 (749, 'Yarr', 7, 4, 'f', 'r', 11, 'When this is scored, discard one opponent''s Snowball card.'),
 (750, 'Keeper of the Boiler Room', 7, 9, 'w', 'g', 11, 'When this is scored, discard one opponent''s Green card.'),
 (801, 'Mobile Fire 1', 8, 0, 'f', 'r', 6, ''),
 (802, 'Mobile Snow 1', 8, 0, 's', 'g', 7, ''),
 (803, 'Mobile Water 1', 8, 0, 'w', 'b', 8, ''),
 (804, 'Mobile Fire Power Card 1', 8, 2, 'f', 'p', 11, 'When this is scored, your card gets +2 for the next round');

INSERT INTO card_starter_deck (item_id, card_id, quantity) VALUES
 (821, 1, 1),
 (821, 6, 1),
 (821, 9, 1),
 (821, 14, 1),
 (821, 17, 1),
 (821, 20, 1),
 (821, 22, 1),
 (821, 23, 1),
 (821, 26, 1),
 (821, 73, 1),
 (821, 89, 1),
 (821, 81, 1),
 (8006, 3, 1),
 (8006, 18, 1),
 (8006, 216, 1),
 (8006, 222, 1),
 (8006, 229, 1),
 (8006, 303, 1),
 (8006, 304, 1),
 (8006, 314, 1),
 (8006, 319, 1),
 (8006, 250, 1),
 (8006, 352, 1),
 (8010, 202, 1),
 (8010, 204, 1),
 (8010, 305, 1),
 (8010, 15, 1),
 (8010, 13, 1),
 (8010, 312, 1),
 (8010, 218, 1),
 (8010, 220, 1),
 (8010, 29, 1),
 (8010, 90, 1);

INSERT INTO flooring (id, name, cost, patched, legacy_inventory, vanilla_inventory) VALUES
 (0, 'Floor Removal', 20, FALSE, FALSE, FALSE),
 (1, 'Terracotta Tile', 680, FALSE, FALSE, FALSE),
 (2, 'Maple Hardwood', 620, FALSE, FALSE, FALSE),
 (3, 'Green Carpet', 530, FALSE, FALSE, FALSE),
 (4, 'Burgundy Carpet', 530, FALSE, FALSE, FALSE),
 (5, 'Black & White Tile', 510, FALSE, FALSE, FALSE),
 (6, 'Linoleum', 540, FALSE, FALSE, FALSE),
 (7, 'Dance Floor', 1000, FALSE, FALSE, FALSE),
 (8, 'Painted Dance Steps', 280, FALSE, FALSE, FALSE),
 (9, 'Bamboo Floor', 370, FALSE, FALSE, FALSE),
 (10, 'Dirt & Leaves', 400, FALSE, FALSE, FALSE),
 (11, 'Blue Turf', 530, FALSE, FALSE, FALSE),
 (12, 'Whirlpool', 750, FALSE, FALSE, FALSE),
 (13, 'Cherry Hardwood', 620, FALSE, FALSE, FALSE),
 (14, 'Phony Lawn', 700, FALSE, FALSE, FALSE),
 (15, 'Black Carpet', 530, FALSE, FALSE, FALSE),
 (16, 'Dark Stone Tile', 800, FALSE, FALSE, FALSE),
 (17, 'Pink Carpet', 530, FALSE, FALSE, FALSE),
 (18, 'Sand Floor', 400, FALSE, FALSE, FALSE),
 (19, 'Sunny Sky Floor', 530, FALSE, FALSE, FALSE),
 (20, 'Cobblestone', 1200, FALSE, FALSE, FALSE),
 (21, 'Snowy Floor', 400, FALSE, FALSE, FALSE),
 (22, 'Lime Green Carpet', 530, FALSE, FALSE, FALSE),
 (23, 'Woven Rice Mat', 750, FALSE, FALSE, FALSE);

INSERT INTO furniture (id, name, type, sort, cost, member, patched, legacy_inventory, vanilla_inventory, bait, max_quantity) VALUES
 (1, 'Pink Chair', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2, 'Pink Beanbag Chair', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (3, 'Pink Table', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (4, 'Pink Dresser', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (5, 'Pink Lamp', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (6, 'Pink Sofa', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (10, 'Log Chair', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (11, 'Log Bench', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (12, 'Log Drawers', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (13, 'Log Stump', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (14, 'Log Table', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (21, 'Blue Chair', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (22, 'Blue Bench', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (23, 'Blue Bookshelf', 1, 1, 450, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (24, 'Blue Table', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (25, 'Blue Deck Chair', 1, 1, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (26, 'Blue Lamp', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (31, 'Spiral Coffee Tree', 1, 1, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (32, 'Coffee Couch', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (33, 'Coffee Table', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (34, 'Java Bag', 1, 1, 120, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (41, 'Book Room Couch', 1, 1, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (42, 'Book Room Lamp', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (43, 'Book Room Arm Chair', 1, 1, 560, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (46, 'Dinner Table', 1, 1, 700, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (47, 'Dinner Chair', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (51, 'Snow Chair', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (52, 'Snow Couch', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (53, 'Snow Clump', 1, 1, 40, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (54, 'Ice Table', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (55, 'Penguin Ice Sculpture', 1, 1, 850, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (56, 'Snow Deck Chair', 1, 1, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (57, 'Ice Sculpture Knight', 1, 1, 900, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (58, 'Snowman', 1, 1, 405, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (59, 'Sled', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (61, 'Princess Armoire', 1, 1, 700, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (62, 'Princess Vanity', 1, 1, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (63, 'Princess Throne', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (65, 'Mermaid Vanity', 1, 1, 735, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (66, 'Shell Chair', 1, 1, 550, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (67, 'Mermaid Clock', 2, 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (71, 'Fridge', 1, 1, 700, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (72, 'Barbecue', 1, 1, 450, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (73, 'Electric Stove', 1, 1, 650, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (74, 'Pantry', 1, 1, 410, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (75, 'Kitchen Sink', 1, 1, 380, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (76, 'Folding Chair', 1, 1, 180, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (77, 'Beach Umbrella', 1, 1, 130, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (78, 'Beach Ball', 1, 1, 110, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (79, 'Inflatable Whale', 1, 1, 190, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (80, 'Inflatable Dragon', 1, 1, 190, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (81, 'Bamboo Couch', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (82, 'Bamboo Chair', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (83, 'Bamboo Table', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (84, 'Bamboo Torch', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (85, 'Palm Tree', 1, 1, 450, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (91, 'Green Plastic Chair', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (92, 'Green Bench', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (93, 'Green Bookcase', 1, 1, 450, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (94, 'Green Vase', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (95, 'Green Deck Chair', 1, 1, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (96, 'Green Couch', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (97, 'Green Plush Chair', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (99, 'Superhero Stage Poster', 2, 2, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (100, 'Moose Head', 2, 2, 2000, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (101, 'Aquarium', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (102, 'Speaker', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (103, 'Large Cactus', 1, 1, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (104, 'Small Cactus', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (105, 'Wood Cabinet', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (106, 'Mona Lisa', 2, 2, 3000, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (107, 'Recycle Bin', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (108, 'Welcome Mat', 3, 3, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (109, 'Big Screen TV', 1, 1, 5000, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (110, 'Tombstone', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (111, 'Happy Jack-O-Lantern', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (112, 'Goofy Jack-O-Lantern', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (113, 'Sad Jack-O-Lantern', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (114, 'Spider Web', 2, 2, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (115, 'Silly Jack-O-Lantern', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (116, 'Puffle Jack-O-Lantern', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (117, 'Cauldron', 1, 1, 630, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (118, 'Pipe Organ', 1, 1, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (119, 'Candelabra', 1, 1, 650, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (120, 'Pumpkin Lights', 2, 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (130, 'Large Christmas Tree', 1, 1, 700, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (131, 'Christmas Wreath', 2, 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (132, 'Fireplace', 2, 2, 900, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (133, 'Small Christmas Tree', 1, 1, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (134, 'Holiday Lights', 2, 2, 30, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (135, 'Giant Candy Cane', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (136, 'Wood Stove', 2, 2, 900, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (137, 'Candle', 1, 1, 120, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (138, 'Nutcracker', 1, 1, 950, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (139, 'Door Garland', 2, 2, 750, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (140, 'Presents', 1, 1, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (143, 'Plush Gray Chair', 1, 1, 560, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (144, 'Watering Can', 1, 1, 105, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (145, 'Rare Flower Pot', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (146, 'Wood Bench', 1, 1, 480, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (147, 'Fan', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (148, 'Lamp', 1, 1, 125, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (149, 'Pinata', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (150, 'Red Lava Lamp', 1, 1, 450, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (151, 'Oval Rug', 3, 3, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (152, 'Rectangle Rug', 3, 3, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (153, 'Coat Rack', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (154, 'Grandfather Clock', 1, 1, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (155, 'Modern Art', 1, 1, 550, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (156, 'Teddy Bear', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (157, 'Student Desk', 1, 1, 125, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (158, 'Sunset Painting', 2, 2, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (159, 'LCD Television', 2, 2, 2500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (160, 'Green Rug', 3, 3, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (161, 'Green Oval Rug', 3, 3, 550, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (162, 'Small House Plant', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (163, 'Party Fountian', 1, 1, 450, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (164, 'Stereo', 1, 1, 650, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (165, 'Wooden Crate', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (166, 'Mullet Fish', 2, 2, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (167, 'Large House Plant', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (168, 'Wall Clock', 2, 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (169, 'Mexican Rug', 3, 3, 550, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (170, 'Pink Balloon', 1, 1, 15, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (171, 'Blue Balloon', 1, 1, 15, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (172, 'Yellow Balloon', 1, 1, 15, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (173, 'Red Balloon', 1, 1, 15, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (174, 'Green Balloon', 1, 1, 15, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (175, 'Party Banner', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (176, 'Birthday Cake', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (177, 'Dance Floor', 3, 3, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (178, 'Jukebox', 1, 1, 775, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (180, 'Umbrella Table', 1, 1, 380, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (181, 'White Plastic Chair', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (182, 'Lounging Deck Chair', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (183, 'Surf Beach Towel', 3, 3, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (184, 'Curved Desk', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (185, 'Classroom Desk', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (186, 'Wall Chalkboard', 2, 2, 410, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (187, 'Classroom Chair', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (200, 'Blue Bed', 1, 4, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (201, 'Pink Bed', 1, 4, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (202, 'Gray Bed', 1, 4, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (203, 'Green Bed', 1, 4, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (204, 'Gray Puffle House', 1, 4, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (205, 'Brown Puffle House', 1, 4, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (206, 'Pink Puffle House', 1, 4, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (207, 'Blue Puffle House', 1, 4, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (208, 'Green Puffle House', 1, 4, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (209, 'Purple Puffle House', 1, 4, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (210, 'Small Scratching Post', 1, 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (211, 'Large Scratching Post', 1, 4, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (212, 'Water Dish', 1, 4, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (213, 'Fish Bowl', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (214, 'Water Bottle', 1, 4, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (215, 'Running Wheel', 1, 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (216, 'Turtle Bowl', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (217, 'Bird Cage', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (218, 'Puffle Tree', 1, 4, 220, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (219, 'Puffle Tent', 1, 4, 460, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (220, 'Puffle Condo', 1, 4, 280, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (221, 'Purple Bed', 1, 4, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (222, 'Red Bed', 1, 4, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (223, 'Red Puffle House', 1, 4, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (224, 'Scratch Tower', 1, 4, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (225, 'Double Dish', 1, 4, 80, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (226, 'Puffle Igloo', 1, 4, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (227, 'Yellow Bed', 1, 4, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (228, 'Yellow Puffle House', 1, 4, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (229, 'Koi Pond', 1, 1, 450, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (230, 'Green Birdhouse', 2, 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (231, 'Blue Birdhouse', 2, 2, 170, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (232, 'White Bed', 1, 4, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (233, 'White Puffle House', 1, 4, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (250, 'Barrel', 1, 1, 320, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (251, 'Piano', 1, 1, 900, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (252, 'Piano Bench', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (253, 'Single Wall Light', 2, 2, 80, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (254, 'Double Wall Light', 2, 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (260, 'Toy Train', 1, 1, 180, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (261, 'Chalk Board', 1, 1, 260, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (270, 'Purple Plastic Chair', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (271, 'Purple Plastic Lawn Chair', 1, 1, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (272, 'Purple Beanbag Chair', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (273, 'Purple Coffee Table', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (274, 'Purple Lava Lamp', 1, 1, 450, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (275, 'Purple Couch', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (280, 'Snowboard Rack', 2, 2, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (281, 'Dart Board', 2, 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (282, 'Ski Rack', 2, 2, 550, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (283, 'Red CP Banner', 2, 2, 170, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (284, 'Blue CP Banner', 2, 2, 170, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (285, 'Straight Racetrack Piece', 3, 3, 80, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (286, 'Curved Racetrack Piece', 3, 3, 120, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (287, 'Badminton Net', 1, 1, 700, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (288, 'Bench', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (289, 'Cricket Wickets', 1, 1, 225, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (290, 'Basketball Net', 2, 2, 320, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (291, 'Score Board', 2, 2, 520, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (292, 'Hockey Net', 1, 1, 360, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (293, 'Weight Bench', 1, 1, 680, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (294, 'Treadmill', 1, 1, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (295, 'Pommel Horse', 1, 1, 220, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (296, 'Red Gym Mat', 3, 3, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (297, 'Blue Gym Mat', 3, 3, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (298, 'Red Biscuit Tube', 1, 1, 120, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (299, 'Court Lines', 3, 3, 260, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (300, 'Exercise Ball', 1, 1, 60, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (301, 'Exercise Bike', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (302, 'Home Plate', 3, 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (303, 'Base', 3, 3, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (304, 'Globe', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (305, 'Treasure Chest', 1, 1, 800, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (306, 'Ship In A Bottle', 2, 2, 355, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (307, 'Pink Flamingo', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (308, 'Porthole', 2, 2, 160, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (309, 'Captain''s Wheel', 2, 2, 210, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (310, 'Wall Map', 2, 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (311, 'Old Penguin Monument', 1, 1, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (312, 'Pirate Ship', 1, 1, 525, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (313, 'Jolly Roger Flag', 2, 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (314, 'Camping Chair', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (315, 'Picnic Table', 1, 1, 800, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (316, 'Lanterns', 2, 2, 175, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (317, 'Fire Pit', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (318, 'Sleeping Bag', 3, 3, 320, TRUE, FALSE, FALSE, FALSE, TRUE, 99),
 (319, 'Bunk Bed', 1, 1, 750, TRUE, FALSE, FALSE, FALSE, TRUE, 99),
 (330, 'Red Coffee Table', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (331, 'Red Couch', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (332, 'Red Plush Chair', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (333, 'Red Bookcase', 1, 1, 450, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (334, 'Red Beanbag Chair', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (335, 'Plastic Deck Chair', 1, 1, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (336, 'Terracotta Sun', 2, 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (337, 'Terracotta Pot', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (338, 'Terracotta Pitcher', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (339, 'Terracotta Vase', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (340, 'Mexican Vase', 1, 1, 320, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (341, 'Medieval Banner', 2, 2, 105, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (342, 'Wall Torch', 2, 2, 85, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (343, 'Royal Throne', 1, 1, 750, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (344, 'Ficus Plant', 1, 1, 320, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (345, 'Snake Grass', 1, 1, 140, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (346, 'Flower Planter', 1, 1, 260, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (347, 'Vacuum', 1, 1, 280, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (348, 'Broom', 1, 1, 30, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (349, 'Mop & Bucket', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (350, 'Drum Kit', 1, 1, 760, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (351, 'Shoe Rack', 1, 1, 120, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (352, 'CD Rack', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (353, 'Microphone', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (354, 'DJ Table', 1, 1, 800, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (355, 'Concert Lights', 2, 2, 650, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (356, 'Pop Art Painting', 2, 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (357, 'Starry Night Painting', 2, 2, 520, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (358, 'Cooler', 1, 1, 360, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (359, 'Pink Plastic Castle', 1, 1, 315, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (360, 'Clam', 1, 1, 305, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (361, 'Sea Weed', 1, 1, 265, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (362, 'Inflatable Chair', 1, 1, 275, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (363, 'Inflatable Sofa', 1, 1, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (364, 'Security Camera', 2, 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (365, 'Blender', 1, 1, 220, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (366, 'Ferris Wheel Chair', 1, 1, 760, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (367, 'Circus Ball', 1, 1, 170, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (368, 'Picket Fence', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (369, 'Penguin Gnome', 1, 1, 220, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (370, 'Garden', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (371, 'Wheelbarrow', 1, 1, 480, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (372, 'Waves', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (373, 'Orange Banner', 2, 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (374, 'Green Banner', 2, 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (375, 'Purple Banner', 2, 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (376, 'Regal Chair', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (377, 'Chess Castle', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (378, 'Chess Knight', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (390, 'Cash Register', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (391, 'Easel', 1, 1, 180, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (392, 'Penguin Band Poster', 2, 2, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (393, 'Orange Inflatable Sofa', 1, 1, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (394, 'Orange Inflatable Chair', 1, 1, 275, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (400, 'Zeus Stage Poster', 2, 2, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (401, 'Peppy Stage Poster', 2, 2, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (402, 'Tate Stage Poster', 2, 2, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (403, 'Space Stage Poster', 2, 2, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (410, 'Harp', 1, 1, 840, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (411, 'Disco Ball', 2, 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (412, 'Music Stand', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (413, 'Guitar Stand', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (414, 'Band Stage', 3, 3, 800, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (415, 'Quarter Note', 2, 2, 120, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (416, 'Eighth Note', 2, 2, 120, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (417, 'Treble Clef', 2, 2, 120, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (418, 'Wall Speaker', 2, 2, 800, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (420, 'Life Ring', 2, 2, 120, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (421, 'Wall Net', 2, 2, 425, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (422, 'Rockhopper Portrait', 2, 2, 1100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (423, 'Lifeboat', 1, 1, 430, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (424, 'Steering Wheel', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (425, 'Desert Island', 1, 1, 850, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (426, 'Coins For Change Banner', 2, 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (430, 'Stainless Steel Stove', 1, 1, 700, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (431, 'Stainless Steel Fridge', 1, 1, 750, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (432, 'Corner Booth Seat', 1, 1, 700, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (433, 'Straight Booth Seat', 1, 1, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (434, 'Coffee Maker', 1, 1, 360, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (435, 'Pizza Oven', 1, 1, 850, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (440, 'Popcorn Machine', 1, 1, 1000, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (441, 'Ticket Booth', 1, 1, 580, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (442, 'Red Carpet', 3, 3, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (443, 'Velvet Rope', 1, 1, 160, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (444, 'Blue Curtain', 2, 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (450, 'Laboratory Desk', 1, 1, 700, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (451, 'Overhead Light', 2, 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (452, 'Plasma Ball', 1, 1, 550, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (453, 'Seismograph', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (454, 'Control Terminal', 1, 1, 800, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (455, 'Moving Eye Painting', 2, 2, 550, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (460, 'Stone Lantern', 1, 1, 575, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (461, 'Rice Paper Wall Screen', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (470, 'Christmas Ribbon', 2, 2, 120, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (471, 'Leaning Tree', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (472, 'Wall Snowflake', 2, 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (473, 'Wrought Iron Lamp Post', 1, 1, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (474, 'Stocking', 2, 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (475, 'Wooden Reindeer', 1, 1, 450, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (480, 'Snow Wall', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (481, 'Sloped Wall', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (482, 'Snow Fortress Wall', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (483, 'Snow Arch', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (484, 'Snow Tower', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (485, 'Icicles', 2, 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (486, 'Snowy Tree', 1, 1, 280, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (487, 'Small Rock', 1, 1, 120, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (488, 'Penguin Mask', 2, 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (489, 'Wall Gecko', 2, 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (490, 'Mexican Blanket', 2, 2, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (491, 'Hand Weights', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (492, 'Climbing Wall', 1, 1, 1000, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (493, 'Stacking Washer and Dryer', 1, 1, 1000, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (494, 'Map Area Rug', 3, 3, 450, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (495, 'Ironing Board', 1, 1, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (496, 'Astro Barrier Poster', 2, 2, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (497, 'Thin Ice Poster', 2, 2, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (500, 'Yellow Puffle Poster', 2, 2, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (501, 'Blue Puffle Poster', 2, 2, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (502, 'Red Puffle Poster', 2, 2, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (503, 'Green Puffle Poster', 2, 2, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (504, 'Black Puffle Poster', 2, 2, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (505, 'Pink Puffle Poster', 2, 2, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (506, 'Purple Puffle Poster', 2, 2, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (507, 'White Puffle Poster', 2, 2, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (511, 'Supernova Shadow Box', 2, 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (512, 'Shadow Guy Shadow Box', 2, 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (513, 'Gamma Gal Shadow Box', 2, 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (514, 'Accoustic Guitar Shadow Box', 2, 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (515, 'Electric Guitar Shadow Box', 2, 2, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (516, 'Formal Table', 1, 1, 550, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (517, 'Formal Chair', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (520, 'Tennis Net', 1, 1, 450, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (521, 'Pitcher''s Mound', 3, 3, 175, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (522, 'Tennis Court', 3, 3, 700, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (525, 'Tea Table', 1, 1, 650, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (526, 'Large Box', 1, 1, 120, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (527, 'Medium Box', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (528, 'Small Box', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (529, 'Portal Box', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (530, 'Bird Bath', 1, 1, 480, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (531, 'Sunset Flowers', 1, 1, 310, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (532, 'Bulrushes', 1, 1, 340, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (533, 'Poodle Plant', 1, 1, 430, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (535, 'Celtic Rug', 3, 3, 550, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (536, 'Ornate Mirror', 2, 2, 680, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (537, 'Archer''s Window', 2, 2, 630, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (538, 'Brazier', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (539, 'Wall Sconce', 2, 2, 140, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (540, 'Sprouting Spectaculous', 1, 1, 540, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (541, 'Plantus Fantasticus', 1, 1, 280, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (542, 'Surprisus Maximus', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (544, 'Tropical Palm', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (545, 'Soccer Net', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (546, 'Stone Arch Ruins', 1, 1, 450, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (547, 'Stone Wall Ruins', 1, 1, 425, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (548, 'Stone Column Ruins', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (549, 'Archaeological Dig Decal', 3, 3, 160, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (550, 'Gramophone', 1, 1, 375, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (551, 'Country Record', 2, 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (552, 'Rock N''Roll Record', 2, 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (553, 'Classical Record', 2, 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (560, 'Window', 2, 2, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (561, 'Cabinet', 2, 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (562, 'Construction Pylon', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (563, 'Construction Barrier', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (564, 'Construction Sign', 1, 1, 145, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (565, 'Bowling Alley', 3, 3, 700, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (566, 'Bowling Pin', 1, 1, 40, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (567, 'Green CP Banner', 2, 2, 170, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (570, 'Computer Desk', 1, 1, 900, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (571, 'Puzzle Floor', 3, 3, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (572, 'Funky Bookshelf', 1, 1, 550, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (573, 'Snack Stand', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (574, 'Giraffe', 1, 1, 450, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (575, 'Puffle Bean Bag', 1, 1, 450, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (576, 'Haunted Mansion Cut-Out', 1, 1, 850, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (577, 'Creepy Cottage Cut-Out', 1, 1, 700, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (578, 'Iron Gate', 1, 1, 650, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (579, 'Spooky Tree', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (580, 'Modern Couch', 1, 1, 850, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (581, 'Modern Chair', 1, 1, 700, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (582, 'Scoop Chair', 1, 1, 650, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (583, 'Large Aquarium', 1, 1, 450, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (584, 'Coffee Table', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (585, 'Coins For Change Donation Station', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (586, 'Santa Hat Snowman', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (587, 'Top Hat Snowman', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (588, 'Orange Scarf Snowman', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (589, 'Icicle Lights', 2, 2, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (590, 'Holiday Bells', 2, 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (591, 'Santa Chair', 1, 1, 15, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (592, 'Stone Chair', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (593, 'Stone Couch', 1, 1, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (594, 'Stone Lamp', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (595, 'Stone Table', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (596, 'HD TV', 2, 2, 1000, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (597, 'Rosewood Chair', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (598, 'Rosewood Dinner Table', 1, 1, 700, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (599, 'Stone Deck Chair', 1, 1, 375, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (600, 'Hot Drink Maker', 1, 1, 275, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (601, 'Mailbox', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (602, 'Puffle Guard', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (603, 'Puffle Washer', 1, 4, 450, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (604, 'Puffle Ball', 1, 4, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (605, 'Blue Couch', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (606, 'Blue Rug', 3, 3, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (607, 'Ice Fishing Decal', 3, 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (608, 'Barrel Chair', 1, 1, 230, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (609, 'Green Clover', 2, 2, 30, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (611, 'Clover Garland', 2, 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (612, 'Clover Balloon', 1, 1, 15, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (613, 'Tree Stump', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (614, 'Tree Stump Chair', 1, 1, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (615, 'Wishing Well', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (616, 'Rainbow with Pot O'' Gold', 2, 2, 450, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (617, 'Salon Chair', 1, 4, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (618, 'Orange Puffle House', 1, 4, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (619, 'Orange Bed', 1, 4, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (620, 'Burgundy Chair', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (621, 'Burgundy Couch', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (622, 'Burgundy Bookshelf', 1, 1, 450, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (623, 'Burgundy Lamp', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (624, 'Burgundy Curtains', 2, 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (625, 'Spikey Plant', 1, 1, 230, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (626, 'Potted Palm', 1, 1, 280, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (627, 'Potted Tree', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (628, 'Flower Pot', 1, 1, 175, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (629, 'Evergreen Plant', 1, 1, 240, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (630, 'Dragon''s Gold', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (632, 'Wall Rack', 2, 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (633, 'Clothes Rack', 1, 1, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (634, 'Penguin Mannequin', 1, 1, 275, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (635, 'Open-Closed Sign', 2, 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (636, 'Guitar Amp', 1, 1, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (637, 'Conga Drum', 1, 1, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (638, 'Lockers', 1, 1, 550, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (639, 'Classroom Bell', 2, 2, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (640, 'Scroll-down Map', 2, 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (641, 'White Board', 2, 2, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (642, 'Yellow CP Banner', 2, 2, 170, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (643, 'Perched Puffle Statue', 2, 2, 275, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (644, 'Crystal Ball', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (646, 'Pile O'' Goo', 3, 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (647, 'Torn Carpet', 3, 3, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (648, 'Pile O'' Candy', 3, 3, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (649, 'Spooky Penguin Statue', 1, 1, 550, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (650, 'Regal Table', 1, 1, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (651, 'Regal Chair', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (652, 'Multi-Pane Window', 2, 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (653, 'Iron Chandelier', 2, 2, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (654, 'Lamp Post', 1, 1, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (655, 'Holiday Tree', 1, 1, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (656, 'Presents', 1, 1, 170, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (657, 'Snowflake', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (658, 'Candy Cane', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (659, 'Stockings', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (660, 'Donation Tube', 1, 1, 70, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (661, 'Tallest Trees', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (662, 'Dream Catcher', 2, 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (663, 'Shiny Red Stove', 1, 1, 650, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (664, 'Shiny Red Fridge', 1, 1, 700, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (665, 'Brown Puffle House', 1, 4, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (666, 'Brown Puffle Picture', 2, 2, 80, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (667, 'Blue Puffle Picture', 2, 2, 80, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (668, 'Red Puffle Picture', 2, 2, 80, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (669, 'Green Puffle Picture', 2, 2, 80, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (670, 'Pink Puffle Picture', 2, 2, 80, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (671, 'Purple Puffle Picture', 2, 2, 80, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (672, 'Yellow Puffle Picture', 2, 2, 80, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (673, 'White Puffle Picture', 2, 2, 80, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (674, 'Black Puffle Picture', 2, 2, 80, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (675, 'Mint Chip Chair', 1, 1, 450, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (676, 'Neapolitan Lamp', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (677, 'Cocoa Cupcake', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (678, 'Marshmallow', 1, 1, 20, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (679, 'Spiral Lollipop', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (680, 'Double Dunk Chair', 1, 1, 450, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (681, 'Candy Stash', 1, 1, 10, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (682, 'Brown Bed', 1, 4, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (683, 'Puffle Cannon', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (684, 'Super Puffle Cannon', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (685, 'Window Basket', 2, 2, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (686, 'Waterfall Pond', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (687, 'Hedge Tree', 1, 1, 325, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (688, 'Hanging Basket', 2, 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (690, 'Orange Puffle Picture', 2, 2, 80, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (691, 'Crystals', 1, 1, 550, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (692, 'Trap Door', 3, 3, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (693, 'Hay Bale', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (694, 'Armor Rack', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (695, 'Elegant Plants', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (696, 'Ye Olde Red Banner', 2, 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (697, 'Ye Olde Blue Banner', 2, 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (698, 'Ye Olde Yellow Banner', 2, 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (699, 'Vines', 2, 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (700, 'Hanging Moss', 2, 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (701, 'Mossy Log', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (702, 'Boiling Cauldron', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (703, 'Island Trinkets', 2, 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (704, 'Blue Hydra Head', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (705, 'Red Hydra Head', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (706, 'Yellow Hydra Head', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (707, 'Single Flare', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (708, 'Flashing Flare', 1, 1, 450, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (710, 'Confetti Blaster', 2, 2, 25, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (711, 'Dynamic Drums', 1, 1, 550, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (712, 'Busking Case', 1, 1, 220, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (713, 'Dance Mat', 3, 3, 135, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (714, 'Party Platter', 1, 1, 180, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (715, 'Fire Extinguisher', 1, 1, 10, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (716, 'Radiant Rocker', 2, 2, 175, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (717, 'Musical Motif', 2, 2, 225, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (718, 'Electric Encore', 2, 2, 175, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (719, 'Laser Lights', 2, 2, 145, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (720, 'Hanging Algae', 2, 2, 30, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (721, 'Sea Streamers', 2, 2, 45, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (722, 'Swinging Vines', 2, 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (723, 'Cream Soda Barrel', 1, 1, 270, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (724, 'Sand Castle Arch', 1, 1, 320, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (725, 'Sand Castle Wall', 1, 1, 325, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (726, 'Rugged Rock', 1, 1, 700, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (727, 'Jutting Rock', 1, 1, 700, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (728, 'X Marks the Spot', 3, 3, 10, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (729, 'Snappy Shark', 2, 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (730, 'In-line Ramp', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (731, 'Arch Ramp', 1, 1, 80, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (732, 'Box Ramp', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (733, 'C Curve Ramp', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (734, 'S Curve Ramp', 1, 1, 230, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (735, 'Winners Podium', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (736, 'Smoothie Machine', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (737, 'Trophy Shelf', 1, 1, 650, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (739, 'Half-pipe', 3, 3, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (740, 'Stair Ramp', 1, 1, 375, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (741, 'Freewheelin'' Foam Pit', 1, 1, 800, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (743, 'Vert Ramp', 3, 3, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (745, 'Weathered Path', 3, 3, 120, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (746, 'Comfy Crab', 1, 1, 275, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (747, 'Fair Fence', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (748, 'Feed Fluffy Trashcan', 1, 1, 175, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (749, 'Balloon Bunch', 1, 1, 125, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (750, 'Fair Flags', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (751, 'Hay', 1, 1, 30, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (752, 'Cream Soda', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (753, 'Water Trough', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (754, 'Feeding Bucket', 1, 1, 60, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (755, 'Ring O'' Fire', 1, 1, 245, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (756, 'Captain Cut-Out', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (757, 'Food Stand', 1, 1, 550, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (758, 'Boss Desk', 1, 1, 700, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (760, 'Writing Desk', 1, 1, 375, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (761, 'The Glowing Grin', 1, 1, 120, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (762, 'The Spooky Surprise', 1, 1, 120, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (763, 'The Laughing Lantern', 1, 1, 120, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (764, 'Trick-or-Treats', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (765, 'Stone Wall', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (766, 'Jack-O-Lights', 2, 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (767, 'Terrifying Tissue Ghost', 2, 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (768, 'Dangly Spider', 2, 2, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (769, 'Wall Pumpkin', 2, 2, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (770, 'Wall Ghosts', 2, 2, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (771, 'Wall Bats', 2, 2, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (772, 'Antique Clock', 1, 1, 575, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (773, 'Swamp Slime', 3, 3, 230, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (774, 'Ninja Gate', 1, 1, 700, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (775, 'Green Paper Lantern', 2, 2, 30, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (776, 'Ninja Cauldron', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (777, 'Bamboo Stalks', 1, 1, 175, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (778, 'Noodle Stand', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (779, 'Paper Screen', 1, 1, 475, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (780, 'Bonsai Tree', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (781, 'Training Dummy', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (782, 'Blue Paper Lantern', 2, 2, 30, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (783, 'Red Paper Lantern', 2, 2, 30, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (784, 'Yellow Paper Lantern', 2, 2, 30, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (785, 'Purple Paper Lantern', 2, 2, 30, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (786, 'Card-Jitsu Mat', 3, 3, 700, TRUE, FALSE, FALSE, FALSE, FALSE, 4),
 (787, 'Arm Chair', 1, 1, 200, TRUE, FALSE, FALSE, TRUE, FALSE, 99),
 (788, 'Couch', 1, 1, 200, TRUE, FALSE, FALSE, TRUE, FALSE, 99),
 (790, 'Fern', 1, 1, 200, TRUE, FALSE, FALSE, TRUE, FALSE, 99),
 (792, 'Door Mat', 1, 1, 200, TRUE, FALSE, FALSE, TRUE, FALSE, 99),
 (793, 'Clock', 2, 2, 200, TRUE, FALSE, FALSE, TRUE, FALSE, 99),
 (794, 'Gingerbread Couch', 1, 1, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (795, 'Gingerbread Chair', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (796, 'Jujubes', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (797, 'Gumdrop Tree', 1, 1, 120, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (798, 'Gingerbread Man', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (799, 'Icing Decorations', 2, 2, 30, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (800, 'Swirly Lollipop', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (801, 'Holiday Star Decoration', 2, 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (802, 'Holiday Tree Decoration', 2, 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (803, 'Festive Coffee Table', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (804, 'Cozy Fireplace', 2, 2, 700, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (805, 'CFC Beanbag Chair', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (806, 'CFC Paper Chain', 2, 2, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (807, 'Tidal Pools', 1, 1, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (808, 'Sea Stones', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (809, 'Ancient Archway', 1, 1, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (810, 'Treasure Chest', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (811, 'Sunken Pillar', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (812, 'Sunken Wall', 1, 1, 425, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (813, 'Sunken Arch', 1, 1, 450, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (814, 'Cavern Couch', 1, 1, 450, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (815, 'Cavern Chair', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (816, 'Dinosaur Skull', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (817, 'Dinosaur Bones', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (818, 'Sewing Table', 1, 1, 450, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (819, 'Mannequin', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (820, 'Catwalk', 1, 1, 575, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (821, 'Violet Velvet Rope', 1, 1, 160, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (822, 'Judge''s Table', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (823, 'Judge''s Chair', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (824, 'Fashion Show Boxes', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (825, 'Fuzzy Green Couch', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (826, 'Happy Lantern', 2, 2, 175, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (827, 'Laughing Lantern', 2, 2, 175, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (828, 'Grumpy Lantern', 2, 2, 175, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (829, 'Cheeky Lantern', 2, 2, 175, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (830, 'Puffle Rug', 3, 3, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (831, 'Puffle Tubes', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (832, 'Fuzzy Blue Couch', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (833, 'Fuzzy Purple Couch', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (834, 'Jungle Bloom', 1, 1, 175, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (835, 'Jungle Fern', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (836, 'Jungle Flora', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (837, 'Watering Hole', 3, 3, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (838, 'Acacia Tree', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (839, 'Desert Stones', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (840, 'Hippopotamus', 1, 1, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (841, 'Zebra', 1, 1, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (842, 'Vine Swing', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (843, 'Bridge of Destiny', 1, 1, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (844, 'Tall Mushrooms', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (845, 'Short Mushrooms', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (846, 'Wishing Well', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (847, 'Fun Fungus', 2, 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (848, 'Wispy Clouds', 2, 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (849, 'Royal Throne', 1, 1, 750, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (850, 'Scorn Statue', 1, 1, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (851, 'Kingdom Flag', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (852, 'Fairy Woods Flag', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (853, 'Wizard Flag', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (854, 'Village Flag', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (855, 'Scorn Flag', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (856, 'Sky Flag', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (857, 'Dragon Flag', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (858, 'Fluffy Shrubbery', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (859, 'Knightly Shrubbery', 1, 1, 550, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (860, 'Maiden Shrubbery', 1, 1, 550, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (861, 'Cityscape', 2, 2, 550, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (862, 'Telephone Box', 1, 1, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (863, 'Manhole', 3, 3, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (864, 'Trashcan', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (865, 'Newspaper Stand', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (866, 'Hydrant', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (867, 'Lamp Post', 1, 1, 175, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (868, 'Searchlight', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (869, 'Hazard Barrel', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (870, 'Green Screen', 1, 1, 550, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (871, 'Guitar Stand', 1, 1, 325, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (872, 'Show Lights', 2, 2, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (873, 'Sound Station', 1, 1, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (874, 'Tall Spotlight', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (875, 'Short Spotlight', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (876, 'Video Camera', 1, 1, 225, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (877, 'Musician''s Chair', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (878, 'Kiwi Seat', 1, 1, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (879, 'Ancient Bench', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (880, 'Lemon Cushion', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (881, 'Ancient Couch', 1, 1, 450, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (882, 'Fruit Pillar', 1, 1, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (883, 'Ancient Recliner', 1, 1, 450, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (884, 'Thatched Awning', 2, 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (885, 'Refreshing Curtain', 2, 2, 65000, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (886, 'Pineapple Table', 1, 1, 65000, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (887, 'Banana Bunch Lamp', 1, 1, 65000, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (888, 'Fruit Vine', 2, 2, 65000, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (889, 'Watermelon Sofa', 1, 1, 65000, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (890, 'Sour Apple Chair', 1, 1, 65000, TRUE, FALSE, FALSE, FALSE, TRUE, 99),
 (891, 'Tropical Hammock', 1, 1, 65000, TRUE, FALSE, FALSE, FALSE, TRUE, 99),
 (892, 'Strawberry Seat', 1, 1, 65000, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (893, 'Banana Couch', 1, 1, 65000, TRUE, FALSE, FALSE, FALSE, TRUE, 99),
 (894, 'Smoothie Stand', 1, 1, 65000, TRUE, FALSE, FALSE, FALSE, TRUE, 99),
 (902, 'Vampire Throne', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (903, 'Batty Chaise Lounge', 1, 1, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (904, 'Ogre Ottoman', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (905, 'Shady Lamp', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (906, 'Gothic Candle Holder', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (907, 'Cursed Volume', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (908, 'Full Moon', 2, 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (909, 'Nightmare Vortex', 3, 3, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (910, 'Black Widow Window', 2, 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (911, 'Ectoplasmic Crevasse', 3, 3, 450, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (912, 'Ornate Tombstone', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (913, 'Headstone', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (914, 'Ectoplasmic Pit', 3, 3, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (915, 'Short Security Laser', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (916, 'Long Security Laser', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (917, 'Cardboard Herbert', 1, 1, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (918, 'Safe', 1, 1, 450, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (919, 'Containment Cell', 1, 1, 700, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (920, 'Emergency Light', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (921, 'Paw Print', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (922, 'Spy Car', 1, 1, 900, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (923, 'Covert Agent Station', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 1),
 (924, 'Holly Jolly Rug', 3, 3, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (925, 'Frost Pillar', 1, 1, 475, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (926, 'Frost Throne', 1, 1, 750, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (927, 'Candy Cane Arch', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (928, 'Candy Fence', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (929, 'Cozy Green House', 1, 1, 800, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (930, 'Cozy Red House', 1, 1, 800, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (931, 'Cozy Blue Door', 1, 1, 800, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (932, 'Tinker Train Engine', 1, 1, 700, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (933, 'Tinker Train Car', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (934, 'Holly Jolly Tree', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (935, 'Holiday Fireplace', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (936, 'Hut Roof', 1, 1, 800, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (937, 'Hollow Tree', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (938, 'Trusty Post', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (939, 'Wooden Walk', 1, 1, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (940, 'Trustier Post', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (941, 'Lava Puddle', 1, 1, 550, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (942, 'Leafy Roof', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (943, 'Lava Pool', 1, 1, 550, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (944, 'Wooden Steps', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (945, 'Short Wooden Steps', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (946, 'Leafy Window', 2, 2, 700, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (947, 'Tall Grass', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (948, 'Leather Canopy', 1, 1, 550, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (949, 'Comfy Stump', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (950, 'Lava Flow', 3, 3, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (951, 'Volcano', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (952, 'Waterfall', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (953, 'Ancient Tree', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (954, 'Festive Chair', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 1),
 (955, 'Holly Jolly Couch', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 1),
 (956, 'Festive Light', 2, 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 1),
 (957, 'Make and Bake Kitchen', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 1),
 (958, 'Santa Desk', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 1),
 (959, 'Space Age Lights', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (960, 'Basketball Court', 3, 3, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (961, 'Grand Piano', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (962, 'DJ Booth', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (963, 'Jumbo Remote', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (964, 'Diner Counter', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (965, 'Jumbo TV', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (966, 'Mars', 2, 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (967, 'Earth', 2, 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (968, 'Basketball Hoop', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (969, 'Director''s Chair', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (970, 'Spaceship', 1, 1, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (971, 'Graffiti Arch Ramp', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (972, 'Graffiti Stair Ramp', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (973, 'Puffle Carrier', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (974, 'Arcade Game', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (975, 'Skating Rink', 3, 3, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (976, 'Rainbow Bridge', 3, 3, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (977, 'Puffle Shop Shelf', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (978, 'Puffle Shop Till', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (979, 'Unicycle Tightrope', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (980, 'Puffle Stage', 3, 3, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (981, 'Puffle Tire Swing', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (982, 'Puffle Patio Lanterns', 2, 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (983, 'Puffle Table', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (984, 'Clinic Entrance', 1, 1, 800, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (985, 'General Store Front', 1, 1, 800, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (986, 'X-Ray Machine', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (987, 'Hospital Chair', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (988, 'Operating Room Lights', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (989, 'Deluxe Tool Chest', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (990, 'Red Bridge Railing', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (991, 'Rice Paper Wall', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (992, 'Floral Paper Screen', 2, 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (993, 'Stone Walk Way', 3, 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (994, 'Circular Archway', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (995, 'Sushi Table', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (996, 'Rock Garden', 3, 3, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (997, 'Blue Floor Cushion', 3, 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (998, 'Blue Tea Set', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (999, 'Sword Display', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1001, 'Custom Furniture', 1, 1, 475, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1002, 'Custom Furniture', 1, 1, 475, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1003, 'Custom Furniture', 1, 1, 475, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1004, 'Custom Furniture', 1, 1, 475, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1005, 'Custom Furniture', 1, 1, 475, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1006, 'Custom Furniture', 1, 1, 475, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1007, 'Custom Furniture', 1, 1, 475, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1008, 'Custom Furniture', 1, 1, 475, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1009, 'Custom Furniture', 1, 1, 475, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1010, 'Custom Furniture', 1, 1, 475, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1011, 'Custom Furniture', 1, 1, 475, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1012, 'Custom Furniture', 1, 1, 475, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1013, 'Custom Furniture', 1, 1, 475, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1014, 'Custom Furniture', 1, 1, 475, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1015, 'Custom Furniture', 1, 1, 475, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1016, 'Custom Furniture', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1017, 'Custom Furniture', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1018, 'Custom Furniture', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1019, 'Custom Furniture', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1020, 'Custom Furniture', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1021, 'Custom Furniture', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1022, 'Custom Furniture', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1023, 'Custom Furniture', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1024, 'Custom Furniture', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1025, 'Custom Furniture', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1026, 'Custom Furniture', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1027, 'Custom Furniture', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1028, 'Custom Furniture', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1029, 'Custom Furniture', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1030, 'Custom Furniture', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1031, 'Custom Furniture', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1032, 'Custom Furniture', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1033, 'Custom Furniture', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1034, 'Custom Furniture', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1035, 'Custom Furniture', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1036, 'Custom Furniture', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1037, 'Custom Furniture', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1038, 'Custom Furniture', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1039, 'Custom Furniture', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1040, 'Custom Furniture', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1041, 'Custom Furniture', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1042, 'Custom Furniture', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1043, 'Custom Furniture', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1044, 'Custom Furniture', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1045, 'Custom Furniture', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1050, 'Custom Furniture', 1, 1, 375, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1051, 'Custom Furniture', 1, 1, 375, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1052, 'Custom Furniture', 1, 1, 375, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1053, 'Custom Furniture', 1, 1, 375, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1054, 'Custom Furniture', 1, 1, 375, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1055, 'Custom Furniture', 1, 1, 375, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1056, 'Custom Furniture', 1, 1, 375, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1057, 'Custom Furniture', 1, 1, 375, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1058, 'Custom Furniture', 1, 1, 375, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1059, 'Custom Furniture', 1, 1, 375, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1060, 'Custom Furniture', 1, 1, 375, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1061, 'Custom Furniture', 1, 1, 375, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1062, 'Custom Furniture', 1, 1, 375, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1063, 'Custom Furniture', 1, 1, 375, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1064, 'Custom Furniture', 1, 1, 375, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1065, 'Custom Furniture', 1, 1, 275, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1066, 'Custom Furniture', 1, 1, 275, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1067, 'Custom Furniture', 1, 1, 275, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1068, 'Custom Furniture', 1, 1, 275, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1069, 'Custom Furniture', 1, 1, 275, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1070, 'Custom Furniture', 1, 1, 275, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1071, 'Custom Furniture', 1, 1, 275, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1072, 'Custom Furniture', 1, 1, 275, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1073, 'Custom Furniture', 1, 1, 275, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1074, 'Custom Furniture', 1, 1, 275, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1075, 'Custom Furniture', 1, 1, 275, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1076, 'Custom Furniture', 1, 1, 275, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1077, 'Custom Furniture', 1, 1, 275, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1078, 'Custom Furniture', 1, 1, 275, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1079, 'Custom Furniture', 1, 1, 275, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1080, 'Custom Furniture', 3, 3, 180, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1081, 'Custom Furniture', 3, 3, 180, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1082, 'Custom Furniture', 3, 3, 180, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1083, 'Custom Furniture', 3, 3, 180, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1084, 'Custom Furniture', 3, 3, 180, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1085, 'Custom Furniture', 3, 3, 180, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1086, 'Custom Furniture', 3, 3, 180, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1087, 'Custom Furniture', 3, 3, 180, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1088, 'Custom Furniture', 3, 3, 180, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1089, 'Custom Furniture', 3, 3, 180, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1090, 'Custom Furniture', 3, 3, 180, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1091, 'Custom Furniture', 3, 3, 180, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1092, 'Custom Furniture', 3, 3, 180, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1093, 'Custom Furniture', 3, 3, 180, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1094, 'Custom Furniture', 3, 3, 180, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1100, 'Custom Furniture', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1101, 'Custom Furniture', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1102, 'Custom Furniture', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1103, 'Custom Furniture', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1104, 'Custom Furniture', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1105, 'Custom Furniture', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1106, 'Custom Furniture', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1107, 'Custom Furniture', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1108, 'Custom Furniture', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1109, 'Custom Furniture', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1110, 'Custom Furniture', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1111, 'Custom Furniture', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1112, 'Custom Furniture', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1113, 'Custom Furniture', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1114, 'Custom Furniture', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1115, 'Custom Furniture', 1, 1, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1116, 'Custom Furniture', 1, 1, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1117, 'Custom Furniture', 1, 1, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1118, 'Custom Furniture', 1, 1, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1119, 'Custom Furniture', 1, 1, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1120, 'Custom Furniture', 1, 1, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1121, 'Custom Furniture', 1, 1, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1122, 'Custom Furniture', 1, 1, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1123, 'Custom Furniture', 1, 1, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1124, 'Custom Furniture', 1, 1, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1125, 'Custom Furniture', 1, 1, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1126, 'Custom Furniture', 1, 1, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1127, 'Custom Furniture', 1, 1, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1128, 'Custom Furniture', 1, 1, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1129, 'Custom Furniture', 1, 1, 600, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1130, 'Custom Furniture', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1131, 'Custom Furniture', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1132, 'Custom Furniture', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1133, 'Custom Furniture', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1134, 'Custom Furniture', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1135, 'Custom Furniture', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1136, 'Custom Furniture', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1137, 'Custom Furniture', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1138, 'Custom Furniture', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1139, 'Custom Furniture', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1140, 'Custom Furniture', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1141, 'Custom Furniture', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1142, 'Custom Furniture', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1143, 'Custom Furniture', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1144, 'Custom Furniture', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1150, 'Custom Furniture', 1, 1, 475, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1151, 'Custom Furniture', 1, 1, 475, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1152, 'Custom Furniture', 1, 1, 475, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1153, 'Custom Furniture', 1, 1, 475, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1154, 'Custom Furniture', 1, 1, 475, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1155, 'Custom Furniture', 1, 1, 475, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1156, 'Custom Furniture', 1, 1, 475, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1157, 'Custom Furniture', 1, 1, 475, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1158, 'Custom Furniture', 1, 1, 475, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1159, 'Custom Furniture', 1, 1, 475, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1160, 'Custom Furniture', 1, 1, 475, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1161, 'Custom Furniture', 1, 1, 475, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1162, 'Custom Furniture', 1, 1, 475, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1163, 'Custom Furniture', 1, 1, 475, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1164, 'Custom Furniture', 1, 1, 475, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1165, 'Custom Furniture', 2, 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1166, 'Custom Furniture', 2, 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1167, 'Custom Furniture', 2, 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1168, 'Custom Furniture', 2, 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1169, 'Custom Furniture', 2, 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1170, 'Custom Furniture', 2, 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1171, 'Custom Furniture', 2, 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1172, 'Custom Furniture', 2, 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1173, 'Custom Furniture', 2, 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1174, 'Custom Furniture', 2, 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1175, 'Custom Furniture', 2, 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1176, 'Custom Furniture', 2, 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1177, 'Custom Furniture', 2, 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1178, 'Custom Furniture', 2, 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1179, 'Custom Furniture', 2, 2, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1181, 'Custom Furniture', 3, 3, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1182, 'Custom Furniture', 3, 3, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1183, 'Custom Furniture', 3, 3, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1184, 'Custom Furniture', 3, 3, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1185, 'Custom Furniture', 3, 3, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1186, 'Custom Furniture', 3, 3, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1187, 'Custom Furniture', 3, 3, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1188, 'Custom Furniture', 3, 3, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1189, 'Custom Furniture', 3, 3, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1190, 'Custom Furniture', 3, 3, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1191, 'Custom Furniture', 3, 3, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1192, 'Custom Furniture', 3, 3, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1193, 'Custom Furniture', 3, 3, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1194, 'Custom Furniture', 3, 3, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (1195, 'Custom Furniture', 3, 3, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2000, 'Monster Bleachers', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2001, 'Green Triangle Pennants', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2002, 'Pink Triangle Pennants', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2003, 'Orange Triangle Pennants', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2004, 'Monster Scoreboard', 2, 2, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2005, 'Monster Archway', 1, 1, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2006, 'OK Pennant', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2007, '3D OK', 2, 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2008, 'PNK Pennant', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2009, '3D PNK', 2, 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2010, 'JOX Pennant', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2011, '3D JOX', 2, 2, 100, TRUE, FALSE, FALSE, FALSE, TRUE, 99),
 (2012, 'Monster Table', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2013, 'Monster Library Shelves', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2014, 'Scare Can', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, TRUE, 99),
 (2015, 'Monster Buffet Food', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2016, 'Monster Rug', 3, 3, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2017, 'Monster Library Table', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2018, 'Monster Eye Pillar', 1, 1, 125, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2019, 'Monster Door Station', 1, 1, 450, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2020, 'Red Triangle Pennants', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2021, 'ROR Pennant', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2022, '3D ROR', 2, 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2023, 'Monster Boombox', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2024, 'Monster Ping Pong Table', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2025, 'Monster Lounge Chair', 1, 1, 350, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2026, 'Imperial Wall Panel', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2027, 'Radar Screen', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2028, 'Droid Cleaning Station', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2029, 'Holonet Terminal', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2030, 'Millennium Falcon Seats', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2031, 'System Readout Terminal', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2032, 'Dejarik Set', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2033, 'GNK Power Droid', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2034, 'Imperial Archway', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2035, 'Computer Console', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2036, 'Rebel House Rug', 3, 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2037, 'Imperial House Rug', 3, 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2038, 'LIN Demolitionmech Droid', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2039, 'TIE Fighter Chair', 1, 1, 450, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2040, 'X-wing Fighter Chair', 1, 1, 450, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2041, 'Tatooine House', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2042, 'Imperial Throne', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2043, 'Holonet Tracking Console', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2044, 'Radar Computer', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2045, 'Death Star', 2, 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2046, 'Swimming Pool', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2047, 'Board Wave', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2048, 'Mermaid Cutout', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2049, 'Merman Cutout', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2050, 'Lit Stage', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2051, 'Beach Beatz', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2052, 'Dressing Table', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2053, 'Retro Jukebox', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2054, 'Inflatable Lounge Chair', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2055, 'Dressing Screen', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2056, 'Lifeguard Chair', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2057, 'Pink Curtains', 2, 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2058, 'Diving Board', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2059, 'Inner Tube', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2060, 'Big Momma''s Diner', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2061, 'Mic', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2062, 'Beach Boards', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2063, 'Motorbike', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2064, 'Gas Pump', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2065, 'Castle Gate', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2066, 'Stone Keep', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2067, 'Watch Tower', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2068, 'Castle Entrance', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2069, 'Battlements', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2070, 'Regal Drapes', 2, 2, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2071, 'Ogre Fence', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2072, 'Ogre Drapes', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2073, 'Ogre Table', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2074, 'Fairy Tree Stump', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2075, 'Fairy Flower', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2076, 'Steel Anvil', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2077, 'Wish You Well', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2078, 'Regal Throne', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2079, 'Potions Table', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2080, 'Magical Garden', 3, 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2081, 'Royal Cannon', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2082, 'Magical Fairy Plant', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2083, 'Barrel Top', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2084, 'Ye Olde Puffle Bowl', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2085, 'Jar Of Eyes', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2086, 'Ogre Puffle Head', 2, 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2087, 'Stone Gatepost', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2088, 'Coffin Cabinet', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2089, 'Gnarled Tree', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2090, 'Eerie Cave', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2091, 'Web Rug', 3, 3, 170, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2092, 'Ghastly Window', 2, 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2093, 'Garland of Sweets', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2094, 'Ghost Garland', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2095, 'Jack-O-Lantern Garland', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2096, 'Candy Garland', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2097, 'Mini Pumpkin Lanterns', 2, 2, 20, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2098, 'Batty Garland', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2099, 'Bat Decal', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2100, 'Neon Bat', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2101, 'Happy Lantern Light', 2, 2, 20, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2102, 'Scary Lantern Light', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2103, 'Wall Webbing', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2104, 'Decrepit Chandelier', 2, 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2105, 'Goblin Eye', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2106, 'Mysterious Mist', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2107, 'Glowing Cauldron', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2108, 'Purple Pit', 3, 3, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2109, 'Gazing Crystal', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2110, 'Abandoned Mine Wall', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2111, 'Abandoned Mine Entrance', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2112, 'Graveyard Plot', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2113, 'Mini Ghost', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2114, 'Spooky Jack-O-Lantern', 1, 1, 20, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2115, 'Candy Cauldron', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2116, 'Slimy Tentacles', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2117, 'Bone Fence', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2118, 'Lazy Bones', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2119, 'Graveyard Gate', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2120, 'Graveyard Fence', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2121, 'Haunted House Entrance', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2122, 'Haunted House Wall', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2123, 'The Count''s Chair', 1, 1, 20, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2124, 'The Count''s Couch', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2125, 'Trick Or Treat Pumpkin', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2126, 'River Pontoon Boat', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2129, 'Gold Walkway', 3, 3, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2130, 'Golden Lava Puddle', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2131, 'Golden Lava Pool', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2132, 'Gold Chandelier', 2, 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2133, 'Snowman', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 1),
 (2134, 'Puffle Freedom Statue', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 1),
 (2135, 'Tall Snowy Tree', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2136, 'Puffle Kennel', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2137, 'Rope Climbing Wall', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2138, 'River''s Edge', 3, 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2139, 'Frozen Ledge', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2140, 'Hot n''Cold Crater', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2141, 'Medic Tent', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2142, 'Log Wall', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2143, 'Wooden Post', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2144, 'Log Cabin', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2145, 'Candy Cane Lantern', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2146, 'Holiday Station Clock', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2147, 'Railroad Corner', 3, 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2148, 'Railroad Crossing', 3, 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2149, 'Train Crossing Sign', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2150, 'Railroad Piece', 3, 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2151, 'Railroad Crossing Sign', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2152, 'Railroad Intersection', 3, 3, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2153, 'Handbag', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2154, 'Luggage Case', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2155, 'Holiday Centerpiece', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2156, 'Holiday Wreath', 2, 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2157, 'Train Car Present', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2158, 'Train Engine Present', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2159, 'Station Bench', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2160, 'Newsstand', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2161, 'Gold Railroad Piece', 3, 3, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 4),
 (2162, 'Gold Railroad Corner', 3, 3, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 4),
 (2163, 'Gold Railroad Crossing', 3, 3, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 1),
 (2164, 'Gold Railroad Intersection', 3, 3, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 1),
 (2165, 'Red Line Tunnel', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 1),
 (2166, 'Table for Two', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 1),
 (2167, 'Green Line Tunnel', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 1),
 (2168, 'Blue Line Tunnel', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 1),
 (2169, 'Bone Bridge', 3, 3, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2170, 'Volcanic Glass Table', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2171, 'Ice Age 3000 BCE', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2172, 'Designer Log Stool', 1, 1, 25, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2173, 'Fish Fossil', 2, 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2174, 'Slab Sofa', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2175, 'Slab Recliner', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2176, 'Sabre-tooth Rug', 3, 3, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2177, 'Tortoise Dining Table', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2178, 'Volcanic Glass Side Table', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2179, 'Dino Dig Site', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 1),
 (2180, 'Hanging Torch', 2, 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2182, 'Primal Shark Arch', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2183, 'Tricera-statue', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2184, 'Prehistoric Gate', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2185, 'Prehistoric Wall', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2186, 'Chieftain''s Seat', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2187, 'Clam Chair', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2188, 'Claw Marks', 1, 1, 25, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2189, 'Popcorn Cart', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2190, 'Carousel Horse', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2191, 'Balloon Pop Booth', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2192, 'Feed-a-Puffle Booth', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2193, 'Lunar Launch Booth', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2194, 'Memory Cards Booth', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2195, 'Puffle Paddle Booth', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2196, 'Puffle Shuffle Booth', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2197, 'Puffle Soaker Booth', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2198, 'Blue Coaster Cart', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2199, 'Orange Coaster Cart', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2200, 'Ticket Hut', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2201, 'Inspiration Station', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2202, 'CP Air Seat', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2203, 'CP Air Pilot''s Chair', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2204, 'Check-in Terminal', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2205, 'Gate Chair', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2206, 'Airport Departures Board', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2207, 'Elastic Barrier', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2208, 'Theater Seat', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2209, 'Stage Curtain', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2210, 'Grand Stage Arch', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2211, 'Footlight Stage', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2212, 'Long Puffle Tube', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2213, 'Short Puffle Tube', 1, 1, 25, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2214, 'Puffle Bubble Tube', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2215, 'Long Solid Tube', 1, 1, 55, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2216, 'Short Solid Tube', 1, 1, 30, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2217, 'Long Window Tube', 1, 1, 55, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2218, 'Short Window Tube', 1, 1, 30, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2219, 'T-joint Puffle Tube', 1, 1, 60, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2220, '4-Way Puffle Tube', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2221, 'Corner Puffle Tube', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2222, 'Puffle Tube Bend', 1, 1, 25, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2223, 'Puffle Tube Box', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2224, 'Puffle Tube Tower', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2225, 'Giant Puffle Hub', 1, 1, 500, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2226, 'Holo-projector', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2227, 'Rocketship Lava Lamp', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2228, 'Futura Stool', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2229, 'Starship Console', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2230, 'Galactic Pod Chair', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2231, 'Galactic Pod Couch', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2232, 'Planet Lamp', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2233, 'Futura Table', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2234, 'Cinema Pod', 1, 1, 125, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2235, 'UFO', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2236, 'Force Field', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2237, 'Soccer Goal', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2238, 'Stadium Bleachers', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2239, 'Multi-Ball', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2240, 'Equipment Rack', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2241, 'Training Hurdle', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2242, 'Team Locker', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2243, 'Go Sharks', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2244, 'Go Space Squids', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2245, 'Go Hot Sauce', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2246, 'Go Fluffies', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2247, 'Soccer Bean Bag', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2248, 'Modern Coffee Table', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2249, 'Modern End Table', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2250, 'Granite Top Dishwasher', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2251, 'Granite Top Corner Cabinet', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2252, 'Granite Top Cabinet', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2253, 'Granite Sink', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2254, 'Granite Top Double Cabinet', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2255, 'Brushed Steel Fridge', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2256, 'Granite Kitchen Island', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2257, 'Upper Cabinets', 2, 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2258, 'Brushed Steel Oven', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2259, 'Kitchen Stool', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2260, 'Wall Cabinet', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2261, 'Double Wall Cabinet', 2, 2, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2262, 'Corner Wall Cabinet', 2, 2, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2263, 'Electric Keyboard', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2264, 'Rock Out Amp', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2265, 'Six String Tune Table', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2266, 'Cozy Porch Swing', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2267, 'Patio Table', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2268, 'Cozy Porch Chair', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2269, 'SoundStudio Booth', 1, 1, 300, TRUE, FALSE, FALSE, FALSE, FALSE, 1),
 (2270, 'Ice Throne', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 5),
 (2271, 'Ice Dining Table', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 3),
 (2272, 'Ice Rug', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 10),
 (2273, 'Ice Block', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2274, 'Stone Fireplace', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2275, 'Painted Rocking Horse', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2276, 'Ice Delivery Sleigh', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2277, 'Beach Party Ball', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2278, 'Sand Bucket', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2279, 'Sunscreen', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2280, 'Patio Parasol', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2281, 'Left Princess Door', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2282, 'Right Princess Door', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2283, 'Ice Lantern', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2284, 'Ice Pillar', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2285, 'Ice Coffee Table', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2286, 'CPU Student Desk', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2287, 'CPU Chair', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2288, 'CPU Locker', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2289, 'Teacher''s Desk', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2290, 'Classy Bookshelf', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2291, 'Wheelie Board', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2292, 'Cafeteria Delights', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2293, 'Magic Crystal Ball', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2294, 'Spooky Hearth', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2295, 'Antique Post Box', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2296, 'Jack-O-Lantern Curtains', 2, 2, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2297, 'O''berry Bush', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2298, 'Wilds Puffle House', 1, 1, 120, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2299, 'Wilds Puffle Treehouse', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2300, 'Puffle Cave', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2301, 'Rustic Puffle Table', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2302, 'Wooden Puffle Swing', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2303, 'Pirate Barrel', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2304, 'Ship Shrouds', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2305, 'Bowsprit', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2306, 'Ship Deck', 1, 1, 25, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2307, 'Mast', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2308, 'Ship Front Left', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2309, 'Ship Front Right', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2310, 'Crow''s Nest Front', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2311, 'Crow''s Nest Back', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2312, 'Ship Hull Side', 1, 1, 25, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2313, 'Ship Hull Corner', 1, 1, 25, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2314, 'Pirate Diving Board', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2315, 'Crab Treasure Chest', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 1),
 (2316, 'Ship''s Wheel', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2317, 'Crystal Candy Cane', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2318, 'Ice Cube Wreath', 2, 2, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2319, 'Blue Ornament Chair', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2320, 'Merry Walrus Snow Sculpture', 1, 1, 250, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2321, 'Potted Poinsettia', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2322, 'Festive Fluffy Feast', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2323, 'Toy Maker''s Desk', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2324, 'Merry Walrus Tree', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2325, 'Merry Walrus Garland', 2, 2, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2326, 'Stable Stall', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2327, 'Speeder Bike', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 50),
 (2328, 'Starship Seats', 1, 1, 125, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2329, 'Droid Conductor', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2330, 'Starship Stool', 1, 1, 25, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2331, 'Starship Panel', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2332, 'Imperial Holoprojector', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2333, 'Imperial Supply Crate', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2334, 'Ezra''s Work Bench', 1, 1, 125, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2335, 'Rebels Area Rug', 3, 3, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2336, 'Judge''s Platform', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2337, 'Puffle Wilds Post', 1, 1, 80, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2338, 'Puffle Wilds Statue', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2339, 'Puffle Wilds Sculpture', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2340, 'Stump Seat', 1, 1, 25, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2341, 'Stump Drawer', 1, 1, 80, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2342, 'Stump Bookcase', 1, 1, 125, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2343, 'Recliner Stump', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2344, 'Wood Ring Table', 1, 1, 150, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2345, 'Multi-berry Bush', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2346, 'Large Multi-berry Bush', 1, 1, 80, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2347, 'Black TV Stand', 1, 1, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2348, 'Gray TV Stand', 1, 1, 400, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2349, 'Black Designer Couch', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2350, 'Brown Designer Couch', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2351, 'Red Designer Couch', 1, 1, 200, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2352, 'Multi Shelves', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2353, 'Wood Shelves', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2354, 'Mounted Speaker', 2, 2, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2355, 'Surround Sound Speaker', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2356, 'Tower Speaker', 1, 1, 80, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2357, 'Spring Flowers Vase', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2358, 'Designer Vase', 1, 1, 50, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2359, 'Costume Trunk', 1, 1, 75, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2360, 'Switch Box 3000', 1, 1, 100, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (2361, 'Baby Grand Piano', 1, 1, 125, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (10076, 'Folding Chair', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (10077, 'Beach Umbrella', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (10079, 'Inflatable Whale', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (10156, 'Teddy Bear', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (10157, 'Student Desk', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (10185, 'Classroom Desk', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 1),
 (10187, 'Classroom Chair', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 1),
 (10214, 'Water Bottle', 4, 4, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (10219, 'Puffle Tent', 1, 4, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (10220, 'Puffle Condo', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (10229, 'Bears Pond', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (10293, 'Weight Bench', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (10304, 'Globe', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 1),
 (10425, 'Desert Island', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (10540, 'Sprouting Spectaculous', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (10544, 'Tropical Palm', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (10545, 'Soccer Net', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (10573, 'Snack Stand', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (10656, 'Presents', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (10701, 'Bears Scratching Post', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (10737, 'Trophy Shelf', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (10774, 'Ninja Gate R', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (10778, 'Noodle Stand R', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (10781, 'Training Dummy R', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (10832, 'Fuzzy Blue Couch', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (10915, 'Short Security Laser', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (10941, 'Lava Puddle', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (10982, 'Puffle Patio Lanterns', 2, 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (12035, 'Computer Console', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (12062, 'Beach Boards', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (12114, 'Spooky Jack-o-lantern', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 1),
 (12123, 'The Count’s Chair', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 1),
 (12127, 'CP Magazine Star Trophy', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (12128, 'CP Magazine Published Reader Trophy', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (12129, 'Matryoshka Doll', 1, 1, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99),
 (12156, 'Holiday Wreath', 2, 2, 0, TRUE, FALSE, FALSE, FALSE, FALSE, 99);

INSERT INTO igloo (id, name, cost, patched, legacy_inventory, vanilla_inventory) VALUES
 (0, 'Igloo Removal', 0, FALSE, FALSE, FALSE),
 (1, 'Basic Igloo', 1500, FALSE, TRUE, TRUE),
 (2, 'Candy Igloo', 1500, FALSE, FALSE, FALSE),
 (3, 'Deluxe Blue Igloo', 4000, FALSE, FALSE, FALSE),
 (4, 'Big Candy Igloo', 4000, FALSE, FALSE, FALSE),
 (5, 'Secret Stone Igloo', 2000, FALSE, FALSE, FALSE),
 (6, 'Snow Igloo', 1000, FALSE, FALSE, FALSE),
 (8, 'Secret Deluxe Stone Igloo', 5000, FALSE, FALSE, FALSE),
 (9, 'Deluxe Snow Igloo', 3000, FALSE, FALSE, FALSE),
 (10, 'Bamboo Hut', 3200, FALSE, FALSE, FALSE),
 (11, 'Log Cabin', 4100, FALSE, FALSE, FALSE),
 (12, 'Gym', 4800, FALSE, FALSE, FALSE),
 (13, 'Split Level Igloo', 4600, FALSE, FALSE, FALSE),
 (14, 'Candy Split Level Igloo', 4600, FALSE, FALSE, FALSE),
 (15, 'Snowglobe', 3700, FALSE, FALSE, FALSE),
 (16, 'Ice Castle', 2400, FALSE, FALSE, FALSE),
 (17, 'Split Level Snow Igloo', 4600, FALSE, FALSE, FALSE),
 (18, 'Fish Bowl', 2400, FALSE, FALSE, FALSE),
 (19, 'Tent', 2700, FALSE, FALSE, FALSE),
 (20, 'Jack O'' Lantern', 2700, FALSE, FALSE, FALSE),
 (21, 'Backyard Igloo', 4200, FALSE, FALSE, FALSE),
 (22, 'Pink Ice Palace', 2400, FALSE, FALSE, FALSE),
 (23, 'Ship Igloo', 4300, FALSE, FALSE, FALSE),
 (24, 'Dojo Igloo', 1300, FALSE, FALSE, FALSE),
 (25, 'Gingerbread House', 2100, FALSE, FALSE, FALSE),
 (26, 'Restaurant Igloo', 4800, FALSE, FALSE, FALSE),
 (27, 'Tree House Igloo', 4500, FALSE, FALSE, FALSE),
 (28, 'Theatre Igloo', 4600, FALSE, FALSE, FALSE),
 (29, 'Circus Tent', 3000, FALSE, FALSE, FALSE),
 (30, 'Snowy Backyard Igloo', 3000, FALSE, FALSE, FALSE),
 (31, 'Cave Igloo', 1500, FALSE, FALSE, FALSE),
 (32, 'Green Clover Igloo', 2050, FALSE, FALSE, FALSE),
 (33, 'Grey Ice Castle', 2400, FALSE, FALSE, FALSE),
 (35, 'Cozy Cottage Igloo', 2500, FALSE, FALSE, FALSE),
 (36, 'Estate Igloo', 2500, FALSE, FALSE, FALSE),
 (37, 'In Half Igloo', 2300, FALSE, FALSE, FALSE),
 (38, 'Shadowy Keep', 1800, FALSE, FALSE, FALSE),
 (39, 'Dragon''s Lair', 3000, FALSE, FALSE, FALSE),
 (40, 'Mermaid Cove', 3030, FALSE, FALSE, FALSE),
 (41, 'Whale''s Mouth', 2700, FALSE, FALSE, FALSE),
 (42, 'Trick-or-Treat Igloo', 2000, FALSE, FALSE, FALSE),
 (43, 'Deluxe Gingerbread House', 0, FALSE, FALSE, FALSE),
 (45, 'Invisible Snowy', 0, FALSE, FALSE, FALSE),
 (46, 'Invisible Beach', 0, FALSE, FALSE, FALSE),
 (47, 'Invisible Forest', 0, FALSE, FALSE, FALSE),
 (48, 'Invisible Mountain', 0, FALSE, FALSE, FALSE),
 (49, 'Shipwreck Igloo', 900, FALSE, FALSE, FALSE),
 (50, 'Wildlife Den', 900, FALSE, FALSE, FALSE),
 (51, 'Medieval Manor', 1200, FALSE, FALSE, FALSE),
 (52, 'Warehouse', 950, FALSE, FALSE, FALSE),
 (53, 'Pineapple Igloo', 0, FALSE, FALSE, FALSE),
 (54, 'Creepy Cavern', 1500, FALSE, FALSE, FALSE),
 (55, 'Frost Bite Palace', 2000, FALSE, FALSE, FALSE),
 (56, 'Fresh Baked Gingerbread House', 2500, FALSE, FALSE, FALSE),
 (57, 'Penthouse', 4000, FALSE, FALSE, FALSE),
 (58, 'VIP Penthouse', 4500, FALSE, FALSE, FALSE),
 (59, 'Invisible Age of Dinosaurs', 0, FALSE, FALSE, FALSE),
 (60, 'Puffle Tree Fort', 0, FALSE, FALSE, FALSE),
 (61, 'Secret Base', 1600, FALSE, FALSE, FALSE),
 (62, 'Imperial Base Igloo', 1000, FALSE, FALSE, FALSE),
 (63, 'Beach Party Igloo', 1500, FALSE, FALSE, FALSE),
 (64, 'Gymnasium Igloo', 0, FALSE, FALSE, FALSE),
 (65, 'Magical Hideout', 1500, FALSE, FALSE, FALSE),
 (66, 'Eerie Castle', 2000, FALSE, FALSE, FALSE),
 (67, 'Sweet Swirl Igloo', 0, FALSE, FALSE, FALSE),
 (68, 'Train Station Igloo', 1100, FALSE, FALSE, FALSE),
 (69, 'Main Event Igloo', 1000, FALSE, FALSE, FALSE),
 (70, 'CP Airliner', 1200, FALSE, FALSE, FALSE),
 (71, 'Puffle Tree House', 1500, FALSE, FALSE, FALSE),
 (72, 'Invisible Distant Planet', 0, FALSE, FALSE, FALSE),
 (73, 'Space Dome Igloo', 2000, FALSE, FALSE, FALSE),
 (74, 'Invisible Soccer Pitch', 0, FALSE, FALSE, FALSE),
 (75, 'Tour Bus Igloo', 1800, FALSE, FALSE, FALSE),
 (76, 'Ice Palace Igloo', 0, FALSE, FALSE, FALSE),
 (77, 'Sharks Gym', 1500, FALSE, FALSE, FALSE),
 (78, 'Schoolhouse Igloo', 0, FALSE, FALSE, FALSE),
 (79, 'Ghostly Cavern Igloo', 1800, FALSE, FALSE, FALSE),
 (80, 'Invisible Undersea', 0, FALSE, FALSE, FALSE),
 (81, 'Invisible Merry Walrus Iceberg', 0, FALSE, FALSE, FALSE),
 (82, 'Ezra''s Hideout', 1800, FALSE, FALSE, FALSE),
 (83, 'Starship Igloo', 0, FALSE, FALSE, FALSE),
 (84, 'Talent Show Stage', 1500, FALSE, FALSE, FALSE),
 (85, 'Red Puffle Tree House', 1000, FALSE, FALSE, FALSE),
 (86, 'Orange Puffle Tree House', 1000, FALSE, FALSE, FALSE),
 (87, 'Yellow Puffle Tree House', 1000, FALSE, FALSE, FALSE),
 (88, 'Green Puffle Tree House', 1000, FALSE, FALSE, FALSE),
 (89, 'Blue Puffle Tree House', 1000, FALSE, FALSE, FALSE),
 (90, 'Purple Puffle Tree House', 1000, FALSE, FALSE, FALSE),
 (91, 'Black Puffle Tree House', 1000, FALSE, FALSE, FALSE),
 (92, 'White Puffle Tree House', 1000, FALSE, FALSE, FALSE),
 (93, 'Brown Puffle Tree House', 1000, FALSE, FALSE, FALSE),
 (94, 'Pink Puffle Tree House', 1000, FALSE, FALSE, FALSE),
 (95, 'Gold Puffle Tree House', 1000, FALSE, FALSE, FALSE),
 (96, 'Rainbow Puffle Tree House', 1000, FALSE, FALSE, FALSE),
 (97, 'Spring Palace', 0, FALSE, FALSE, FALSE),
 (98, 'Stage Igloo', 1500, FALSE, FALSE, FALSE);

INSERT INTO location (id, name, cost, patched, legacy_inventory, vanilla_inventory) VALUES
 (1, 'Default', 0, FALSE, TRUE, TRUE),
 (2, 'Beach', 2800, FALSE, FALSE, FALSE),
 (3, 'Forest', 2700, FALSE, FALSE, FALSE),
 (4, 'Mountain', 2800, FALSE, FALSE, FALSE),
 (5, 'Age Of Dinosaurs', 0, FALSE, FALSE, FALSE),
 (6, 'Distant Planet', 1800, FALSE, FALSE, FALSE),
 (7, 'Soccer Pitch', 1800, FALSE, FALSE, FALSE),
 (8, 'Undersea', 1800, FALSE, FALSE, FALSE);

INSERT INTO postcard (id, name, cost, enabled) VALUES
 (1, 'Re-decorated Igloo', 10, TRUE),
 (2, 'Puffle Performance', 10, TRUE),
 (3, 'Penguins Time Forgot', 10, TRUE),
 (4, 'Coffee Shop', 10, TRUE),
 (5, 'Pizza Parlor', 10, TRUE),
 (6, 'Happy Birthday', 10, TRUE),
 (7, 'Costume Try-Outs', 10, TRUE),
 (8, 'Norman Swarm Has Been Transformed', 10, TRUE),
 (9, 'Hide and Seek', 10, TRUE),
 (10, 'Mancala', 10, TRUE),
 (11, 'Happy Christmas Postcard', 10, TRUE),
 (12, 'Welcome to Club Penguin', 10, TRUE),
 (13, 'Thanks for Gift', 10, TRUE),
 (14, 'Beach', 10, TRUE),
 (15, 'Donate', 10, TRUE),
 (16, 'Ice Rink', 10, TRUE),
 (17, 'Cool Igloo', 10, TRUE),
 (19, 'Great Game', 10, TRUE),
 (20, 'Snow Forts', 10, TRUE),
 (21, 'Pool', 10, TRUE),
 (22, 'Happy Halloween', 10, TRUE),
 (23, 'Trick or Treat', 10, TRUE),
 (24, 'Cool Name', 10, TRUE),
 (25, 'Awesome Outfit', 10, TRUE),
 (26, 'I Like Your Outfit', 10, TRUE),
 (27, 'Party at my Igloo', 10, TRUE),
 (28, 'Bamboo Forest', 10, TRUE),
 (29, 'Tip The Berg', 10, TRUE),
 (30, 'Hidden Lake', 10, TRUE),
 (31, 'Invited to the Penguin Play Awards', 10, TRUE),
 (32, 'Terrific Friend', 10, TRUE),
 (33, 'Interview', 10, TRUE),
 (34, 'Puffle Rescue', 10, TRUE),
 (35, 'Long Time No See', 10, TRUE),
 (36, 'You''re Funny', 10, TRUE),
 (37, 'Turn Green', 10, TRUE),
 (38, 'Community Garden', 10, TRUE),
 (39, 'Meet me for a sled race', 10, TRUE),
 (40, 'Band Practice', 10, TRUE),
 (41, 'Watch My Band Play', 10, TRUE),
 (42, 'Wanna join my band', 10, TRUE),
 (43, 'Happy Puffles', 10, TRUE),
 (44, 'Best Friends for Life!', 10, TRUE),
 (45, 'Cool Puffles', 10, TRUE),
 (46, 'Aloha', 10, TRUE),
 (47, 'EPF Recruitment', 10, TRUE),
 (48, 'Follow me!', 10, TRUE),
 (49, 'Dance Party', 10, TRUE),
 (50, 'Your Biggest Fan', 10, TRUE),
 (51, 'Jet Pack', 10, TRUE),
 (52, 'Ski Hill', 10, TRUE),
 (53, 'Lighthouse', 10, TRUE),
 (54, 'Stage', 10, TRUE),
 (55, 'Dojo', 10, TRUE),
 (56, 'Hey There', 10, TRUE),
 (57, 'Be My Buddy', 10, TRUE),
 (58, 'Cart Surfer', 10, TRUE),
 (59, 'Find Four', 10, TRUE),
 (60, 'Ice Hockey', 10, TRUE),
 (61, 'Igloo Party', 10, TRUE),
 (62, 'Let''s Hang Out', 10, TRUE),
 (63, 'Meet me at the Nightclub', 10, TRUE),
 (64, 'Stopping By', 10, TRUE),
 (65, 'Soccer', 10, TRUE),
 (66, 'Ruby and the Ruby', 10, TRUE),
 (67, 'Planet Y', 10, TRUE),
 (68, 'Treasure Hunt', 10, TRUE),
 (69, 'Aqua Grabber', 10, TRUE),
 (70, 'Cool Outfit', 10, TRUE),
 (71, 'Actors Needed', 10, TRUE),
 (72, 'Spin some tunes', 10, TRUE),
 (73, 'Season''s Greetings', 10, TRUE),
 (74, 'Star in a play', 10, TRUE),
 (75, 'Card-Jitsu', 10, TRUE),
 (76, 'Donate', 10, TRUE),
 (77, 'Merry Christmas', 10, TRUE),
 (78, 'Happy New Year', 10, TRUE),
 (79, 'Winter Fiesta', 10, TRUE),
 (80, 'Dance Off', 10, TRUE),
 (81, 'Squidzoid', 10, TRUE),
 (82, 'Puffle Party', 10, TRUE),
 (83, 'Snow Sculpture Showcase', 10, TRUE),
 (84, 'I met Rockhopper', 10, TRUE),
 (85, 'Join Team Blue', 10, TRUE),
 (86, 'Join Team Red', 10, TRUE),
 (87, 'St Patrick''s Day', 10, TRUE),
 (88, 'My Tour', 10, TRUE),
 (89, 'Tree Forts', 10, TRUE),
 (90, 'The Haunting of the Viking Opera', 10, TRUE),
 (91, 'Fairy Fables', 10, TRUE),
 (92, 'Let''s Explore', 10, TRUE),
 (93, 'Band practice in my igloo', 10, TRUE),
 (94, 'Vote Lavender', 10, TRUE),
 (95, 'Vote Aqua', 10, TRUE),
 (96, 'Vote Maroon', 10, TRUE),
 (97, 'Book Room', 10, TRUE),
 (98, 'Underwater Adventure', 10, TRUE),
 (99, 'Find the Pin', 10, TRUE),
 (100, 'Blue Ran', 0, FALSE),
 (101, 'Pink Ran', 0, FALSE),
 (102, 'Black Ran', 0, FALSE),
 (103, 'Green Ran', 0, FALSE),
 (104, 'Purple Ran', 0, FALSE),
 (105, 'Red Ran', 0, FALSE),
 (106, 'Yellow Ran', 0, FALSE),
 (107, 'Halloween Igloo Contest Winner', 0, FALSE),
 (108, '4th Anniversary', 0, FALSE),
 (109, 'Orange Ran', 0, FALSE),
 (110, 'Feed me I''m hungry', 0, FALSE),
 (111, 'Adoption', 0, FALSE),
 (112, 'Attention', 0, FALSE),
 (120, 'Penguin Mail', 0, FALSE),
 (121, 'Membership Badge', 0, FALSE),
 (122, 'Membership expires', 0, FALSE),
 (123, 'Membership expires', 0, FALSE),
 (124, 'Membership expired', 0, FALSE),
 (125, 'Welcome to Club Penguin', 0, FALSE),
 (126, 'Congratulations Tour Guide', 0, FALSE),
 (127, 'Welcome to the PSA', 0, FALSE),
 (128, 'Thanks CPIP', 0, FALSE),
 (129, 'PSA Wants You', 0, FALSE),
 (130, 'Attention Agent', 0, FALSE),
 (131, '3rd Anniversary Party Invitation', 0, FALSE),
 (132, 'Halloween Decorating Contest', 0, FALSE),
 (133, 'Attention Agent', 0, FALSE),
 (134, 'Find your Amulet Postcard', 0, FALSE),
 (135, 'Igloo Contest', 0, FALSE),
 (156, 'Music Jam Backstage', 0, FALSE),
 (157, 'Christmas Decorating Contest', 0, FALSE),
 (158, 'You''re a Published Penguin', 0, FALSE),
 (159, 'Thanks for Donating', 10, TRUE),
 (160, 'Dance Party', 0, FALSE),
 (161, 'Published in the book room', 0, FALSE),
 (162, 'Published Sculpture', 0, FALSE),
 (163, 'Membership expires', 0, FALSE),
 (164, 'Membership Badge', 0, FALSE),
 (165, 'Invited to Penguin Play Awards', 0, FALSE),
 (166, 'Invited to Penguin Play Awards', 0, FALSE),
 (167, 'Penguin Awards Week 2', 0, FALSE),
 (168, 'Penguin Awards Week 3', 0, FALSE),
 (169, 'White Ran', 0, FALSE),
 (170, 'April Fool''s', 10, TRUE),
 (171, 'Jobs: Secret Agent', 0, FALSE),
 (172, 'Jobs: Tour Guide', 0, FALSE),
 (173, 'Easter', 0, FALSE),
 (174, 'Medieval Party', 0, FALSE),
 (175, 'Quest Awaits You', 0, FALSE),
 (176, 'Card-Jitsu', 0, FALSE),
 (177, 'Card-Jitsu White Belt', 0, FALSE),
 (178, 'Card-Jitsu Blue Belt', 0, FALSE),
 (179, 'Card-Jitsu Black Belt', 0, FALSE),
 (180, 'Dojo Contest 2nd Place', 0, FALSE),
 (181, 'Rockhopper''s Plants', 0, FALSE),
 (182, 'Dojo Contest 1st Place', 0, FALSE),
 (183, 'Sensei Mascot Invite', 0, FALSE),
 (184, 'Jobs: EPF Agent', 0, FALSE),
 (185, 'Brown Ran', 0, FALSE),
 (186, 'Earn Stamps for your Stamp Book', 0, FALSE),
 (187, 'Happy Birthday', 10, TRUE),
 (188, 'Coins for Change', 10, TRUE),
 (189, 'Keep up the good work', 10, TRUE),
 (190, 'You''re awesome!', 0, FALSE),
 (191, 'You''re a Published Penguin', 0, FALSE),
 (192, 'You are this month''s VIP', 0, FALSE),
 (200, 'Campfire Story', 10, TRUE),
 (201, 'Thank You', 10, TRUE),
 (202, 'Check out my igloo', 10, TRUE),
 (203, 'Join Team Green', 10, TRUE),
 (204, 'Join Team Yellow', 10, TRUE),
 (205, 'The Fair', 10, TRUE),
 (206, 'Stamp Book', 10, TRUE),
 (207, 'Join Team Blue', 10, TRUE),
 (208, 'Join Team Red', 10, TRUE),
 (209, 'Happy Holidays', 10, TRUE),
 (210, 'Season''s Greetings', 10, TRUE),
 (211, 'Let It Snow!', 10, TRUE),
 (212, 'Let''s Play Hockey', 10, TRUE),
 (213, 'Coins for Change', 10, TRUE),
 (216, 'Happy Holidays', 10, TRUE),
 (217, 'Let''s Explore', 10, TRUE),
 (218, 'My Stamp Book', 10, TRUE),
 (219, 'Best Friends for Life!', 10, TRUE),
 (220, 'Black Puffle', 10, TRUE),
 (221, 'Blue Puffle', 10, TRUE),
 (222, 'Bring your puffles!', 10, TRUE),
 (223, 'Brown Puffle', 10, TRUE),
 (224, 'Green Puffle', 10, TRUE),
 (225, 'Orange Puffle', 10, TRUE),
 (226, 'Pink Puffle', 10, TRUE),
 (227, 'Purple Puffle', 10, TRUE),
 (228, 'Red Puffle', 10, TRUE),
 (229, 'White Puffle', 10, TRUE),
 (230, 'Yellow Puffle', 10, TRUE),
 (231, 'Cool Puffles', 10, TRUE),
 (232, 'Medieval Party', 10, TRUE),
 (233, 'Medieval Quest Invite', 10, TRUE),
 (234, 'Band practise in my igloo', 10, TRUE),
 (235, 'Penguin Band Rocks', 10, TRUE),
 (236, 'Happy Holidays', 10, TRUE),
 (237, 'Thanks for my gift!', 10, TRUE),
 (238, 'You''ve earned a GeoPalz award', 0, FALSE),
 (239, 'You were published', 0, FALSE),
 (240, 'You''ve been published 3 times', 0, FALSE),
 (241, 'You are this month''s VIP', 0, FALSE);

INSERT INTO puffle_item (id, parent_id, name, type, play_external, cost, quantity, member, food_effect, rest_effect, play_effect, clean_effect) VALUES
 (1, 1, 'Brush', 'care', 'none', 0, 1, FALSE, -2, -2, 5, 5),
 (3, 3, 'Puffle O''s', 'food', 'none', 0, 1, FALSE, 50, -5, -5, -1),
 (4, 4, 'Cookie', 'food', 'none', 10, 1, FALSE, 15, -7, -20, -1),
 (5, 5, 'Bubblegum', 'food', 'bubblegum', 5, 1, FALSE, 0, -2, -2, -3),
 (6, 6, 'Carrot', 'food', 'none', 4, 1, FALSE, 50, -5, -5, 0),
 (7, 7, 'Pizza', 'food', 'none', 5, 1, FALSE, 30, -8, -10, -1),
 (8, 8, 'Bath', 'care', 'none', 0, 1, FALSE, -5, -5, 10, 100),
 (9, 6, 'Bunch of 5 Carrots', 'food', 'none', 20, 5, FALSE, 0, 0, 0, 0),
 (10, 6, '10 Carrots', 'food', 'none', 50, 10, FALSE, 0, 0, 0, 0),
 (11, 7, 'Box of Pizza (8 Slices)', 'food', 'none', 40, 8, FALSE, 0, 0, 0, 0),
 (12, 7, '10 Slices of Pizza', 'food', 'none', 40, 10, FALSE, 0, 0, 0, 0),
 (13, 7, '3 Boxes of Pizza (24 Slices)', 'food', 'none', 120, 24, FALSE, 0, 0, 0, 0),
 (14, 14, 'Red Puffle''s Cannon', 'play', 'superplay', 200, 1, TRUE, -5, -5, 40, -10),
 (15, 15, 'Blue Puffle''s Bouncy Ball', 'play', 'superplay', 200, 1, TRUE, -5, -5, 40, -10),
 (16, 16, 'Green Puffle''s Propellor Cap', 'play', 'superplay', 200, 1, TRUE, -5, -5, 40, -10),
 (17, 17, 'Purple Puffle''s Disco Ball', 'play', 'superplay', 200, 1, TRUE, -5, -5, 40, -10),
 (18, 18, 'Yellow Puffle''s Director''s Chair & Camera', 'play', 'superplay', 200, 1, TRUE, -5, -5, 40, -10),
 (19, 19, 'Black Puffle''s Hot Sauce', 'play', 'superplay', 200, 1, TRUE, -5, -5, 40, -10),
 (20, 20, 'White Puffle''s Skate', 'play', 'superplay', 200, 1, TRUE, -5, -5, 40, -10),
 (21, 21, 'Pink Puffle''s Trampoline', 'play', 'superplay', 200, 1, TRUE, -5, -5, 40, -10),
 (22, 22, 'Brown Puffle''s Rocket', 'play', 'superplay', 200, 1, TRUE, -5, -5, 40, -10),
 (23, 23, 'Orange Puffle''s Wagon', 'play', 'superplay', 200, 1, TRUE, -5, -5, 40, -10),
 (24, 5, 'Pack of Bubble Gum (10 pieces)', 'food', 'bubblegum', 20, 10, FALSE, 0, 0, 0, 0),
 (25, 4, '10 Cookies', 'food', 'none', 30, 10, FALSE, 0, 0, 0, 0),
 (26, 3, '10 Puffle O''s', 'food', 'none', 50, 10, FALSE, 0, 0, 0, 0),
 (27, 27, 'Ball', 'play', 'play', 0, 1, FALSE, -2, -2, 30, -5),
 (28, 28, 'Skipping Rope', 'play', 'play', 0, 1, FALSE, -2, -2, 30, -5),
 (29, 29, 'Bowling Pins', 'play', 'play', 0, 1, FALSE, -2, -2, 30, -5),
 (30, 30, 'Unicycle', 'play', 'play', 0, 1, FALSE, -2, -2, 30, -5),
 (31, 31, 'Skateboard', 'play', 'play', 0, 1, FALSE, -2, -2, 30, -5),
 (32, 32, 'Paint Brush', 'play', 'play', 0, 1, FALSE, -2, -2, 30, -5),
 (33, 33, 'Snowflake', 'play', 'play', 0, 1, FALSE, -2, -2, 30, -5),
 (34, 34, 'Hoolahoop', 'play', 'play', 0, 1, FALSE, -2, -2, 30, -5),
 (35, 35, 'Bubble Wand', 'play', 'play', 0, 1, FALSE, -2, -2, 30, -5),
 (36, 36, 'Plasma Ball', 'play', 'play', 0, 1, FALSE, -2, -2, 30, -5),
 (37, 37, 'Sleep', 'sleep', 'sleep', 0, 1, FALSE, -5, 80, -10, 0),
 (38, 38, 'Game', 'game', 'play', 0, 1, FALSE, -10, -10, 70, -15),
 (39, 39, 'Franken Hat', 'head', 'none', 200, 1, FALSE, 0, 0, 0, 0),
 (46, 46, 'Pharaoh Hat', 'head', 'none', 200, 1, FALSE, 0, 0, 0, 0),
 (47, 47, 'Sombrero', 'head', 'none', 200, 1, TRUE, 0, 0, 0, 0),
 (48, 48, 'Pirate Hat', 'head', 'none', 200, 1, TRUE, 0, 0, 0, 0),
 (49, 49, 'Princess Cap', 'head', 'none', 200, 1, TRUE, 0, 0, 0, 0),
 (50, 50, 'Night Cap', 'head', 'none', 200, 1, FALSE, 0, 0, 0, 0),
 (51, 51, 'Propeller Cap', 'head', 'none', 200, 1, TRUE, 0, 0, 0, 0),
 (52, 52, 'Tiara', 'head', 'none', 200, 1, TRUE, 0, 0, 0, 0),
 (53, 53, 'Headphones', 'head', 'none', 200, 1, TRUE, 0, 0, 0, 0),
 (54, 54, 'Gear Hat', 'head', 'none', 200, 1, TRUE, 0, 0, 0, 0),
 (55, 55, 'Jester Hat', 'head', 'none', 200, 1, TRUE, 0, 0, 0, 0),
 (56, 56, 'Crown', 'head', 'none', 200, 1, TRUE, 0, 0, 0, 0),
 (58, 58, 'Snowflake Helmet', 'head', 'none', 200, 1, FALSE, 0, 0, 0, 0),
 (59, 59, 'Gold Viking Helmet', 'head', 'none', 200, 1, FALSE, 0, 0, 0, 0),
 (60, 60, 'Top Hat', 'head', 'none', 200, 1, FALSE, 0, 0, 0, 0),
 (61, 61, 'Heavy Metal Hair', 'head', 'none', 200, 1, FALSE, 0, 0, 0, 0),
 (62, 62, 'Jolly Roger Bandanna', 'head', 'none', 200, 1, FALSE, 0, 0, 0, 0),
 (63, 63, 'The Big Bang', 'head', 'none', 200, 1, FALSE, 0, 0, 0, 0),
 (64, 64, 'The Overgrown', 'head', 'none', 200, 1, FALSE, 0, 0, 0, 0),
 (65, 65, 'The Scene-Stealer', 'head', 'none', 200, 1, FALSE, 0, 0, 0, 0),
 (66, 66, 'The Rad Scientist', 'head', 'none', 200, 1, FALSE, 0, 0, 0, 0),
 (67, 67, 'Disco Dome', 'head', 'none', 200, 1, FALSE, 0, 0, 0, 0),
 (68, 68, 'Candy Cane Cap', 'head', 'none', 200, 1, TRUE, 0, 0, 0, 0),
 (69, 69, 'Turquoise Toque', 'head', 'none', 200, 1, FALSE, 0, 0, 0, 0),
 (70, 70, 'Mini Polka Dot Puffle Hat', 'head', 'none', 0, 1, FALSE, 0, 0, 0, 0),
 (71, 71, 'Blue Earmuffs', 'head', 'none', 200, 1, TRUE, 0, 0, 0, 0),
 (72, 72, 'Pink Bow', 'head', 'none', 200, 1, TRUE, 0, 0, 0, 0),
 (73, 73, 'Mini Squid Lid', 'head', 'none', 200, 1, TRUE, 0, 0, 0, 0),
 (74, 77, '10 Slices of Cake', 'food', 'none', 40, 10, FALSE, 0, 0, 0, 0),
 (75, 78, '10 Wedges of Cheese', 'food', 'none', 40, 10, FALSE, 0, 0, 0, 0),
 (76, 79, '10 Apples', 'food', 'none', 50, 10, FALSE, 0, 0, 0, 0),
 (77, 77, 'Slice of Cake', 'food', 'none', 8, 1, FALSE, 30, -9, -20, -1),
 (78, 78, 'Wedge of Cheese', 'food', 'none', 8, 1, FALSE, 30, -5, -8, -1),
 (79, 79, 'Apple', 'food', 'none', 4, 1, FALSE, 50, -5, -5, 0),
 (80, 80, 'Jingle Jangle Hat', 'head', 'none', 0, 1, FALSE, 0, 0, 0, 0),
 (81, 81, 'Knight Helmet', 'head', 'none', 0, 1, FALSE, 0, 0, 0, 0),
 (82, 82, 'Sherwood Hat', 'head', 'none', 0, 1, FALSE, 0, 0, 0, 0),
 (83, 83, 'Wizard Hat', 'head', 'none', 0, 1, FALSE, 0, 0, 0, 0),
 (84, 84, 'Princess Braid', 'head', 'none', 0, 1, FALSE, 0, 0, 0, 0),
 (85, 85, 'Surf Swoop', 'head', 'none', 0, 1, FALSE, 0, 0, 0, 0),
 (86, 86, 'Sundae Swirl', 'head', 'none', 0, 1, FALSE, 0, 0, 0, 0),
 (87, 87, 'All-Star Curls', 'head', 'none', 0, 1, FALSE, 0, 0, 0, 0),
 (88, 88, 'Shock of Hair', 'head', 'none', 0, 1, FALSE, 0, 0, 0, 0),
 (89, 89, 'Chocolate Beret', 'head', 'none', 0, 1, FALSE, 0, 0, 0, 0),
 (90, 90, 'Touchdown Dome', 'head', 'none', 0, 1, FALSE, 0, 0, 0, 0),
 (91, 91, 'Rainbow Fro', 'head', 'none', 0, 1, FALSE, 0, 0, 0, 0),
 (92, 92, 'Green 180 Cap', 'head', 'none', 0, 1, FALSE, 0, 0, 0, 0),
 (93, 93, 'Blast Off Cap', 'head', 'none', 0, 1, FALSE, 0, 0, 0, 0),
 (94, 94, 'Swing Batta Batta Hat', 'head', 'none', 0, 1, FALSE, 0, 0, 0, 0),
 (95, 95, 'The High Flyer', 'head', 'none', 0, 1, FALSE, 0, 0, 0, 0),
 (96, 96, 'The Workout Curl', 'head', 'none', 0, 1, FALSE, 0, 0, 0, 0),
 (97, 97, 'Ninja Mask', 'head', 'none', 0, 1, FALSE, 0, 0, 0, 0),
 (98, 98, 'The Master Hat', 'head', 'none', 0, 1, FALSE, 0, 0, 0, 0),
 (99, 99, 'Polka Dot Pirate', 'head', 'none', 4000, 1, TRUE, 0, 0, 0, 0),
 (100, 100, 'First Mate Bandana', 'head', 'none', 0, 1, FALSE, 0, 0, 0, 0),
 (101, 101, 'The Stone-Age Roller', 'head', 'none', 200, 1, TRUE, 0, 0, 0, 0),
 (102, 102, 'The Warrior', 'head', 'none', 200, 1, TRUE, 0, 0, 0, 0),
 (103, 103, 'Super Cloud Toy', 'play', 'play', 0, 1, TRUE, -2, -2, 30, -5),
 (104, 104, 'R2-D2 Helmet', 'head', 'none', 200, 1, TRUE, 0, 0, 0, 0),
 (105, 105, 'Pretzel', 'food', 'none', 65000, 1, FALSE, 30, -3, -5, -3),
 (106, 106, 'Stinky Cheese', 'food', 'none', 65000, 1, FALSE, 20, -2, -10, -5),
 (107, 107, 'Watermelon', 'food', 'none', 65000, 1, FALSE, 50, -1, -1, -10),
 (108, 108, 'Sock', 'food', 'none', 65000, 1, FALSE, 40, -1, -10, -10),
 (109, 109, 'Tacos', 'food', 'none', 65000, 1, FALSE, 40, -3, -5, -5),
 (110, 110, 'Hummus and Pita', 'food', 'none', 65000, 1, FALSE, 20, -5, -1, -1),
 (111, 111, 'Yogurt Parfait', 'food', 'none', 65000, 1, FALSE, 10, -3, -2, -2),
 (112, 112, 'Fish Burger', 'food', 'none', 65000, 1, FALSE, 50, -5, -5, -10),
 (113, 113, 'Ice Cream Sandwich', 'food', 'none', 65000, 1, FALSE, 20, -3, -5, -1),
 (114, 114, 'Popcorn', 'food', 'none', 65000, 1, FALSE, 10, -1, -3, -1),
 (115, 115, 'Rainbow Lollipop', 'food', 'none', 65000, 1, FALSE, 20, -10, 5, -1),
 (117, 117, 'Deluxe', 'food', 'none', 10, 1, TRUE, 100, 0, 0, 0),
 (119, 119, 'Candy Deluxe', 'food', 'none', 10, 1, TRUE, 100, 0, 0, 0),
 (121, 121, 'Pepperoni', 'food', 'none', 5, 1, FALSE, 30, -8, -10, -1),
 (122, 122, 'R - Dig Helmet', 'head', 'none', 0, 1, FALSE, 0, 0, 0, 0),
 (123, 123, 'Pumpkin Lid', 'head', 'none', 0, 1, TRUE, 0, 0, 0, 0),
 (125, 125, 'Gold Jackhammer', 'play', 'play', 0, 1, FALSE, -2, -2, 30, -5),
 (126, 126, 'Gold O''berry', 'food', 'none', 10, 1, TRUE, 0, 0, 0, 0),
 (127, 127, 'Golden Headphones', 'head', 'none', 0, 1, TRUE, 0, 0, 0, 0),
 (128, 128, 'Chocolate Coin', 'food', 'none', 65000, 1, FALSE, 30, -9, -20, -1),
 (129, 129, 'Eat Station', 'care', 'none', 0, 1, FALSE, 15, 0, 0, 0),
 (130, 130, 'Play Station', 'care', 'none', 0, 1, FALSE, 0, 0, 20, 0),
 (131, 131, 'Clean Station', 'care', 'none', 0, 1, FALSE, 0, 0, 0, 50),
 (132, 132, 'Rest Station', 'care', 'none', 0, 1, FALSE, 0, 50, 0, 0),
 (133, 133, 'Gold Puffle''s Dump Truck', 'play', 'superplay', 200, 1, TRUE, -5, -5, 40, -10),
 (134, 134, 'Rainbow Puffle''s Royal Throne', 'play', 'superplay', 200, 1, TRUE, -5, -5, 40, -10),
 (135, 146, '10 Pickles', 'food', 'none', 40, 10, FALSE, 0, 0, 0, 0),
 (136, 145, '10 Pieces of Broccoli', 'food', 'none', 50, 10, FALSE, 0, 0, 0, 0),
 (137, 144, '10 Dragon Fruits', 'food', 'none', 50, 10, FALSE, 0, 0, 0, 0),
 (138, 143, '10 Cayenne Peppers', 'food', 'none', 40, 10, FALSE, 0, 0, 0, 0),
 (139, 139, 'Mining Helmet', 'head', 'none', 200, 1, TRUE, 0, 0, 0, 0),
 (140, 140, 'Safari Hat', 'head', 'none', 200, 1, TRUE, 0, 0, 0, 0),
 (141, 141, 'Adventurer''s Fedora', 'head', 'none', 200, 1, TRUE, 0, 0, 0, 0),
 (142, 142, 'Polka Puffle Hat', 'head', 'none', 0, 1, FALSE, 0, 0, 0, 0),
 (143, 143, 'Cayenne Pepper', 'food', 'none', 4, 1, FALSE, 20, -5, -5, -1),
 (144, 144, 'Dragon Fruit', 'food', 'none', 5, 1, FALSE, 50, -5, -5, 0),
 (145, 145, 'Broccoli', 'food', 'none', 5, 1, FALSE, 50, -5, -5, 0),
 (146, 146, 'Pickle', 'food', 'none', 4, 1, FALSE, 30, -5, -8, -1),
 (147, 147, 'Sheriff Stetson', 'head', 'none', 4000, 1, TRUE, 0, 0, 0, 0),
 (148, 148, 'Quasar Helmet', 'head', 'none', 4000, 1, TRUE, 0, 0, 0, 0),
 (149, 149, 'Aviator Hat', 'head', 'none', 200, 1, TRUE, 0, 0, 0, 0),
 (150, 150, '4 Quart Pot', 'head', 'none', 200, 1, TRUE, 0, 0, 0, 0),
 (151, 151, 'Flap Hat', 'head', 'none', 200, 1, TRUE, 0, 0, 0, 0),
 (152, 152, 'Chef''s Hat', 'head', 'none', 200, 1, TRUE, 0, 0, 0, 0),
 (153, 153, 'Shrimp', 'food', 'none', 4, 1, FALSE, 20, -5, -5, -1),
 (154, 153, '10 Shrimp', 'food', 'none', 40, 10, FALSE, 0, 0, 0, 0),
 (155, 155, 'Biscuit Treats', 'food', 'none', 4, 1, FALSE, 20, -5, 0, -1),
 (156, 155, '10 Biscuit Treats', 'food', 'none', 40, 10, FALSE, 0, 0, 0, 0),
 (157, 157, 'Rainbow Matrix Hat', 'head', 'none', 0, 40, TRUE, 0, 0, 0, 0),
 (158, 158, 'Drum Roll', 'head', 'none', 0, 1, TRUE, 0, 0, 0, 0),
 (159, 159, 'Swashbuckler Hat', 'head', 'none', 0, 1, TRUE, 0, 0, 0, 0);

INSERT INTO puffle (id, parent_id, name, cost, member, favourite_food, favourite_toy, runaway_postcard) VALUES
 (0, NULL, 'Blue', 400, FALSE, 101, 27, 100),
 (1, NULL, 'Pink', 400, TRUE, 107, 28, 101),
 (2, NULL, 'Black', 400, TRUE, 112, 31, 102),
 (3, NULL, 'Green', 400, TRUE, 109, 30, 103),
 (4, NULL, 'Purple', 400, TRUE, 110, 35, 104),
 (5, NULL, 'Red', 400, FALSE, 106, 29, 105),
 (6, NULL, 'Yellow', 400, TRUE, 114, 32, 106),
 (7, NULL, 'White', 400, TRUE, 111, 33, 169),
 (8, NULL, 'Orange', 400, TRUE, 108, 34, 109),
 (9, NULL, 'Brown', 400, TRUE, 113, 36, NULL),
 (10, NULL, 'Rainbow', 0, TRUE, 115, 103, NULL),
 (11, NULL, 'Gold', 0, TRUE, 128, 125, NULL),
 (1000, 2, 'Black T-Rex', 0, TRUE, 112, NULL, NULL),
 (1001, 4, 'Purple T-Rex', 0, TRUE, 110, NULL, NULL),
 (1002, 5, 'Red Triceratops', 0, TRUE, 106, NULL, NULL),
 (1003, 0, 'Blue Triceratops', 0, TRUE, 101, NULL, NULL),
 (1004, 6, 'Yellow Stegasaurus', 0, TRUE, 114, NULL, NULL),
 (1005, 1, 'Pink Stegasaurus', 0, TRUE, 107, NULL, NULL),
 (1006, 0, 'Blue Dog', 800, TRUE, 101, NULL, NULL),
 (1007, 8, 'Orange Cat', 800, TRUE, 108, NULL, NULL),
 (1008, 3, 'Green Raccoon', 800, TRUE, 109, NULL, NULL),
 (1009, 8, 'Orange Raccoon', 800, TRUE, 108, NULL, NULL),
 (1010, 1, 'Pink Raccoon', 800, TRUE, 107, NULL, NULL),
 (1011, 0, 'Blue Raccoon', 800, TRUE, 101, NULL, NULL),
 (1012, 3, 'Green Rabbit', 800, TRUE, 109, NULL, NULL),
 (1013, 1, 'Pink Rabbit', 800, TRUE, 107, NULL, NULL),
 (1014, 7, 'White Rabbit', 800, TRUE, 111, NULL, NULL),
 (1015, 5, 'Red Rabbit', 800, TRUE, 106, NULL, NULL),
 (1016, 0, 'Blue Deer', 800, TRUE, 101, NULL, NULL),
 (1017, 2, 'Black Deer', 800, TRUE, 112, NULL, NULL),
 (1018, 5, 'Red Deer', 800, TRUE, 106, NULL, NULL),
 (1019, 4, 'Purple Deer', 800, TRUE, 110, NULL, NULL),
 (1020, 6, 'Yellow Unicorn', 800, TRUE, 114, NULL, NULL),
 (1021, 7, 'Snowman', 0, TRUE, 111, NULL, NULL),
 (1022, 4, 'Ghost', 0, TRUE, 110, NULL, NULL),
 (1023, 0, 'Crystal', 0, TRUE, 101, NULL, NULL),
 (1024, 3, 'Green Alien', 0, FALSE, 109, NULL, NULL),
 (1025, 8, 'Orange Alien', 0, TRUE, 108, NULL, NULL),
 (1026, 6, 'Yellow Alien', 0, TRUE, 114, NULL, NULL),
 (1027, 4, 'Purple Alien', 0, TRUE, 110, NULL, NULL);
 
INSERT INTO puffle_treasure_puffle_item (puffle_id, puffle_item_id) VALUES
 (0, 115), (1, 115), (2, 115), (3, 115), (4, 115), (5, 115), (6, 115), (7, 115), (8, 115), (9, 115), (10, 115), (11, 115), (1000, 115), (1001, 115), (1002, 115), (1003, 115), (1004, 115), (1005, 115), (1006, 115), (1007, 115), (1008, 115), (1009, 115), (1010, 115), (1011, 115), (1012, 115), (1013, 115), (1014, 115), (1015, 115), (1016, 115), (1017, 115), (1018, 115), (1019, 115), (1020, 115), (1021, 115), (1022, 115), (1023, 115), (1024, 115), (1025, 115), (1026, 115), (1027, 115),
 (0, 114), (1, 114), (2, 114), (3, 114), (4, 114), (5, 114), (6, 114), (7, 114), (8, 114), (9, 114), (10, 114), (11, 114), (1000, 114), (1001, 114), (1002, 114), (1003, 114), (1004, 114), (1005, 114), (1006, 114), (1007, 114), (1008, 114), (1009, 114), (1010, 114), (1011, 114), (1012, 114), (1013, 114), (1014, 114), (1015, 114), (1016, 114), (1017, 114), (1018, 114), (1019, 114), (1020, 114), (1021, 114), (1022, 114), (1023, 114), (1024, 114), (1025, 114), (1026, 114), (1027, 114),
 (0, 109), (1, 109), (2, 109), (3, 109), (4, 109), (5, 109), (6, 109), (7, 109), (8, 109), (9, 109), (10, 109), (11, 109), (1000, 109), (1001, 109), (1002, 109), (1003, 109), (1004, 109), (1005, 109), (1006, 109), (1007, 109), (1008, 109), (1009, 109), (1010, 109), (1011, 109), (1012, 109), (1013, 109), (1014, 109), (1015, 109), (1016, 109), (1017, 109), (1018, 109), (1019, 109), (1020, 109), (1021, 109), (1022, 109), (1023, 109), (1024, 109), (1025, 109), (1026, 109), (1027, 109),
 (0, 112), (1, 112), (2, 112), (3, 112), (4, 112), (5, 112), (6, 112), (7, 112), (8, 112), (9, 112), (10, 112), (11, 112), (1000, 112), (1001, 112), (1002, 112), (1003, 112), (1004, 112), (1005, 112), (1006, 112), (1007, 112), (1008, 112), (1009, 112), (1010, 112), (1011, 112), (1012, 112), (1013, 112), (1014, 112), (1015, 112), (1016, 112), (1017, 112), (1018, 112), (1019, 112), (1020, 112), (1021, 112), (1022, 112), (1023, 112), (1024, 112), (1025, 112), (1026, 112), (1027, 112),
 (0, 110), (1, 110), (2, 110), (3, 110), (4, 110), (5, 110), (6, 110), (7, 110), (8, 110), (9, 110), (10, 110), (11, 110), (1000, 110), (1001, 110), (1002, 110), (1003, 110), (1004, 110), (1005, 110), (1006, 110), (1007, 110), (1008, 110), (1009, 110), (1010, 110), (1011, 110), (1012, 110), (1013, 110), (1014, 110), (1015, 110), (1016, 110), (1017, 110), (1018, 110), (1019, 110), (1020, 110), (1021, 110), (1022, 110), (1023, 110), (1024, 110), (1025, 110), (1026, 110), (1027, 110),
 (0, 105), (1, 105), (2, 105), (3, 105), (4, 105), (5, 105), (6, 105), (7, 105), (8, 105), (9, 105), (10, 105), (11, 105), (1000, 105), (1001, 105), (1002, 105), (1003, 105), (1004, 105), (1005, 105), (1006, 105), (1007, 105), (1008, 105), (1009, 105), (1010, 105), (1011, 105), (1012, 105), (1013, 105), (1014, 105), (1015, 105), (1016, 105), (1017, 105), (1018, 105), (1019, 105), (1020, 105), (1021, 105), (1022, 105), (1023, 105), (1024, 105), (1025, 105), (1026, 105), (1027, 105),
 (0, 113), (1, 113), (2, 113), (3, 113), (4, 113), (5, 113), (6, 113), (7, 113), (8, 113), (9, 113), (10, 113), (11, 113), (1000, 113), (1001, 113), (1002, 113), (1003, 113), (1004, 113), (1005, 113), (1006, 113), (1007, 113), (1008, 113), (1009, 113), (1010, 113), (1011, 113), (1012, 113), (1013, 113), (1014, 113), (1015, 113), (1016, 113), (1017, 113), (1018, 113), (1019, 113), (1020, 113), (1021, 113), (1022, 113), (1023, 113), (1024, 113), (1025, 113), (1026, 113), (1027, 113),
 (0, 106), (1, 106), (2, 106), (3, 106), (4, 106), (5, 106), (6, 106), (7, 106), (8, 106), (9, 106), (10, 106), (11, 106), (1000, 106), (1001, 106), (1002, 106), (1003, 106), (1004, 106), (1005, 106), (1006, 106), (1007, 106), (1008, 106), (1009, 106), (1010, 106), (1011, 106), (1012, 106), (1013, 106), (1014, 106), (1015, 106), (1016, 106), (1017, 106), (1018, 106), (1019, 106), (1020, 106), (1021, 106), (1022, 106), (1023, 106), (1024, 106), (1025, 106), (1026, 106), (1027, 106),
 (0, 108), (1, 108), (2, 108), (3, 108), (4, 108), (5, 108), (6, 108), (7, 108), (8, 108), (9, 108), (10, 108), (11, 108), (1000, 108), (1001, 108), (1002, 108), (1003, 108), (1004, 108), (1005, 108), (1006, 108), (1007, 108), (1008, 108), (1009, 108), (1010, 108), (1011, 108), (1012, 108), (1013, 108), (1014, 108), (1015, 108), (1016, 108), (1017, 108), (1018, 108), (1019, 108), (1020, 108), (1021, 108), (1022, 108), (1023, 108), (1024, 108), (1025, 108), (1026, 108), (1027, 108),
 (0, 107), (1, 107), (2, 107), (3, 107), (4, 107), (5, 107), (6, 107), (7, 107), (8, 107), (9, 107), (10, 107), (11, 107), (1000, 107), (1001, 107), (1002, 107), (1003, 107), (1004, 107), (1005, 107), (1006, 107), (1007, 107), (1008, 107), (1009, 107), (1010, 107), (1011, 107), (1012, 107), (1013, 107), (1014, 107), (1015, 107), (1016, 107), (1017, 107), (1018, 107), (1019, 107), (1020, 107), (1021, 107), (1022, 107), (1023, 107), (1024, 107), (1025, 107), (1026, 107), (1027, 107),
 (0, 111), (1, 111), (2, 111), (3, 111), (4, 111), (5, 111), (6, 111), (7, 111), (8, 111), (9, 111), (10, 111), (11, 111), (1000, 111), (1001, 111), (1002, 111), (1003, 111), (1004, 111), (1005, 111), (1006, 111), (1007, 111), (1008, 111), (1009, 111), (1010, 111), (1011, 111), (1012, 111), (1013, 111), (1014, 111), (1015, 111), (1016, 111), (1017, 111), (1018, 111), (1019, 111), (1020, 111), (1021, 111), (1022, 111), (1023, 111), (1024, 111), (1025, 111), (1026, 111), (1027, 111),
 (0, 128), (1, 128), (2, 128), (3, 128), (4, 128), (5, 128), (6, 128), (7, 128), (8, 128), (9, 128), (10, 128), (11, 128), (1000, 128), (1001, 128), (1002, 128), (1003, 128), (1004, 128), (1005, 128), (1006, 128), (1007, 128), (1008, 128), (1009, 128), (1010, 128), (1011, 128), (1012, 128), (1013, 128), (1014, 128), (1015, 128), (1016, 128), (1017, 128), (1018, 128), (1019, 128), (1020, 128), (1021, 128), (1022, 128), (1023, 128), (1024, 128), (1025, 128), (1026, 128), (1027, 128);

INSERT INTO puffle_treasure_furniture (puffle_id, furniture_id) VALUES
 (0, 305), (1, 305), (2, 305), (3, 305), (4, 305), (5, 305), (6, 305), (7, 305), (8, 305), (9, 305), (10, 305), (11, 305), (1000, 305), (1001, 305), (1002, 305), (1003, 305), (1004, 305), (1005, 305), (1006, 305), (1007, 305), (1008, 305), (1009, 305), (1010, 305), (1011, 305), (1012, 305), (1013, 305), (1014, 305), (1015, 305), (1016, 305), (1017, 305), (1018, 305), (1019, 305), (1020, 305), (1021, 305), (1022, 305), (1023, 305), (1024, 305), (1025, 305), (1026, 305), (1027, 305),
 (0, 313), (1, 313), (2, 313), (3, 313), (4, 313), (5, 313), (6, 313), (7, 313), (8, 313), (9, 313), (10, 313), (11, 313), (1000, 313), (1001, 313), (1002, 313), (1003, 313), (1004, 313), (1005, 313), (1006, 313), (1007, 313), (1008, 313), (1009, 313), (1010, 313), (1011, 313), (1012, 313), (1013, 313), (1014, 313), (1015, 313), (1016, 313), (1017, 313), (1018, 313), (1019, 313), (1020, 313), (1021, 313), (1022, 313), (1023, 313), (1024, 313), (1025, 313), (1026, 313), (1027, 313),
 (0, 504), (1, 504), (2, 504), (3, 504), (4, 504), (5, 504), (6, 504), (7, 504), (8, 504), (9, 504), (10, 504), (11, 504), (1000, 504), (1001, 504), (1002, 504), (1003, 504), (1004, 504), (1005, 504), (1006, 504), (1007, 504), (1008, 504), (1009, 504), (1010, 504), (1011, 504), (1012, 504), (1013, 504), (1014, 504), (1015, 504), (1016, 504), (1017, 504), (1018, 504), (1019, 504), (1020, 504), (1021, 504), (1022, 504), (1023, 504), (1024, 504), (1025, 504), (1026, 504), (1027, 504),
 (0, 506), (1, 506), (2, 506), (3, 506), (4, 506), (5, 506), (6, 506), (7, 506), (8, 506), (9, 506), (10, 506), (11, 506), (1000, 506), (1001, 506), (1002, 506), (1003, 506), (1004, 506), (1005, 506), (1006, 506), (1007, 506), (1008, 506), (1009, 506), (1010, 506), (1011, 506), (1012, 506), (1013, 506), (1014, 506), (1015, 506), (1016, 506), (1017, 506), (1018, 506), (1019, 506), (1020, 506), (1021, 506), (1022, 506), (1023, 506), (1024, 506), (1025, 506), (1026, 506), (1027, 506),
 (0, 500), (1, 500), (2, 500), (3, 500), (4, 500), (5, 500), (6, 500), (7, 500), (8, 500), (9, 500), (10, 500), (11, 500), (1000, 500), (1001, 500), (1002, 500), (1003, 500), (1004, 500), (1005, 500), (1006, 500), (1007, 500), (1008, 500), (1009, 500), (1010, 500), (1011, 500), (1012, 500), (1013, 500), (1014, 500), (1015, 500), (1016, 500), (1017, 500), (1018, 500), (1019, 500), (1020, 500), (1021, 500), (1022, 500), (1023, 500), (1024, 500), (1025, 500), (1026, 500), (1027, 500),
 (0, 503), (1, 503), (2, 503), (3, 503), (4, 503), (5, 503), (6, 503), (7, 503), (8, 503), (9, 503), (10, 503), (11, 503), (1000, 503), (1001, 503), (1002, 503), (1003, 503), (1004, 503), (1005, 503), (1006, 503), (1007, 503), (1008, 503), (1009, 503), (1010, 503), (1011, 503), (1012, 503), (1013, 503), (1014, 503), (1015, 503), (1016, 503), (1017, 503), (1018, 503), (1019, 503), (1020, 503), (1021, 503), (1022, 503), (1023, 503), (1024, 503), (1025, 503), (1026, 503), (1027, 503),
 (0, 501), (1, 501), (2, 501), (3, 501), (4, 501), (5, 501), (6, 501), (7, 501), (8, 501), (9, 501), (10, 501), (11, 501), (1000, 501), (1001, 501), (1002, 501), (1003, 501), (1004, 501), (1005, 501), (1006, 501), (1007, 501), (1008, 501), (1009, 501), (1010, 501), (1011, 501), (1012, 501), (1013, 501), (1014, 501), (1015, 501), (1016, 501), (1017, 501), (1018, 501), (1019, 501), (1020, 501), (1021, 501), (1022, 501), (1023, 501), (1024, 501), (1025, 501), (1026, 501), (1027, 501),
 (0, 507), (1, 507), (2, 507), (3, 507), (4, 507), (5, 507), (6, 507), (7, 507), (8, 507), (9, 507), (10, 507), (11, 507), (1000, 507), (1001, 507), (1002, 507), (1003, 507), (1004, 507), (1005, 507), (1006, 507), (1007, 507), (1008, 507), (1009, 507), (1010, 507), (1011, 507), (1012, 507), (1013, 507), (1014, 507), (1015, 507), (1016, 507), (1017, 507), (1018, 507), (1019, 507), (1020, 507), (1021, 507), (1022, 507), (1023, 507), (1024, 507), (1025, 507), (1026, 507), (1027, 507),
 (0, 505), (1, 505), (2, 505), (3, 505), (4, 505), (5, 505), (6, 505), (7, 505), (8, 505), (9, 505), (10, 505), (11, 505), (1000, 505), (1001, 505), (1002, 505), (1003, 505), (1004, 505), (1005, 505), (1006, 505), (1007, 505), (1008, 505), (1009, 505), (1010, 505), (1011, 505), (1012, 505), (1013, 505), (1014, 505), (1015, 505), (1016, 505), (1017, 505), (1018, 505), (1019, 505), (1020, 505), (1021, 505), (1022, 505), (1023, 505), (1024, 505), (1025, 505), (1026, 505), (1027, 505),
 (0, 502), (1, 502), (2, 502), (3, 502), (4, 502), (5, 502), (6, 502), (7, 502), (8, 502), (9, 502), (10, 502), (11, 502), (1000, 502), (1001, 502), (1002, 502), (1003, 502), (1004, 502), (1005, 502), (1006, 502), (1007, 502), (1008, 502), (1009, 502), (1010, 502), (1011, 502), (1012, 502), (1013, 502), (1014, 502), (1015, 502), (1016, 502), (1017, 502), (1018, 502), (1019, 502), (1020, 502), (1021, 502), (1022, 502), (1023, 502), (1024, 502), (1025, 502), (1026, 502), (1027, 502),
 (0, 616), (1, 616), (2, 616), (3, 616), (4, 616), (5, 616), (6, 616), (7, 616), (8, 616), (9, 616), (10, 616), (11, 616), (1000, 616), (1001, 616), (1002, 616), (1003, 616), (1004, 616), (1005, 616), (1006, 616), (1007, 616), (1008, 616), (1009, 616), (1010, 616), (1011, 616), (1012, 616), (1013, 616), (1014, 616), (1015, 616), (1016, 616), (1017, 616), (1018, 616), (1019, 616), (1020, 616), (1021, 616), (1022, 616), (1023, 616), (1024, 616), (1025, 616), (1026, 616), (1027, 616),
 (0, 542), (1, 542), (2, 542), (3, 542), (4, 542), (5, 542), (6, 542), (7, 542), (8, 542), (9, 542), (10, 542), (11, 542), (1000, 542), (1001, 542), (1002, 542), (1003, 542), (1004, 542), (1005, 542), (1006, 542), (1007, 542), (1008, 542), (1009, 542), (1010, 542), (1011, 542), (1012, 542), (1013, 542), (1014, 542), (1015, 542), (1016, 542), (1017, 542), (1018, 542), (1019, 542), (1020, 542), (1021, 542), (1022, 542), (1023, 542), (1024, 542), (1025, 542), (1026, 542), (1027, 542),
 (0, 340), (1, 340), (2, 340), (3, 340), (4, 340), (5, 340), (6, 340), (7, 340), (8, 340), (9, 340), (10, 340), (11, 340), (1000, 340), (1001, 340), (1002, 340), (1003, 340), (1004, 340), (1005, 340), (1006, 340), (1007, 340), (1008, 340), (1009, 340), (1010, 340), (1011, 340), (1012, 340), (1013, 340), (1014, 340), (1015, 340), (1016, 340), (1017, 340), (1018, 340), (1019, 340), (1020, 340), (1021, 340), (1022, 340), (1023, 340), (1024, 340), (1025, 340), (1026, 340), (1027, 340),
 (0, 150), (1, 150), (2, 150), (3, 150), (4, 150), (5, 150), (6, 150), (7, 150), (8, 150), (9, 150), (10, 150), (11, 150), (1000, 150), (1001, 150), (1002, 150), (1003, 150), (1004, 150), (1005, 150), (1006, 150), (1007, 150), (1008, 150), (1009, 150), (1010, 150), (1011, 150), (1012, 150), (1013, 150), (1014, 150), (1015, 150), (1016, 150), (1017, 150), (1018, 150), (1019, 150), (1020, 150), (1021, 150), (1022, 150), (1023, 150), (1024, 150), (1025, 150), (1026, 150), (1027, 150),
 (0, 149), (1, 149), (2, 149), (3, 149), (4, 149), (5, 149), (6, 149), (7, 149), (8, 149), (9, 149), (10, 149), (11, 149), (1000, 149), (1001, 149), (1002, 149), (1003, 149), (1004, 149), (1005, 149), (1006, 149), (1007, 149), (1008, 149), (1009, 149), (1010, 149), (1011, 149), (1012, 149), (1013, 149), (1014, 149), (1015, 149), (1016, 149), (1017, 149), (1018, 149), (1019, 149), (1020, 149), (1021, 149), (1022, 149), (1023, 149), (1024, 149), (1025, 149), (1026, 149), (1027, 149),
 (0, 369), (1, 369), (2, 369), (3, 369), (4, 369), (5, 369), (6, 369), (7, 369), (8, 369), (9, 369), (10, 369), (11, 369), (1000, 369), (1001, 369), (1002, 369), (1003, 369), (1004, 369), (1005, 369), (1006, 369), (1007, 369), (1008, 369), (1009, 369), (1010, 369), (1011, 369), (1012, 369), (1013, 369), (1014, 369), (1015, 369), (1016, 369), (1017, 369), (1018, 369), (1019, 369), (1020, 369), (1021, 369), (1022, 369), (1023, 369), (1024, 369), (1025, 369), (1026, 369), (1027, 369),
 (0, 370), (1, 370), (2, 370), (3, 370), (4, 370), (5, 370), (6, 370), (7, 370), (8, 370), (9, 370), (10, 370), (11, 370), (1000, 370), (1001, 370), (1002, 370), (1003, 370), (1004, 370), (1005, 370), (1006, 370), (1007, 370), (1008, 370), (1009, 370), (1010, 370), (1011, 370), (1012, 370), (1013, 370), (1014, 370), (1015, 370), (1016, 370), (1017, 370), (1018, 370), (1019, 370), (1020, 370), (1021, 370), (1022, 370), (1023, 370), (1024, 370), (1025, 370), (1026, 370), (1027, 370),
 (0, 300), (1, 300), (2, 300), (3, 300), (4, 300), (5, 300), (6, 300), (7, 300), (8, 300), (9, 300), (10, 300), (11, 300), (1000, 300), (1001, 300), (1002, 300), (1003, 300), (1004, 300), (1005, 300), (1006, 300), (1007, 300), (1008, 300), (1009, 300), (1010, 300), (1011, 300), (1012, 300), (1013, 300), (1014, 300), (1015, 300), (1016, 300), (1017, 300), (1018, 300), (1019, 300), (1020, 300), (1021, 300), (1022, 300), (1023, 300), (1024, 300), (1025, 300), (1026, 300), (1027, 300),
 (11, 2132),
 (11, 2131),
 (11, 2130),
 (11, 2129),
 (1006, 2180),
 (1006, 2182),
 (1006, 2183),
 (1000, 2180), (1001, 2180), (1002, 2180), (1003, 2180), (1004, 2180), (1005, 2180),
 (1000, 2182), (1001, 2182), (1002, 2182), (1003, 2182), (1004, 2182), (1005, 2182),
 (1000, 2183), (1001, 2183), (1002, 2183), (1003, 2183), (1004, 2183), (1005, 2183);

INSERT INTO puffle_treasure_item (puffle_id, item_id) VALUES
 (0, 3028), (1, 3028), (2, 3028), (3, 3028), (4, 3028), (5, 3028), (6, 3028), (7, 3028), (8, 3028), (9, 3028), (10, 3028), (11, 3028), (1000, 3028), (1001, 3028), (1002, 3028), (1003, 3028), (1004, 3028), (1005, 3028), (1006, 3028), (1007, 3028), (1008, 3028), (1009, 3028), (1010, 3028), (1011, 3028), (1012, 3028), (1013, 3028), (1014, 3028), (1015, 3028), (1016, 3028), (1017, 3028), (1018, 3028), (1019, 3028), (1020, 3028), (1021, 3028), (1022, 3028), (1023, 3028), (1024, 3028), (1025, 3028), (1026, 3028), (1027, 3028),
 (0, 232), (1, 232), (2, 232), (3, 232), (4, 232), (5, 232), (6, 232), (7, 232), (8, 232), (9, 232), (10, 232), (11, 232), (1000, 232), (1001, 232), (1002, 232), (1003, 232), (1004, 232), (1005, 232), (1006, 232), (1007, 232), (1008, 232), (1009, 232), (1010, 232), (1011, 232), (1012, 232), (1013, 232), (1014, 232), (1015, 232), (1016, 232), (1017, 232), (1018, 232), (1019, 232), (1020, 232), (1021, 232), (1022, 232), (1023, 232), (1024, 232), (1025, 232), (1026, 232), (1027, 232),
 (0, 412), (1, 412), (2, 412), (3, 412), (4, 412), (5, 412), (6, 412), (7, 412), (8, 412), (9, 412), (10, 412), (11, 412), (1000, 412), (1001, 412), (1002, 412), (1003, 412), (1004, 412), (1005, 412), (1006, 412), (1007, 412), (1008, 412), (1009, 412), (1010, 412), (1011, 412), (1012, 412), (1013, 412), (1014, 412), (1015, 412), (1016, 412), (1017, 412), (1018, 412), (1019, 412), (1020, 412), (1021, 412), (1022, 412), (1023, 412), (1024, 412), (1025, 412), (1026, 412), (1027, 412),
 (0, 112), (1, 112), (2, 112), (3, 112), (4, 112), (5, 112), (6, 112), (7, 112), (8, 112), (9, 112), (10, 112), (11, 112), (1000, 112), (1001, 112), (1002, 112), (1003, 112), (1004, 112), (1005, 112), (1006, 112), (1007, 112), (1008, 112), (1009, 112), (1010, 112), (1011, 112), (1012, 112), (1013, 112), (1014, 112), (1015, 112), (1016, 112), (1017, 112), (1018, 112), (1019, 112), (1020, 112), (1021, 112), (1022, 112), (1023, 112), (1024, 112), (1025, 112), (1026, 112), (1027, 112),
 (0, 184), (1, 184), (2, 184), (3, 184), (4, 184), (5, 184), (6, 184), (7, 184), (8, 184), (9, 184), (10, 184), (11, 184), (1000, 184), (1001, 184), (1002, 184), (1003, 184), (1004, 184), (1005, 184), (1006, 184), (1007, 184), (1008, 184), (1009, 184), (1010, 184), (1011, 184), (1012, 184), (1013, 184), (1014, 184), (1015, 184), (1016, 184), (1017, 184), (1018, 184), (1019, 184), (1020, 184), (1021, 184), (1022, 184), (1023, 184), (1024, 184), (1025, 184), (1026, 184), (1027, 184),
 (0, 1056), (1, 1056), (2, 1056), (3, 1056), (4, 1056), (5, 1056), (6, 1056), (7, 1056), (8, 1056), (9, 1056), (10, 1056), (11, 1056), (1000, 1056), (1001, 1056), (1002, 1056), (1003, 1056), (1004, 1056), (1005, 1056), (1006, 1056), (1007, 1056), (1008, 1056), (1009, 1056), (1010, 1056), (1011, 1056), (1012, 1056), (1013, 1056), (1014, 1056), (1015, 1056), (1016, 1056), (1017, 1056), (1018, 1056), (1019, 1056), (1020, 1056), (1021, 1056), (1022, 1056), (1023, 1056), (1024, 1056), (1025, 1056), (1026, 1056), (1027, 1056),
 (0, 6012), (1, 6012), (2, 6012), (3, 6012), (4, 6012), (5, 6012), (6, 6012), (7, 6012), (8, 6012), (9, 6012), (10, 6012), (11, 6012), (1000, 6012), (1001, 6012), (1002, 6012), (1003, 6012), (1004, 6012), (1005, 6012), (1006, 6012), (1007, 6012), (1008, 6012), (1009, 6012), (1010, 6012), (1011, 6012), (1012, 6012), (1013, 6012), (1014, 6012), (1015, 6012), (1016, 6012), (1017, 6012), (1018, 6012), (1019, 6012), (1020, 6012), (1021, 6012), (1022, 6012), (1023, 6012), (1024, 6012), (1025, 6012), (1026, 6012), (1027, 6012),
 (0, 118), (1, 118), (2, 118), (3, 118), (4, 118), (5, 118), (6, 118), (7, 118), (8, 118), (9, 118), (10, 118), (11, 118), (1000, 118), (1001, 118), (1002, 118), (1003, 118), (1004, 118), (1005, 118), (1006, 118), (1007, 118), (1008, 118), (1009, 118), (1010, 118), (1011, 118), (1012, 118), (1013, 118), (1014, 118), (1015, 118), (1016, 118), (1017, 118), (1018, 118), (1019, 118), (1020, 118), (1021, 118), (1022, 118), (1023, 118), (1024, 118), (1025, 118), (1026, 118), (1027, 118),
 (0, 774), (1, 774), (2, 774), (3, 774), (4, 774), (5, 774), (6, 774), (7, 774), (8, 774), (9, 774), (10, 774), (11, 774), (1000, 774), (1001, 774), (1002, 774), (1003, 774), (1004, 774), (1005, 774), (1006, 774), (1007, 774), (1008, 774), (1009, 774), (1010, 774), (1011, 774), (1012, 774), (1013, 774), (1014, 774), (1015, 774), (1016, 774), (1017, 774), (1018, 774), (1019, 774), (1020, 774), (1021, 774), (1022, 774), (1023, 774), (1024, 774), (1025, 774), (1026, 774), (1027, 774),
 (0, 366), (1, 366), (2, 366), (3, 366), (4, 366), (5, 366), (6, 366), (7, 366), (8, 366), (9, 366), (10, 366), (11, 366), (1000, 366), (1001, 366), (1002, 366), (1003, 366), (1004, 366), (1005, 366), (1006, 366), (1007, 366), (1008, 366), (1009, 366), (1010, 366), (1011, 366), (1012, 366), (1013, 366), (1014, 366), (1015, 366), (1016, 366), (1017, 366), (1018, 366), (1019, 366), (1020, 366), (1021, 366), (1022, 366), (1023, 366), (1024, 366), (1025, 366), (1026, 366), (1027, 366),
 (0, 103), (1, 103), (2, 103), (3, 103), (4, 103), (5, 103), (6, 103), (7, 103), (8, 103), (9, 103), (10, 103), (11, 103), (1000, 103), (1001, 103), (1002, 103), (1003, 103), (1004, 103), (1005, 103), (1006, 103), (1007, 103), (1008, 103), (1009, 103), (1010, 103), (1011, 103), (1012, 103), (1013, 103), (1014, 103), (1015, 103), (1016, 103), (1017, 103), (1018, 103), (1019, 103), (1020, 103), (1021, 103), (1022, 103), (1023, 103), (1024, 103), (1025, 103), (1026, 103), (1027, 103),
 (0, 498), (1, 498), (2, 498), (3, 498), (4, 498), (5, 498), (6, 498), (7, 498), (8, 498), (9, 498), (10, 498), (11, 498), (1000, 498), (1001, 498), (1002, 498), (1003, 498), (1004, 498), (1005, 498), (1006, 498), (1007, 498), (1008, 498), (1009, 498), (1010, 498), (1011, 498), (1012, 498), (1013, 498), (1014, 498), (1015, 498), (1016, 498), (1017, 498), (1018, 498), (1019, 498), (1020, 498), (1021, 498), (1022, 498), (1023, 498), (1024, 498), (1025, 498), (1026, 498), (1027, 498),
 (0, 469), (1, 469), (2, 469), (3, 469), (4, 469), (5, 469), (6, 469), (7, 469), (8, 469), (9, 469), (10, 469), (11, 469), (1000, 469), (1001, 469), (1002, 469), (1003, 469), (1004, 469), (1005, 469), (1006, 469), (1007, 469), (1008, 469), (1009, 469), (1010, 469), (1011, 469), (1012, 469), (1013, 469), (1014, 469), (1015, 469), (1016, 469), (1017, 469), (1018, 469), (1019, 469), (1020, 469), (1021, 469), (1022, 469), (1023, 469), (1024, 469), (1025, 469), (1026, 469), (1027, 469),
 (0, 1082), (1, 1082), (2, 1082), (3, 1082), (4, 1082), (5, 1082), (6, 1082), (7, 1082), (8, 1082), (9, 1082), (10, 1082), (11, 1082), (1000, 1082), (1001, 1082), (1002, 1082), (1003, 1082), (1004, 1082), (1005, 1082), (1006, 1082), (1007, 1082), (1008, 1082), (1009, 1082), (1010, 1082), (1011, 1082), (1012, 1082), (1013, 1082), (1014, 1082), (1015, 1082), (1016, 1082), (1017, 1082), (1018, 1082), (1019, 1082), (1020, 1082), (1021, 1082), (1022, 1082), (1023, 1082), (1024, 1082), (1025, 1082), (1026, 1082), (1027, 1082),
 (0, 5196), (1, 5196), (2, 5196), (3, 5196), (4, 5196), (5, 5196), (6, 5196), (7, 5196), (8, 5196), (9, 5196), (10, 5196), (11, 5196), (1000, 5196), (1001, 5196), (1002, 5196), (1003, 5196), (1004, 5196), (1005, 5196), (1006, 5196), (1007, 5196), (1008, 5196), (1009, 5196), (1010, 5196), (1011, 5196), (1012, 5196), (1013, 5196), (1014, 5196), (1015, 5196), (1016, 5196), (1017, 5196), (1018, 5196), (1019, 5196), (1020, 5196), (1021, 5196), (1022, 5196), (1023, 5196), (1024, 5196), (1025, 5196), (1026, 5196), (1027, 5196),
 (0, 790), (1, 790), (2, 790), (3, 790), (4, 790), (5, 790), (6, 790), (7, 790), (8, 790), (9, 790), (10, 790), (11, 790), (1000, 790), (1001, 790), (1002, 790), (1003, 790), (1004, 790), (1005, 790), (1006, 790), (1007, 790), (1008, 790), (1009, 790), (1010, 790), (1011, 790), (1012, 790), (1013, 790), (1014, 790), (1015, 790), (1016, 790), (1017, 790), (1018, 790), (1019, 790), (1020, 790), (1021, 790), (1022, 790), (1023, 790), (1024, 790), (1025, 790), (1026, 790), (1027, 790),
 (0, 4039), (1, 4039), (2, 4039), (3, 4039), (4, 4039), (5, 4039), (6, 4039), (7, 4039), (8, 4039), (9, 4039), (10, 4039), (11, 4039), (1000, 4039), (1001, 4039), (1002, 4039), (1003, 4039), (1004, 4039), (1005, 4039), (1006, 4039), (1007, 4039), (1008, 4039), (1009, 4039), (1010, 4039), (1011, 4039), (1012, 4039), (1013, 4039), (1014, 4039), (1015, 4039), (1016, 4039), (1017, 4039), (1018, 4039), (1019, 4039), (1020, 4039), (1021, 4039), (1022, 4039), (1023, 4039), (1024, 4039), (1025, 4039), (1026, 4039), (1027, 4039),
 (0, 326), (1, 326), (2, 326), (3, 326), (4, 326), (5, 326), (6, 326), (7, 326), (8, 326), (9, 326), (10, 326), (11, 326), (1000, 326), (1001, 326), (1002, 326), (1003, 326), (1004, 326), (1005, 326), (1006, 326), (1007, 326), (1008, 326), (1009, 326), (1010, 326), (1011, 326), (1012, 326), (1013, 326), (1014, 326), (1015, 326), (1016, 326), (1017, 326), (1018, 326), (1019, 326), (1020, 326), (1021, 326), (1022, 326), (1023, 326), (1024, 326), (1025, 326), (1026, 326), (1027, 326),
 (0, 105), (1, 105), (2, 105), (3, 105), (4, 105), (5, 105), (6, 105), (7, 105), (8, 105), (9, 105), (10, 105), (11, 105), (1000, 105), (1001, 105), (1002, 105), (1003, 105), (1004, 105), (1005, 105), (1006, 105), (1007, 105), (1008, 105), (1009, 105), (1010, 105), (1011, 105), (1012, 105), (1013, 105), (1014, 105), (1015, 105), (1016, 105), (1017, 105), (1018, 105), (1019, 105), (1020, 105), (1021, 105), (1022, 105), (1023, 105), (1024, 105), (1025, 105), (1026, 105), (1027, 105),
 (0, 122), (1, 122), (2, 122), (3, 122), (4, 122), (5, 122), (6, 122), (7, 122), (8, 122), (9, 122), (10, 122), (11, 122), (1000, 122), (1001, 122), (1002, 122), (1003, 122), (1004, 122), (1005, 122), (1006, 122), (1007, 122), (1008, 122), (1009, 122), (1010, 122), (1011, 122), (1012, 122), (1013, 122), (1014, 122), (1015, 122), (1016, 122), (1017, 122), (1018, 122), (1019, 122), (1020, 122), (1021, 122), (1022, 122), (1023, 122), (1024, 122), (1025, 122), (1026, 122), (1027, 122),
 (0, 5080), (1, 5080), (2, 5080), (3, 5080), (4, 5080), (5, 5080), (6, 5080), (7, 5080), (8, 5080), (9, 5080), (10, 5080), (11, 5080), (1000, 5080), (1001, 5080), (1002, 5080), (1003, 5080), (1004, 5080), (1005, 5080), (1006, 5080), (1007, 5080), (1008, 5080), (1009, 5080), (1010, 5080), (1011, 5080), (1012, 5080), (1013, 5080), (1014, 5080), (1015, 5080), (1016, 5080), (1017, 5080), (1018, 5080), (1019, 5080), (1020, 5080), (1021, 5080), (1022, 5080), (1023, 5080), (1024, 5080), (1025, 5080), (1026, 5080), (1027, 5080),
 (0, 111), (1, 111), (2, 111), (3, 111), (4, 111), (5, 111), (6, 111), (7, 111), (8, 111), (9, 111), (10, 111), (11, 111), (1000, 111), (1001, 111), (1002, 111), (1003, 111), (1004, 111), (1005, 111), (1006, 111), (1007, 111), (1008, 111), (1009, 111), (1010, 111), (1011, 111), (1012, 111), (1013, 111), (1014, 111), (1015, 111), (1016, 111), (1017, 111), (1018, 111), (1019, 111), (1020, 111), (1021, 111), (1022, 111), (1023, 111), (1024, 111), (1025, 111), (1026, 111), (1027, 111),
 (11, 2139),
 (11, 2137),
 (11, 5385),
 (11, 3185),
 (11, 5384),
 (11, 5386),
 (11, 6209),
 (11, 2138),
 (11, 1735),
 (11, 3186),
 (11, 1734),
 (11, 2136),
 (11, 4994),
 (11, 4993),
 (11, 3187),
 (1006, 24073),
 (1006, 24075),
 (1006, 24078),
 (1006, 24074),
 (1006, 24080),
 (1006, 24076),
 (1006, 24081),
 (1006, 24071),
 (1006, 24072),
 (1006, 24077),
 (1006, 24079),
 (1006, 24070),
 (1006, 4414),
 (1000, 24031), (1001, 24031), (1002, 24031), (1003, 24031), (1004, 24031), (1005, 24031),
 (1000, 24030), (1001, 24030), (1002, 24030), (1003, 24030), (1004, 24030), (1005, 24030),
 (1000, 24033), (1001, 24033), (1002, 24033), (1003, 24033), (1004, 24033), (1005, 24033),
 (1000, 24029), (1001, 24029), (1002, 24029), (1003, 24029), (1004, 24029), (1005, 24029);

INSERT INTO stamp_group (id, name, parent_id) VALUES
 (5, 'Events', NULL), 
 (6, 'Characters', 5), 
 (7, 'Activities', NULL), 
 (8, 'Games', NULL), 
 (11, 'Jet Pack Adventure', 8), 
 (13, 'Aqua Grabber', 8), 
 (14, 'Astro Barrier', 8), 
 (15, 'Catchin'' Waves', 8), 
 (16, 'Thin Ice', 8), 
 (19, 'Puffle Rescue', 8), 
 (22, 'Missions', 8), 
 (23, 'Party', 5), 
 (25, 'Video Games', NULL), 
 (26, 'Game Day', 25), 
 (28, 'Cart Surfer', 8), 
 (32, 'Card-Jitsu : Fire', 8), 
 (34, 'Card-Jitsu : Water', 8), 
 (38, 'Card-Jitsu', 8), 
 (46, 'System Defender', 8), 
 (48, 'Puffle Launch', 8), 
 (52, 'Ice Fishing', 8), 
 (54, 'Pizzatron 3000', 8), 
 (56, 'Treasure Hunt', 8), 
 (57, 'Pufflescape', 8), 
 (58, 'Smoothie Smash', 8), 
 (60, 'Card-Jitsu : Snow', 8);
 
INSERT INTO stamp (id, name, group_id, member, rank, description) VALUES
 (9, 'Stage Crew', 7, FALSE, 1, 'Operate the Switchbox 3000'),
 (10, 'Underground', 7, FALSE, 1, 'Find the secret entrance to the underground'),
 (11, 'Snapshot', 7, FALSE, 1, 'Use a camera on top of the Ski Hill'),
 (12, 'Go Swimming', 7, FALSE, 1, 'Swim in water with a rubber duck'),
 (13, 'Clock Target', 7, FALSE, 1, 'Hit the Clock target 10 times in a row'),
 (14, '183 days! ', 7, FALSE, 2, 'Log in with a penguin 183 days old or older'),
 (15, 'Going Places', 7, FALSE, 2, 'Waddle around 30 rooms'),
 (16, 'Dance Party', 7, FALSE, 2, 'Party in the Night Club with 10 penguins'),
 (17, 'Igloo Party', 7, FALSE, 2, 'Throw a party for 10 penguins in your igloo.'),
 (18, 'Coffee Server', 7, FALSE, 2, 'Serve 5 coffees, using the apron and emote'),
 (19, 'Pizza Waiter', 7, FALSE, 2, 'Serve 5 pizzas, using the apron and emote'),
 (20, '365 days!', 7, FALSE, 3, 'Log in with a penguin 365 days old or older'),
 (21, 'Puffle Owner', 7, FALSE, 3, 'Adopt and care for 16 puffles'),
 (22, 'Floor Filler', 7, FALSE, 3, 'Dance in the Night Club with 25 penguins'),
 (23, 'Full House', 7, TRUE, 3, 'Fill your igloo with 99 furniture items'),
 (24, 'Play It Loud! ', 7, FALSE, 3, 'Form a full band at the Lighthouse'),
 (25, 'Soccer Team ', 7, FALSE, 4, 'Form a team with 5 penguins in the same jersey'),
 (26, 'Berg Drill!', 7, FALSE, 3, 'Dance with 30 penguins wearing hard hats at the Berg'),
 (27, 'Fort Battle', 7, FALSE, 3, 'Throw snow at the Forts with 5 penguins of the same color'),
 (28, 'Party Host ', 7, FALSE, 4, 'Throw a party for 30 penguins in your igloo.'),
 (29, 'Hockey Team', 7, FALSE, 3, 'Form a team with 5 penguins in the same jersey'),
 (30, 'Ninja Meeting', 7, FALSE, 4, 'Meet 10 black belts in the Dojo'),
 (197, 'Field Agent', 7, FALSE, 1, 'Earn 1 EPF medal'),
 (198, 'Special Agent', 7, FALSE, 2, 'Earn 5 EPF medals'),
 (199, 'Special Forces', 7, FALSE, 2, 'Earn 10 EPF medals'),
 (200, 'Elite Protector', 7, FALSE, 3, 'Earn 25 EPF medals'),
 (201, 'Island Guardian', 7, FALSE, 4, 'Earn 50 EPF medals'),
 (488, '3 Gems', 7, FALSE, 4, 'Earn the Fire, Water, and Snow Gems'),
 (489, 'Puffle Dig', 7, FALSE, 1, 'Walk a puffle and find treasure'),
 (490, 'First Dig', 7, FALSE, 2, 'See another player''s puffle find their first treasure'),
 (491, 'Every Color', 7, FALSE, 2, 'Dig up treasure with 11 different color puffles'),
 (492, 'Dig All Day', 7, FALSE, 2, 'Find puffle treasure 5 times in a day'),
 (493, 'Big Dig', 7, FALSE, 3, 'Find over 50 coins in a puffle dig'),
 (494, 'Treasure Box', 7, TRUE, 3, 'Find an item in a puffle treasure dig'),
 (495, 'Tasty Treasure', 7, TRUE, 1, 'Find your puffle''s favorite food in a treasure dig'),
 (7, 'Rockhopper ', 6, FALSE, 2, 'Be in the same room as Rockhopper'),
 (8, 'Gary', 6, FALSE, 3, 'Be in the same room as Gary the Gadget Guy'),
 (31, 'Cadence ', 6, FALSE, 3, 'Be in the same room as Cadence'),
 (32, 'Franky', 6, FALSE, 3, 'Be in the same room as Franky'),
 (33, 'Aunt Arctic', 6, FALSE, 4, 'Be in the same room as Aunt Arctic'),
 (34, 'G Billy', 6, FALSE, 3, 'Be in the same room as G Billy'),
 (35, 'Petey K', 6, FALSE, 3, 'Be in the same room as Petey K'),
 (36, 'Stompin'' Bob', 6, FALSE, 3, 'Be in the same room as Stompin'' Bob'),
 (290, 'Sensei', 6, FALSE, 4, 'Be in the same room as Sensei'),
 (358, 'Rookie', 6, FALSE, 4, 'Be in the same room as Rookie'),
 (448, 'PH', 6, FALSE, 1, 'Be in the same room as PH'),
 (466, 'Herbert', 6, FALSE, 4, 'Be in the same room as Herbert.'),
 (182, 'Happy Room', 23, FALSE, 2, 'Make 10 penguins smile in a room'),
 (183, 'Party Puzzle', 23, FALSE, 2, 'Solve a puzzle at a party'),
 (184, 'Snack Shack', 23, FALSE, 1, 'Serve snacks from a booth, using any food emote'),
 (185, 'Target Champion', 23, FALSE, 2, 'Hit 50 targets in a task at a party'),
 (186, 'Celebration', 23, FALSE, 1, 'Blow out the candles on the Anniversary Cake'),
 (187, 'Monster Mash', 23, FALSE, 2, 'Wear a monster outfit at the Halloween Party'),
 (188, 'Scavenger Hunt', 23, FALSE, 2, 'Complete a Scavenger Hunt'),
 (189, 'Construction', 23, FALSE, 1, 'Join in construction at a party with your jackhammer'),
 (190, 'Explorer', 23, TRUE, 1, 'Visit all the decorated party rooms'),
 (191, 'Volunteer', 23, FALSE, 1, 'Donate to a Coins For Change cause'),
 (193, 'Path Finder', 23, FALSE, 2, 'Complete a maze at a party'),
 (292, 'Out At Sea', 23, FALSE, 1, 'Sail away from the island in a boat'),
 (294, 'Top Volunteer', 23, FALSE, 3, 'Give a 5000 coin donation to Coins For Change'),
 (296, 'Epic Volunteer', 23, FALSE, 4, 'Give a 10,000 coin donation to Coins For Change.'),
 (330, 'Party Puffle', 23, FALSE, 0, 'Walk your puffle to their party room'),
 (332, 'Food Fight', 23, FALSE, 1, 'Throw your meal in a place where food is found'),
 (360, 'Noble Knight', 23, TRUE, 2, 'Wear a knight costume at the Medieval party'),
 (362, 'Go Green', 23, FALSE, 1, 'Recycle 10 objects at the Recycling Plant'),
 (364, 'Tree Mob', 23, FALSE, 2, 'Get 10 or more friends to dress up as trees'),
 (426, 'Music Maestro ', 23, FALSE, 2, 'Solve a musical puzzle at a party.'),
 (438, 'Stunt Penguin', 23, FALSE, 3, 'Complete an obstacle course'),
 (439, 'Mountaineer', 23, FALSE, 3, 'Reach a mountain peak'),
 (440, 'Snowboarder', 23, TRUE, 1, 'Do a snowboard dance at a party'),
 (444, 'Trick-or-treat', 23, FALSE, 2, 'See 10 Trick-or-Treat igloos at the Halloween Party'),
 (72, 'Squid Spotter', 13, FALSE, 1, 'Spot the squid'),
 (73, 'Aqua Puffle', 13, FALSE, 1, 'Take your pink puffle for a deep-sea dive'),
 (74, 'Soda Success', 13, TRUE, 2, 'Complete Soda Seas levels'),
 (75, 'Clam Success', 13, FALSE, 2, 'Complete Clam Waters levels'),
 (76, 'Soda Master', 13, TRUE, 2, 'Complete Soda Seas levels without losing a life'),
 (77, 'Clam Master', 13, FALSE, 2, 'Complete Clam Waters levels without losing a life'),
 (78, 'Get Fluffy', 13, TRUE, 2, 'Catch the yellow fish and return to your net'),
 (79, 'Get the Worm', 13, TRUE, 2, 'Catch a worm and return to your net'),
 (80, 'Bubble Catch', 13, FALSE, 2, 'Collect a pink puffle''s bubble'),
 (81, 'Pearl Capture', 13, FALSE, 2, 'Collect all the white pearls in Clam Waters levels'),
 (82, 'Clam Treasure', 13, FALSE, 3, 'Discover secret treasure in Clam Waters levels'),
 (83, 'Soda Treasure', 13, TRUE, 3, 'Discover secret treasure in Soda Seas levels'),
 (84, 'Clam Compress', 13, FALSE, 3, 'Complete the Clam Waters compressed air mode'),
 (85, 'Soda Compress', 13, TRUE, 3, 'Complete the Soda Seas compressed air mode'),
 (86, 'Clam Pressure', 13, FALSE, 3, 'Master the Clam Waters compressed air mode without losing a life'),
 (87, 'Soda Pressure', 13, TRUE, 3, 'Master the Soda Seas compressed air mode without losing a life'),
 (88, 'Clam Timer', 13, FALSE, 3, 'Complete the Clam Waters time trial'),
 (89, 'Soda Timer', 13, TRUE, 3, 'Complete the Soda Seas time trial'),
 (91, 'Crab''s Treasure', 13, TRUE, 3, 'Capture all the crab''s treasure'),
 (92, 'Mullet Capture', 13, TRUE, 4, 'Capture the large red fish and return to the net'),
 (51, 'Astro5', 14, FALSE, 1, 'Finish 5 levels'),
 (52, 'Astro5 Max', 14, FALSE, 1, 'Finish 5 levels without missing a shot'),
 (53, 'Astro40', 14, TRUE, 2, 'Complete levels 1-40'),
 (54, 'Ship Blast', 14, TRUE, 2, 'Use a Blaster to shoot a ship'),
 (55, '1-up Blast', 14, TRUE, 2, 'Use a Blaster to shoot a 1-up'),
 (56, 'Astro Secret', 14, TRUE, 2, 'Complete all Secret Levels'),
 (57, 'Astro10 Max', 14, TRUE, 2, 'Complete levels 1-10 without missing a shot '),
 (58, 'Astro20 Max', 14, TRUE, 2, 'Complete levels 1-20 without missing a shot'),
 (59, 'Astro Expert', 14, TRUE, 3, 'Complete the Expert Levels'),
 (60, 'Astro30 Max', 14, TRUE, 3, 'Complete levels 1-30 without missing a shot'),
 (61, 'Astro 1-up', 14, TRUE, 3, 'Collect 8 1-ups'),
 (62, 'Astro Master', 14, TRUE, 4, 'Complete 25 levels + Secret and Expert'),
 (230, 'Grasshopper', 38, FALSE, 1, 'Begin your journey. Earn your white belt'),
 (232, 'Fine Student ', 38, FALSE, 2, 'Reach half way. Earn your blue belt'),
 (234, 'True Ninja ', 38, FALSE, 3, 'Prove your ninja skills. Earn your black belt'),
 (236, 'Ninja Master ', 38, FALSE, 3, 'Defeat Sensei and become a Ninja'),
 (238, 'Flawless Victory ', 38, FALSE, 2, 'Win without letting your opponent gain any cards'),
 (240, 'Match Master', 38, FALSE, 3, 'Win 25 matches'),
 (242, 'Elemental Win', 38, FALSE, 1, 'Win a match with 3 different elements'),
 (244, 'One Element ', 38, FALSE, 2, 'Win a match with 3 cards of the same element'),
 (246, 'Sensei Card', 38, FALSE, 4, 'See the Sensei Power Card'),
 (248, 'Full Dojo', 38, FALSE, 4, 'Score 9 cards before you win'),
 (252, 'Warm Up', 32, FALSE, 1, 'Win 10 Fire matches'),
 (254, 'Score Fire ', 32, FALSE, 1, 'Win 1 energy point in a match'),
 (256, 'Fire Midway', 32, FALSE, 2, 'Earn the coat. Finish 50% of your Fire Ninja journey'),
 (260, 'Strong Defence', 32, FALSE, 3, 'Win a match without losing any energy'),
 (262, 'Fire Suit', 32, FALSE, 3, 'Complete your Fire Suit'),
 (264, 'Fire Ninja', 32, FALSE, 3, 'Defeat Sensei and become a Fire Ninja'),
 (266, 'Max Energy', 32, FALSE, 4, 'Win 3 energy points in a match'),
 (268, 'Fire Expert ', 32, FALSE, 4, 'Win 50 matches'),
 (467, '3 Ninja Combo', 60, FALSE, 2, 'Play a Power Card in the same turn as all ninjas'),
 (468, '4 Ninja Combo', 60, FALSE, 3, 'Play a Power Card in the same turn as all ninjas & Sensei'),
 (469, 'Snow Ninja', 60, FALSE, 1, 'Win 3 Rounds as a Snow Ninja'),
 (470, 'Fire Ninja', 60, FALSE, 0, 'Win 3 Rounds as a Fire Ninja'),
 (471, 'Water Ninja', 60, FALSE, 1, 'Win 3 Rounds as a Water Ninja'),
 (472, 'Full Health', 60, FALSE, 4, 'Reach a Full Health Bonus Round'),
 (473, 'Bonus Win', 60, FALSE, 4, 'Defeat all Snow Minions in a Bonus Round'),
 (474, 'Revive', 60, FALSE, 1, 'Revive a ninja'),
 (475, 'Up and at ''em', 60, FALSE, 2, 'Get revived and still win the battle'),
 (476, 'Team Revival', 60, FALSE, 3, 'Win after all ninjas have fallen and been revived'),
 (477, 'Heal 15', 60, FALSE, 3, 'Heal a ninja 15 times in a battle'),
 (478, 'Huge Heal', 60, FALSE, 2, 'Heal all ninjas with 1 Snow Power Card'),
 (479, 'Snow Shield', 60, FALSE, 2, 'Shield other ninjas in a Snow Power Card Combo'),
 (480, 'Tidal Wave', 60, FALSE, 3, 'Damage 3 Snow Minions with 1 Water Power Card'),
 (481, 'Wave Boost', 60, FALSE, 2, 'Boost all ninja attacks in a Water Power Card Combo'),
 (482, 'Fire Blast', 60, FALSE, 3, 'Blast 3 Snow Minions with 1 Fire Power Card'),
 (483, 'Fire Blast Combo', 60, FALSE, 3, 'Blast 3 Snow Minions with 1 Fire Power Card in a Combo'),
 (484, 'Power Card Pro', 60, FALSE, 2, 'Play 3 Power Cards in a battle'),
 (485, '3 Combos', 60, FALSE, 3, 'Play 3 Power Card Combos in a battle'),
 (486, 'Final Battle', 60, FALSE, 4, 'Defeat the master of Snow'),
 (487, 'Snow Pro', 60, FALSE, 2, 'Earn the gem. Finish over half of your Snow journey.'),
 (270, 'Gong!', 34, FALSE, 1, 'Win a match and sound the gong'),
 (274, 'Watery Fall', 34, FALSE, 1, 'Take the plunge! Fall off the Waterfall'),
 (276, 'Water Expert', 34, FALSE, 2, 'Win 100 matches'),
 (278, 'Water Midway', 34, FALSE, 2, 'Earn the coat. Finish 50% of your Water Ninja journey'),
 (282, 'Water Suit', 34, FALSE, 3, 'Complete your Water Suit'),
 (284, 'Water Ninja', 34, FALSE, 3, 'Defeat Sensei and become a Water Ninja'),
 (286, 'Two Close', 34, FALSE, 4, 'Drift to the edge twice and still win the match'),
 (288, 'Skipping Stones', 34, FALSE, 4, 'Clear 28 stones of any element in a match'),
 (206, 'Cart Pro', 28, FALSE, 2, 'Earn 150 coins in one game'),
 (208, 'Cart Expert', 28, FALSE, 3, 'Earn 250 coins in one game'),
 (210, 'Cart Master', 28, FALSE, 4, 'Earn 350 coins in one game'),
 (212, 'Great Balance', 28, FALSE, 1, 'Recover from a wobble'),
 (214, 'Mine Marvel', 28, FALSE, 2, 'Perform 10 tricks in the cart'),
 (216, 'Mine Mission', 28, FALSE, 2, 'Perform 10 tricks out of the cart'),
 (218, 'Trick Master', 28, FALSE, 2, 'Perform 14 different tricks'),
 (220, 'Mine Grind', 28, FALSE, 3, 'Grind around 8 corners'),
 (222, 'Surf''s Up', 28, FALSE, 3, 'Surf around 8 corners'),
 (224, 'Flip Mania', 28, FALSE, 3, 'Flip 20 times in a row without crashing'),
 (226, 'Ultimate Duo', 28, FALSE, 3, 'Perform 20 tricks with your puffle'),
 (228, 'Puffle Power', 28, FALSE, 2, 'Recover from a wobble with a puffle'),
 (93, 'First Trick', 15, FALSE, 1, 'Master your first trick on the board'),
 (94, 'Puffle Surfin''', 15, FALSE, 1, 'Take your red puffle for a surf lesson'),
 (95, 'Easy Flip', 15, FALSE, 1, 'Perform 1 flip'),
 (96, 'Easy Tube', 15, FALSE, 1, 'Hit 100 point combo while shooting the tube'),
 (97, 'Easy Grind', 15, FALSE, 1, 'Hit 100 point combo while grinding the wave'),
 (98, 'Graduate', 15, FALSE, 1, 'Finish the surf lesson'),
 (99, 'Trick Star', 15, FALSE, 2, 'Master 13 different tricks in a surf'),
 (100, 'Podium Puffle', 15, TRUE, 2, 'Finish in 1st, 2nd or 3rd place with your puffle'),
 (101, 'Flip Star', 15, FALSE, 2, 'Jump and flip 3 times'),
 (102, 'Easy Jump', 15, FALSE, 2, 'Jump the height of 8 penguins'),
 (103, 'Super Tube', 15, FALSE, 2, 'Hit 1000 point combo while shooting the tube'),
 (104, 'Easy Spin', 15, FALSE, 1, 'Perform a 3 spin jump'),
 (105, 'Super Spin', 15, FALSE, 2, 'Perform a 7 spin jump'),
 (106, 'Super Grind', 15, FALSE, 2, 'Hit 700 point combo while grinding the wave'),
 (107, 'Max Grind', 15, TRUE, 3, 'Hit 1200 point combo while grinding the wave'),
 (108, 'First Place', 15, TRUE, 3, 'Win first place in a competition'),
 (109, 'Max Flips', 15, TRUE, 3, 'Flip 10 times'),
 (110, 'High Jump', 15, TRUE, 3, 'Jump the height of 20 penguins'),
 (111, 'Max Tube', 15, FALSE, 3, 'Hit 5000 point combo while shooting the tube'),
 (112, 'Max Spin', 15, FALSE, 3, 'Do a 10 spin jump'),
 (113, 'Shark!', 15, TRUE, 4, 'Spot the shark'),
 (114, 'Survivor', 15, TRUE, 4, 'Complete Survival Mode'),
 (372, 'Snack Attack', 52, FALSE, 1, 'Feed a fish to a shark'),
 (374, 'Shock King', 52, FALSE, 2, 'Get 3 shocks from jellyfish and finish the game'),
 (376, 'Fishtastic', 52, FALSE, 2, 'Catch 15 fish without any mistakes'),
 (378, 'Worm Win', 52, FALSE, 3, 'Finish the game without losing a worm'),
 (380, 'Crab Cuts', 52, FALSE, 3, 'Have 3 crabs cut your line and finish the game'),
 (382, 'Afishionado', 52, FALSE, 3, 'Catch 45 fish without any mistakes'),
 (384, 'Gray Goodies', 52, TRUE, 3, 'Catch 15 gray fish'),
 (386, 'Prize Mullet', 52, FALSE, 3, 'Capture Mullet'),
 (388, 'Fly Fisher', 52, FALSE, 3, 'Catch 63 fish in under 5 minutes'),
 (390, 'Ace Angler', 52, TRUE, 4, 'Hook 15 gray fish and Mullet with no worm lost'),
 (37, 'Lift Off!', 11, FALSE, 1, 'Take off and return to the launch pad'),
 (38, 'Fuel Rank 1', 11, FALSE, 1, 'Collect all fuel cans in Level 1'),
 (39, 'Jet Pack 5', 11, TRUE, 2, 'Complete 5 levels'),
 (40, 'Crash!', 11, TRUE, 2, 'Crash land your jet pack at the Forest'),
 (41, 'Fuel Rank 2', 11, TRUE, 2, 'Collect all fuel cans in Level 2'),
 (42, 'Fuel Rank 3', 11, TRUE, 2, 'Collect all fuel cans in Level 3'),
 (43, 'Fuel Rank 4', 11, TRUE, 2, 'Collect all fuel cans in Level 4'),
 (44, 'Fuel Rank 5', 11, TRUE, 2, 'Collect all fuel cans in Level 5'),
 (45, '1-up Leader', 11, TRUE, 2, 'Collect 2 x 1-ups'),
 (46, 'Kerching!', 11, TRUE, 3, 'Collect 650 coins'),
 (47, 'Fuel Command', 11, TRUE, 3, 'Collect all fuel cans in 1 game'),
 (48, 'Fuel Wings', 11, FALSE, 3, 'Collect a fuel can while falling'),
 (49, '1-up Captain', 11, TRUE, 3, 'Find all 1-ups in one game'),
 (50, 'Ace Pilot ', 11, TRUE, 4, 'Complete the game without collecting coins'),
 (202, 'Puffle Pilot ', 11, FALSE, 1, 'Bring a green puffle into the game'),
 (203, 'Puffle Bonus ', 11, TRUE, 2, 'Your green puffle collects 200 coins'),
 (204, 'Puffle Plus ', 11, TRUE, 3, 'Your green puffle collects 450 coins'),
 (205, 'Puffle Boost', 11, FALSE, 4, 'Your green puffle gets a gas can when you run out of fuel'),
 (159, 'Mission 1 Medal', 22, FALSE, 0, 'Complete Case of the Missing Puffles'),
 (160, 'Aunt Arctic Letter', 22, FALSE, 0, 'Complete the secret task in Mission 1'),
 (161, 'Mission 2 Medal', 22, FALSE, 0, 'Complete G''s Secret Mission'),
 (162, 'G''s Letter', 22, FALSE, 0, 'Complete the secret task in Mission 2'),
 (163, 'Mission 3 Medal', 22, FALSE, 0, 'Complete Case of the Missing Coins'),
 (164, 'Thank You Card', 22, FALSE, 0, 'Complete the secret task in Mission 3'),
 (165, 'Mission 4 Medal', 22, FALSE, 0, 'Complete Avalanche Rescue'),
 (166, 'Handy Award', 22, FALSE, 0, 'Complete the secret task in Mission 4'),
 (167, 'Mission 5 Medal', 22, FALSE, 0, 'Complete Secret of the Fur'),
 (168, 'Pizza Box', 22, FALSE, 0, 'Complete the secret task in Mission 5'),
 (169, 'Mission 6 Medal', 22, FALSE, 0, 'Complete Questions for a Crab'),
 (170, 'Blueprint', 22, FALSE, 0, 'Complete the secret task in Mission 6'),
 (171, 'Mission 7 Medal', 22, FALSE, 0, 'Complete Clockwork Repairs'),
 (172, 'Pennant', 22, FALSE, 0, 'Complete the secret task in Mission 7'),
 (173, 'Mission 8 Medal', 22, FALSE, 0, 'Complete Mysterious Tremors'),
 (174, 'Cool Gift', 22, FALSE, 0, 'Complete the secret task in Mission 8'),
 (175, 'Mission 9 Medal', 22, FALSE, 0, 'Complete Operation: Spy and Seek'),
 (176, 'Chocolate Box', 22, FALSE, 0, 'Complete the secret task in Mission 9'),
 (177, 'Mission 10 Medal', 22, FALSE, 0, 'Complete Waddle Squad'),
 (178, 'Employee Award', 22, FALSE, 0, 'Complete the secret task in Mission 10'),
 (179, 'Mission 11 Medal', 22, FALSE, 0, 'Complete The Veggie Villain'),
 (180, 'Snow Globe', 22, FALSE, 0, 'Complete the secret task in Mission 11'),
 (392, 'Food Fiasco', 54, FALSE, 1, 'Make a mess of the kitchen with 3 wrong pizzas'),
 (394, 'Just Dessert', 54, FALSE, 1, 'Play Pizzatron in Candy Mode'),
 (396, 'Chef''s Hat', 54, FALSE, 2, 'Make 20 pizzas without any mistakes'),
 (398, 'Spice Sea', 54, FALSE, 2, 'Make 3 hot sauce and shrimp pizzas to order'),
 (400, 'Cocoa Beans', 54, FALSE, 2, 'Make 3 jellybean and chocolate pizzas to order'),
 (402, 'Fiery Squids', 54, FALSE, 3, 'Make 3 hot sauce and squid pizzas to order'),
 (404, 'Candy Land', 54, FALSE, 3, 'Make 3 pink sauce and marshmallow pizzas to order'),
 (406, 'Pizza Chef', 54, FALSE, 3, 'Make 30 pizzas without any mistakes'),
 (408, 'Pizza Master', 54, FALSE, 4, 'Make 40 pizzas without any mistakes'),
 (410, 'Dessert Chef', 54, FALSE, 4, 'Make 40 candy pizzas without any mistakes'),
 (334, 'Begin Build', 48, FALSE, 1, 'Collect your first cannon piece'),
 (336, 'Launch Ready', 48, FALSE, 2, 'Build a cannon'),
 (338, 'Epic Cannon ', 48, TRUE, 3, 'Build the Epic Cannon'),
 (340, 'Crab Attack', 48, FALSE, 2, 'Defeat a crab boss once'),
 (342, 'Crab Crash', 48, TRUE, 3, 'Defeat 3 crab bosses in a single play session'),
 (344, 'Crab Battle', 48, TRUE, 3, 'Defeat 6 crab bosses in a single play session'),
 (346, 'Quick Launch', 48, TRUE, 2, 'Complete all 36 levels in under 35 minutes'),
 (348, 'Supersonic Launch', 48, TRUE, 3, 'Complete all 36 levels in under 25 minutes'),
 (350, 'Light Speed Launch', 48, TRUE, 4, 'Complete all 36 levels in under 18 minutes'),
 (352, 'Turbo Time', 48, TRUE, 2, 'Complete a level in turbo mode'),
 (354, 'Turbo Battle', 48, TRUE, 3, 'Defeat the crab in turbo mode'),
 (356, 'Turbo Master', 48, TRUE, 4, 'Complete all 36 levels in turbo mode'),
 (130, 'SOS 30', 19, FALSE, 2, 'Rescue 30 puffles in snow levels'),
 (131, 'SOS 60', 19, FALSE, 3, 'Rescue 60 Puffles in snow levels'),
 (132, '1 Coin Bag', 19, FALSE, 1, 'Find 1 coin bag in the snow levels'),
 (133, '2 Coin Bags ', 19, FALSE, 2, 'Find 2 coin bags in the snow levels'),
 (134, '3 Coin Bags', 19, FALSE, 3, 'Find 3 coin bags in the snow levels'),
 (135, 'Snow Student', 19, FALSE, 2, 'Complete 3 snow levels in no more than 70 moves'),
 (136, 'Snow Master', 19, FALSE, 3, 'Complete 6 snow levels in no more than 160 moves'),
 (137, 'Snow Hero', 19, FALSE, 4, 'Complete 9 snow levels in no more than 270 moves'),
 (138, 'Cave Coins', 19, TRUE, 2, 'Find hidden coins on level 1 in the caves'),
 (139, 'Cave Coins Plus', 19, TRUE, 2, 'Find hidden coins on level 4 in the caves'),
 (140, 'Cave Coins Max', 19, TRUE, 3, 'Find hidden coins on Level 7 in the caves'),
 (141, 'Quick Catch', 19, TRUE, 1, 'Catch 1 coin bag before it falls'),
 (142, 'Super Catch', 19, TRUE, 2, 'Catch 2 coin bags before they fall'),
 (144, 'Expert Catch', 19, TRUE, 3, 'Catch 3 coin bags before they fall'),
 (145, 'Easy Cannon', 19, TRUE, 2, 'Destroy 8 snowball cannons in cave levels'),
 (146, 'Super Cannon', 19, TRUE, 3, 'Destroy 12 snowball cannons in cave levels'),
 (147, 'Extreme Cannon', 19, TRUE, 4, 'Destroy 15 snowball cannons in cave levels'),
 (148, 'Close Call', 19, TRUE, 2, 'Dodge the giant snowball'),
 (149, 'Rapid Rescue', 19, TRUE, 1, 'Complete 5 sea levels in 30 seconds each'),
 (150, 'Express Rescue', 19, TRUE, 2, 'Complete 10 sea levels in 30 seconds each'),
 (151, 'Extreme Rescue', 19, TRUE, 3, 'Complete 20 sea levels in 30 seconds each'),
 (152, '10 Sea Levels', 19, TRUE, 2, 'Reach 10 sea levels with 5 full lives'),
 (153, '20 Sea Levels', 19, TRUE, 3, 'Reach 20 sea levels with 5 full lives'),
 (154, '30 Sea Levels', 19, TRUE, 4, 'Reach 30 sea levels with 5 full lives'),
 (155, '1 Coin Bubble', 19, TRUE, 1, 'Find 1 coin bag in a bubble in the sea levels'),
 (156, '2 Coin Bubbles', 19, TRUE, 2, 'Find 2 coin bags in a bubble in the sea levels'),
 (157, '3 Coin Bubbles', 19, TRUE, 3, 'Find 3 coin bags in a bubble in the sea levels'),
 (427, 'Bonus Snack', 57, FALSE, 1, 'Eat a bonus Puffle O'),
 (428, 'Puffle O Feast', 57, FALSE, 2, 'Eat all regular Puffle O''s in one set of levels'),
 (429, 'Bite in Time', 57, FALSE, 2, 'Eat a disappearing Puffle O'),
 (430, 'Quick Snack', 57, FALSE, 3, 'Eat all disappearing O''s in one level'),
 (431, 'Fast Food', 57, FALSE, 3, 'Eat all disappearing O''s in one set of levels'),
 (432, 'On a Roll', 57, FALSE, 1, 'Complete level 4'),
 (433, 'Puzzle Pro', 57, FALSE, 1, 'Complete level 8'),
 (434, 'Extreme Puzzler', 57, TRUE, 2, 'Complete level 12'),
 (435, 'Master Roller', 57, TRUE, 2, 'Complete level 17'),
 (436, 'Epic Roller', 57, TRUE, 3, 'Complete level 22'),
 (437, 'Ice Master', 57, TRUE, 3, 'Complete the final level'),
 (450, 'Fruit Squeeze', 58, FALSE, 1, 'Complete 5 recipes in one game'),
 (451, 'Shipshake', 58, FALSE, 1, 'Serve Rockhopper'),
 (452, 'Berry Smart', 58, FALSE, 1, 'Serve Aunt Arctic'),
 (453, 'Beary Berry', 58, FALSE, 1, 'Serve Herbert'),
 (454, 'Tasty Treat', 58, FALSE, 2, 'Complete 10 recipes in one game'),
 (455, 'Fruit Splatter', 58, FALSE, 2, 'Smash 10 correct fruit in a row'),
 (456, 'Fruit Power', 58, FALSE, 2, 'Last 2 minutes in survival mode'),
 (457, 'Deflector', 58, FALSE, 2, 'Take no damage for 2 mins in survival mode'),
 (458, 'Energizer', 58, FALSE, 3, 'Complete 20 recipes in one game'),
 (459, 'Fruit Stomper', 58, FALSE, 3, 'Smash 20 correct fruit in a row'),
 (460, 'Smoothie Warrior', 58, FALSE, 3, 'Last 4 minutes in survival mode'),
 (461, 'Smoothie Survivor ', 58, FALSE, 3, 'Take no damage for 3 mins in survival mode'),
 (462, 'Smoothie Zing', 58, FALSE, 4, 'Complete 30 recipes in one game'),
 (463, 'Fruit Smasher', 58, FALSE, 4, 'Smash 50 correct fruit in a row'),
 (464, 'Smoothie Master', 58, FALSE, 4, 'Last 8 minutes in survival mode'),
 (465, 'Smoothie Hero', 58, FALSE, 4, 'Take no damage for 4 mins in survival mode'),
 (298, 'Ready For Duty', 46, FALSE, 1, 'Finish G''s tutorial'),
 (300, 'Garbage Disposal', 46, FALSE, 1, 'Destroy 100 enemies in 1 level'),
 (302, 'Strategic Success', 46, FALSE, 2, 'Destroy 100 enemies without upgrading cannons'),
 (304, 'Mono Mechanic', 46, FALSE, 2, 'Complete a game with only one type of cannon'),
 (306, 'Tactical Pro', 46, FALSE, 2, 'Destroy 100 enemies without taking damage'),
 (308, 'Master Mechanic', 46, FALSE, 2, 'Fill every open socket with a cannon'),
 (310, 'Strategic Master', 46, FALSE, 3, 'Destroy 250 enemies without upgrading'),
 (312, 'Tactical Ace', 46, FALSE, 3, 'Destroy 250 enemies without taking damage'),
 (320, 'Bug Overload', 46, FALSE, 1, 'Protect the EPF mainframe from rogue bugs'),
 (322, 'Herbert Attacks', 46, FALSE, 2, 'Complete the level ''Herbert Attacks'''),
 (324, 'Klutzy Attack', 46, FALSE, 2, 'Protect the EPF mainframe from Klutzy'),
 (326, 'Test Bot Trio', 46, TRUE, 3, 'Protect the EPF mainframe from the Test Bots'),
 (328, 'Protobot Attack', 46, TRUE, 3, 'Protect the EPF mainframe from Ultimate Protobot'),
 (443, 'Track Herbert ', 46, FALSE, 4, 'Complete the level ''Track Herbert'''),
 (63, '1 Coin Bag', 16, FALSE, 1, 'Collect 1 coin bag'),
 (64, '3 Coin Bags', 16, FALSE, 2, 'Collect 3 coin bags'),
 (65, '6 Coin Bags', 16, FALSE, 2, 'Collect 6 coin bags'),
 (66, 'Iced Treasure', 16, TRUE, 2, 'Find the hidden treasure room'),
 (67, '10 Coin Bags', 16, TRUE, 3, 'Collect 10 coin bags'),
 (68, 'Ice Bonus', 16, FALSE, 3, 'Completely melt 480 ice tiles'),
 (69, 'Ice Trekker', 16, TRUE, 3, 'Push all blocks to the correct position'),
 (70, 'All Coin Bags', 16, TRUE, 4, 'Collect all coin bags on every level'),
 (71, 'Ice Master', 16, TRUE, 4, 'Master all the mazes'),
 (414, 'Gem Skills', 56, FALSE, 2, 'Uncover 2 gems in a single game'),
 (416, 'Collector', 56, FALSE, 2, 'Collect 8 coins in a single game'),
 (418, 'In the Rough', 56, FALSE, 3, 'Find a rare treasure'),
 (420, 'Gem Expert', 56, FALSE, 3, 'Uncover a gem in only 3 moves'),
 (422, 'Gem Pro', 56, FALSE, 4, 'Uncover 2 gems in only 7 moves'),
 (115, 'Red Win', 26, FALSE, 1, 'Conquer the island as Team Red'),
 (116, 'Blue Win', 26, FALSE, 1, 'Conquer the island as Team Blue'),
 (117, 'Green Win', 26, FALSE, 1, 'Conquer the island as Team Green'),
 (118, 'Yellow Win', 26, FALSE, 1, 'Conquer the island as Team Yellow'),
 (119, 'Conquer the Island', 26, FALSE, 2, 'Conquer the island as each team'),
 (120, 'Collector', 26, FALSE, 3, 'Collect all the items'),
 (121, 'Remote Upload', 26, FALSE, 1, 'Upload another penguin to your Wii remote'),
 (122, 'Sumo Smash', 26, FALSE, 4, 'Win a 4-player game without being knocked out'),
 (123, 'Puffle Paddle', 26, FALSE, 2, 'Play 4-player mode without dropping any puffles'),
 (124, 'Bean Balance', 26, FALSE, 2, 'Balance 30 bags in a 4-player game'),
 (125, '2 Vs. 2', 26, FALSE, 3, 'Come 1st in all 2 vs. 2 matches'),
 (126, '2 Vs. 2 Max', 26, FALSE, 4, 'Win all 2 vs. 2 matches against the Wii at hard level'),
 (127, 'White Puffle', 26, FALSE, 1, 'Feed a white puffle in Feed a Puffle'),
 (128, 'Goalie', 26, FALSE, 4, 'Win a game without letting a puck in your net');
 
INSERT INTO room (id, name, member, max_users, required_item, game, blackhole, spawn, stamp_group) VALUES
 (100, 'Town Center', FALSE, 120, NULL, FALSE, FALSE, TRUE, NULL),
 (110, 'Coffee Shop', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (111, 'Book Room', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (120, 'Night Club', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (121, 'Lounge', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (122, 'school', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (130, 'Clothes Shop', FALSE, 80, NULL, FALSE, FALSE, TRUE, NULL),
 (200, 'Ski Village', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (210, 'Sports Shop', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (211, 'Agent Lobby Solo', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (212, 'Agent Lobby Multi', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (213, 'Agent VR', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (220, 'Ski Lodge', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (221, 'Lodge Attic', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (230, 'Ski Hill', FALSE, 80, NULL, FALSE, FALSE, TRUE, NULL),
 (300, 'The Plaza', FALSE, 80, NULL, FALSE, FALSE, TRUE, NULL),
 (310, 'Pet Shop', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (320, 'The Dojo', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (321, 'Dojo Exterior', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (322, 'Dojo Hide', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (323, 'Agent Command', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (330, 'Pizza Parlor', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (340, 'The Mall', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (400, 'The Beach', FALSE, 80, NULL, FALSE, FALSE, TRUE, NULL),
 (410, 'Lighthouse', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (411, 'Beacon', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (420, 'Pirate Ship', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (421, 'Ship''s Hold', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (422, 'Captain''s Quarters', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (423, 'The Crows Nest', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (800, 'The Dock', FALSE, 80, NULL, FALSE, FALSE, TRUE, NULL),
 (801, 'Snow Forts', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (802, 'Ice Rink', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (803, 'HQ', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (804, 'Boiler Room', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (805, 'Ice Berg', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (806, 'Underground Pool', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (807, 'Mine Shack', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (808, 'The Mine', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (809, 'Forest', FALSE, 80, NULL, FALSE, FALSE, TRUE, NULL),
 (810, 'Cove', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (811, 'Box Dimension', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (812, 'Dojo Fire', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (813, 'Gold Mine', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (814, 'Lake', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (815, 'Underwater', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (816, 'Dojo Water', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (851, 'Party1', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (852, 'Party2', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (853, 'Party3', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (854, 'Party4', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (855, 'Party5', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (856, 'Party6', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (857, 'Party7', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (858, 'Party8', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (859, 'Party9', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (860, 'Party10', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (861, 'Party11', TRUE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (862, 'Party12', TRUE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (863, 'Party13', TRUE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (899, 'Party', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (900, 'Astro', FALSE, 800, NULL, TRUE, TRUE, FALSE, 14),
 (901, 'Beans', FALSE, 800, NULL, TRUE, FALSE, FALSE, NULL),
 (902, 'Puffle', FALSE, 800, NULL, TRUE, FALSE, FALSE, NULL),
 (903, 'Biscuit', FALSE, 800, NULL, TRUE, FALSE, FALSE, NULL),
 (904, 'Fish', FALSE, 800, NULL, TRUE, FALSE, FALSE, 52),
 (905, 'MineCar1', FALSE, 800, NULL, TRUE, FALSE, FALSE, 28),
 (906, 'JetPack', FALSE, 800, NULL, TRUE, FALSE, FALSE, 11),
 (907, 'Mission1', FALSE, 800, NULL, TRUE, FALSE, FALSE, 22),
 (908, 'Mission2', FALSE, 800, NULL, TRUE, FALSE, FALSE, 22),
 (909, 'ThinIce', FALSE, 800, NULL, TRUE, TRUE, FALSE, 16),
 (910, 'PizzaTron', FALSE, 800, NULL, TRUE, FALSE, FALSE, 54),
 (911, 'Mission3', FALSE, 800, NULL, TRUE, FALSE, FALSE, 22),
 (912, 'Wave', FALSE, 800, NULL, TRUE, FALSE, FALSE, 15),
 (913, 'Mission4', FALSE, 800, NULL, TRUE, FALSE, FALSE, 22),
 (914, 'Mission5', FALSE, 800, NULL, TRUE, FALSE, FALSE, 22),
 (915, 'Mission6', FALSE, 800, NULL, TRUE, FALSE, FALSE, 22),
 (916, 'SubGame', FALSE, 800, NULL, TRUE, FALSE, FALSE, 13),
 (917, 'BookGame1', FALSE, 800, NULL, TRUE, FALSE, FALSE, NULL),
 (918, 'BookGame2', FALSE, 800, NULL, TRUE, FALSE, FALSE, NULL),
 (919, 'BookGame3', FALSE, 800, NULL, TRUE, FALSE, FALSE, NULL),
 (920, 'Mission7', FALSE, 800, NULL, TRUE, FALSE, FALSE, 22),
 (921, 'Mission8', FALSE, 800, NULL, TRUE, FALSE, FALSE, 22),
 (922, 'Mission9', FALSE, 800, NULL, TRUE, FALSE, FALSE, 22),
 (923, 'Mission10', FALSE, 800, NULL, TRUE, FALSE, FALSE, 22),
 (926, 'MixMaster', FALSE, 800, NULL, TRUE, FALSE, FALSE, NULL),
 (927, 'Mission11', FALSE, 800, NULL, TRUE, FALSE, FALSE, 22),
 (941, 'Puffle Soaker', FALSE, 800, NULL, TRUE, FALSE, FALSE, NULL),
 (942, 'Balloon Pop', FALSE, 800, NULL, TRUE, FALSE, FALSE, NULL),
 (943, 'Ring the Bell', FALSE, 800, NULL, TRUE, FALSE, FALSE, NULL),
 (944, 'Feed a Puffle', FALSE, 800, NULL, TRUE, FALSE, FALSE, NULL),
 (945, 'Memory Card', FALSE, 800, NULL, TRUE, FALSE, FALSE, NULL),
 (946, 'Puffle paddle', FALSE, 800, NULL, TRUE, FALSE, FALSE, NULL),
 (947, 'Puffle Shuffle', FALSE, 800, NULL, TRUE, FALSE, FALSE, NULL),
 (948, 'Grab and Spin', FALSE, 800, NULL, TRUE, FALSE, FALSE, NULL),
 (949, 'Puffle Rescue', FALSE, 800, NULL, TRUE, FALSE, FALSE, 19),
 (950, 'System Defender', FALSE, 800, NULL, TRUE, TRUE, FALSE, 46),
 (951, 'Sensei', FALSE, 80, NULL, TRUE, FALSE, FALSE, NULL),
 (952, 'Dancing', FALSE, 800, NULL, TRUE, FALSE, FALSE, NULL),
 (953, 'Fire Sensei', FALSE, 800, NULL, TRUE, FALSE, FALSE, NULL),
 (954, 'Water Sensei', FALSE, 800, NULL, TRUE, FALSE, FALSE, NULL),
 (960, 'Ice Jam', FALSE, 800, NULL, TRUE, FALSE, FALSE, NULL),
 (994, 'igloo_card', FALSE, 80, NULL, TRUE, FALSE, FALSE, NULL),
 (995, 'Water', FALSE, 80, NULL, TRUE, FALSE, FALSE, 34),
 (997, 'Fire', FALSE, 80, NULL, TRUE, FALSE, FALSE, 32),
 (998, 'Card Jitsu', FALSE, 80, NULL, TRUE, FALSE, FALSE, 38),
 (999, 'Sled Race', TRUE, 80, NULL, TRUE, FALSE, FALSE, NULL),
 (324, 'Dojo Exterior Solo', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (326, 'Dojo Snow', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (430, 'Puffle Hotel Lobby', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (431, 'Puffle Hotel Spa', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (432, 'Puffle Hotel Roof', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (433, 'Cloud Forest', TRUE, 800, NULL, FALSE, FALSE, FALSE, NULL),
 (434, 'Puffle Park', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (435, 'Skatepark', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (436, 'Puffle Wild', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (437, 'UFO', FALSE, 800, NULL, FALSE, FALSE, FALSE, NULL),
 (864, 'Party14', TRUE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (865, 'Party15', TRUE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (866, 'Party16', TRUE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (867, 'Party17', TRUE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (868, 'Party18', TRUE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (869, 'Party19', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (874, 'Party24', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (875, 'Party25', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (876, 'Party26', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (877, 'Party27', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (878, 'Party28', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (890, 'Party99', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL),
 (891, 'partysolo1', FALSE, 800, NULL, FALSE, FALSE, FALSE, NULL),
 (898, 'Sound Room', FALSE, 800, NULL, FALSE, FALSE, FALSE, NULL),
 (924, 'Game24', FALSE, 80, NULL, TRUE, FALSE, FALSE, NULL),
 (925, 'Game25', FALSE, 80, NULL, TRUE, FALSE, FALSE, NULL),
 (955, 'Puffle Launch', FALSE, 800, NULL, TRUE, FALSE, FALSE, 48),
 (956, 'Bits and Bolts', FALSE, 800, NULL, TRUE, TRUE, FALSE, NULL),
 (957, 'Rollerscape', FALSE, 800, NULL, TRUE, FALSE, FALSE, 57),
 (958, 'Scorn Battle', FALSE, 800, NULL, TRUE, FALSE, FALSE, NULL),
 (959, 'Smoothie', FALSE, 800, NULL, TRUE, FALSE, FALSE, 58),
 (961, 'Dino Treasure Hunt', FALSE, 800, NULL, TRUE, FALSE, FALSE, NULL),
 (962, 'Party Game', FALSE, 800, NULL, TRUE, FALSE, FALSE, NULL),
 (963, 'Spy Drills', FALSE, 800, NULL, TRUE, TRUE, FALSE, NULL),
 (996, 'Snow', FALSE, 2000, NULL, TRUE, FALSE, FALSE, 60),
 (1100, 'My Penguin', FALSE, 0, NULL, FALSE, FALSE, FALSE, NULL),
 (1104, 'Gold Mine', FALSE, 80, NULL, FALSE, FALSE, FALSE, NULL);
 
INSERT INTO room_waddle (id, room_id, seats, game) VALUES
 (100, 230, 4, 'sled'),
 (101, 230, 3, 'sled'),
 (102, 230, 2, 'sled'),
 (103, 230, 2, 'sled'),
 (200, 320, 2, 'card'),
 (201, 320, 2, 'card'),
 (202, 320, 2, 'card'),
 (203, 320, 2, 'card'),
 (300, 812, 2, 'fire'),
 (301, 812, 2, 'fire'),
 (302, 812, 3, 'fire'),
 (303, 812, 4, 'fire'),
 (300, 816, 2, 'water'),
 (301, 816, 2, 'water'),
 (302, 816, 3, 'water'),
 (303, 816, 4, 'water');
 
INSERT INTO room_table (id, room_id, game) VALUES
 (205, 220, 'four'),
 (206, 220, 'four'),
 (207, 220, 'four'),
 (200, 221, 'four'),
 (201, 221, 'four'),
 (202, 221, 'four'),
 (203, 221, 'four'),
 (204, 221, 'four'),
 (100, 111, 'mancala'),
 (101, 111, 'mancala'),
 (102, 111, 'mancala'),
 (103, 111, 'mancala'),
 (104, 111, 'mancala'),
 (300, 422, 'treasure'),
 (301, 422, 'treasure'),
 (302, 422, 'treasure'),
 (303, 422, 'treasure'),
 (304, 422, 'treasure'),
 (305, 422, 'treasure'),
 (306, 422, 'treasure'),
 (307, 422, 'treasure');

INSERT INTO quest (id, name) VALUES (1, 'shopping');
INSERT INTO quest (id, name) VALUES (2, 'puffle');
INSERT INTO quest (id, name) VALUES (3, 'igloo');

INSERT INTO quest_award_item (quest_id, item_id) VALUES (1, 24023);
INSERT INTO quest_award_furniture (quest_id, furniture_id) VALUES (3, 2166);
INSERT INTO quest_award_puffle_item (quest_id, puffle_item_id) VALUES (2, 70);

INSERT INTO quest_task (quest_id, description, room_id) VALUES (1, 'Visit the Clothes Shop', 130);
INSERT INTO quest_task (quest_id, description, room_id) VALUES (2, 'Visit the Pet Shop', 310);
INSERT INTO quest_task (quest_id, description) VALUES (3, 'Visit your Igloo');

INSERT INTO redemption_book (id, name) VALUES
 (1, 'The Ultimate Official Guide to Club Penguin'),
 (2, 'Club Penguin Stowaway! Adventures at Sea'),
 (3, 'The Inventor''s Apprentice'),
 (4, 'Secret Agent Handbook'),
 (5, 'Star Reporter'),
 (6, 'The Official Stage Playbook'),
 (7, 'The Great Puffle Switch'),
 (8, 'O Aprendiz De Inventor'),
 (9, 'O Passageiro Clandestino'),
 (10, 'O Guia Oficial Do Club Penguin'),
 (11, 'The Awesome Official Guide to Club Penguin'),
 (12, 'O Super-Repórter'),
 (13, 'Um Dia De Puffle'),
 (14, 'Dancing with Candence'),
 (15, 'Club Penguin: Official Annual 2012'),
 (16, 'The Card-Jitsu Handbook'),
 (17, 'Before Card-Jitsu: The Ninja Quest'),
 (18, 'Dançando com Candence'),
 (19, 'La Fantástica Guía Oficial de Club Penguin'),
 (20, 'Club Penguin: Official Annual 2013'),
 (21, 'The Epic Official Guide to Club Penguin'),
 (22, 'Puffle Whisperer'),
 (23, 'Antes do Dojo: A Jornada Ninja'),
 (24, 'Manual do Desafio Ninja'),
 (25, 'PH''s Great Puffle Search'),
 (26, 'Agent Rookie''s Secret Mission'),
 (27, 'Kit de Montagem: Reconstruindo o Plaza'),
 (28, 'Lo que tu Puffle dice de ti'),
 (29, 'Aprendiz de inventor'),
 (30, 'Bienvenido a la isla'),
 (31, 'Aventuras en alta mar'),
 (32, 'A Grande Expedição dos Puffles'),
 (33, 'A Missão Secreta do Agente Rookie');

 
INSERT INTO redemption_book_word (book_id, page, line, word_number, answer) VALUES
 (1, 9, 15, 4, 'account'),
 (1, 11, 5, 2, 'always'),
 (1, 11, 19, 5, 'birthday'),
 (1, 15, 4, 3, 'click'),
 (1, 15, 10, 5, 'activate'),
 (1, 35, 12, 1, 'misses'),
 (1, 35, 21, 1, 'sound'),
 (1, 37, 6, 6, 'catalog'),
 (1, 38, 8, 3, 'scavenger'),
 (1, 38, 3, 5, 'express'),
 (1, 40, 10, 10, 'examples'),
 (1, 40, 3, 5, 'creating'),
 (1, 60, 6, 3, 'puffle'),
 (1, 61, 13, 1, 'inside'),
 (1, 61, 8, 5, 'decide'),
 (1, 62, 5, 3, 'picture'),
 (1, 63, 11, 3, 'snacks'),
 (1, 63, 7, 4, 'cookies'),
 (1, 64, 9, 3, 'fishing'),
 (1, 64, 1, 8, 'explore'),
 (1, 65, 10, 5, 'storage'),
 (1, 70, 9, 3, 'script'),
 (1, 70, 7, 6, 'theater'),
 (1, 71, 19, 2, 'musical'),
 (1, 71, 13, 7, 'seats'),
 (1, 72, 7, 3, 'snowball'),
 (1, 72, 3, 5, 'space'),
 (1, 73, 4, 2, 'search'),
 (1, 73, 8, 8, 'school'),
 (1, 75, 11, 6, 'comfy'),
 (1, 75, 5, 8, 'tossing'),
 (1, 77, 3, 5, 'buying'),
 (1, 97, 7, 2, 'serious'),
 (1, 97, 11, 3, 'protect'),
 (1, 101, 14, 1, 'light'),
 (1, 101, 8, 4, 'donated'),
 (1, 105, 8, 1, 'platform'),
 (1, 105, 12, 3, 'climb'),
 (1, 106, 6, 4, 'spied'),
 (1, 117, 12, 2, 'uniforms'),
 (1, 117, 3, 8, 'jerseys'),
 (1, 118, 8, 1, 'squads'),
 (1, 118, 4, 4, 'fashion'),
 (1, 131, 3, 5, 'crowd'),
 (1, 131, 9, 8, 'gather'),
 (1, 133, 2, 4, 'controls'),
 (1, 133, 8, 5, 'arrow'),
 (1, 135, 8, 4, 'hidden'),
 (1, 135, 3, 7, 'party'),
 (1, 137, 6, 4, 'document'),
 (1, 137, 8, 4, 'latest'),
 (1, 139, 11, 2, 'button'),
 (1, 139, 5, 5, 'antenna'),
 (1, 140, 7, 4, 'screen'),
 (1, 140, 4, 6, 'breeze'),
 (1, 141, 11, 1, 'combined'),
 (1, 141, 2, 7, 'objects'),
 (1, 143, 18, 2, 'secret'),
 (1, 143, 5, 4, 'simply'),
 (1, 149, 10, 4, 'duplex'),
 (1, 149, 4, 5, 'reveal'),
 (1, 152, 4, 6, 'chance'),
 (1, 153, 12, 3, 'blazing'),
 (1, 153, 7, 4, 'owners'),
 (1, 154, 8, 4, 'arena'),
 (1, 154, 9, 5, 'buddies'),
 (1, 155, 8, 5, 'pumpkin'),
 (1, 155, 2, 6, 'queen'),
 (1, 156, 8, 5, 'small'),
 (1, 156, 6, 7, 'extreme'),
 (1, 159, 4, 1, 'guests'),
 (1, 159, 11, 2, 'chat'),
 (1, 163, 12, 3, 'friend'),
 (1, 164, 6, 7, 'place'),
 (1, 165, 6, 4, 'toolbar'),
 (1, 166, 6, 7, 'emote'),
 (1, 167, 2, 9, 'keys'),
 (1, 169, 16, 1, 'start'),
 (1, 171, 7, 5, 'annual'),
 (1, 172, 10, 2, 'twinkled'),
 (1, 174, 3, 4, 'hovering'),
 (1, 175, 8, 2, 'crown'),
 (1, 176, 9, 1, 'whistle'),
 (1, 177, 20, 2, 'snorkel'),
 (1, 179, 17, 4, 'penguin'),
 (1, 180, 8, 9, 'how'),
 (1, 182, 7, 4, 'actions'),
 (1, 183, 10, 5, 'break'),
 (2, 5, 6, 8, 'telescope'),
 (2, 6, 17, 5, 'mysterious'),
 (2, 7, 13, 2, 'rockhopper'),
 (2, 8, 16, 11, 'furniture'),
 (2, 9, 12, 3, 'hearties'),
 (2, 10, 19, 1, 'friends'),
 (2, 11, 6, 1, 'hide'),
 (2, 12, 6, 1, 'zucchini'),
 (2, 13, 22, 8, 'jungle'),
 (2, 14, 6, 2, 'directions'),
 (2, 15, 13, 8, 'barrels'),
 (2, 16, 12, 2, 'panicked'),
 (2, 17, 17, 4, 'patch'),
 (2, 18, 5, 3, 'swimming'),
 (2, 19, 19, 5, 'positive'),
 (2, 20, 7, 3, 'preserver'),
 (2, 21, 12, 2, 'migrator'),
 (2, 22, 14, 6, 'horizon'),
 (2, 23, 13, 5, 'turtles'),
 (2, 24, 7, 2, 'fuzzy'),
 (2, 25, 20, 3, 'yarr'),
 (2, 26, 9, 2, 'stowaway'),
 (2, 27, 12, 2, 'heart'),
 (2, 28, 10, 3, 'saltwater'),
 (2, 29, 13, 7, 'venture'),
 (2, 30, 6, 2, 'iguanas'),
 (2, 31, 9, 2, 'sparkly'),
 (2, 32, 9, 5, 'quiet'),
 (2, 33, 19, 6, 'picture'),
 (2, 34, 11, 2, 'reward'),
 (2, 36, 16, 3, 'flipper'),
 (2, 37, 19, 3, 'tiptoe'),
 (2, 38, 7, 4, 'anchor'),
 (2, 39, 17, 4, 'branches'),
 (2, 40, 1, 2, 'waterfall'),
 (2, 41, 6, 3, 'onboard'),
 (2, 42, 4, 2, 'porthole'),
 (2, 43, 15, 10, 'green'),
 (2, 44, 18, 5, 'forward'),
 (2, 45, 2, 1, 'rails'),
 (2, 46, 3, 5, 'riddle'),
 (2, 47, 23, 4, 'instrument'),
 (2, 48, 13, 1, 'magnifying'),
 (2, 49, 13, 6, 'balance'),
 (2, 50, 6, 4, 'sand'),
 (2, 51, 8, 3, 'beard'),
 (2, 52, 4, 10, 'cave'),
 (2, 53, 15, 5, 'cookie'),
 (2, 54, 11, 4, 'good'),
 (2, 55, 16, 4, 'matey'),
 (2, 56, 2, 4, 'starboard'),
 (2, 57, 13, 2, 'grateful'),
 (2, 59, 24, 6, 'deck'),
 (2, 60, 12, 5, 'crew'),
 (2, 61, 10, 5, 'stones'),
 (2, 62, 3, 2, 'waddle'),
 (2, 63, 18, 5, 'cheer'),
 (2, 64, 5, 2, 'guide'),
 (2, 65, 14, 4, 'steer'),
 (2, 66, 6, 4, 'excitedly'),
 (2, 67, 6, 6, 'hammock'),
 (2, 68, 13, 2, 'hundreds'),
 (2, 69, 22, 4, 'tempting'),
 (2, 70, 3, 4, 'little'),
 (2, 71, 6, 7, 'mine'),
 (2, 72, 7, 5, 'snow'),
 (2, 73, 13, 3, 'gathered'),
 (2, 74, 16, 9, 'wristwatch'),
 (2, 75, 10, 2, 'legendary'),
 (2, 76, 6, 6, 'ornament'),
 (2, 77, 17, 6, 'leaves'),
 (2, 78, 6, 5, 'thrilled'),
 (2, 79, 15, 4, 'dolphins'),
 (2, 80, 20, 4, 'penguin'),
 (3, 5, 3, 4, 'board'),
 (3, 5, 10, 5, 'good'),
 (3, 5, 8, 7, 'beat'),
 (3, 6, 5, 5, 'printed'),
 (3, 6, 17, 5, 'plans'),
 (3, 6, 5, 6, 'in'),
 (3, 6, 14, 7, 'extreme'),
 (3, 7, 12, 2, 'find'),
 (3, 7, 8, 6, 'yellow'),
 (3, 7, 12, 6, 'staff'),
 (3, 7, 17, 7, 'where'),
 (3, 8, 11, 2, 'smiles'),
 (3, 8, 17, 3, 'shortage'),
 (3, 8, 4, 7, 'round'),
 (3, 10, 5, 1, 'power'),
 (3, 10, 13, 5, 'back'),
 (3, 10, 8, 6, 'rockets'),
 (3, 11, 3, 3, 'penguin'),
 (3, 11, 10, 4, 'grip'),
 (3, 11, 4, 7, 'panic'),
 (3, 13, 2, 4, 'safely'),
 (3, 13, 8, 4, 'land'),
 (3, 13, 15, 6, 'music'),
 (3, 15, 11, 2, 'around'),
 (3, 15, 8, 4, 'only'),
 (3, 15, 6, 5, 'buttons'),
 (3, 17, 14, 4, 'stains'),
 (3, 17, 7, 4, 'wall'),
 (3, 17, 19, 5, 'steaming'),
 (3, 18, 8, 6, 'the'),
 (3, 18, 14, 8, 'of'),
 (3, 21, 6, 5, 'tingle'),
 (3, 21, 4, 6, 'box'),
 (3, 22, 8, 2, 'put'),
 (3, 22, 10, 6, 'igloo'),
 (3, 22, 4, 6, 'reverse'),
 (3, 24, 11, 7, 'anything'),
 (3, 24, 6, 7, 'booth'),
 (3, 24, 14, 8, 'curtain'),
 (3, 25, 4, 1, 'warning'),
 (3, 25, 20, 5, 'probably'),
 (3, 26, 2, 3, 'drawer'),
 (3, 26, 8, 4, 'hang'),
 (3, 26, 4, 5, 'token'),
 (3, 27, 7, 4, 'cheering'),
 (3, 27, 17, 5, 'takes'),
 (3, 27, 5, 7, 'snowball'),
 (3, 28, 7, 5, 'ingredients'),
 (3, 28, 1, 8, 'hops'),
 (3, 28, 4, 9, 'red'),
 (3, 29, 14, 5, 'attach'),
 (3, 30, 9, 4, 'turn'),
 (3, 30, 5, 6, 'time'),
 (3, 31, 11, 1, 'still'),
 (3, 31, 17, 6, 'reach'),
 (3, 31, 5, 6, 'freeze'),
 (3, 32, 7, 3, 'frozen'),
 (3, 32, 17, 7, 'never'),
 (3, 32, 11, 9, 'sport'),
 (3, 35, 15, 3, 'gear'),
 (3, 35, 3, 5, 'across'),
 (3, 35, 9, 5, 'swings'),
 (3, 36, 25, 3, 'moment'),
 (3, 36, 14, 4, 'detour'),
 (3, 36, 9, 7, 'stuffed'),
 (3, 39, 6, 4, 'dangling'),
 (3, 39, 14, 6, 'underneath'),
 (3, 40, 6, 3, 'flip'),
 (3, 40, 18, 5, 'test'),
 (3, 40, 2, 9, 'stick'),
 (3, 41, 7, 2, 'thrilled'),
 (3, 41, 19, 3, 'down'),
 (3, 41, 11, 6, 'graze'),
 (3, 42, 3, 2, 'goes'),
 (3, 42, 10, 4, 'normal'),
 (3, 42, 6, 4, 'burns'),
 (3, 43, 8, 6, 'piano'),
 (3, 43, 2, 7, 'learn'),
 (3, 45, 4, 3, 'slipper'),
 (3, 45, 7, 8, 'crunch'),
 (3, 45, 18, 9, 'dust'),
 (3, 47, 19, 6, 'melt'),
 (3, 49, 19, 8, 'bottom'),
 (3, 51, 18, 5, 'machine'),
 (3, 52, 12, 6, 'though'),
 (3, 53, 20, 5, 'sees'),
 (3, 56, 14, 6, 'cozy'),
 (3, 59, 15, 4, 'interest'),
 (3, 61, 18, 2, 'starts'),
 (3, 62, 13, 9, 'castle'),
 (3, 64, 20, 3, 'broken'),
 (3, 64, 6, 4, 'surges'),
 (3, 65, 16, 6, 'walls'),
 (3, 69, 14, 4, 'manager'),
 (3, 71, 20, 3, 'politely'),
 (3, 72, 17, 4, 'puddle'),
 (3, 74, 18, 6, 'science'),
 (3, 76, 8, 9, 'going'),
 (4, 5, 5, 3, 'fashions'),
 (5, 6, 5, 5, 'printed'),
 (5, 7, 12, 6, 'staff'),
 (5, 8, 19, 2, 'assures'),
 (5, 8, 19, 6, 'reporter'),
 (5, 9, 13, 3, 'penguin'),
 (5, 10, 19, 7, 'games'),
 (5, 10, 9, 9, 'blue'),
 (5, 13, 7, 3, 'bulb'),
 (5, 13, 15, 6, 'play'),
 (5, 14, 6, 1, 'prove'),
 (5, 14, 8, 5, 'corner'),
 (5, 18, 9, 4, 'shout'),
 (5, 19, 4, 7, 'office'),
 (5, 22, 21, 4, 'proud'),
 (5, 22, 6, 6, 'bunch'),
 (5, 22, 13, 9, 'order'),
 (5, 24, 10, 4, 'rush'),
 (5, 24, 3, 8, 'yellow'),
 (5, 28, 1, 7, 'ahead'),
 (5, 28, 4, 7, 'smile'),
 (5, 30, 15, 5, 'walk'),
 (5, 30, 10, 8, 'tip'),
 (5, 31, 3, 2, 'looking'),
 (5, 31, 15, 3, 'really'),
 (5, 34, 17, 8, 'poem'),
 (5, 36, 12, 3, 'over'),
 (5, 36, 6, 4, 'want'),
 (5, 36, 2, 7, 'cool'),
 (5, 38, 19, 3, 'landed'),
 (5, 38, 5, 4, 'flash'),
 (5, 38, 10, 4, 'crowd'),
 (5, 38, 2, 7, 'raises'),
 (5, 39, 3, 6, 'camera'),
 (5, 39, 6, 6, 'her'),
 (5, 41, 18, 8, 'at'),
 (5, 42, 14, 3, 'make'),
 (5, 42, 11, 5, 'turn'),
 (5, 42, 2, 7, 'pool'),
 (5, 42, 18, 8, 'both'),
 (5, 44, 13, 4, 'ruined'),
 (5, 44, 2, 5, 'move'),
 (5, 44, 9, 8, 'tough'),
 (5, 44, 6, 10, 'snow'),
 (5, 46, 1, 1, 'hands'),
 (5, 46, 5, 6, 'igloo'),
 (5, 46, 3, 8, 'flipper'),
 (5, 48, 2, 6, 'have'),
 (5, 50, 11, 2, 'action'),
 (5, 50, 4, 8, 'pearls'),
 (5, 51, 4, 4, 'investigate'),
 (5, 51, 6, 4, 'crew'),
 (5, 53, 15, 3, 'discovery'),
 (5, 53, 1, 6, 'paint'),
 (5, 53, 4, 8, 'hear'),
 (5, 55, 3, 5, 'wiggle'),
 (5, 55, 11, 7, 'believe'),
 (5, 57, 7, 7, 'sled'),
 (5, 60, 8, 5, 'article'),
 (5, 60, 1, 6, 'cookie'),
 (5, 60, 4, 10, 'feed'),
 (5, 61, 10, 5, 'check'),
 (5, 61, 5, 7, 'close'),
 (5, 61, 2, 9, 'changed'),
 (5, 62, 1, 4, 'cheerfully'),
 (5, 62, 3, 5, 'director'),
 (5, 62, 8, 7, 'clumsy'),
 (5, 64, 2, 4, 'fort'),
 (5, 64, 6, 4, 'plan'),
 (5, 64, 10, 4, 'snowball'),
 (5, 64, 16, 6, 'story'),
 (5, 66, 17, 7, 'furniture'),
 (5, 66, 5, 8, 'material'),
 (5, 68, 4, 2, 'quickly'),
 (5, 68, 10, 4, 'phone'),
 (5, 68, 19, 4, 'caller'),
 (5, 68, 13, 8, 'clock'),
 (5, 70, 1, 3, 'slips'),
 (5, 70, 7, 5, 'register'),
 (5, 70, 2, 6, 'clues'),
 (5, 70, 12, 7, 'history'),
 (5, 72, 18, 3, 'regular'),
 (5, 72, 9, 4, 'shakes'),
 (5, 72, 13, 5, 'suggests'),
 (5, 72, 17, 6, 'lever'),
 (5, 75, 2, 1, 'print'),
 (5, 75, 11, 3, 'search'),
 (5, 75, 4, 7, 'chocolate'),
 (5, 75, 8, 8, 'next'),
 (5, 79, 7, 2, 'glance'),
 (5, 79, 20, 3, 'faster'),
 (5, 79, 15, 5, 'promise'),
 (5, 79, 6, 6, 'members'),
 (6, 3, 8, 2, 'great'),
 (6, 3, 12, 2, 'acting'),
 (6, 3, 14, 5, 'flipper'),
 (6, 3, 3, 5, 'bright'),
 (6, 3, 15, 5, 'answer'),
 (6, 3, 5, 7, 'show'),
 (6, 6, 5, 3, 'wonderful'),
 (6, 6, 4, 7, 'work'),
 (6, 7, 8, 5, 'through'),
 (6, 7, 12, 2, 'around'),
 (6, 7, 5, 6, 'penguins'),
 (6, 7, 10, 6, 'hanging'),
 (6, 7, 4, 8, 'special'),
 (6, 8, 4, 4, 'fun'),
 (6, 8, 8, 4, 'director'),
 (6, 8, 9, 7, 'decide'),
 (6, 8, 14, 7, 'script'),
 (6, 8, 18, 9, 'role'),
 (6, 9, 18, 1, 'other'),
 (6, 9, 6, 2, 'music'),
 (6, 9, 9, 3, 'black'),
 (6, 9, 16, 3, 'needs'),
 (6, 9, 2, 5, 'booth'),
 (6, 9, 22, 5, 'actions'),
 (6, 10, 4, 3, 'rehearse'),
 (6, 10, 12, 3, 'outside'),
 (6, 10, 15, 5, 'stars'),
 (6, 10, 6, 7, 'time'),
 (6, 10, 10, 9, 'fancy'),
 (6, 11, 6, 4, 'success'),
 (6, 11, 3, 6, 'your'),
 (6, 11, 9, 8, 'everyone'),
 (6, 12, 5, 1, 'golden'),
 (6, 12, 13, 2, 'the'),
 (6, 12, 9, 3, 'arrived'),
 (6, 12, 15, 3, 'including'),
 (6, 12, 4, 5, 'treasure'),
 (6, 18, 7, 2, 'charming'),
 (6, 18, 3, 5, 'magical'),
 (6, 19, 12, 3, 'causing'),
 (6, 19, 3, 6, 'forest'),
 (6, 19, 7, 7, 'hero'),
 (6, 20, 8, 5, 'jumped'),
 (6, 20, 11, 5, 'teeth'),
 (6, 20, 19, 6, 'flying'),
 (6, 20, 15, 7, 'blew'),
 (6, 20, 5, 8, 'lovely'),
 (6, 21, 9, 4, 'snowball'),
 (6, 21, 19, 5, 'wool'),
 (6, 21, 11, 6, 'some'),
 (6, 21, 2, 8, 'long'),
 (6, 24, 7, 2, 'about'),
 (6, 24, 4, 3, 'careful'),
 (6, 24, 14, 3, 'saying'),
 (6, 25, 6, 4, 'dinosaurs'),
 (6, 25, 10, 9, 'control'),
 (6, 26, 11, 5, 'stuck'),
 (6, 26, 19, 8, 'use'),
 (6, 27, 5, 4, 'wearing'),
 (6, 27, 18, 10, 'ten'),
 (6, 30, 4, 2, 'stop'),
 (6, 30, 22, 2, 'unwrapped'),
 (6, 30, 7, 7, 'loyal'),
 (6, 31, 3, 8, 'statue'),
 (6, 32, 8, 5, 'middle'),
 (6, 32, 3, 8, 'little'),
 (6, 33, 16, 2, 'dark'),
 (6, 33, 3, 4, 'give'),
 (6, 33, 5, 4, 'think'),
 (6, 33, 10, 6, 'chocolate'),
 (6, 36, 19, 3, 'questioned'),
 (6, 36, 6, 4, 'glamorous'),
 (6, 36, 4, 5, 'sniffing'),
 (6, 36, 22, 7, 'solved'),
 (6, 36, 17, 8, 'track'),
 (6, 37, 6, 6, 'hidden'),
 (6, 37, 13, 7, 'very'),
 (6, 37, 5, 8, 'leads'),
 (6, 38, 18, 5, 'here'),
 (6, 38, 10, 7, 'outside'),
 (6, 38, 21, 7, 'answer'),
 (6, 38, 3, 8, 'storm'),
 (6, 38, 19, 8, 'few'),
 (6, 39, 14, 10, 'numbers'),
 (6, 39, 7, 4, 'flowers'),
 (6, 39, 2, 9, 'needed'),
 (6, 42, 5, 2, 'aliens'),
 (6, 42, 8, 2, 'under'),
 (6, 42, 17, 3, 'brains'),
 (6, 42, 4, 4, 'landings'),
 (6, 42, 15, 6, 'friend'),
 (6, 43, 14, 2, 'yellow'),
 (6, 43, 2, 3, 'began'),
 (6, 43, 15, 3, 'flying'),
 (6, 43, 8, 7, 'plan'),
 (6, 44, 10, 4, 'survey'),
 (6, 44, 19, 8, 'rebuild'),
 (6, 45, 3, 5, 'clearing'),
 (6, 45, 7, 6, 'excellent'),
 (6, 45, 14, 7, 'join'),
 (6, 46, 10, 4, 'thrown'),
 (6, 46, 2, 4, 'entry'),
 (6, 46, 14, 4, 'tried'),
 (6, 47, 9, 3, 'own'),
 (6, 47, 18, 4, 'course'),
 (6, 47, 19, 5, 'again'),
 (6, 47, 3, 9, 'spacecraft'),
 (6, 50, 13, 2, 'everything'),
 (6, 50, 16, 2, 'reporter'),
 (6, 50, 8, 3, 'punch'),
 (6, 50, 20, 4, 'battle'),
 (6, 50, 4, 5, 'watch'),
 (6, 51, 5, 3, 'changed'),
 (6, 51, 3, 6, 'eats'),
 (6, 51, 11, 6, 'smashing'),
 (6, 52, 9, 4, 'heroes'),
 (6, 52, 16, 6, 'their'),
 (6, 53, 17, 3, 'city'),
 (6, 53, 5, 5, 'let'),
 (6, 56, 16, 3, 'winner'),
 (6, 56, 12, 3, 'graceful'),
 (6, 56, 4, 4, 'energetic'),
 (6, 56, 8, 5, 'parrot'),
 (6, 57, 10, 1, 'unstoppable'),
 (6, 57, 7, 1, 'sports'),
 (6, 57, 18, 1, 'without'),
 (6, 57, 3, 2, 'rooting'),
 (6, 57, 15, 5, 'ninja'),
 (6, 58, 22, 1, 'skills'),
 (6, 58, 4, 5, 'silly'),
 (6, 62, 25, 7, 'than'),
 (6, 62, 22, 8, 'look'),
 (6, 63, 4, 8, 'alone'),
 (6, 63, 2, 9, 'cool'),
 (6, 64, 4, 1, 'believe'),
 (6, 64, 9, 8, 'cheer'),
 (6, 65, 1, 5, 'sure'),
 (6, 65, 8, 6, 'play'),
 (6, 65, 12, 6, 'shine'),
 (6, 65, 5, 7, 'broken'),
 (6, 66, 20, 9, 'ultra'),
 (6, 67, 16, 7, 'words'),
 (6, 67, 22, 10, 'powerful'),
 (6, 68, 6, 2, 'misses'),
 (6, 72, 7, 3, 'sweet'),
 (6, 74, 6, 3, 'plan'),
 (6, 74, 20, 4, 'empty'),
 (6, 76, 6, 8, 'unique'),
 (6, 77, 3, 4, 'set'),
 (6, 78, 8, 1, 'cave'),
 (6, 78, 3, 4, 'entirely'),
 (6, 79, 4, 1, 'recipe'),
 (6, 79, 7, 2, 'house'),
 (6, 79, 1, 4, 'random'),
 (6, 80, 2, 6, 'applause'),
 (7, 5, 4, 6, 'green'),
 (7, 5, 18, 7, 'right'),
 (7, 6, 21, 5, 'bright'),
 (7, 7, 7, 4, 'anything'),
 (7, 7, 19, 4, 'looks'),
 (7, 8, 8, 2, 'realize'),
 (7, 8, 17, 6, 'zapped'),
 (7, 9, 4, 6, 'cool'),
 (7, 9, 12, 9, 'workshop'),
 (7, 10, 9, 6, 'every'),
 (7, 10, 22, 6, 'door'),
 (7, 11, 2, 2, 'notice'),
 (7, 11, 4, 5, 'grumble'),
 (7, 13, 18, 3, 'head'),
 (7, 13, 10, 5, 'penguin'),
 (7, 16, 11, 4, 'special'),
 (7, 16, 15, 4, 'important'),
 (7, 16, 8, 5, 'lost'),
 (7, 16, 16, 6, 'matter'),
 (7, 18, 16, 5, 'telling'),
 (7, 18, 3, 6, 'inside'),
 (7, 18, 11, 7, 'good'),
 (7, 20, 16, 3, 'fun'),
 (7, 20, 21, 4, 'visit'),
 (7, 24, 13, 4, 'lights'),
 (7, 24, 4, 5, 'measures'),
 (7, 24, 19, 7, 'shyly'),
 (7, 26, 8, 10, 'great'),
 (7, 26, 10, 10, 'bubble'),
 (7, 26, 3, 2, 'pretty'),
 (7, 28, 7, 4, 'rumbling'),
 (7, 28, 11, 5, 'sizzles'),
 (7, 30, 4, 6, 'heart'),
 (7, 30, 23, 9, 'still'),
 (7, 33, 18, 10, 'puffle'),
 (7, 33, 5, 6, 'solve'),
 (7, 34, 16, 6, 'go'),
 (7, 36, 3, 5, 'another'),
 (7, 36, 17, 7, 'examines'),
 (7, 38, 7, 4, 'proud'),
 (7, 38, 23, 6, 'crate'),
 (7, 42, 16, 5, 'four'),
 (7, 42, 3, 6, 'electrical'),
 (7, 43, 5, 2, 'snowballs'),
 (7, 44, 21, 3, 'home'),
 (7, 44, 5, 6, 'fluffy'),
 (7, 44, 13, 6, 'little'),
 (7, 48, 5, 4, 'suggestions'),
 (7, 48, 17, 4, 'hesitate'),
 (7, 48, 12, 5, 'switch'),
 (7, 50, 17, 2, 'and'),
 (7, 50, 7, 5, 'but'),
 (7, 50, 3, 9, 'tempted'),
 (7, 52, 7, 3, 'lasso'),
 (7, 52, 2, 5, 'black'),
 (7, 52, 17, 5, 'mission'),
 (7, 54, 12, 2, 'wait'),
 (7, 54, 19, 4, 'brightens'),
 (7, 54, 4, 6, 'charge'),
 (7, 56, 6, 4, 'shocked'),
 (7, 56, 12, 5, 'went'),
 (7, 56, 18, 5, 'possible'),
 (7, 60, 4, 2, 'turns'),
 (7, 60, 19, 6, 'waves'),
 (7, 61, 3, 4, 'got'),
 (7, 61, 9, 5, 'bouncing'),
 (7, 63, 3, 3, 'waiting'),
 (7, 63, 16, 6, 'feels'),
 (7, 66, 2, 4, 'into'),
 (7, 66, 9, 5, 'where'),
 (7, 66, 15, 5, 'going'),
 (7, 68, 8, 3, 'off'),
 (7, 68, 19, 5, 'welding'),
 (7, 68, 1, 9, 'bobbing'),
 (7, 69, 3, 3, 'metal'),
 (7, 69, 18, 3, 'safe'),
 (7, 69, 9, 5, 'away'),
 (7, 71, 2, 2, 'telescope'),
 (7, 71, 7, 3, 'mighty'),
 (7, 71, 11, 3, 'approaches'),
 (7, 73, 15, 4, 'flight'),
 (7, 73, 11, 5, 'lurches'),
 (7, 73, 6, 7, 'wood'),
 (7, 76, 5, 1, 'desired'),
 (7, 76, 3, 5, 'look'),
 (7, 76, 8, 5, 'will'),
 (7, 76, 13, 5, 'nervous'),
 (7, 79, 8, 2, 'big'),
 (7, 79, 2, 3, 'help'),
 (7, 79, 13, 4, 'whizzes'),
 (7, 79, 5, 8, 'ordinary'),
 (8, 13, 12, 2, 'trancinhas'),
 (8, 21, 4, 1, 'rapidamente'),
 (8, 40, 18, 5, 'tornou'),
 (8, 45, 18, 6, 'grita'),
 (9, 6, 17, 5, 'arruma'),
 (9, 11, 4, 3, 'desconfiar'),
 (9, 12, 6, 2, 'adiante'),
 (9, 15, 13, 3, 'esconder'),
 (9, 16, 12, 5, 'brilhando'),
 (9, 18, 5, 3, 'carregar'),
 (9, 21, 12, 6, 'migrator'),
 (9, 24, 7, 1, 'apresentadora'),
 (9, 27, 12, 2, 'coracao'),
 (9, 30, 6, 2, 'iguanas'),
 (9, 33, 19, 4, 'desenhos'),
 (9, 36, 1, 3, 'fazemos'),
 (9, 46, 3, 4, 'confusa'),
 (9, 47, 23, 6, 'capitao'),
 (9, 49, 13, 6, 'equilibra'),
 (9, 53, 15, 5, 'boceja'),
 (9, 63, 17, 6, 'forma'),
 (9, 68, 13, 5, 'lamenta'),
 (9, 71, 6, 7, 'club'),
 (10, 62, 9, 11, 'pesca'),
 (10, 106, 5, 4, 'originalmente'),
 (10, 117, 12, 5, 'amigos'),
 (10, 118, 9, 1, 'maior'),
 (10, 149, 4, 5, 'trena'),
 (10, 153, 10, 1, 'travessuras'),
 (10, 171, 7, 3, 'abobora'),
 (11, 8, 3, 1, 'share'),
 (11, 9, 4, 1, 'information'),
 (11, 10, 3, 6, 'snowflake'),
 (11, 14, 14, 2, 'circle'),
 (11, 15, 7, 8, 'buddy'),
 (11, 16, 7, 2, 'special'),
 (11, 16, 3, 5, 'sign'),
 (11, 17, 10, 6, 'igloo'),
 (11, 19, 18, 2, 'island'),
 (11, 20, 5, 3, 'making'),
 (11, 22, 5, 4, 'quiz'),
 (11, 23, 2, 1, 'there'),
 (11, 29, 2, 7, 'always'),
 (11, 33, 1, 2, 'what'),
 (11, 36, 4, 6, 'headphones'),
 (11, 37, 3, 4, 'music'),
 (11, 39, 19, 8, 'free'),
 (11, 41, 13, 4, 'correct'),
 (11, 41, 3, 4, 'to'),
 (11, 44, 3, 4, 'upstairs'),
 (11, 47, 7, 4, 'reach'),
 (11, 49, 8, 4, 'instruction'),
 (11, 51, 4, 6, 'catalog'),
 (11, 52, 9, 4, 'available'),
 (11, 53, 3, 4, 'that'),
 (11, 54, 3, 6, 'combine'),
 (11, 57, 3, 4, 'surface'),
 (11, 58, 4, 5, 'fourth'),
 (11, 60, 4, 2, 'hangout'),
 (11, 61, 4, 3, 'pressing'),
 (11, 62, 15, 2, 'stalactites'),
 (11, 63, 5, 4, 'hunting'),
 (11, 65, 3, 6, 'tricky'),
 (11, 69, 8, 6, 'bite'),
 (11, 71, 6, 8, 'care'),
 (11, 72, 6, 2, 'stories'),
 (11, 80, 4, 6, 'clicking'),
 (11, 83, 11, 4, 'food'),
 (11, 89, 8, 5, 'ticket'),
 (11, 92, 6, 3, 'snowball'),
 (11, 93, 5, 3, 'have'),
 (11, 99, 3, 9, 'pizzas'),
 (11, 102, 6, 10, 'creative'),
 (11, 105, 3, 1, 'binoculars'),
 (11, 108, 4, 3, 'pink'),
 (11, 112, 10, 2, 'gears'),
 (11, 115, 5, 2, 'stands'),
 (11, 121, 4, 4, 'catch'),
 (11, 123, 4, 3, 'purchase'),
 (11, 125, 3, 5, 'guitar'),
 (11, 126, 6, 8, 'adventurous'),
 (11, 127, 4, 2, 'switch'),
 (11, 129, 4, 2, 'coin'),
 (11, 131, 11, 2, 'an'),
 (11, 131, 3, 6, 'beside'),
 (11, 133, 7, 2, 'autumn'),
 (11, 139, 7, 7, 'rather'),
 (11, 140, 7, 3, 'background'),
 (11, 141, 6, 5, 'uncover'),
 (11, 143, 5, 2, 'whistles'),
 (11, 144, 4, 4, 'fashion'),
 (11, 147, 7, 5, 'postcard'),
 (11, 153, 5, 7, 'hook'),
 (11, 154, 9, 13, 'castle'),
 (11, 155, 2, 7, 'rock'),
 (11, 161, 14, 10, 'side'),
 (11, 165, 2, 8, 'beautiful'),
 (11, 166, 4, 6, 'earning'),
 (11, 169, 18, 2, 'fortune'),
 (11, 169, 3, 3, 'young'),
 (11, 171, 13, 5, 'opponent'),
 (11, 171, 2, 9, 'highest'),
 (11, 174, 6, 5, 'lava'),
 (11, 177, 5, 2, 'battle'),
 (11, 179, 4, 5, 'popcorn'),
 (11, 180, 4, 6, 'breeze'),
 (11, 181, 6, 7, 'decoding'),
 (11, 183, 13, 6, 'parties'),
 (11, 183, 4, 8, 'agents'),
 (11, 183, 24, 9, 'in'),
 (11, 184, 10, 1, 'pizzatron'),
 (11, 184, 1, 2, 'greatest'),
 (11, 185, 1, 2, 'analyzer'),
 (11, 185, 6, 4, 'gadget'),
 (11, 186, 13, 4, 'to'),
 (11, 186, 3, 4, 'where'),
 (11, 186, 9, 4, 'stalagmite'),
 (11, 186, 6, 5, 'flour'),
 (11, 189, 3, 2, 'section'),
 (11, 190, 19, 10, 'storage'),
 (11, 190, 5, 3, 'your'),
 (11, 192, 6, 5, 'placed'),
 (11, 194, 4, 6, 'chance'),
 (11, 194, 11, 7, 'brightly'),
 (11, 195, 3, 6, 'ninjas'),
 (11, 196, 16, 2, 'loves'),
 (11, 196, 9, 3, 'tables'),
 (11, 197, 8, 3, 'comfy'),
 (11, 197, 2, 4, 'towers'),
 (11, 198, 18, 2, 'decorate'),
 (11, 198, 4, 8, 'whole'),
 (11, 201, 2, 3, 'games'),
 (11, 202, 16, 3, 'access'),
 (11, 202, 2, 9, 'suddenly'),
 (11, 204, 5, 5, 'computer'),
 (11, 205, 11, 5, 'speech'),
 (11, 209, 7, 6, 'advice'),
 (11, 211, 10, 4, 'fuel'),
 (11, 212, 9, 3, 'managers'),
 (11, 213, 9, 6, 'games'),
 (11, 215, 4, 8, 'popular'),
 (11, 216, 12, 9, 'high'),
 (11, 220, 6, 5, 'comic'),
 (11, 227, 10, 4, 'pin'),
 (11, 227, 15, 5, 'yearbooks'),
 (12, 38, 2, 7, 'pronto'),
 (12, 72, 2, 1, 'expresso'),
 (12, 34, 17, 8, 'poema'),
 (12, 38, 19, 3, 'entrou'),
 (13, 5, 3, 4, 'passear'),
 (13, 7, 11, 1, 'gigante'),
 (13, 8, 17, 2, 'verdade'),
 (13, 14, 6, 2, 'palavras'),
 (13, 22, 24, 4, 'fazer'),
 (13, 24, 16, 3, 'amigo'),
 (13, 28, 13, 2, 'raio'),
 (13, 32, 11, 2, 'redondos'),
 (13, 42, 12, 6, 'tempestade'),
 (13, 42, 4, 7, 'eletrica'),
 (13, 50, 7, 3, 'conserta'),
 (13, 68, 12, 3, 'lado'),
 (13, 72, 5, 5, 'relatos'),
 (13, 75, 12, 6, 'troca'),
 (13, 79, 19, 8, 'sera'),
 (13, 80, 9, 1, 'creio'),
 (13, 80, 18, 3, 'concorda'),
 (14, 5, 1, 4, 'please'),
 (14, 5, 9, 5, 'reading'),
 (14, 5, 3, 8, 'favorite'),
 (14, 6, 2, 5, 'paper'),
 (14, 6, 8, 5, 'rest'),
 (14, 7, 3, 5, 'dancing'),
 (14, 7, 13, 5, 'watch'),
 (14, 8, 9, 3, 'come'),
 (14, 8, 6, 5, 'sporty'),
 (14, 10, 2, 3, 'groove'),
 (14, 10, 17, 3, 'happen'),
 (14, 12, 5, 3, 'imagine'),
 (14, 13, 5, 4, 'caught'),
 (14, 15, 3, 2, 'hands'),
 (14, 15, 14, 3, 'really'),
 (14, 16, 3, 1, 'you'),
 (14, 19, 2, 1, 'advanced'),
 (14, 19, 21, 3, 'competing'),
 (14, 19, 11, 4, 'jukebox'),
 (14, 19, 16, 6, 'ninja'),
 (14, 20, 10, 2, 'talented'),
 (14, 20, 2, 5, 'rooftop'),
 (14, 20, 19, 8, 'pink'),
 (14, 21, 5, 7, 'three'),
 (14, 21, 13, 7, 'hurt'),
 (14, 24, 3, 3, 'rhythm'),
 (14, 26, 4, 2, 'without'),
 (14, 26, 15, 6, 'promise'),
 (14, 27, 1, 4, 'envelope'),
 (14, 27, 4, 5, 'clue'),
 (14, 29, 3, 4, 'walk'),
 (14, 29, 8, 5, 'stand'),
 (14, 30, 5, 3, 'grass'),
 (14, 30, 12, 3, 'finish'),
 (14, 30, 1, 7, 'might'),
 (14, 32, 2, 3, 'quickly'),
 (14, 32, 13, 6, 'kitchen'),
 (14, 32, 8, 8, 'movement'),
 (14, 33, 12, 1, 'probably'),
 (14, 33, 1, 5, 'blond'),
 (14, 33, 5, 5, 'excited'),
 (14, 36, 7, 2, 'shake'),
 (14, 36, 2, 4, 'toasty'),
 (14, 39, 16, 3, 'stripes'),
 (14, 39, 7, 6, 'thrill'),
 (14, 41, 15, 1, 'flipper'),
 (14, 41, 2, 2, 'bouncing'),
 (14, 42, 4, 5, 'little'),
 (14, 42, 14, 5, 'floor'),
 (14, 43, 2, 6, 'more'),
 (14, 43, 12, 7, 'help'),
 (14, 45, 14, 3, 'realize'),
 (14, 45, 7, 4, 'yellow'),
 (14, 47, 3, 3, 'plucks'),
 (14, 47, 13, 8, 'curve'),
 (14, 49, 16, 4, 'face'),
 (14, 49, 2, 5, 'long'),
 (14, 51, 1, 7, 'speed'),
 (14, 53, 3, 5, 'notice'),
 (14, 53, 19, 5, 'them'),
 (14, 55, 9, 7, 'idea'),
 (14, 55, 3, 8, 'made'),
 (14, 57, 13, 3, 'could'),
 (14, 57, 5, 6, 'coaching'),
 (14, 61, 1, 2, 'advice'),
 (14, 61, 6, 3, 'remember'),
 (14, 61, 14, 4, 'cry'),
 (14, 61, 20, 4, 'waves'),
 (14, 62, 13, 3, 'sunglasses'),
 (14, 62, 4, 4, 'ten'),
 (14, 64, 2, 3, 'black'),
 (14, 64, 15, 5, 'hops'),
 (14, 65, 15, 10, 'leading'),
 (14, 65, 3, 4, 'starts'),
 (14, 66, 1, 7, 'green'),
 (14, 68, 3, 2, 'round'),
 (14, 68, 7, 2, 'bump'),
 (14, 68, 16, 4, 'music'),
 (14, 68, 13, 6, 'perfect'),
 (14, 68, 1, 8, 'judges'),
 (14, 69, 3, 5, 'win'),
 (14, 69, 14, 5, 'outside'),
 (14, 74, 12, 1, 'blast'),
 (14, 74, 4, 2, 'playing'),
 (14, 74, 9, 4, 'dangerous'),
 (14, 74, 1, 5, 'now'),
 (14, 76, 14, 3, 'toward'),
 (14, 76, 1, 6, 'decide'),
 (14, 76, 8, 6, 'practiced'),
 (14, 78, 12, 1, 'swim'),
 (14, 78, 1, 5, 'gears'),
 (14, 78, 8, 6, 'watery'),
 (14, 80, 5, 2, 'eyes'),
 (14, 80, 1, 4, 'fuel'),
 (14, 80, 10, 6, 'puffle'),
 (15, 4, 4, 4, 'adopt'),
 (15, 4, 3, 7, 'play'),
 (15, 6, 2, 2, 'rockhopper'),
 (15, 6, 4, 14, 'pirates'),
 (15, 7, 3, 2, 'postcard'),
 (15, 8, 2, 7, 'remote'),
 (15, 8, 4, 11, 'answers'),
 (15, 9, 2, 5, 'voyage'),
 (15, 10, 2, 12, 'treasures'),
 (15, 12, 6, 5, 'igloo'),
 (15, 12, 4, 9, 'party'),
 (15, 12, 2, 10, 'penguin'),
 (15, 13, 5, 8, 'clues'),
 (15, 15, 4, 2, 'rhyme'),
 (15, 15, 2, 7, 'limericks'),
 (15, 17, 4, 1, 'newspaper'),
 (15, 17, 5, 3, 'unscramble'),
 (15, 21, 5, 1, 'football'),
 (15, 23, 5, 1, 'bunny'),
 (15, 23, 2, 4, 'claws'),
 (15, 26, 2, 2, 'with'),
 (15, 26, 6, 2, 'penguin'),
 (15, 26, 5, 9, 'equipment'),
 (15, 27, 5, 3, 'imagination'),
 (15, 27, 3, 7, 'pizzatron'),
 (15, 28, 4, 1, 'delicious'),
 (15, 28, 6, 6, 'recipe'),
 (15, 29, 4, 1, 'correctly'),
 (15, 29, 2, 8, 'gadget'),
 (15, 32, 2, 11, 'coffee'),
 (15, 32, 3, 12, 'beans'),
 (15, 34, 2, 5, 'dojo'),
 (15, 35, 3, 7, 'logic'),
 (15, 35, 4, 7, 'puffle'),
 (15, 36, 2, 5, 'sensei'),
 (15, 36, 2, 15, 'lucky'),
 (15, 37, 4, 2, 'pictures'),
 (15, 37, 2, 8, 'powers'),
 (15, 38, 1, 2, 'surprise'),
 (15, 46, 4, 2, 'mixing'),
 (15, 46, 2, 7, 'grooves'),
 (15, 47, 2, 6, 'spin'),
 (15, 48, 2, 3, 'simple'),
 (15, 48, 3, 3, 'dance'),
 (15, 48, 4, 11, 'drawing'),
 (15, 57, 1, 4, 'gleamed'),
 (15, 57, 6, 5, 'machine'),
 (16, 3, 7, 5, 'epic'),
 (16, 3, 3, 9, 'journey'),
 (16, 4, 2, 8, 'footsteps'),
 (16, 5, 1, 3, 'appear'),
 (16, 6, 6, 8, 'snow'),
 (16, 6, 7, 8, 'volcano'),
 (16, 7, 1, 8, 'ancient'),
 (16, 9, 5, 5, 'teach'),
 (16, 10, 3, 10, 'learn'),
 (16, 11, 4, 5, 'strongest'),
 (16, 11, 2, 10, 'battle'),
 (16, 12, 2, 8, 'opponent'),
 (16, 14, 3, 11, 'different'),
 (16, 14, 4, 11, 'blue'),
 (16, 16, 4, 3, 'glow'),
 (16, 16, 6, 12, 'move'),
 (16, 18, 8, 6, 'mighty'),
 (16, 18, 3, 9, 'pebbles'),
 (16, 20, 4, 6, 'earn'),
 (16, 20, 8, 12, 'belts'),
 (16, 22, 7, 5, 'fire'),
 (16, 22, 6, 12, 'maze'),
 (16, 24, 8, 11, 'challenge'),
 (16, 24, 2, 12, 'play'),
 (16, 25, 3, 12, 'pillow'),
 (16, 25, 9, 13, 'white'),
 (16, 26, 3, 1, 'bowing'),
 (16, 26, 7, 7, 'tease'),
 (16, 26, 3, 9, 'respect'),
 (16, 30, 5, 1, 'element'),
 (16, 30, 3, 8, 'master'),
 (16, 34, 4, 3, 'trials'),
 (16, 34, 2, 8, 'proud'),
 (16, 36, 5, 1, 'card'),
 (16, 36, 3, 4, 'freezes'),
 (16, 36, 4, 12, 'predict'),
 (16, 38, 10, 4, 'difficult'),
 (16, 40, 2, 6, 'mighty'),
 (16, 40, 7, 7, 'knowledge'),
 (16, 40, 5, 10, 'examine'),
 (16, 42, 4, 2, 'starter'),
 (16, 42, 4, 3, 'deck'),
 (16, 42, 7, 7, 'snowball'),
 (16, 46, 2, 3, 'impressed'),
 (16, 46, 4, 7, 'confident'),
 (16, 46, 6, 10, 'memory'),
 (16, 48, 3, 7, 'unlock'),
 (16, 48, 2, 9, 'collect'),
 (16, 48, 5, 13, 'ninja'),
 (16, 50, 3, 1, 'end'),
 (16, 50, 2, 3, 'water'),
 (16, 52, 2, 7, 'chill'),
 (16, 54, 2, 1, 'bundle'),
 (16, 54, 3, 1, 'forecast'),
 (16, 56, 3, 1, 'next'),
 (16, 56, 5, 1, 'pink'),
 (16, 56, 6, 6, 'logic'),
 (16, 58, 2, 11, 'defeated'),
 (16, 62, 2, 10, 'stamps'),
 (16, 62, 2, 12, 'playing'),
 (16, 66, 6, 2, 'haikus'),
 (16, 66, 4, 8, 'tackle'),
 (16, 66, 5, 8, 'puzzle'),
 (16, 66, 2, 9, 'bravely'),
 (16, 68, 6, 2, 'pride'),
 (16, 68, 7, 2, 'goal'),
 (16, 68, 4, 7, 'thrill'),
 (16, 68, 2, 8, 'match'),
 (16, 72, 3, 2, 'giant'),
 (16, 72, 2, 4, 'sense'),
 (16, 72, 2, 5, 'power'),
 (16, 72, 3, 6, 'pretty'),
 (16, 72, 6, 6, 'walk'),
 (16, 72, 2, 7, 'determination'),
 (16, 72, 3, 10, ''),
 (16, 74, 3, 8, 'day'),
 (16, 74, 2, 10, 'skills'),
 (16, 74, 3, 11, 'worthy'),
 (16, 76, 4, 2, 'sensei'),
 (16, 76, 6, 2, 'practicing'),
 (16, 76, 4, 4, 'guide'),
 (16, 76, 5, 7, 'mountains'),
 (16, 76, 2, 9, 'experience'),
 (16, 76, 3, 9, 'intense'),
 (17, 5, 3, 1, 'flippers'),
 (17, 5, 6, 2, 'boots'),
 (17, 5, 2, 3, 'crunches'),
 (17, 5, 14, 4, 'whispers'),
 (17, 5, 5, 6, 'warm'),
 (17, 5, 9, 8, 'map'),
 (17, 6, 11, 1, 'smoke'),
 (17, 6, 19, 2, 'pond'),
 (17, 6, 2, 4, 'unusual'),
 (17, 6, 2, 7, 'sky'),
 (17, 7, 18, 1, 'master'),
 (17, 7, 6, 4, 'waddling'),
 (17, 7, 11, 4, 'reading'),
 (17, 8, 9, 1, 'smells'),
 (17, 8, 1, 2, 'heart'),
 (17, 8, 5, 4, 'teapot'),
 (17, 8, 8, 6, 'cup'),
 (17, 8, 6, 10, 'ground'),
 (17, 9, 2, 1, 'location'),
 (17, 9, 8, 1, 'natural'),
 (17, 9, 5, 2, 'unrolls'),
 (17, 9, 5, 4, 'paper'),
 (17, 10, 7, 2, 'water'),
 (17, 10, 4, 4, 'decide'),
 (17, 10, 8, 7, 'narrow'),
 (17, 11, 2, 4, 'pleased'),
 (17, 11, 6, 7, 'enjoy'),
 (17, 11, 4, 8, 'say'),
 (17, 12, 2, 2, 'believes'),
 (17, 13, 8, 1, 'climb'),
 (17, 13, 2, 3, 'rope'),
 (17, 13, 5, 3, 'hook'),
 (17, 14, 1, 3, 'slides'),
 (17, 14, 2, 4, 'swinging'),
 (17, 14, 11, 4, 'mountain'),
 (17, 15, 2, 2, 'motorboat'),
 (17, 15, 4, 3, 'slow'),
 (17, 15, 5, 5, 'riverbank'),
 (17, 15, 2, 6, 'journey'),
 (17, 16, 2, 2, 'squint'),
 (17, 16, 5, 2, 'love'),
 (17, 16, 2, 5, 'sunlight'),
 (17, 16, 4, 7, 'puffle'),
 (17, 17, 3, 1, 'penguins'),
 (17, 17, 2, 3, 'treasure'),
 (17, 18, 3, 2, 'emerald'),
 (17, 18, 2, 5, 'gold'),
 (17, 20, 4, 1, 'gold'),
 (17, 20, 9, 1, 'melt'),
 (17, 20, 3, 5, 'massive'),
 (17, 21, 5, 2, 'black'),
 (17, 22, 4, 5, 'random'),
 (17, 22, 3, 8, 'pack'),
 (17, 23, 3, 3, 'anxious'),
 (17, 23, 11, 4, 'stumble'),
 (17, 23, 7, 7, 'shimmy'),
 (17, 24, 3, 10, 'hurry'),
 (17, 25, 8, 3, 'tunnel'),
 (17, 25, 3, 4, 'whip'),
 (17, 25, 2, 6, 'snowballs'),
 (17, 26, 4, 1, 'pretending'),
 (17, 27, 4, 1, 'ledge'),
 (17, 28, 4, 3, 'path'),
 (17, 28, 2, 5, 'volcano'),
 (17, 29, 2, 5, 'patience'),
 (17, 30, 6, 1, 'wooden'),
 (17, 30, 4, 6, 'flashlight'),
 (17, 31, 4, 2, 'ninja'),
 (17, 32, 5, 1, 'chance'),
 (17, 32, 2, 9, 'interested'),
 (17, 33, 2, 4, 'blasting'),
 (17, 34, 1, 6, 'hot'),
 (17, 36, 5, 9, 'backpack'),
 (17, 37, 5, 1, 'farther'),
 (17, 37, 4, 2, 'lighting'),
 (17, 39, 3, 7, 'frozen'),
 (17, 40, 2, 1, 'construction'),
 (17, 40, 9, 1, 'deep'),
 (17, 46, 3, 3, 'notice'),
 (17, 48, 2, 2, 'spring'),
 (17, 48, 3, 3, 'carefully'),
 (17, 52, 2, 5, 'hairy'),
 (17, 54, 4, 7, 'gives'),
 (17, 55, 2, 9, 'rocky'),
 (17, 56, 11, 2, 'plummet'),
 (17, 58, 5, 1, 'strong'),
 (17, 58, 1, 6, 'mistaken'),
 (17, 62, 7, 2, 'would'),
 (17, 64, 4, 6, 'perfect'),
 (17, 64, 5, 7, 'news'),
 (17, 66, 4, 2, 'photo'),
 (17, 68, 8, 2, 'flowing'),
 (17, 70, 8, 4, 'searching'),
 (17, 72, 2, 6, 'shine'),
 (17, 75, 1, 3, 'dry'),
 (17, 75, 5, 8, 'reflective'),
 (17, 77, 4, 4, 'easier'),
 (17, 78, 5, 4, 'rumbling'),
 (17, 79, 5, 8, 'feet'),
 (17, 79, 2, 2, 'avalanche'),
 (20, 4, 4, 1, 'citizens'),
 (20, 4, 12, 2, 'invitation'),
 (20, 4, 7, 3, 'protects'),
 (20, 4, 2, 8, 'confidential'),
 (20, 4, 14, 12, 'stealth'),
 (20, 5, 8, 3, 'plans'),
 (20, 8, 8, 2, 'best'),
 (20, 8, 5, 11, 'help'),
 (20, 8, 7, 11, 'abilities'),
 (20, 8, 13, 11, 'gadgets'),
 (20, 16, 17, 3, 'jet'),
 (20, 16, 3, 3, 'cool'),
 (20, 16, 8, 6, 'assignment'),
 (20, 16, 12, 8, 'sunglasses'),
 (20, 17, 7, 2, 'beak'),
 (20, 17, 28, 4, 'bear'),
 (20, 17, 30, 5, 'pesky'),
 (20, 17, 2, 7, 'evidence'),
 (20, 18, 7, 1, 'land'),
 (20, 18, 25, 3, 'loose'),
 (20, 18, 12, 4, 'flashlight'),
 (20, 18, 9, 4, 'investigating'),
 (20, 18, 15, 6, 'cart'),
 (20, 19, 26, 3, 'suit'),
 (20, 19, 12, 5, 'track'),
 (20, 20, 6, 2, 'secret'),
 (20, 20, 9, 2, 'pizzatron'),
 (20, 20, 11, 4, 'claw'),
 (20, 20, 27, 4, 'games'),
 (20, 20, 20, 5, 'spy'),
 (20, 20, 12, 6, 'cannons'),
 (20, 20, 5, 8, 'cave'),
 (20, 21, 15, 1, 'discovered'),
 (20, 21, 6, 1, 'invention'),
 (20, 21, 10, 2, 'technology'),
 (20, 21, 17, 2, 'grinned'),
 (20, 21, 22, 2, 'team'),
 (20, 21, 4, 2, 'blue'),
 (20, 28, 2, 3, 'solve'),
 (20, 28, 2, 5, 'tricky'),
 (20, 30, 3, 5, 'code'),
 (20, 32, 4, 1, 'costume'),
 (20, 36, 5, 3, 'sparkly'),
 (20, 36, 29, 4, 'character'),
 (20, 36, 8, 4, 'script'),
 (20, 36, 26, 6, 'stage'),
 (20, 36, 9, 9, 'play'),
 (20, 37, 16, 5, 'leads'),
 (20, 37, 24, 5, 'hat'),
 (20, 37, 27, 6, 'surprise'),
 (20, 37, 12, 7, 'story'),
 (20, 38, 14, 3, 'ruby'),
 (20, 38, 30, 4, 'actors'),
 (20, 38, 32, 4, 'laughing'),
 (20, 38, 16, 8, 'lobby'),
 (20, 39, 2, 2, 'job'),
 (20, 39, 5, 3, 'headquarters'),
 (20, 39, 25, 4, 'mystery'),
 (20, 46, 12, 1, 'drawings'),
 (20, 46, 13, 1, 'happy'),
 (20, 46, 19, 6, 'missions'),
 (20, 48, 3, 6, 'investigate'),
 (20, 49, 3, 9, 'jellybean'),
 (20, 50, 30, 1, 'snowball'),
 (20, 50, 4, 1, 'together'),
 (20, 50, 5, 1, 'pizza'),
 (20, 50, 10, 10, 'propeller'),
 (20, 50, 39, 10, 'special'),
 (20, 50, 18, 3, 'aprons'),
 (20, 50, 29, 4, 'brown'),
 (20, 50, 17, 6, 'couches'),
 (20, 50, 18, 7, 'cocoa'),
 (20, 50, 33, 7, 't-shirt'),
 (20, 50, 24, 8, 'beans'),
 (20, 51, 12, 10, 'safety'),
 (20, 51, 7, 2, 'pink'),
 (20, 51, 7, 3, 'penguin'),
 (20, 51, 3, 5, 'signal'),
 (20, 52, 25, 1, 'electronic'),
 (20, 52, 2, 3, 'device'),
 (20, 52, 36, 9, 'skating'),
 (20, 53, 5, 11, 'plan'),
 (20, 53, 14, 2, 'room'),
 (20, 53, 11, 5, 'agent'),
 (20, 53, 18, 7, 'ray'),
 (20, 53, 20, 8, 'computer'),
 (20, 53, 2, 9, 'flying'),
 (20, 54, 10, 9, 'blast'),
 (20, 55, 10, 1, 'communication'),
 (20, 55, 39, 5, 'flippers'),
 (20, 58, 3, 14, 'quiz'),
 (20, 58, 2, 5, 'book'),
 (20, 59, 6, 11, 'strange'),
 (20, 59, 3, 3, 'challenges'),
 (21, 6, 2, 3, 'reader'),
 (21, 6, 4, 3, 'extraordinary'),
 (21, 6, 6, 4, 'journey'),
 (21, 6, 8, 4, 'secrets'),
 (21, 9, 3, 5, 'exploring'),
 (21, 10, 2, 3, 'grabber'),
 (21, 11, 4, 1, 'catalog'),
 (21, 11, 5, 3, 'special'),
 (21, 11, 11, 3, 'friend'),
 (21, 11, 8, 5, 'puffle'),
 (21, 11, 9, 6, 'moments'),
 (21, 12, 7, 2, 'password'),
 (21, 12, 8, 2, 'igloo'),
 (21, 12, 3, 3, 'online'),
 (21, 14, 6, 1, 'personality'),
 (21, 14, 13, 1, 'snowball'),
 (21, 14, 2, 2, 'penguins'),
 (21, 14, 7, 3, 'favorite'),
 (21, 14, 9, 3, 'clothing'),
 (21, 14, 10, 3, 'background'),
 (21, 14, 14, 3, 'friendly'),
 (21, 14, 12, 4, 'dancing'),
 (21, 14, 3, 5, 'snowflake'),
 (21, 14, 13, 6, 'playful'),
 (21, 14, 14, 6, 'helping'),
 (21, 15, 6, 2, 'internet'),
 (21, 15, 10, 2, 'sports'),
 (21, 15, 6, 3, 'safety'),
 (21, 15, 7, 3, 'imagination'),
 (21, 15, 3, 5, 'forever'),
 (21, 15, 2, 6, 'important'),
 (21, 15, 12, 6, 'numbers'),
 (21, 15, 13, 6, 'different'),
 (21, 15, 9, 8, 'animal'),
 (21, 17, 7, 1, 'games'),
 (21, 17, 1, 7, 'unique'),
 (21, 17, 1, 8, 'style'),
 (21, 18, 8, 2, 'here'),
 (21, 18, 13, 4, 'screen'),
 (21, 18, 2, 6, 'helpful'),
 (21, 18, 15, 9, 'messages'),
 (21, 18, 17, 9, 'chatting'),
 (21, 18, 12, 10, 'target'),
 (21, 18, 10, 11, 'make'),
 (21, 18, 2, 13, 'toolbar'),
 (21, 19, 10, 4, 'music'),
 (21, 19, 1, 5, 'member'),
 (21, 19, 1, 9, 'badge'),
 (21, 20, 3, 1, 'opportunities'),
 (21, 20, 8, 1, 'discover'),
 (21, 20, 14, 1, 'backstage'),
 (21, 20, 16, 1, 'celebrate'),
 (21, 20, 4, 2, 'exclusive'),
 (21, 20, 12, 3, 'puffles'),
 (21, 20, 4, 4, 'parties'),
 (21, 20, 1, 5, 'membership'),
 (21, 20, 7, 5, 'decorate'),
 (21, 20, 3, 6, 'customize'),
 (21, 21, 2, 2, 'waddle'),
 (21, 21, 3, 10, 'island'),
 (21, 31, 1, 2, 'enjoy'),
 (21, 31, 4, 2, 'cheer'),
 (21, 31, 1, 5, 'games'),
 (21, 31, 6, 7, 'action'),
 (21, 31, 1, 8, 'friends'),
 (21, 32, 1, 2, 'gallery'),
 (21, 34, 3, 1, 'speakers'),
 (21, 34, 4, 1, 'headphones'),
 (21, 34, 9, 1, 'arcade'),
 (21, 34, 10, 1, 'puffle'),
 (21, 34, 7, 4, 'challenge'),
 (21, 34, 6, 7, 'dancing'),
 (21, 34, 3, 9, 'dance'),
 (21, 34, 4, 11, 'funky'),
 (21, 34, 10, 11, 'groove'),
 (21, 52, 10, 1, 'lifeguard'),
 (21, 52, 15, 3, 'backgrounds'),
 (21, 52, 6, 7, 'seasonal'),
 (21, 52, 8, 8, 'different'),
 (21, 53, 2, 2, 'catalog'),
 (21, 53, 1, 4, 'hidden'),
 (21, 58, 2, 1, 'underground'),
 (21, 58, 5, 1, 'hangout'),
 (21, 58, 9, 1, 'costumes'),
 (21, 58, 9, 6, 'snorkels'),
 (21, 58, 5, 7, 'flippers'),
 (21, 58, 13, 8, 'splash'),
 (21, 58, 4, 9, 'underwater'),
 (21, 58, 8, 10, 'mermaid'),
 (21, 59, 1, 10, 'relax'),
 (21, 60, 6, 1, 'bravery'),
 (21, 60, 18, 7, 'trivia'),
 (21, 60, 4, 9, 'adventure'),
 (21, 64, 3, 4, 'community'),
 (21, 65, 1, 8, 'mysterious'),
 (21, 65, 2, 8, 'up'),
 (21, 70, 2, 5, 'treasure'),
 (21, 72, 5, 6, 'performer'),
 (21, 76, 1, 4, 'discovery'),
 (22, 6, 5, 5, 'igloo'),
 (22, 6, 2, 11, 'puffles'),
 (22, 7, 5, 1, 'purple'),
 (22, 8, 6, 2, 'coins'),
 (22, 8, 2, 8, 'adopt'),
 (22, 9, 3, 8, 'cookies'),
 (22, 9, 5, 9, 'princess'),
 (22, 10, 5, 3, 'fun'),
 (22, 11, 4, 1, 'penguins'),
 (22, 11, 8, 3, 'seaweed'),
 (22, 12, 11, 2, 'skate'),
 (22, 13, 8, 2, 'pink'),
 (22, 13, 4, 4, 'chariot'),
 (22, 14, 5, 3, 'jester'),
 (22, 15, 9, 6, 'dancers'),
 (22, 16, 2, 3, 'imagine'),
 (22, 16, 19, 4, 'colours'),
 (22, 16, 24, 4, 'yellow'),
 (22, 16, 5, 5, 'furry'),
 (22, 18, 3, 2, 'friend'),
 (22, 19, 2, 3, 'play'),
 (22, 19, 8, 3, 'ball'),
 (22, 20, 10, 5, 'happy'),
 (22, 20, 2, 5, 'companions'),
 (22, 21, 2, 4, 'furniture'),
 (22, 22, 8, 4, 'treasure'),
 (22, 23, 16, 1, 'bubbles'),
 (22, 23, 12, 8, 'items'),
 (22, 24, 7, 4, 'snacks'),
 (22, 25, 16, 1, 'funny'),
 (22, 25, 4, 2, 'unicycle'),
 (22, 26, 4, 7, 'game'),
 (22, 27, 4, 8, 'coffee'),
 (22, 27, 3, 10, 'fly'),
 (22, 29, 14, 5, 'cool'),
 (22, 29, 3, 6, 'skateboard'),
 (22, 30, 2, 7, 'mine'),
 (22, 32, 10, 1, 'chocolate'),
 (22, 33, 9, 3, 'disco'),
 (22, 35, 3, 4, 'beat'),
 (22, 35, 6, 8, 'waddle'),
 (22, 36, 4, 5, 'surfing'),
 (22, 36, 10, 5, 'pizza'),
 (22, 37, 3, 7, 'pins'),
 (22, 38, 17, 1, 'tricks'),
 (22, 40, 4, 6, 'song'),
 (22, 41, 13, 8, 'music'),
 (22, 42, 5, 11, 'grooving'),
 (22, 44, 4, 3, 'powerful'),
 (22, 44, 7, 8, 'frozen'),
 (22, 45, 4, 2, 'creature'),
 (22, 45, 3, 3, 'snow'),
 (22, 46, 4, 2, 'excitement'),
 (22, 47, 2, 2, 'zany'),
 (22, 47, 14, 2, 'wacky'),
 (22, 48, 3, 1, 'science'),
 (22, 49, 9, 9, 'goggles'),
 (22, 50, 11, 11, 'landscape'),
 (22, 51, 3, 5, 'level'),
 (22, 52, 14, 2, 'treats'),
 (22, 53, 3, 2, 'balloons'),
 (22, 55, 9, 13, 'squid'),
 (22, 57, 6, 4, 'maze'),
 (22, 59, 10, 5, 'earn'),
 (22, 60, 3, 8, 'island'),
 (22, 61, 7, 1, 'missions'),
 (22, 64, 7, 7, 'mysterious'),
 (22, 65, 5, 6, 'book'),
 (22, 66, 3, 4, 'training'),
 (22, 68, 3, 6, 'snowballs'),
 (22, 71, 7, 7, 'agent'),
 (22, 72, 2, 5, 'unique'),
 (22, 73, 10, 3, 'party'),
 (22, 73, 3, 6, 'smile'),
 (22, 74, 2, 6, 'dress'),
 (22, 76, 6, 2, 'draw'),
 (22, 77, 1, 4, 'cute'),
 (22, 78, 11, 4, 'cheese'),
 (22, 79, 7, 1, 'personalities'),
 (22, 80, 4, 3, 'journey'),
 (22, 80, 11, 11, 'expert'),
 (25, 5, 5, 1, 'friend'),
 (25, 7, 2, 2, 'discovered'),
 (25, 8, 1, 6, 'cheerful'),
 (25, 9, 4, 1, 'starstruck'),
 (25, 10, 3, 11, 'expedition'),
 (25, 11, 2, 8, 'awesome'),
 (25, 11, 3, 9, 'secret'),
 (25, 15, 5, 7, 'horizon'),
 (25, 21, 3, 2, 'dance'),
 (25, 22, 17, 2, 'dancing'),
 (25, 22, 1, 11, 'igloo'),
 (25, 23, 5, 6, 'wonder'),
 (25, 24, 4, 8, 'rowboat'),
 (25, 27, 6, 7, 'whistle'),
 (25, 28, 6, 7, 'waddle'),
 (25, 29, 2, 3, 'habitat'),
 (25, 30, 8, 1, 'adventure'),
 (25, 30, 16, 2, 'expert'),
 (25, 30, 10, 3, 'learning'),
 (25, 30, 12, 3, 'studying'),
 (25, 30, 14, 9, 'smile'),
 (25, 31, 22, 2, 'traveling'),
 (25, 31, 10, 3, 'snow'),
 (25, 35, 5, 4, 'penguins'),
 (25, 35, 2, 5, 'iceberg'),
 (25, 37, 2, 1, 'evergreen'),
 (25, 37, 4, 2, 'celebrations'),
 (25, 37, 7, 2, 'discovery'),
 (25, 37, 6, 4, 'exciting'),
 (25, 37, 3, 7, 'parties'),
 (25, 38, 6, 5, 'flipper'),
 (25, 40, 4, 1, 'funny'),
 (25, 41, 6, 4, 'pirate'),
 (25, 41, 3, 5, 'lighthouse'),
 (25, 41, 4, 5, 'migrator'),
 (25, 41, 8, 5, 'rockhopper'),
 (25, 42, 4, 4, 'snowballs'),
 (25, 43, 13, 1, 'costume'),
 (25, 43, 5, 9, 'bridge'),
 (25, 44, 1, 5, 'arcade'),
 (25, 47, 6, 2, 'fascinating'),
 (25, 48, 2, 5, 'cannon'),
 (25, 49, 3, 9, 'bubble'),
 (25, 50, 6, 4, 'exercise'),
 (25, 50, 1, 5, 'sports'),
 (25, 50, 4, 8, 'three'),
 (25, 50, 6, 8, 'outdoor'),
 (25, 51, 4, 3, 'yarr'),
 (25, 51, 5, 4, 'stadium'),
 (25, 52, 6, 1, 'friendly'),
 (25, 52, 5, 3, 'playful'),
 (25, 52, 6, 3, 'loyal'),
 (25, 52, 12, 3, 'arctict'),
 (25, 53, 9, 4, 'trampoline'),
 (25, 53, 22, 5, 'inventive'),
 (25, 53, 19, 9, 'safety'),
 (25, 54, 1, 3, 'thoughtful'),
 (25, 54, 3, 6, 'personality'),
 (25, 56, 10, 1, 'hockey'),
 (25, 56, 4, 2, 'skating'),
 (25, 57, 2, 4, 'puffle'),
 (25, 57, 8, 6, 'imagine'),
 (25, 57, 13, 6, 'waddling'),
 (25, 57, 6, 8, 'skateboarding'),
 (25, 60, 9, 9, 'study'),
 (25, 61, 1, 7, 'berries'),
 (25, 62, 23, 1, 'wilderness'),
 (25, 63, 1, 3, 'interesting'),
 (25, 67, 4, 9, 'discovery'),
 (25, 67, 3, 11, 'puffle'),
 (25, 68, 2, 2, 'curious'),
 (25, 68, 11, 8, 'helmet'),
 (25, 69, 12, 5, 'happy'),
 (25, 71, 14, 1, 'pizza'),
 (25, 71, 7, 2, 'music'),
 (25, 71, 15, 4, 'piano'),
 (25, 71, 2, 5, 'creative'),
 (25, 71, 4, 6, 'inventing'),
 (25, 71, 8, 6, 'onstage'),
 (25, 71, 6, 7, 'paint'),
 (25, 72, 4, 4, 'three'),
 (25, 72, 3, 7, 'grin'),
 (25, 73, 3, 2, 'doesn’t'),
 (25, 73, 5, 7, 'mountains'),
 (25, 74, 2, 1, 'sunglasses'),
 (25, 74, 4, 8, 'epf'),
 (25, 76, 2, 5, 'path'),
 (25, 77, 9, 1, 'agents'),
 (25, 77, 11, 1, 'machine'),
 (25, 77, 17, 4, 'herbert'),
 (25, 77, 6, 5, 'spy'),
 (25, 77, 15, 7, 'backpack'),
 (25, 77, 14, 8, 'together'),
 (25, 78, 3, 1, 'headquarters'),
 (25, 78, 18, 2, 'elite'),
 (25, 78, 4, 3, 'computer'),
 (25, 78, 4, 7, 'activity'),
 (25, 78, 12, 8, 'reward'),
 (25, 80, 9, 3, 'great'),
 (25, 80, 16, 3, 'cheer'),
 (26, 5, 23, 3, 'sunglasses'),
 (26, 5, 5, 4, 'rockhopper'),
 (26, 5, 13, 6, 'pirate'),
 (26, 5, 1, 7, 'snow'),
 (26, 5, 8, 7, 'friends'),
 (26, 6, 8, 1, 'rookie'),
 (26, 6, 2, 2, 'penguin'),
 (26, 6, 3, 5, 'wonder'),
 (26, 6, 19, 6, 'enthusiasm'),
 (26, 6, 8, 7, 'herbert'),
 (26, 7, 15, 1, 'adventure'),
 (26, 7, 11, 5, 'reading'),
 (26, 7, 9, 7, 'superstar'),
 (26, 8, 6, 1, 'puffle'),
 (26, 8, 4, 2, 'excitement'),
 (26, 8, 4, 6, 'super'),
 (26, 9, 7, 1, 'traveler'),
 (26, 9, 10, 1, 'lighthouse'),
 (26, 9, 6, 3, 'gary'),
 (26, 9, 2, 4, 'waddle'),
 (26, 10, 2, 4, 'stage'),
 (26, 11, 10, 7, 'triumphantly'),
 (26, 12, 4, 4, 'wish'),
 (26, 13, 2, 1, 'pizza'),
 (26, 14, 7, 3, 'spy'),
 (26, 14, 10, 4, 'flipper'),
 (26, 16, 6, 8, 'tree'),
 (26, 18, 1, 1, 'underwater'),
 (26, 18, 5, 6, 'paddle'),
 (26, 18, 1, 7, 'jet'),
 (26, 19, 2, 1, 'cheer'),
 (26, 19, 3, 7, 'town'),
 (26, 19, 2, 8, 'great'),
 (26, 20, 2, 5, 'confidence'),
 (26, 21, 5, 1, 'victory'),
 (26, 21, 2, 2, 'bravely'),
 (26, 22, 4, 1, 'jump'),
 (26, 22, 4, 4, 'swim'),
 (26, 22, 12, 4, 'gold'),
 (26, 22, 10, 7, 'sunlight'),
 (26, 24, 3, 5, 'surprise'),
 (26, 24, 4, 7, 'iceberg'),
 (26, 24, 6, 7, 'ruby'),
 (26, 25, 9, 2, 'races'),
 (26, 26, 1, 4, 'binoculars'),
 (26, 26, 2, 9, 'shiny'),
 (26, 27, 3, 2, 'inventions'),
 (26, 27, 1, 3, 'gadgets'),
 (26, 28, 1, 1, 'perfect'),
 (26, 29, 2, 5, 'apple'),
 (26, 30, 3, 1, 'mission'),
 (26, 31, 6, 2, 'awesome'),
 (26, 31, 2, 4, 'coins'),
 (26, 31, 2, 7, 'secret'),
 (26, 32, 4, 1, 'location'),
 (26, 33, 3, 1, 'waddle'),
 (26, 33, 6, 2, 'waves'),
 (26, 34, 16, 1, 'island'),
 (26, 35, 2, 3, 'music'),
 (26, 35, 1, 6, 'message'),
 (26, 35, 5, 6, 'play'),
 (26, 41, 8, 1, 'vintage'),
 (26, 41, 3, 5, 'plaza'),
 (26, 45, 4, 1, 'splash'),
 (26, 45, 5, 3, 'sparks'),
 (26, 47, 8, 7, 'baseball'),
 (26, 49, 3, 2, 'try'),
 (26, 50, 9, 2, 'happy'),
 (26, 50, 12, 5, 'hero'),
 (26, 50, 6, 6, 'newspaper'),
 (26, 51, 5, 3, 'mountains'),
 (26, 53, 5, 1, 'thoughtw'),
 (26, 53, 7, 2, 'waddle'),
 (26, 54, 3, 6, 'favorite'),
 (26, 54, 3, 7, 'plays'),
 (26, 55, 3, 9, 'ocean'),
 (26, 56, 3, 3, 'starfish'),
 (26, 60, 6, 5, 'cheerfully'),
 (26, 61, 9, 4, 'surf'),
 (26, 62, 2, 2, 'colorful'),
 (26, 62, 2, 3, 'surfboards'),
 (26, 63, 2, 5, 'mysterious'),
 (26, 64, 4, 1, 'backstage'),
 (26, 68, 2, 4, 'clues'),
 (26, 69, 2, 4, 'cove'),
 (26, 70, 7, 1, 'animals'),
 (26, 70, 7, 3, 'environments'),
 (26, 70, 7, 7, 'world'),
 (26, 70, 4, 10, 'cool'),
 (26, 74, 2, 4, 'agent'),
 (26, 74, 3, 7, 'beak'),
 (26, 75, 7, 8, 'dot'),
 (26, 77, 10, 4, 'front'),
 (26, 78, 11, 3, 'sunshine'),
 (26, 78, 8, 4, 'race');
 
INSERT INTO dance_song (id, name, song_length_millis, song_length, millis_per_bar) VALUES
 (0, 'Penguin Band Boogie', 123000, 241, 2000),
 (1, 'The Generic Way', 117000, 221, 2070),
 (2, 'Epic Win', 124000, 176, 2666),
 (3, 'Lets Bounce', 130000, 288, 1714),
 (4, 'Go West', 139000, 248, 2181),
 (5, 'Patrick''s Jig', 118000, 168, 2790);
 
INSERT INTO character (id, name, gift_id, stamp_id) VALUES
 (1, 'Rockhopper', 9215, 7),
 (2, 'Aunt Arctic', 9196, 33),
 (3, 'Cadence', 9217, 31),
 (4, 'Gary', 9190, 8),
 (5, 'Franky', 9251, 32),
 (6, 'Petey K', 9251, 35),
 (7, 'G Billy', 9251, 34),
 (8, 'Stompin Bob', 9251, 36),
 (9, 'Sensei', 9200, 290),
 (10, 'Herbert P Bear', 9239, 466),
 (11, 'Wheel Bot', NULL, NULL),
 (12, 'Jet Bot', NULL, NULL),
 (13, 'Snow Bot', NULL, NULL),
 (14, 'Protobot', NULL, NULL),
 (15, 'Dot', 9303, NULL),
 (16, 'Rookie', 9250, 358),
 (17, 'Jet Pack Guy', NULL, NULL),
 (18, 'Director', NULL, NULL),
 (19, 'Klutzy', NULL, NULL),
 (25, 'G', 9190, NULL),
 (26, 'Happy77', NULL, NULL),
 (27, 'Billybob', NULL, NULL),
 (28, 'PH', 9193, 488),
 (31, 'Brady', 9211, NULL),
 (32, 'McKenzie', 9211, NULL),
 (33, 'Kermit The Frog', 9252, NULL),
 (35, 'Sam', 9277, NULL);
 
INSERT INTO penguin (id, username, nickname, approval_en, password, email, color, character) VALUES
 (1, 'rockhopper', 'Rockhopper', TRUE, '$2b$12$CCYijGFRZyymIJWWNpkmP.pysAEN5E1mRwPtrjIDmTR3LnhKdJeBK', '', 5, 1),
 (2, 'aunt arctic', 'Aunt Arctic', TRUE, '$2b$12$CCYijGFRZyymIJWWNpkmP.pysAEN5E1mRwPtrjIDmTR3LnhKdJeBK', '', 2, 2),
 (3, 'cadence', 'Cadence', TRUE, '$2b$12$CCYijGFRZyymIJWWNpkmP.pysAEN5E1mRwPtrjIDmTR3LnhKdJeBK', '', 10, 3),
 (4, 'gary', 'Gary', TRUE, '$2b$12$CCYijGFRZyymIJWWNpkmP.pysAEN5E1mRwPtrjIDmTR3LnhKdJeBK', '', 1, 4),
 (5, 'franky', 'Franky', TRUE, '$2b$12$CCYijGFRZyymIJWWNpkmP.pysAEN5E1mRwPtrjIDmTR3LnhKdJeBK', '', 7, 5),
 (6, 'petey k', 'Petey K', TRUE, '$2b$12$CCYijGFRZyymIJWWNpkmP.pysAEN5E1mRwPtrjIDmTR3LnhKdJeBK', '', 2, 6),
 (7, 'g billy', 'G Billy', TRUE, '$2b$12$CCYijGFRZyymIJWWNpkmP.pysAEN5E1mRwPtrjIDmTR3LnhKdJeBK', '', 1, 7),
 (8, 'stompin bob', 'Stompin Bob', TRUE, '$2b$12$CCYijGFRZyymIJWWNpkmP.pysAEN5E1mRwPtrjIDmTR3LnhKdJeBK', '', 5, 8),
 (9, 'sensei', 'Sensei', TRUE, '$2b$12$CCYijGFRZyymIJWWNpkmP.pysAEN5E1mRwPtrjIDmTR3LnhKdJeBK', '', 14, 9),
 (10, 'herbert', 'Herbert P Bear', TRUE, '$2b$12$CCYijGFRZyymIJWWNpkmP.pysAEN5E1mRwPtrjIDmTR3LnhKdJeBK', '', 1, 10),
 (11, 'wheel bot', 'Wheel Bot', TRUE, '$2b$12$CCYijGFRZyymIJWWNpkmP.pysAEN5E1mRwPtrjIDmTR3LnhKdJeBK', '', 1, 11),
 (12, 'jet bot', 'Jet Bot', TRUE, '$2b$12$CCYijGFRZyymIJWWNpkmP.pysAEN5E1mRwPtrjIDmTR3LnhKdJeBK', '', 1, 12),
 (13, 'snow bot', 'Snow Bot', TRUE, '$2b$12$CCYijGFRZyymIJWWNpkmP.pysAEN5E1mRwPtrjIDmTR3LnhKdJeBK', '', 1, 13),
 (14, 'protobot', 'Protobot', TRUE, '$2b$12$CCYijGFRZyymIJWWNpkmP.pysAEN5E1mRwPtrjIDmTR3LnhKdJeBK', '', 1, 14),
 (15, 'dot', 'Dot', TRUE, '$2b$12$CCYijGFRZyymIJWWNpkmP.pysAEN5E1mRwPtrjIDmTR3LnhKdJeBK', '', 1, 15),
 (16, 'rookie', 'Rookie', TRUE, '$2b$12$CCYijGFRZyymIJWWNpkmP.pysAEN5E1mRwPtrjIDmTR3LnhKdJeBK', '', 11, 16),
 (17, 'jet pack guy', 'Jet Pack Guy', TRUE, '$2b$12$CCYijGFRZyymIJWWNpkmP.pysAEN5E1mRwPtrjIDmTR3LnhKdJeBK', '', 1, 17),
 (18, 'director', 'Director', TRUE, '$2b$12$CCYijGFRZyymIJWWNpkmP.pysAEN5E1mRwPtrjIDmTR3LnhKdJeBK', '', 1, 18),
 (19, 'klutzy', 'Klutzy', TRUE, '$2b$12$CCYijGFRZyymIJWWNpkmP.pysAEN5E1mRwPtrjIDmTR3LnhKdJeBK', '', 1, 19),
 (20, 'g', 'G', TRUE, '$2b$12$CCYijGFRZyymIJWWNpkmP.pysAEN5E1mRwPtrjIDmTR3LnhKdJeBK', '', 1, 25),
 (21, 'ph', 'PH', TRUE, '$2b$12$CCYijGFRZyymIJWWNpkmP.pysAEN5E1mRwPtrjIDmTR3LnhKdJeBK', '', 9, 28),
 (22, 'brady', 'Brady', TRUE, '$2b$12$CCYijGFRZyymIJWWNpkmP.pysAEN5E1mRwPtrjIDmTR3LnhKdJeBK', '', 7, 31),
 (23, 'mckenzie', 'McKenzie', TRUE, '$2b$12$CCYijGFRZyymIJWWNpkmP.pysAEN5E1mRwPtrjIDmTR3LnhKdJeBK', '', 15, 32),
 (24, 'kermit', 'Kermit The Frog', TRUE, '$2b$12$CCYijGFRZyymIJWWNpkmP.pysAEN5E1mRwPtrjIDmTR3LnhKdJeBK', '', 2, 33),
 (25, 'sam', 'Sam', TRUE, '$2b$12$CCYijGFRZyymIJWWNpkmP.pysAEN5E1mRwPtrjIDmTR3LnhKdJeBK', '', 1, 35),
 (26, 'rocky', 'Rocky', TRUE, '$2b$12$CCYijGFRZyymIJWWNpkmP.pysAEN5E1mRwPtrjIDmTR3LnhKdJeBK', '', 15, NULL),
 (27, 'cece', 'CeCe', TRUE, '$2b$12$CCYijGFRZyymIJWWNpkmP.pysAEN5E1mRwPtrjIDmTR3LnhKdJeBK', '', 7, NULL),
 (28, 'merry walrus', 'Merry Walrus', TRUE, '$2b$12$CCYijGFRZyymIJWWNpkmP.pysAEN5E1mRwPtrjIDmTR3LnhKdJeBK', '', 1, NULL);
 
INSERT INTO penguin (username, nickname, approval_en, active, password, email, color) VALUES
  ('basil', 'Basil', TRUE, TRUE, '$2b$12$CCYijGFRZyymIJWWNpkmP.pysAEN5E1mRwPtrjIDmTR3LnhKdJeBK', 'basil@solero.me', 1);
  
INSERT INTO penguin_item (penguin_id, item_id) VALUES
  (101, 1);
  
INSERT INTO penguin_item (penguin_id, item_id) VALUES 
 (1, 5), (1, 1692), (1, 4712), (1, 442), (1, 1476), (1, 3194), (1, 4946), (1, 1521), (1, 1753), (1, 24016), (1, 161), (1, 152), (1, 5020),
 (2, 2), (2, 2007), (2, 1044), (2, 1562), (2, 4814),
 (3, 10), (3, 1032), (3, 3011), (3, 1033), (3, 1460), (3, 1701), (3, 5023), (3, 1235), (3, 4955), (3, 1034),
 (4, 1), (4, 1875), (4, 1924), (4, 115), (4, 2087), (4, 2102), (4, 2128), (4, 2113), (4, 4022),
 (5, 7), (5, 4382), (5, 6079), (5, 1000), (5, 5024), (5, 1918), (5, 6000),
 (6, 2), (6, 1921), (6, 3082), (6, 4381), (6, 2034), (6, 1273), (6, 1003), (6, 2000), (6, 6078), (6, 3016),
 (7, 1), (7, 6080), (7, 5107), (7, 1275), (7, 4384), (7, 1001), (7, 1919), (7, 5000),
 (8, 5), (8, 5025), (8, 4383), (8, 5106), (8, 5105), (8, 1274), (8, 1920), (8, 1002),
 (9, 14), (9, 1622), (9, 4885), (9, 6177), (9, 4148), (9, 1068), (9, 4281), (9, 2009),
 (10, 1), (10, 1654), (10, 1737), (10, 4752),
 (15, 1), (15, 17), (15, 24315), (15, 21032),
 (16, 11), (16, 4969), (16, 1783), (16, 1720), (16, 4366), (16, 1257), (16, 4365), (16, 2030),
 (21, 9), (21, 1384), (21, 4555), (21, 1974), (21, 1947), (21, 1736),
 (22, 7), (22, 1682),
 (23, 15), (23, 1683),
 (24, 2), (24, 1805),
 (25, 1), (25, 1917),
 (26, 15), (26, 5168), (26, 1458), (26, 3130), (26, 4690), (26, 6127),
 (27, 7), (27, 5169), (27, 1459), (27, 3131), (27, 4691), (27, 6128),
 (28, 1), (28, 1964);
